#ifndef __COMMON_H_INCLUDED
#define __COMMON_H_INCLUDED
#include "gddcommon.h"
/*#define MULTI*/

// real([0:256.0))
double drand( void )
{
  return (double)(rand()%256);
}
// complex([0:256.0),[0:256.0))
cuDoubleComplex zrand( void )
{
  return make_cuDoubleComplex((double)(rand()%256),(double)(rand()%256));
}

double dzero( void )
{
  return 0.0;
}
cuDoubleComplex zzero( void )
{
  return make_cuDoubleComplex(0.0,0.0);
}

void doutput( double A[], int m, int lda )
{
  int i, j;
  for( j=0; j<lda; j++ ){
    for( i=0; i<m; i++ ){
      printf(" %f", A[j+i*lda] );
    }
    printf("\n");
  }
  printf("\n");
}

void zoutput( cuDoubleComplex A[], int m, int lda )
{
  int i, j;
  for( j=0; j<lda; j++ ){
    for( i=0; i<m; i++ ){
      printf(" %f,%f", cuCreal(A[j+i*lda]), cuCimag(A[j+i*lda]) );
    }
    printf("\n");
  }
  printf("\n");
}

int dcheck
( double Ch[], double Dh[],
  int m, int lda )
{
  int i,j;
  int err=0;

  printf("check\n");
  for( j=0; j<lda; j++ ){
    for( i=0; i<m; i++ ){
      int k = j+i*lda;
      if( fabs(Ch[k]-Dh[k])>1.0e-11*(fabs(Ch[k])+fabs(Dh[k])) ){
	if( err++>16 ) break;

	printf("%3d %3d %20.15e %20.15e %e %e\n", i, j,
	       Ch[k],  Dh[k],
	       Ch[k] - Dh[k] ,fabs((Ch[k]-Dh[k])/(fabs(Ch[k])+fabs(Dh[k]))));
      }
    }
  }

  if( err == 0 ){
    printf("Success\n");
  }
  else{
    printf("Failed\n");
  }

  return err;
}

int ddcheck
( double Ch[], double Cl[],
 double Dh[], double Dl[],
  int m, int lda )
{
  int i,j;
  int err=0;

  printf("check\n");
  for( j=0; j<lda; j++ ){
    for( i=0; i<m; i++ ){
      int k = j+i*lda;

      double addresulth,addresultl,diffresulth,diffresultl;
      gdd_diff(Ch+k,Cl+k,Dh+k,Dl+k,&diffresulth,&diffresultl);
      gdd_add(Ch+k,Cl+k,Dh+k,Dl+k ,&addresulth ,&addresultl);
      if( fabs(diffresulth)>1.0e-30*fabs(addresulth) ){
	if( err++>16 ) break;

	printf("%3d %3d (%20.15e,%20.15e) (%20.15e,%20.15e) %e %e\n", i, j,
	       Ch[k],  Cl[k],
	       Dh[k],  Dl[k],
	       addresulth,diffresulth );
      }
    }
  }

  if( err == 0 ){
    printf("Success\n");
  }
  else{
    printf("Failed\n");
  }

  return err;
}






int zcheck
( cuDoubleComplex Ch[], cuDoubleComplex Dh[],
  int m, int lda )
{
  int i, j;
  int err=0;

  for( j=0; j<lda; j++ ){
    for( i=0; i<m; i++ ){
      int k = j+i*lda;
      if( cuCabs(cuCsub(Ch[k],Dh[k]))>1.0e-15*(cuCabs(Ch[k])+cuCabs(Dh[k])) ){
	if( err++>16 ) break;

	printf("%3d %3d (%e,%e) (%e,%e)\n", i, j,
	       cuCreal(Ch[k]),  cuCimag(Ch[k]),
	       cuCreal(Dh[k]),  cuCimag(Dh[k]) );
      }
    }
  }

  if( err == 0 ){
    printf("Success\n");
  }
  else{
    printf("Failed\n");
  }

  return err;
}

char getRandomChar(char*s,int n)
{
	int m = rand()%n;
	return s[m];
}


#endif // __COMMON_H_INCLUDED
