#include <stdio.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"
#include <time.h>


int
main( int argc, char const* argv[] )
{
    srand( (unsigned)time( NULL ) );
    int      i, j, k;
    int      MAX_DIM = 256;
    int  m= rand( )%MAX_DIM+1;
    int  n= rand( )%MAX_DIM+1;
    int      kl = rand()%m;
    int      ku = rand()%(m-kl-1);

    int      lda = m;


    char trans = getRandomChar("nNTt",4);
/*    char     trans = 'N';*/

    double * Ah, *Al, *bAh, *bAl, *hAh, *hAl;
    int  incx = rand( )%MAX_DIM+1;
    int  incy = rand( )%MAX_DIM+1;
/*    int      incx = 1;*/
/*    int      incy = 1;*/
    int      sizex, sizey;

    if( trans == 'N' || trans == 'n' )
    {
        sizex = incx*n;
        sizey = incy*m;
    }
    else
    {
        sizex = incx*m;
        sizey = incy*n;
    }

    int sizeA = lda*n;

    if(
        cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    int sizehA = m*n;
    if(
        cudaSuccess != cudaMallocHost( (void **)&hAh, sizeA*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&hAl, sizeA*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    int ldba = ( kl+ku+1 );
    int sizebA = ldba*n;

    if(
        cudaSuccess != cudaMallocHost( (void **)&bAh, sizebA*sizeof( double ) )
      || cudaSuccess !=
        cudaMallocHost( (void **)&bAl, sizebA*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    for( i = 0; i < m; i++ )
    {
        for( j = 0; j < n; j++ )
        {
            Ah[i+j*lda] = (double) (i+1+j*lda);
            Al[i+j*lda] = 0.0;
        }
    }


/*    int x, y;*/
/*    for( x = 0; x < m; x++ )*/
/*    {*/
/*        for( y = 0; y < n; y++ )*/
/*        {*/
/*            printf( "%f,", Ah[x+y*lda] );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/

/*    printf( "****************************\n" );*/


    copyToBandMatrix( m, n, lda, kl, ku, Ah, Al, bAh, bAl );

/*    for( x = 0; x < kl+ku+1; x++ )*/
/*    {*/
/*        for( y = 0; y < n; y++ )*/
/*        {*/
/*            printf( "%f,", bAh[x+y*( kl+ku+1 )] );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/

/*    printf( "****************************\n" );*/


    double * xh, *xl;


    if(
        cudaSuccess != cudaMallocHost( (void **)&xh, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&xl, sizex*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    for( i = 0; i < sizex; i++ )
    {
/*        xh[i] = (double) i+1;*/
	xh[i] = drand();
        xl[i] = 0.0;
    }

/*    for( i = 0; i < sizex; i++ )*/
/*    {*/
/*        printf( "%e,", xh[i] );*/
/*    }*/
/*    printf( "\n\n" );*/

    double * yh, *yl;
    double * yh0, *yl0;


    if(
        cudaSuccess != cudaMallocHost( (void **) &yh, sizey*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&yl, sizey*sizeof( double ) )
      || cudaSuccess !=
        cudaMallocHost( (void **) &yh0, sizey*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&yl0, sizey*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    for( i = 0; i < sizey; i++ )
    {
	yh[i] = drand( );
/*        yh[i] = (double) (i+1)*2;*/
        yl[i] = 0.0;
        yh0[i] = yh[i];
        yl0[i] = yl[i];
    }

/*    for( i = 0; i < sizey; i++ )*/
/*    {*/
/*        printf( "%e,", yh[i] );*/
/*    }*/
/*    printf( "\n\n" );*/


    double alphah = drand();
    double alphal = 0.0;
    double betah = drand();
    double betal = 0.0;


#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddgbmv_host_multigpu_( &trans, &m,  &n,  &kl, &ku, &alphah,  &alphal, bAh,  bAl,
                  &ldba, xh,
                  xl,  &incx,  &betah, &betal,  yh,  yl,
                  &incy );
#else
    printf("calling single-gpu function\n");
    gddgbmv_host_( &trans, &m,  &n,  &kl, &ku, &alphah,  &alphal, bAh,  bAl,
                  &ldba, xh,
                  xl,  &incx,  &betah, &betal,  yh,  yl,
                  &incy );

/*    for( i = 0; i < sizey; i++ )*/
/*    {*/
/*        printf( "%e,", yh[i] );*/
/*    }*/
/*    printf( "\n\n" );*/

#endif
    printf("calling single-cpu function\n");
    dgbmv_( &trans, &m,  &n,  &kl, &ku, &alphah, bAh,
            &ldba, xh,
            &incx,  &betah,  yh0,
            &incy );


/*    for( i = 0; i < sizey; i++ )*/
/*    {*/
/*        printf( "%e,", yh0[i] );*/
/*    }*/
/*    printf( "\n\n" );*/

    printf( "trans:%c,kl:%d,ku:%d,m:%d,n:%d,incx:%d,incy:%d\n", trans, kl,
            ku, m, n, incx,
            incy );
    dcheck( yh, yh0, sizey, 1 );

    cudaFreeHost( Ah );
    cudaFreeHost( Al );
    cudaFreeHost( xh );
    cudaFreeHost( xl );
    cudaFreeHost( yh );
    cudaFreeHost( yl );

    return 0;
}

