// test for gddgemm

#include <stdio.h>
#include <stdlib.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"
#include <time.h>

int main( int argc, char const* argv[] )
{
	srand((unsigned)time(NULL));
  // A(nrowA,ncolA), B(nrowB,ncolB), C(nrowC,ncolC)
  // lda>=nrowA ldb>=nrowB, ldc>=nrowC

  // dimmension of the operation
  int MAX_DIM = 256;
  int m = rand()%MAX_DIM+1; // nrowC and nrow op(A)
  int n = rand()%MAX_DIM+1; // ncolC and ncol op(B)
  int k = rand()%MAX_DIM+1; // ncol op(A) and nrow op(B)

  char transA = getRandomChar("nNtT",4); // 'N' or 'T'
  char transB = getRandomChar("nNtT",4); // 'N' or 'T'

  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;
  int lda, ldb, ldc;
  int sizeA, sizeB, sizeC;

  if( transA == 'N' || transA == 'n'){
    nrowA = m, ncolA = k;
  }
  else{
    nrowA = k, ncolA = m;
  }
  lda = nrowA;

  if( transB == 'N' || transB == 'n'){
    nrowB = k, ncolB = n;
  }
  else{
    nrowB = n, ncolB = k;
  }
  ldb = nrowB;

  {
    nrowC = m, ncolC = n;
  }
  ldc = nrowC;


  double *Ah, *Al, *Bh, *Bl, *Ch, *Cl, *Dh;

  sizeA = lda*ncolA;
  sizeB = ldb*ncolB;
  sizeC = ldc*ncolC;

  if(
     cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Bh, sizeB*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Bl, sizeB*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Ch, sizeC*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Cl, sizeC*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Dh, sizeC*sizeof( double ) ) ){
    printf("Host Malloc Error\n");
    return -1;
  }

  int i, j;

  for( j=0; j<ncolA; j++ ){
    for( i=0; i<nrowA; i++ ){
      Ah[i+j*lda] = drand();
      Al[i+j*lda] = dzero();
    }
  }

  for( j=0; j<ncolB; j++ ){
    for( i=0; i<nrowB; i++ ){
      Bh[i+j*ldb] = drand();
      Bl[i+j*ldb] = dzero();
    }
  }

  for( j=0; j<ncolC; j++ ){
    for( i=0; i<nrowC; i++ ){
      Ch[i+j*ldc] = drand();
      Cl[i+j*ldc] = dzero();

      Dh[i+j*ldc] = Ch[i+j*ldc];
    }
  }

  double alphah = drand();
  double alphal = dzero();

  double betah = drand();
  double betal = dzero();

  {
    printf("dgemm C(%d,%d) += A(%d,%d)^%c * B(%d,%d)^%c\n",
	   nrowC, ncolC, nrowA, ncolA, transA,
	   nrowB, ncolB, transB );

#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddgemm_host_multigpu_
      ( &transA, &transB, &m, &n, &k, &alphah, &alphal, Ah, Al, &lda,
	Bh, Bl, &ldb, &betah, &betal, Ch, Cl, &ldc );
#else
    printf("calling single-gpu function\n");
    gddgemm_host_
      ( &transA, &transB, &m, &n, &k, &alphah, &alphal, Ah, Al, &lda,
	Bh, Bl, &ldb, &betah, &betal, Ch, Cl, &ldc );
#endif
    printf("calling single-cpu function\n");
    dgemm_
      ( &transA, &transB, &m, &n, &k, &alphah, Ah, &lda,
	Bh, &ldb, &betah, Dh, &ldc );

    dcheck(Ch,Dh,ncolC,ldc);
  }

  /*
  doutput(Ah, nrowA, lda );
  doutput(Bh, nrowB, ldb );
  doutput(Ch, nrowC, ldc );
  doutput(Dh, nrowC, ldc );
  */

  cudaFreeHost( Ah );
  cudaFreeHost( Al );
  cudaFreeHost( Bh );
  cudaFreeHost( Bl );
  cudaFreeHost( Ch );
  cudaFreeHost( Cl );
  cudaFreeHost( Dh );

  return 0;
}

