#include <stdio.h>
#include <stdlib.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"
#include <time.h>

int
main( int argc, char const* argv[] )
{
    srand( (unsigned)time( NULL ) );
    /* dimmension of the operation */
    int      MAX_DIM = 256;
    int      m = rand( )%MAX_DIM+1; /* nrowC and nrow op(A) */
    int      n = rand( )%MAX_DIM+1; /* ncolC and ncol op(B) */
    int      i, j;
    int      lda = m;

    char     trans = getRandomChar( "nNtT", 4 ); /* 'N' or 'T' */

    double * Ah, *Al;
    int      incx = rand( )%MAX_DIM+1;
    int      incy = rand( )%MAX_DIM+1;
    int      sizex, sizey;

    if( trans == 'N' || trans == 'n' )
    {
        sizex = incx*n;
        sizey = incy*m;
    }
    else
    {
        sizex = incx*m;
        sizey = incy*n;
    }

    int sizeA = lda*n;

    if(
        cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    for( i = 0; i < m; i++ )
    {
        for( j = 0; j < n; j++ )
        {
            Ah[i+j*lda] = drand( );
            Al[i+j*lda] = dzero( );
        }
    }


/*    int x, y;*/
/*    for( x = 0; x < m; x++ )*/
/*    {*/
/*        for( y = 0; y < n; y++ )*/
/*        {*/
/*            printf( "%f,", Ah[x+y*lda] );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/

/*    printf( "****************************\n" );*/


    double * xh, *xl;


    if(
        cudaSuccess != cudaMallocHost( (void **)&xh, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&xl, sizex*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    for( i = 0; i < sizex; i++ )
    {
        xh[i] = drand();
        xl[i] = dzero();
    }

/*    for( i = 0; i < sizex; i++ )*/
/*    {*/
/*        printf( "%e,", xh[i] );*/
/*    }*/
/*    printf( "\n\n" );*/

    double * yh, *yl;
    double * yh2;


    if(
        cudaSuccess != cudaMallocHost( (void **) &yh, sizey*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&yl, sizey*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&yh2, sizey*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    for( i = 0; i < sizey; i++ )
    {
        yh[i] = drand();
        yl[i] = dzero();
        yh2[i] = yh[i];
    }

/*    for( i = 0; i < sizey; i++ )*/
/*    {*/
/*        printf( "%e,", yh[i] );*/
/*    }*/
/*    printf( "\n\n" );*/


    double alphah = drand();
    double alphal = dzero();
    double betah = drand();
    double betal = dzero();


/*    gddgemv_host_(&trans,  &m,  &n, &alphah,  &alphal, Ah,  Al,  &lda, xh,*/
/*        xl,  &incx,  &betah, &betal,  yh,  yl, &incy );*/
#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddgemv_host_multigpu_( &trans,  &m,  &n, &alphah,  &alphal, Ah,  Al,
                           &lda, xh,
                           xl,  &incx,  &betah, &betal,  yh,  yl,
                           &incy );
#else
    printf("calling single-gpu function\n");
    gddgemv_host_( &trans,  &m,  &n, &alphah,  &alphal, Ah,  Al,
                           &lda, xh,
                           xl,  &incx,  &betah, &betal,  yh,  yl,
                           &incy );
#endif
    printf("calling single-cpu function\n");
    dgemv_( &trans,  &m,  &n, &alphah,  Ah, 
                           &lda, xh,
                           &incx,  &betah,   yh2, 
                           &incy );

    printf("trans:%c,n:%d,incx:%d,incy:%d\n",trans,n,incx,incy);
    dcheck(yh,yh2,sizey,1);


/*    for( i = 0; i < sizey; i++ )*/
/*    {*/
/*        printf( "%e,", yh[i] );*/
/*    }*/
/*    printf( "\n\n" );*/



    cudaFreeHost( Ah );
    cudaFreeHost( Al );
    cudaFreeHost( xh );
    cudaFreeHost( xl );

    return 0;
}

