#include <stdio.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"
#include <time.h>

int
main( int argc, char const* argv[] )
{
    srand( (unsigned)time( NULL ) );
    int      MAX_DIM = 256;
    int      n = rand( )%MAX_DIM+1; /* ncolC and ncol op(B) */

    double * xh, *xl;
    double *yh;
    int      incx = rand( )%MAX_DIM+1;

    if(
        cudaSuccess != cudaMallocHost( (void **)&xh, incx*n*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&xl, incx*n*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&yh, incx*n*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    int i;
    for( i = 0; i < n*incx; i++ )
    {
	xh[i] = drand();
        xl[i] = dzero();
	yh[i] = xh[i];
    }



/*    for( i = 0; i < n*incx; i++ )*/
/*    {*/
/*        printf( "x::%d,%e,%e\n", i, xh[i], xl[i] );*/
/*    }*/
/*    printf( "###############\n" );*/

#ifdef MULTI
    printf("calling multi-gpu function\n");
   int t1 = gddidamax_host_multigpu_( &n, xh, xl, &incx);
#else
    printf("calling single-gpu function\n");
    int t1 = gddidamax_host_( &n, xh, xl, &incx );
#endif
    printf("calling single-cpu function\n");
    int t2= idamax_( &n, yh,  &incx);


    printf("n:%d,incx:%d\n",n,incx);
    if(t1 == t2) 
    {
    printf("Success: t1:%d,t2:%d\n",t1,t2);
    }
    else
    {
    printf("Fail: t1:%d,t2:%d\n",t1,t2);
    }


    cudaFreeHost( xh );
    cudaFreeHost( xl );

    return 0;
}

