#include <stdio.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"
#include <time.h>



double dnrm2_( int *n, double*xh,  double*incx);
int
main( int argc, char const* argv[] )
{
	srand((unsigned)time(NULL));
	int MAX_DIM = 256;
    int      n = rand()%MAX_DIM;

    double * xh, *xl;
    int      incx = rand()%MAX_DIM;

    if(
        cudaSuccess != cudaMallocHost( (void **)&xh, incx*n*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&xl, incx*n*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    int i;
    for( i = 0; i < n*incx; i++ )
    {
        xh[i] = drand();
        xl[i] = drand()*1e-17 ;
    }


    double reth = 1.0;
    double retl = 0.0;
    double reth0 = 1.0;
    double retl0 = 0.0;

#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddnrm2_host_multigpu_( &n, xh, xl, &incx, &reth, &retl );
#else
    printf("calling single-gpu function\n");
    gddnrm2_host_( &n, xh, xl, &incx, &reth, &retl );
#endif
    printf("calling single-cpu function\n");
    reth0 = dnrm2_( &n, xh,  &incx);


    printf( "n:%d,incx:%d\n",n,incx );
    dcheck( &reth,  &reth0,  1, 1 );

    cudaFreeHost( xh );
    cudaFreeHost( xl );

    return 0;
}

