#include <stdio.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"
#include <time.h>

int
main( int argc, char const* argv[] )
{
    srand( (unsigned)time( NULL ) );
    int      MAX_DIM = 256;
    int      n = rand( )%MAX_DIM;

    double * xh, *xl, *yh, *yl, *xh0, *xl0, *yh0, *yl0;
    int      incx = rand( )%MAX_DIM;
    int      incy = rand( )%MAX_DIM;


    int      sizex = incx*n;
    int      sizey = incy*n;

    if(
         cudaSuccess != cudaMallocHost( (void **)&xh, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&xl, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&yh, sizey*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&yl, sizey*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&xh0, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&xl0, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&yh0, sizey*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&yl0, sizey*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    int i;
    for( i = 0; i < sizex; i++ )
    {
        xh[i] = drand( );
        xl[i] = drand( )*1e-17;
        xh0[i] = xh[i];
        xl0[i] = xl[i];

/*        printf("xh0,%e\n",xh[i]);*/
    }

    for( i = 0; i < sizey; i++ )
    {
        yh[i] = drand( );
        yl[i] = drand( )*1e-17;
        yh0[i] = yh[i];
        yl0[i] = yl[i];
/*        printf("yh0,%e\n",yh[i]);*/
    }


    double ch = drand( );          
    double cl = drand( )*1e-17;    
    double sh = drand( );          
    double sl = drand( )*1e-17;    

/*    for( i = 0; i < n*incx; i++ )*/
/*    {*/
/*        printf( "x::%d,%e,%e\n", i, xh[i], xl[i] );*/
/*    }*/
/*    printf( "###############\n" );*/
/*    for( i = 0; i < n*incy; i++ )*/
/*    {*/
/*        printf( "x::%d,%e,%e\n", i, yh[i], yl[i] );*/
/*    }*/
/*    printf( "###############\n" );*/

/*   int t = gddaxpy_host_multigpu_( &n, &alphah,&alphal,xh, xl, &incx, yh, yl, &incy );*/
/*   printf("t=%d\n",t);*/
#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddrot_host_multigpu_( &n, xh, xl, &incx, yh, yl, &incy,&ch, &cl, &sh, &sl );
#else
    printf("calling single-gpu function\n");
    gddrot_host_( &n, xh, xl, &incx, yh, yl, &incy,&ch, &cl, &sh, &sl );
#endif
    printf("calling single-cpu function\n");
    drot_( &n, xh0, &incx, yh0, &incy,&ch,  &sh );

    printf("n:%d,incx:%d,incy:%d\n",n,incx,incy);
    dcheck( xh, xh0,  sizex, 1 );
    dcheck( yh, yh0,  sizey, 1 );



    cudaFreeHost( xh );
    cudaFreeHost( xl );
    cudaFreeHost( yh );
    cudaFreeHost( yl );

    return 0;
}

