// test for gddsymm

#include <stdio.h>
#include <stdlib.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"

int main( int argc, char const* argv[] )
{
	srand((unsigned)time(NULL));
  // A(nrowA,ncolA), B(nrowB,ncolB), C(nrowC,ncolC)
  // lda>=nrowA ldb>=nrowB, ldc>=nrowC

  // dimmension of the operation
  int MAX_DIM = 256;
  int m = rand()%MAX_DIM; // nrowC and nrow op(A)
  int n = rand()%MAX_DIM; // ncolC and ncol op(B)
/*  int m = 3; // nrowC and nrow op(A)*/
/*  int n = 4; // ncolC and ncol op(B)*/
  int k = rand()%MAX_DIM; // ncol op(A) and nrow op(B)

/*  char side = 'L'; // 'L' or 'R'*/
/*  char uplo = 'L'; // 'L' or 'U'*/
  char side = getRandomChar("lLrR",4); // 'L' or 'R' 
  char uplo = getRandomChar("uUlL",4); // 'L' or 'U' 

  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;
  int lda, ldb, ldc;
  int sizeA, sizeB, sizeC;

  /*
    A(M,M)*B(M,N), A is symmetric, B is general
   */
  if( side == 'L' || side == 'l' ){
    nrowA = m, ncolA = m;
    nrowB = m, ncolB = n;
    nrowC = m, ncolC = n;
  }
  /*
    B(M,N)*A(N,N), A is symmetric, B is general
   */
  else{
    nrowB = m, ncolB = n;
    nrowA = n, ncolA = n;
    nrowC = m, ncolC = n;
  }
  lda = nrowA;
  ldb = nrowB;
  ldc = nrowC;

  double *Ah, *Al, *Bh, *Bl, *Ch, *Cl, *Dh;

  sizeA = lda*ncolA;
  sizeB = ldb*ncolB;
  sizeC = ldc*ncolC;

  if(
     cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Bh, sizeB*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Bl, sizeB*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Ch, sizeC*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Cl, sizeC*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Dh, sizeC*sizeof( double ) ) ){
    printf("Host Malloc Error\n");
    return -1;
  }

  int i, j;

  // ignore symmetricity
  for( j=0; j<ncolA; j++ ){
    for( i=0; i<nrowA; i++ ){
      Ah[i+j*lda] = drand();
      Al[i+j*lda] = dzero();
    }
  }

  for( j=0; j<ncolB; j++ ){
    for( i=0; i<nrowB; i++ ){
      Bh[i+j*ldb] = drand();
      Bl[i+j*ldb] = dzero();
    }
  }

  for( j=0; j<ncolC; j++ ){
    for( i=0; i<nrowC; i++ ){
      Ch[i+j*ldc] = drand();
      Cl[i+j*ldc] = dzero();

      Dh[i+j*ldc] = Ch[i+j*ldc];
    }
  }


  double alphah = drand();
  double alphal = dzero();

  double betah = drand();
  double betal = dzero();


  {
    if( side == 'L'  || side == 'l'){
      printf("dsymm C(%d,%d) += A(%d,%d) * B(%d,%d)\n",
	     nrowC, ncolC, nrowA, ncolA, nrowB, ncolB );
      printf("A is symmetric, %c side is used.\n", uplo );
    }
    else{
      printf("dsymm C(%d,%d) += B(%d,%d) * A(%d,%d)\n",
	     nrowC, ncolC, nrowB, ncolB, nrowA, ncolA );
      printf("A is symmetric, %c side is used.\n", uplo );
    }

#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddsymm_host_multigpu_
      ( &side, &uplo, &m, &n, &alphah, &alphal, Ah, Al, &lda,
	Bh, Bl, &ldb, &betah, &betal, Ch, Cl, &ldc );
#else
    printf("calling single-gpu function\n");
    gddsymm_host_
      ( &side, &uplo, &m, &n, &alphah, &alphal, Ah, Al, &lda,
	Bh, Bl, &ldb, &betah, &betal, Ch, Cl, &ldc );
#endif

    printf("calling single-cpu function\n");
    dsymm_
      ( &side, &uplo, &m, &n, &alphah, Ah, &lda,
	Bh, &ldb, &betah, Dh, &ldc );

    dcheck(Ch,Dh,ncolC,ldc);
  }

  /*
  doutput(Ah, nrowA, lda );
  doutput(Bh, nrowB, ldb );
  doutput(Ch, nrowC, ldc );
  doutput(Dh, nrowC, ldc );
  */

  cudaFreeHost( Ah );
  cudaFreeHost( Al );
  cudaFreeHost( Bh );
  cudaFreeHost( Bl );
  cudaFreeHost( Ch );
  cudaFreeHost( Cl );
  cudaFreeHost( Dh );

  return 0;
}

