#include <stdio.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"
#include <time.h>

int
main( int argc, char const* argv[] )
{
    srand( (unsigned)time( NULL ) );
    int      MAX_DIM = 256;
    int      n = rand( )%MAX_DIM+1; /* ncolC and ncol op(B) */

    double * xh, *xl, *yh, *yl, *Ah, *Al,*Bh,*Bl;
    int      incx = rand( )%MAX_DIM+1;
    int      incy = rand( )%MAX_DIM+1;

    int      lda = n;

    char uplo = getRandomChar( "uUlL", 4 ); /* 'U' or 'L' */

    int      sizeA =  lda*n;
    int      sizex = incx*n;
    int      sizey = incy*n;


    if(
         cudaSuccess != cudaMallocHost( (void **)&xh, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&xl, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&yh, sizey*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&yl, sizey*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&Bh, sizeA*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&Bl, sizeA*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    int i;
    for( i = 0; i < sizex; i++ )
    {
        xh[i] = drand();
        xl[i] = dzero();
    }
    for( i = 0; i < sizey; i++ )
    {
        yh[i] = drand();
        yl[i] = dzero();
    }

    for( i = 0; i < sizeA; i++ )
    {
        Ah[i] = drand();
        Al[i] = 0.0;
	Bh[i] = Ah[i];
    }


/*    for( i = 0; i < sizex; i++ )*/
/*    {*/
/*        printf( "x::%d,%e,%e\n", i, xh[i], xl[i] );*/
/*    }*/
/*    printf( "###############\n" );*/
/*    for( i = 0; i < sizey; i++ )*/
/*    {*/
/*        printf( "x::%d,%e,%e\n", i, yh[i], yl[i] );*/
/*    }*/
/*    printf( "###############\n" );*/

    double alphah = drand();
    double alphal = 0.0;


#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddsyr2_host_multigpu_( &uplo, &n, &alphah, &alphal, xh, xl, &incx, yh, yl, &incy, Ah,
		 Al,&lda );
#else
    printf("calling single-gpu function\n");
    gddsyr2_host_( &uplo, &n, &alphah, &alphal, xh, xl, &incx, yh, yl, &incy, Ah,
		 Al,&lda );
#endif
    printf("calling single-cpu function\n");
    dsyr2_( &uplo, &n, &alphah,  xh, &incx, yh, &incy, Bh,&lda );

    dcheck(Ah,Bh,lda,n);
/*    gddsyr2_host_multigpu_( &uplo, &n, &alphah, &alphal, xh, xl, &incx, yh, yl, &incy, Ah,*/
/*                 Al,&lda );*/

/*    int    x, y;*/
/*    for( x = 0; x < n; x++ )*/
/*    {*/
/*        for( y = 0; y < n; y++ )*/
/*        {*/
/*            printf( "(%e,%e),", *( Ah+x+lda*y ), *( Al+x+lda*y ) );*/
/*        }*/

/*        printf( "\n" );*/
/*    }*/

    cudaFreeHost( xh );
    cudaFreeHost( xl );
    cudaFreeHost( yh );
    cudaFreeHost( yl );
    cudaFreeHost( Ah );
    cudaFreeHost( Al );

    return 0;
}

