#include <stdio.h>
#include <stdlib.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"

int main( int argc, char const* argv[] )
{
	srand((unsigned)time(NULL));
  // A(nrowA,ncolA), B(nrowB,ncolB), C(nrowC,ncolC)
  // lda>=nrowA ldb>=nrowB, ldc>=nrowC

  // dimmension of the operation
  int MAX_DIM = 256;
  int n = rand()%MAX_DIM; // nrow op(A) and nrow op(B)
  int k = rand()%MAX_DIM; // ncol op(A) and ncol op(B)
/*  int n = 50; // nrow op(A) and nrow op(B)*/
/*  int k = 20; // ncol op(A) and ncol op(B)*/

/*  char uplo  = 'U'; // 'L' or 'U'*/
/*  char trans = 'T'; // 'N' or 'T'*/
  char uplo  = getRandomChar("lLuU",4); // 'L' or 'U'
  char trans = getRandomChar("nNtT",4); // 'N' or 'T'

  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;
  int lda, ldb, ldc;
  int sizeA, sizeB, sizeC;

  /*
    C(N,N) += A(N,K) * B(N,K)^T
   */
  if( trans == 'N' || trans == 'n' ){
    nrowA = n, ncolA = k;
    nrowB = n, ncolB = k;
    nrowC = n, ncolC = n;
  }
  /*
    C(N,N) += A(K,N)^T * B(K,N)
   */
  else{
    nrowA = k, ncolA = n;
    nrowB = k, ncolB = n;
    nrowC = n, ncolC = n;
  }

  lda = nrowA;
  ldb = nrowB;
  ldc = nrowC;

  sizeA = lda*ncolA;
  sizeB = ldb*ncolB;
  sizeC = ldc*ncolC;

  double *Ah, *Al, *Bh, *Bl, *Ch, *Cl, *Dh;

  if(
     cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof(double) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof(double) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Bh, sizeB*sizeof(double) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Bl, sizeB*sizeof(double) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Ch, sizeC*sizeof(double) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Cl, sizeC*sizeof(double) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Dh, sizeC*sizeof(double) ) ){
    printf("Host Malloc Error\n");
    return -1;
  }

  int i, j;

  // ignore symmetricity
  for( j=0; j<ncolA; j++ ){
    for( i=0; i<nrowA; i++ ){
      Ah[i+j*lda] = (double) i+lda*j;
      Al[i+j*lda] = dzero();
    }
  }

  for( j=0; j<ncolB; j++ ){
    for( i=0; i<nrowB; i++ ){
      Bh[i+j*ldb] = (double) (2*i+lda*j-3);
      Bl[i+j*ldb] = dzero();
    }
  }

  for( j=0; j<ncolC; j++ ){
    for( i=0; i<nrowC; i++ ){
      Ch[i+j*ldc] = dzero();
      Cl[i+j*ldc] = dzero();

      Dh[i+j*ldc] = Ch[i+j*ldc];
    }
  }


/*  double alphah = drand();*/
  double alphah = 1.0;
  double alphal = dzero();

  double betah = 1.0;
  double betal = dzero();

  {
    if( trans == 'N' || trans == 'n' ){
      printf("dsyr2k C(%d,%d) += A(%d,%d) * B(%d,%d)^T\n",
	     nrowC, ncolC, nrowA, ncolA, nrowB, ncolB );
    }
    else{
      printf("dsyr2k C(%d,%d) += A(%d,%d)^T * B(%d,%d)\n",
	     nrowC, ncolC, nrowA, ncolA, nrowB, ncolB );
    }

#define MULTI
#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddsyr2k_host_multigpu_
      ( &uplo, &trans, &n, &k, &alphah, &alphal,
	Ah, Al, &lda, Bh, Bl, &ldb, &betah, &betal, Ch, Cl, &ldc );
#else
    printf("calling single-gpu function\n");
    gddsyr2k_host_
      ( &uplo, &trans, &n, &k, &alphah, &alphal,
	Ah, Al, &lda, Bh, Bl, &ldb, &betah, &betal, Ch, Cl, &ldc );
#endif

    printf("calling single-cpu function\n");
    dsyr2k_
      ( &uplo, &trans, &n, &k, &alphah,
	Ah, &lda, Bh, &ldb, &betah, Dh, &ldc );


    dcheck(Ch,Dh,ncolC,ldc);
  }

  /*
  doutput(Ah, nrowA, lda );
  doutput(Bh, nrowB, ldb );
  doutput(Ch, nrowC, ldc );
  doutput(Dh, nrowC, ldc );
  */

  cudaFreeHost( Ah );
  cudaFreeHost( Al );
  cudaFreeHost( Bh );
  cudaFreeHost( Bl );
  cudaFreeHost( Ch );
  cudaFreeHost( Cl );
  cudaFreeHost( Dh );

  return 0;
}

