#include <stdio.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"
#include <time.h>

int
main( int argc, char const* argv[] )
{
    srand( (unsigned)time( NULL ) );
    int      MAX_DIM = 256;
    int      n = rand( )%MAX_DIM+1; /* ncolC and ncol op(B) */

    double * xh, *xl, *Ah, *Al, *Bh, *Bl;
    int      incx = rand( )%MAX_DIM+1;

    int      lda = n;
    char uplo = getRandomChar( "uUlL", 4 ); /* 'U' or 'L' */

    int      incy = rand( )%MAX_DIM+1;

    int      sizeA =  lda*n;
    int      sizex = incx*n;


    if(
        cudaSuccess != cudaMallocHost( (void **)&xh,  sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&xl, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&Bh, sizeA*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&Bl, sizeA*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    int i;
    for( i = 0; i < sizex; i++ )
    {
        xh[i] = drand( );
        xl[i] = dzero( );
    }

    for( i = 0; i < sizeA; i++ )
    {
        Ah[i] = dzero( );
        Al[i] = dzero( );
        Bh[i] = Ah[i];
    }


/*    for( i = 0; i < sizex; i++ )*/
/*    {*/
/*        printf( "x::%d,%e,%e\n", i, xh[i], xl[i] );*/
/*    }*/
/*    printf( "###############\n" );*/

    double alphah = drand( );
    double alphal = 0.0;


/*    gddsyr_host_( &uplo, &n, &alphah, &alphal, xh, xl, &incx, Ah,*/
/*                 Al,&lda );*/
#ifdef MULTI
    printf( "calling multi-gpu function\n" );
    gddsyr_host_multigpu_( &uplo, &n, &alphah, &alphal, xh, xl, &incx, Ah,
                           Al, &lda );
#else
    printf( "calling single-gpu function\n" );
    gddsyr_host_( &uplo, &n, &alphah, &alphal, xh, xl, &incx, Ah,
                  Al, &lda );
#endif
    printf( "calling single-cpu function\n" );
    dsyr_( &uplo, &n, &alphah,  xh,  &incx, Bh, &lda );

    printf("n:%d,incx:%d,uplo:%c\n",n,incx,uplo);
    dcheck(Ah,Bh,lda,n);
/*    int    x, y;*/
/*    for( x = 0; x < n; x++ )*/
/*    {*/
/*        for( y = 0; y < n; y++ )*/
/*        {*/
/*            printf( "(%e,%e),", *( Ah+x+lda*y ), *( Al+x+lda*y ) );*/
/*        }*/

/*        printf( "\n" );*/
/*    }*/

    cudaFreeHost( xh );
    cudaFreeHost( xl );
    cudaFreeHost( Ah );
    cudaFreeHost( Al );
    cudaFreeHost( Bh );
    cudaFreeHost( Bl );

    return 0;
}

