// test for gddsyrk

#include <stdio.h>
#include <stdlib.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"

int main( int argc, char const* argv[] )
{
	srand((unsigned)time(NULL));
  // A(nrowA,ncolA), B(nrowB,ncolB), C(nrowC,ncolC)
  // lda>=nrowA ldb>=nrowB, ldc>=nrowC

  // dimmension of the operation
  int MAX_DIM = 256;
  int n = rand()%MAX_DIM+1; // nrow op(A) and nrow op(B)
  int k = rand()%MAX_DIM+1; // ncol op(A) and ncol op(B)

/*  char uplo  = 'L'; // 'L' or 'U'*/
/*  char trans = 'N'; // 'N' or 'T'*/
  char uplo  = getRandomChar("lLuU",4); // 'L' or 'U'
  char trans = getRandomChar("nNtT",4); // 'N' or 'T'

  int nrowA, ncolA;
  int nrowC, ncolC;
  int lda, ldb, ldc;
  int sizeA, sizeC;

  /*
    C(N,N) += A(N,K) * A(N,K)^T
   */
  if( trans == 'N' || trans == 'n' ){
    nrowA = n, ncolA = k;
    nrowC = n, ncolC = n;
  }
  /*
    C(N,N) += A(K,N)^T * A(K,N)
   */
  else{
    nrowA = k, ncolA = n;
    nrowC = n, ncolC = n;
  }

  lda = nrowA;
  ldc = nrowC;

  sizeA = lda*ncolA;
  sizeC = ldc*ncolC;

  double *Ah, *Al, *Ch, *Cl, *Dh;

  if(
     cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof(double) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof(double) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Ch, sizeC*sizeof(double) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Cl, sizeC*sizeof(double) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Dh, sizeC*sizeof(double) ) ){
    printf("Host Malloc Error\n");
    return -1;
  }

  int i, j;

  // ignore symmetricity
  for( j=0; j<ncolA; j++ ){
    for( i=0; i<nrowA; i++ ){
      Ah[i+j*lda] = drand();
      Al[i+j*lda] = dzero();
    }
  }

  for( j=0; j<ncolC; j++ ){
    for( i=0; i<nrowC; i++ ){
      Ch[i+j*ldc] = drand();
      Cl[i+j*ldc] = dzero();

      Dh[i+j*ldc] = Ch[i+j*ldc];
    }
  }

  double alphah = drand();
  double alphal = dzero();

  double betah = drand();
  double betal = dzero();

  {
    if( trans == 'N' ){
      printf("dsyrk C(%d,%d) += A(%d,%d) * A(%d,%d)^T\n",
	     nrowC, ncolC, nrowA, ncolA, nrowA, ncolA );
    }
    else{
      printf("dsyrk C(%d,%d) += A(%d,%d)^T * A(%d,%d)\n",
	     nrowC, ncolC, nrowA, ncolA, nrowA, ncolA );
    }

#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddsyrk_host_multigpu_
      ( &uplo, &trans, &n, &k, &alphah, &alphal,
	Ah, Al, &lda, &betah, &betal, Ch, Cl, &ldc );
#else
    printf("calling single-gpu function\n");
    gddsyrk_host_
      ( &uplo, &trans, &n, &k, &alphah, &alphal,
	Ah, Al, &lda, &betah, &betal, Ch, Cl, &ldc );
#endif

    printf("calling single-cpu function\n");
    dsyrk_
      ( &uplo, &trans, &n, &k, &alphah,
	Ah, &lda, &betah, Dh, &ldc );

    printf("n:%d,k:%d,uplo:%c,trans:%c\n",n,k,uplo,trans);
    dcheck(Ch,Dh,ncolC,ldc);
  }

  /*
  doutput(Ah, nrowA, lda );
  doutput(Ch, nrowC, ldc );
  doutput(Dh, nrowC, ldc );
  */

  cudaFreeHost( Ah );
  cudaFreeHost( Al );
  cudaFreeHost( Ch );
  cudaFreeHost( Cl );
  cudaFreeHost( Dh );

  return 0;
}

