#include <stdio.h>
#include <stdlib.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"
#include <time.h>

int
main( int argc, char const* argv[] )
{
    srand( (unsigned)time( NULL ) );
    int  MAX_DIM = 256;
    int  m = rand( )%MAX_DIM+1;
/*    int  m = 5;*/
    int  n = m;

    int  lda = m;

    int  i, j, k;

    char    trans = getRandomChar( "nNtT", 4 ); /* 'N' or 'T' */
    char    uplo  = getRandomChar( "uUlL", 4 ); /* 'U' or 'L' */
    char    diag  = getRandomChar( "nNuU", 4 ); /* 'N' or 'U' */

    int  kl, ku, kb;


    kb = rand()%m;

    if( uplo == 'u' || uplo == 'U' )
    {
        ku = kb;
        kl = 0;
    }
    else
    {
        ku = 0;
        kl = kb;
    }





    double * Ah, *Al, *bAh, *bAl, *hAh, *hAl;
    int  incx = rand( )%MAX_DIM+1;
    int  incy = rand( )%MAX_DIM+1;
    int      sizex, sizey;

    if( trans == 'N' || trans == 'n' )
    {
        sizex = incx*n;
        sizey = incy*m;
    }
    else
    {
        sizex = incx*m;
        sizey = incy*n;
    }

    int sizeA = lda*n;

    if(
        cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    int sizehA = m*n;
    if(
        cudaSuccess != cudaMallocHost( (void **)&hAh, sizeA*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&hAl, sizeA*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    int ldba = ( kl+ku+1 );
    int sizebA = ldba*n;

    if(
        cudaSuccess != cudaMallocHost( (void **)&bAh, sizebA*sizeof( double ) )
      || cudaSuccess !=
        cudaMallocHost( (void **)&bAl, sizebA*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    for( i = 0; i < m; i++ )
    {
        for( j = 0; j < n; j++ )
        {
            Ah[i+j*lda] = drand();
            Al[i+j*lda] = dzero();
        }
    }


    int x, y;
/*    for( x = 0; x < m; x++ )*/
/*    {*/
/*        for( y = 0; y < n; y++ )*/
/*        {*/
/*            printf( "%f,", Ah[x+y*lda] );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/

/*    printf( "****************************\n" );*/


    copyToBandMatrix( m, n, lda, kl, ku, Ah, Al, bAh, bAl );

/*    for( x = 0; x < kl+ku+1; x++ )*/
/*    {*/
/*        for( y = 0; y < n; y++ )*/
/*        {*/
/*            printf( "%f,", bAh[x+y*( kl+ku+1 )] );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/

/*    printf( "****************************\n" );*/


    double * xh, *xl,*zh,*zl;


    if(
        cudaSuccess != cudaMallocHost( (void **)&xh, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&xl, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&zh, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&zl, sizex*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    for( i = 0; i < sizex; i++ )
    {
        xh[i] = drand();
        xl[i] = dzero();
	zh[i] = xh[i];
	zl[i] = zl[i];
    }

/*    for( i = 0; i < sizex; i++ )*/
/*    {*/
/*        printf( "%e,", xh[i] );*/
/*    }*/
/*    printf( "\n\n" );*/

    double * yh, *yl;


    if(
        cudaSuccess != cudaMallocHost( (void **) &yh, sizey*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&yl, sizey*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    for( i = 0; i < sizey; i++ )
    {
        yh[i] = dzero();
        yl[i] = dzero();
    }

    double alphah = drand();
    double alphal = 0.0;
    double betah = drand();
    double betal = 0.0;


    printf("uplo:%c,trans:%c,diag:%c,n:%d,kb:%d,lda:%d\n",uplo,trans,diag,n,kb,ldba);
#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddtbmv_host_multigpu_( &uplo, &trans, &diag, &n, &kb, bAh,  bAl,  &ldba, zh, zl,
                  &incx );
#else
    printf("calling single-gpu function\n");
    gddtbmv_host_( &uplo, &trans, &diag, &n, &kb, bAh,  bAl,  &ldba, zh, zl,
                  &incx );
#endif
    printf("calling single-cpu function\n");
    dtbmv_( &uplo, &trans, &diag, &n, &kb, bAh,  &ldba, xh, 
                  &incx );
    dcheck(xh,zh,sizey,1);

/*    for( i = 0; i < sizey; i++ )*/
/*    {*/
/*        printf( "%e,", xh[i]-zh[i] );*/
/*    }*/
/*    printf( "\n\n" );*/


    cudaFreeHost( Ah );
    cudaFreeHost( Al );
    cudaFreeHost( xh );
    cudaFreeHost( xl );
    cudaFreeHost( yh );
    cudaFreeHost( yl );

    return 0;
}

