// test for gddtrmm

#include <stdio.h>
#include <stdlib.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"

int main( int argc, char const* argv[] )
{
	srand((unsigned)time(NULL));
  // A(nrowA,ncolA), B(nrowB,ncolB), C(nrowC,ncolC)
  // lda>=nrowA ldb>=nrowB, ldc>=nrowC

  // dimmension of the operation
  int MAX_DIM = 256;
  int m = rand()%MAX_DIM+1; // nrowC and nrow op(A)
  int n = rand()%MAX_DIM+1; // ncolC and ncol op(B)
/*  int m = 3; // nrowC and nrow op(A)*/
/*  int n = 2; // ncolC and ncol op(B)*/

  char side  = getRandomChar("lLrR",4); // 'L' or 'R'
  char uplo  = getRandomChar("uUlL",4); // 'L' or 'U'
  char trans = getRandomChar("nNtT",4); // 'N' or 'T'
  char diag  = getRandomChar("uUnN",4); // 'U' or 'N'
/*  char side  = 'R'; // 'L' or 'R'*/
/*  char uplo  = 'L'; // 'L' or 'U'*/
/*  char trans = 'T'; // 'N' or 'T'*/
/*  char diag  = 'N'; // 'U' or 'N'*/

  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;
  int lda, ldb, ldc;
  int sizeA, sizeB, sizeC;

  /*
    A(M,M)*B(M,N), A is symmetric, B is general
   */
  if( side == 'L' || side == 'l' ){
    nrowA = m, ncolA = m;
    nrowB = m, ncolB = n;
    nrowC = m, ncolC = n;
  }
  /*
    B(M,N)*A(N,N), A is symmetric, B is general
   */
  else{
    nrowA = n, ncolA = n;
    nrowB = m, ncolB = n;
    nrowC = m, ncolC = n;
  }
  lda = nrowA;
  ldb = nrowB;
  ldc = nrowC;

  double *Ah, *Al, *Bh, *Bl, *Ch, *Cl, *Dh;

  sizeA = lda*ncolA;
  sizeB = ldb*ncolB;
  sizeC = ldc*ncolC;

  if(
     cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Bh, sizeB*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Bl, sizeB*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Ch, sizeC*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Cl, sizeC*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Dh, sizeC*sizeof( double ) ) ){
    printf("Host Malloc Error\n");
    return -1;
  }

  int i, j;

  // ignore triangular matrix
  for( j=0; j<ncolA; j++ ){
    for( i=0; i<nrowA; i++ ){
      Ah[i+j*lda] = drand();
/*      Ah[i+j*lda] = (double) (i + nrowA*j+1);*/
      Al[i+j*lda] = dzero();
    }
  }

  for( j=0; j<ncolB; j++ ){
    for( i=0; i<nrowB; i++ ){
      Bh[i+j*ldb] = drand();
/*      Bh[i+j*ldb] = (double) (i+nrowB*j+1)*2;*/
      Bl[i+j*ldb] = dzero();
    }
  }

  // copy B to C and D to backup B
  for( j=0; j<ncolC; j++ ){
    for( i=0; i<nrowC; i++ ){
      Ch[i+j*ldc] = Bh[i+j*ldc];
      Cl[i+j*ldc] = dzero();

      Dh[i+j*ldc] = Bh[i+j*ldb];
    }
  }

/*  double alphah = drand();*/
  double alphah = 1.0;
  double alphal = dzero();

  {
    if( side == 'L' || side == 'l' ){
      printf("dtrmm C(%d,%d) += A(%d,%d) * B(%d,%d)\n",
	     nrowC, ncolC, nrowA, ncolA, nrowB, ncolB );
    }
    else{
      printf("dtrmm C(%d,%d) += B(%d,%d) * A(%d,%d)\n",
	     nrowC, ncolC, nrowB, ncolB, nrowA, ncolA );
    }


/*    printf("A:\n");*/
/*    for (j = 0; j < nrowA; j++) {*/
/*    for (i = 0; i < ncolA; i++) {*/
/*            printf("%f",Ah[j+i*lda]);*/
/*    }*/
/*    printf("\n");*/
/*    }*/


/*    printf("C:\n");*/
/*    for (j = 0; j < nrowC; j++) {*/
/*    for (i = 0; i < ncolC; i++) {*/
/*            printf("%f",Ch[j+i*ldc]);*/
/*    }*/
/*    printf("\n");*/
/*    }*/


#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddtrmm_host_multigpu_
      ( &side, &uplo, &trans, &diag, &m, &n, &alphah, &alphal,
	Ah, Al, &lda, Ch, Cl, &ldc );
#else
    printf("calling single-gpu function\n");
    gddtrmm_host_
      ( &side, &uplo, &trans, &diag, &m, &n, &alphah, &alphal,
	Ah, Al, &lda, Ch, Cl, &ldc );



#endif


/*    printf("C:\n");*/
/*    for (j = 0; j < nrowC; j++) {*/
/*    for (i = 0; i < ncolC; i++) {*/
/*            printf("%f",Ch[j+i*ldc]);*/
/*    }*/
/*    printf("\n");*/
/*    }*/



/*    printf("D:\n");*/
/*    for (j = 0; j < nrowC; j++) {*/
/*    for (i = 0; i < ncolC; i++) {*/
/*            printf("%f",Dh[j+i*ldc]);*/
/*    }*/
/*    printf("\n");*/
/*    }*/



    printf("calling single-cpu function\n");
    dtrmm_
      ( &side, &uplo, &trans, &diag, &m, &n, &alphah,
	Ah, &lda, Dh, &ldc );



/*    printf("D:\n");*/
/*    for (j = 0; j < nrowC; j++) {*/
/*    for (i = 0; i < ncolC; i++) {*/
/*            printf("%f",Dh[j+i*lda]);*/
/*    }*/
/*    printf("\n");*/
/*    }*/



    printf("side:%c,uplo:%c,trans:%c,diag:%c\n",side,uplo,trans,diag);
    dcheck(Ch,Dh,ncolC,ldc);
  }

  /*
  doutput(Ah, nrowA, lda );
  doutput(Bh, nrowB, ldb );
  doutput(Ch, nrowC, ldc );
  doutput(Dh, nrowC, ldc );
  */

  cudaFreeHost( Ah );
  cudaFreeHost( Al );
  cudaFreeHost( Bh );
  cudaFreeHost( Bl );
  cudaFreeHost( Ch );
  cudaFreeHost( Cl );
  cudaFreeHost( Dh );

  return 0;
}

