#include <stdio.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"
#include <stdlib.h>

int
main( int argc, char const* argv[] )
{
    srand( (unsigned)time( NULL ) );
  int MAX_DIM = 256;
    int      i, j;
    int      n = rand( )%MAX_DIM+1; /* nrowC and nrow op(A) */

    int      lda = n;

    char     uplo  = getRandomChar( "uUlL", 4 ); /* 'L' or 'U' */
    char     trans = getRandomChar( "nNtT", 4 ); /* 'N' or 'T' */
    char     diag  = getRandomChar( "uUnN", 4 ); /* 'U' or 'N' */

    double * Ah, *Al;
    int      incx = rand( )%MAX_DIM+1; /* ncolC and ncol op(B) */
    int      sizex;

    sizex = incx*n;

    int      sizeA = lda*n;

    if(
        cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    for( i = 0; i < n; i++ )
    {
        for( j = 0; j < n; j++ )
        {
            Ah[i+j*lda] = drand();
            Al[i+j*lda] = 0.0;
        }
    }


/*    int x, y;*/
/*    for( x = 0; x < n; x++ )*/
/*    {*/
/*        for( y = 0; y < n; y++ )*/
/*        {*/
/*            printf( "%f,", Ah[x+y*lda] );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/

/*    printf( "****************************\n" );*/


    double * xh, *xl;
    double * xh0, *xl0;


    if(
        cudaSuccess != cudaMallocHost( (void **)&xh, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&xl, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&xh0, sizex*sizeof( double ) )
      || cudaSuccess != cudaMallocHost( (void **)&xl0, sizex*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    for( i = 0; i < sizex; i++ )
    {
        xh[i] = drand();
        xl[i] = 0.0;
        xh0[i] = xh[i];
        xl0[i] = xl[i];
    }

/*    for( i = 0; i < sizex; i++ )*/
/*    {*/
/*        printf( "%e,", xh[i] );*/
/*    }*/
/*    printf( "\n\n" );*/


/*    gddtrmv_host_(&uplo,&trans,&diag, &n, Ah,  Al,  &lda, xh,*/
/*        xl,  &incx);*/

#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddtrmv_host_multigpu_( &uplo, &trans, &diag, &n, Ah,  Al,  &lda, xh,
                            xl,  &incx );
#else
    printf("calling single-gpu function\n");
    gddtrmv_host_( &uplo, &trans, &diag, &n, Ah,  Al,  &lda, xh,
                            xl,  &incx );
#endif
    printf("calling single-cpu function\n");
    printf("uplo:%c,trans:%c,diag:%c,n:%d,incx:%d\n",uplo,trans,diag,n,incx);
    dtrmv_( &uplo, &trans, &diag, &n, Ah,  &lda, xh0,
            &incx );

    dcheck( xh, xh0, sizex, 1 );

    cudaFreeHost( Ah );
    cudaFreeHost( Al );
    cudaFreeHost( xh );
    cudaFreeHost( xl );

    return 0;
}

