#include <stdio.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include <stdlib.h>
#include "common_test.h"

int main( int argc, char const* argv[] )
{
	srand((unsigned)time(NULL));
  // A(nrowA,ncolA), B(nrowB,ncolB), C(nrowC,ncolC)
  // lda>=nrowA ldb>=nrowB, ldc>=nrowC

  // dimmension of the operation
  int MAX_DIM = 256;
  int m = rand()%MAX_DIM+5; // nrowC and nrow op(A)
/*  int n = rand()%MAX_DIM+5; // ncolC and ncol op(B)*/
/*  int m = 32; // nrowC and nrow op(A)*/
/*  int n = 23; // ncolC and ncol op(B)*/
/*  int m = 20; // nrowC and nrow op(A)*/
  int n = 1; // ncolC and ncol op(B)
/*  int k = rand()%MAX_DIM; // ncol op(A) and nrow op(B)*/


  char side  = 'l'; // 'L' or 'R'
/*  char uplo  = 'L'; // 'L' or 'U'*/
/*  char trans = 'T'; // 'N' or 'T'*/
  char diag  = 'N'; // 'U' or 'N'
/*  char side  = getRandomChar("lLrR",4); // 'L' or 'R'*/
  char uplo  = getRandomChar("lLuU",4); // 'L' or 'U'
  char trans = getRandomChar("nNtT",4); // 'N' or 'T'
/*  char diag  = getRandomChar("uUnN",4); // 'U' or 'N'*/

  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;
  int lda, ldb, ldc;
  int sizeA, sizeB, sizeC;

  /*
    A(M,M)*B(M,N), A is symmetric, B is general
   */
  if( side == 'L' || side == 'l' ){
    nrowA = m, ncolA = m;
    nrowB = m, ncolB = n;
    nrowC = m, ncolC = n;
  }
  /*
    B(M,N)*A(N,N), A is symmetric, B is general
   */
  else{
    nrowB = m, ncolB = n;
    nrowA = n, ncolA = n;
    nrowC = m, ncolC = n;
  }
  lda = nrowA;
  ldb = nrowB;
  ldc = nrowC;

  double *Ah, *Al, *Bh, *Bl, *Ch, *Cl, *Dh;

  sizeA = lda*ncolA;
  sizeB = ldb*ncolB;
  sizeC = ldc*ncolC;

  if(
     cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Bh, sizeB*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Bl, sizeB*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Ch, sizeC*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Cl, sizeC*sizeof( double ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Dh, sizeC*sizeof( double ) ) ){
    printf("Host Malloc Error\n");
    return -1;
  }

  int i, j;

  // ignore triangular matrix
  for( j=0; j<ncolA; j++ ){
    for( i=0; i<nrowA; i++ ){
/*      Ah[i+j*lda] = drand();*/
      Ah[i+j*lda] = (double) i+j*lda+1;
      Al[i+j*lda] = dzero();
    }
  }

/*    int x, y;*/
/*    for( x = 0; x < nrowA; x++ )*/
/*    {*/
/*        for( y = 0; y < ncolA; y++ )*/
/*        {*/
/*            printf( "%f,", Ah[x+y*lda] );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/

/*    printf( "****************************\n" );*/




  for( j=0; j<ncolB; j++ ){
    for( i=0; i<nrowB; i++ ){
/*      Bh[i+j*ldb] = drand();*/
      Bh[i+j*ldb] = i+j*ldb+1;
      int idx  = i+j*ldb;
      Bl[i+j*ldb] = dzero();
    }
  }

/*    for( x = 0; x < nrowB; x++ )*/
/*    {*/
/*        for( y = 0; y < ncolB; y++ )*/
/*        {*/
/*            printf( "%f,", Bh[x+y*ldb] );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/

/*    printf( "****************************\n" );*/



  // copy B to C and D to backup B
  for( j=0; j<ncolC; j++ ){
    for( i=0; i<nrowC; i++ ){
      Ch[i+j*ldc] = Bh[i+j*ldc];
      Cl[i+j*ldc] = dzero();

      Dh[i+j*ldc] = Bh[i+j*ldc];
    }
  }

/*  double alphah = drand();*/
  double alphah = 1.0;
  double alphal = dzero();

  {
    if( side == 'L' ){
      printf("dtrsm B(%d,%d) = A^(-1)(%d,%d) * B(%d,%d)\n",
	     nrowC, ncolC, nrowA, ncolA, nrowB, ncolB );
    }
    else{
      printf("dtrsm B(%d,%d) = B(%d,%d) * A^(-1)(%d,%d)\n",
	     nrowC, ncolC, nrowB, ncolB, nrowA, ncolA );
    }

#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddtrsm_host_multigpu_
      ( &side, &uplo, &trans, &diag, &m, &n, &alphah, &alphal,
	Ah, Al, &lda, Ch, Cl, &ldc );
#else
    printf("calling single-gpu function\n");
    gddtrsm_host_
      ( &side, &uplo, &trans, &diag, &m, &n, &alphah, &alphal,
	Ah, Al, &lda, Ch, Cl, &ldc );
#endif
    printf("calling single-cpu function\n");
    dtrsm_
      ( &side, &uplo, &trans, &diag, &m, &n, &alphah,
	Ah, &lda, Dh, &ldc );

  printf("side:%c,uplo:%c,trans:%c,diag:%c,m:%d,n:%d\n",side,uplo,trans,diag,m,n);
    dcheck(Ch,Dh,ncolC,nrowC);

/*    for( x = 0; x < nrowB; x++ )*/
/*    {*/
/*        for( y = 0; y < ncolB; y++ )*/
/*        {*/
/*            printf( "%e,", Ch[x+y*ldb] );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/
/*    printf( "****************************\n" );*/



/*    for( x = 0; x < nrowB; x++ )*/
/*    {*/
/*        for( y = 0; y < ncolB; y++ )*/
/*        {*/
/*            printf( "%e,", Dh[x+y*ldb] );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/
/*    printf( "****************************\n" );*/


  }


  /*
  doutput(Ah, nrowA, lda );
  doutput(Bh, nrowB, ldb );
  doutput(Ch, nrowC, ldc );
  doutput(Dh, nrowC, ldc );
  */

  cudaFreeHost( Ah );
  cudaFreeHost( Al );
  cudaFreeHost( Bh );
  cudaFreeHost( Bl );
  cudaFreeHost( Ch );
  cudaFreeHost( Cl );
  cudaFreeHost( Dh );

  return 0;
}

