#include <stdio.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"
#include <time.h>

int
main( int argc, char const* argv[] )
{
	srand((unsigned)time(NULL));
    int  i, j, k;
   
    int  MAX_DIM = 256;
    int  m = rand( )%MAX_DIM+1;
/*    int  m = 10;*/
    int  n = m;

    char   uplo = getRandomChar( "uUlL", 4 ); /* 'U' or 'L' */
    int  kl, ku;
    int kb = rand()%m;
/*        int kb = 4;*/

    {
        if( uplo == 'U' || uplo == 'u' )
        {
            ku = kb;
            kl = 0;
        }
        else
        {
            ku = 0;
            kl = kb;
        }
    }

    int      lda = n;

    cuDoubleComplex * Ah, *Al, *bAh, *bAl;
    int      incx = rand()%MAX_DIM+1;
    int      incy = rand()%MAX_DIM+1;
/*    int      incx = 1;*/
/*    int      incy = 1;*/

    int      sizex = incx*n;
    int      sizey = incy*n;
    int      sizeA = lda*n;

    if(
        cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof( cuDoubleComplex ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    int ldba = ( kl+ku+1 );
    int sizebA = ldba*n;

    if(
        cudaSuccess != cudaMallocHost( (void **)&bAh, sizebA*sizeof( cuDoubleComplex ) )
      || cudaSuccess !=
        cudaMallocHost( (void **)&bAl, sizebA*sizeof( cuDoubleComplex ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    for( i = 0; i < n; i++ )
    {
        for( j = 0; j < n; j++ )
        {
            Ah[i+j*lda] = zrand();
            Al[i+j*lda] = zzero();
        }
    }


/*    int x, y;*/
/*    for( x = 0; x < n; x++ )*/
/*    {*/
/*        for( y = 0; y < n; y++ )*/
/*        {*/
/*            printf( "%f,", Ah[x+y*lda] );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/

/*    printf( "****************************\n" );*/


    zcopyToBandMatrix( n, n, lda, kl, ku, Ah, Al, bAh, bAl );

/*    for( x = 0; x < kl+ku+1; x++ )*/
/*    {*/
/*        for( y = 0; y < n; y++ )*/
/*        {*/
/*            printf( "%f,", bAh[x+y*( kl+ku+1 )] );*/
/*        }*/
/*        printf( "\n" );*/
/*    }*/

/*    printf( "****************************\n" );*/


    cuDoubleComplex * xh, *xl;


    if(
        cudaSuccess != cudaMallocHost( (void **)&xh, sizex*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&xl, sizex*sizeof( cuDoubleComplex ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    for( i = 0; i < sizex; i++ )
    {
        xh[i] = zrand() ;
        xl[i] = zzero() ;
    }

/*    for( i = 0; i < sizex; i++ )*/
/*    {*/
/*        printf( "%e,", xh[i] );*/
/*    }*/
/*    printf( "\n\n" );*/

    cuDoubleComplex * yh, *yl;
    cuDoubleComplex * yh0, *yl0;


    if(
        cudaSuccess != cudaMallocHost( (void **) &yh, sizey*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&yl, sizey*sizeof( cuDoubleComplex ) )
        || cudaSuccess != cudaMallocHost( (void **) &yh0, sizey*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&yl0, sizey*sizeof( cuDoubleComplex ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    for( i = 0; i < sizey; i++ )
    {
        yh[i] = zrand() ;   
        yl[i] = zzero() ;   
	yh0[i] = yh[i];
	yl0[i] = yl[i];
    }

/*    for( i = 0; i < sizey; i++ )*/
/*    {*/
/*        printf( "%e,", yh[i] );*/
/*    }*/
/*    printf( "\n\n" );*/


    cuDoubleComplex alphah = zrand() ;   
    cuDoubleComplex alphal = zzero() ;   
    cuDoubleComplex betah = zrand() ;   
    cuDoubleComplex betal = zzero() ;   


#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddzhbmv_host_multigpu_( &uplo,  &n,  &kb, &alphah,  &alphal, bAh,  bAl,
		  &ldba, xh,
		  xl,  &incx,  &betah, &betal,  yh,  yl,
		  &incy );
#else
    printf("calling single-gpu function\n");
    gddzhbmv_host_( &uplo,  &n,  &kb, &alphah,  &alphal, bAh,  bAl,
		  &ldba, xh,
		  xl,  &incx,  &betah, &betal,  yh,  yl,
		  &incy );
#endif
    printf("calling single-cpu function\n");
    zhbmv_( &uplo,  &n,  &kb, &alphah, bAh,
		  &ldba, xh,
		  &incx,  &betah,  yh0,
		  &incy );

    printf("uplo:%c,k:%d,n:%d,incx:%d,incy:%d\n",uplo,kb,n,incx,incy);
    zcheck(yh,yh0,sizey,1);

    cudaFreeHost( Ah );
    cudaFreeHost( Al );
    cudaFreeHost( xh );
    cudaFreeHost( xl );
    cudaFreeHost( yh );
    cudaFreeHost( yl );

    return 0;
}

