// test for gddsymm

#include <stdio.h>
#include <stdlib.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"

int main( int argc, char const* argv[] )
{
	srand((unsigned)time(NULL));
  // A(nrowA,ncolA), B(nrowB,ncolB), C(nrowC,ncolC)
  // lda>=nrowA ldb>=nrowB, ldc>=nrowC

  // dimmension of the operation
  int MAX_DIM = 256;
  int m = rand()%MAX_DIM+1; // nrowC and nrow op(A)
  int n = rand()%MAX_DIM+1; // ncolC and ncol op(B)
/*  int m = 151; // nrowC and nrow op(A)*/
/*  int n = 67; // ncolC and ncol op(B)*/

/*  char side = 'R'; // 'L' or 'R'*/
/*  char uplo = 'L'; // 'L' or 'U'*/
  char side  = getRandomChar("lLrR",4); // 'L' or 'R'
  char uplo  = getRandomChar("uUlL",4); // 'L' or 'U'

  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;
  int lda, ldb, ldc;
  int sizeA, sizeB, sizeC;

  /*
    A(M,M)*B(M,N), A is hermite, B is general
   */
  if( side == 'L' || side == 'l' ){
    nrowA = m, ncolA = m;
    nrowB = m, ncolB = n;
    nrowC = m, ncolC = n;
  }
  /*
    B(M,N)*A(N,N), A is hermite, B is general
   */
  else{
    nrowB = m, ncolB = n;
    nrowA = n, ncolA = n;
    nrowC = m, ncolC = n;
  }
  lda = nrowA;
  ldb = nrowB;
  ldc = nrowC;

  cuDoubleComplex *Ah, *Al, *Bh, *Bl, *Ch, *Cl, *Dh;

  sizeA = lda*ncolA;
  sizeB = ldb*ncolB;
  sizeC = ldc*ncolC;

  if(
     cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof( cuDoubleComplex ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof( cuDoubleComplex ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Bh, sizeB*sizeof( cuDoubleComplex ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Bl, sizeB*sizeof( cuDoubleComplex ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Ch, sizeC*sizeof( cuDoubleComplex ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Cl, sizeC*sizeof( cuDoubleComplex ) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Dh, sizeC*sizeof( cuDoubleComplex ) ) ){
    printf("Host Malloc Error\n");
    return -1;
  }

  int i, j;

  // ignore hermite
  for( j=0; j<ncolA; j++ ){
    for( i=0; i<nrowA; i++ ){
/*        Ah[i+j*lda] = zrand();*/
	Ah[i+j*lda] = make_cuDoubleComplex((double)i+j*lda+1,i+j*lda+1);
	Al[i+j*lda] = zzero();
    }
  }

  for( j=0; j<ncolB; j++ ){
    for( i=0; i<nrowB; i++ ){
      Bh[i+j*ldb] = make_cuDoubleComplex((double)i,(double)-i);
      Bl[i+j*ldb] = zzero();
    }
  }

  for( j=0; j<ncolC; j++ ){
    for( i=0; i<nrowC; i++ ){
      Ch[i+j*ldc] = make_cuDoubleComplex(1.0,0.0);
      Cl[i+j*ldc] = zzero();

      Dh[i+j*ldc] = Ch[i+j*ldc];
    }
  }


  cuDoubleComplex alphah = zrand();
  cuDoubleComplex alphal = zzero();

  cuDoubleComplex betah = zrand();
  cuDoubleComplex betal = zzero();

  {
    if( side == 'L' || side == 'l' ){
      printf("zhemm C(%d,%d) += A(%d,%d) * B(%d,%d)\n",
	     nrowC, ncolC, nrowA, ncolA, nrowB, ncolB );
      printf("A is hermite, %c side is used.\n", uplo );
    }
    else{
      printf("zhemm C(%d,%d) += B(%d,%d) * A(%d,%d)\n",
	     nrowC, ncolC, nrowB, ncolB, nrowA, ncolA );
      printf("A is hermite, %c side is used.\n", uplo );
    }

/*    printf("A:\n");*/
/*    for (j = 0; j < nrowA; j++) {*/
/*    for (i = 0; i < ncolA; i++) {*/
/*            printf("(%e,%e)",Ah[j+i*lda].x,Ah[j+i*lda].y);*/
/*    }*/
/*    printf("\n");*/
/*    }*/

/*    printf("B:\n");*/
/*    for (j = 0; j < nrowB; j++) {*/
/*    for (i = 0; i < ncolB; i++) {*/
/*            printf("(%e,%e)",Bh[j+i*lda].x,Bh[j+i*lda].y);*/
/*    }*/
/*    printf("\n");*/
/*    }*/


/*    printf("C:\n");*/
/*    for (j = 0; j < nrowC; j++) {*/
/*    for (i = 0; i < ncolC; i++) {*/
/*            printf("(%e,%e)",Ch[j+i*lda].x,Ch[j+i*lda].y);*/
/*    }*/
/*    printf("\n");*/
/*    }*/





#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddzhemm_host_multigpu_
      ( &side, &uplo, &m, &n, &alphah, &alphal, Ah, Al, &lda,
	Bh, Bl, &ldb, &betah, &betal, Ch, Cl, &ldc );
#else
    printf("calling single-gpu function\n");
    gddzhemm_host_
      ( &side, &uplo, &m, &n, &alphah, &alphal, Ah, Al, &lda,
	Bh, Bl, &ldb, &betah, &betal, Ch, Cl, &ldc );
#endif


/*    printf("C:\n");*/
/*    for (j = 0; j < nrowC; j++) {*/
/*    for (i = 0; i < ncolC; i++) {*/
/*            printf("(%e,%e)",Ch[j+i*lda].x,Ch[j+i*lda].y);*/
/*    }*/
/*    printf("\n");*/
/*    }*/




    printf("calling single-cpu function\n");
    zhemm_
      ( &side, &uplo, &m, &n, &alphah, Ah, &lda,
	Bh, &ldb, &betah, Dh, &ldc );



/*    printf("D:\n");*/
/*    for (j = 0; j < nrowC; j++) {*/
/*    for (i = 0; i < ncolC; i++) {*/
/*            printf("(%e,%e)",Dh[j+i*lda].x,Dh[j+i*lda].y);*/
/*    }*/
/*    printf("\n");*/
/*    }*/



    printf("side:%c,uplo:%c,m:%d,n,%d\n",side,uplo,m,n);
    zcheck(Ch,Dh,ncolC,ldc);
  }

  /*
  zoutput(Ah, nrowA, lda );
  zoutput(Bh, nrowB, ldb );
  zoutput(Ch, nrowC, ldc );
  zoutput(Dh, nrowC, ldc );
  */

  cudaFreeHost( Ah );
  cudaFreeHost( Al );
  cudaFreeHost( Bh );
  cudaFreeHost( Bl );
  cudaFreeHost( Ch );
  cudaFreeHost( Cl );
  cudaFreeHost( Dh );

  return 0;
}

