#include <stdio.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"
#include <time.h>
int
main( int argc, char const* argv[] )
{
    srand( (unsigned)time( NULL ) );
    int      MAX_DIM = 256;
    int      n = rand( )%MAX_DIM+1; /* ncolC and ncol op(B) */

    cuDoubleComplex * xh, *xl, *yh, *yl, *Ah, *Al, *Bh, *Bl;
    int      incx = rand( )%MAX_DIM+1;
    int      incy = rand( )%MAX_DIM+1;

    int      lda = n;

  char uplo = getRandomChar("LlUu",4); // 'L' or 'U'
/*  char uplo = 'U'; // 'L' or 'U'*/

    int      sizeA =  lda*n;
    int      sizex = incx*n;
    int      sizey = incy*n;


    if(
        cudaSuccess != cudaMallocHost( (void **)&xh,  sizex*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&xl, sizex*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&yh, sizey*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&yl, sizey*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&Bh, sizeA*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&Bl, sizeA*sizeof( cuDoubleComplex ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }


    int i;
    for( i = 0; i < sizex; i++ )
    {
        xh[i] = zrand();
        xl[i] = zzero();
    }
    for( i = 0; i < sizey; i++ )
    {
        yh[i] = zrand();
        yl[i] = zzero();
    }

    for( i = 0; i < sizeA; i++ )
    {
        Ah[i] = zrand();
        Al[i] = zzero();
        Bh[i] = Ah[i];
        Bl[i] = Al[i];
    }


    cuDoubleComplex alphah = zrand();
    cuDoubleComplex alphal = zzero();

#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddzher2_host_multigpu_( &uplo, &n, &alphah, &alphal, xh, xl, &incx, yh, yl, &incy, Ah,
		 Al,&lda );
#else
    printf("calling single-gpu function\n");
    gddzher2_host_( &uplo, &n, &alphah, &alphal, xh, xl, &incx, yh, yl, &incy, Ah,
		 Al,&lda );
#endif
    printf("calling single-cpu function\n");
    zher2_( &uplo, &n, &alphah, xh, &incx, yh, &incy, Bh,&lda );

    printf("uplo:%c,n:%d,incx:%d,incy:%d\n",uplo,n,incx,incy);
    zcheck(Ah,Bh,lda,n);


    cudaFreeHost( xh );
    cudaFreeHost( xl );
    cudaFreeHost( yh );
    cudaFreeHost( yl );
    cudaFreeHost( Ah );
    cudaFreeHost( Al );
    cudaFreeHost( Bh );
    cudaFreeHost( Bl );

    return 0;
}

