#include <stdio.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"
#include <time.h>

int
main( int argc, char const* argv[] )
{
    srand( (unsigned)time( NULL ) );
    int      MAX_DIM = 256;
    int      n = rand( )%MAX_DIM+1; /* ncolC and ncol op(B) */

    cuDoubleComplex * xh, *xl, *Ah, *Al,*Bh,*Bl;
    int     incx = rand( )%MAX_DIM+1; /* ncolC and ncol op(B) */

    int      lda = n;

  char uplo = getRandomChar("LlUu",4); // 'L' or 'U'

    int      sizeA =  lda*n;
    int      sizex = incx*n;


    if(
        cudaSuccess != cudaMallocHost( (void **)&xh,  sizex*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&xl, sizex*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&Bh, sizeA*sizeof( cuDoubleComplex ) )
      || cudaSuccess != cudaMallocHost( (void **)&Bl, sizeA*sizeof( cuDoubleComplex ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    int i;
    for( i = 0; i < sizex; i++ )
    {
        xh[i] = zrand();
        xl[i] = zzero();
    }

    for( i = 0; i < sizeA; i++ )
    {
        Ah[i] = zrand();
        Al[i] = zzero();
        Bh[i] = Ah[i];
    }


/*    for( i = 0; i < sizex; i++ )*/
/*    {*/
/*        printf( "x::%d,(%e,%e),(%e,%e)\n", i, xh[i].x,xh[i].y, xl[i].x,xl[i].y );*/
/*    }*/
/*    printf( "###############\n" );*/

    double alphah = drand();
    double alphal = dzero();


#ifdef MULTI
    printf("calling multi-gpu function\n");
    gddzher_host_multigpu_( &uplo, &n, &alphah, &alphal, xh, xl, &incx, Ah,Al,&lda );
#else
    printf("calling single-gpu function\n");
    gddzher_host_( &uplo,&n, &alphah, &alphal, xh, xl, &incx,  Ah,Al,&lda );
#endif
    printf("calling single-cpu function\n");
    zher_( &uplo,&n, &alphah, xh, &incx, Bh,&lda );

    printf("uplo:%c,n:%d,incx:%d\n",uplo,n,incx);
    zcheck(Ah,Bh,lda,n);


    cudaFreeHost( xh );
    cudaFreeHost( xl );
    cudaFreeHost( Ah );
    cudaFreeHost( Al );

    return 0;
}

