// test for gddsyrk

#include <stdio.h>
#include <stdlib.h>
#include "gddcommon.h"
#include "gddblas_host.h"
#include "common_test.h"

int main( int argc, char const* argv[] )
{
	srand((unsigned)time(NULL));
  // A(nrowA,ncolA), B(nrowB,ncolB), C(nrowC,ncolC)
  // lda>=nrowA ldb>=nrowB, ldc>=nrowC

  // dimmension of the operation
  int MAX_DIM = 256;
  int n = rand()%MAX_DIM; // nrow op(A) and nrow op(B)
  int k = rand()%MAX_DIM; // ncol op(A) and ncol op(B)

  char uplo = getRandomChar("uUlL",4); // 'U' or 'L'
  char trans = getRandomChar("nNtT",4); // 'N' or 'T'

  int nrowA, ncolA;
  int nrowC, ncolC;
  int lda, ldb, ldc;
  int sizeA, sizeC;

  /*
    C(N,N) += A(N,K) * A(N,K)^T
   */
  if( trans == 'N' || trans == 'n' ){
    nrowA = n, ncolA = k;
    nrowC = n, ncolC = n;
  }
  /*
    C(N,N) += A(K,N)^T * A(K,N)
   */
  else{
    nrowA = k, ncolA = n;
    nrowC = n, ncolC = n;
  }

  lda = nrowA;
  ldc = nrowC;

  sizeA = lda*ncolA;
  sizeC = ldc*ncolC;

  cuDoubleComplex *Ah, *Al, *Ch, *Cl, *Dh;

  if(
     cudaSuccess != cudaMallocHost( (void **)&Ah, sizeA*sizeof(cuDoubleComplex) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Al, sizeA*sizeof(cuDoubleComplex) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Ch, sizeC*sizeof(cuDoubleComplex) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Cl, sizeC*sizeof(cuDoubleComplex) ) ||
     cudaSuccess != cudaMallocHost( (void **)&Dh, sizeC*sizeof(cuDoubleComplex) ) ){
    printf("Host Malloc Error\n");
    return -1;
  }

  int i, j;

  // ignore symmetricity
  for( j=0; j<ncolA; j++ ){
    for( i=0; i<nrowA; i++ ){
      Ah[i+j*lda] = zrand();
      Al[i+j*lda] = zzero();
    }
  }

  for( j=0; j<ncolC; j++ ){
    for( i=0; i<nrowC; i++ ){
      Ch[i+j*ldc] = zrand();
      Cl[i+j*ldc] = zzero();

      Dh[i+j*ldc] = Ch[i+j*ldc];
    }
  }

  cuDoubleComplex alphah = zrand();
  cuDoubleComplex alphal = zzero();

  cuDoubleComplex betah = zrand();
  cuDoubleComplex betal = zzero();

  {
    if( trans == 'N' ){
      printf("zsyrk C(%d,%d) += A(%d,%d) * A(%d,%d)^T\n",
	     nrowC, ncolC, nrowA, ncolA, nrowA, ncolA );
    }
    else{
      printf("zsyrk C(%d,%d) += A(%d,%d)^T * A(%d,%d)\n",
	     nrowC, ncolC, nrowA, ncolA, nrowA, ncolA );
    }

    printf("calling single-gpu function\n");
    gddzsyrk_host_
      ( &uplo, &trans, &n, &k, &alphah, &alphal,
	Ah, Al, &lda, &betah, &betal, Ch, Cl, &ldc );

    printf("calling single-cpu function\n");
    zsyrk_
      ( &uplo, &trans, &n, &k, &alphah,
	Ah, &lda, &betah, Dh, &ldc );

    zcheck(Ch,Dh,nrowC,ldc);
  }

  /*
  zoutput(Ah, nrowA, lda );
  zoutput(Ch, nrowC, ldc );
  zoutput(Dh, nrowC, ldc );
  */

  cudaFreeHost( Ah );
  cudaFreeHost( Al );
  cudaFreeHost( Ch );
  cudaFreeHost( Cl );
  cudaFreeHost( Dh );

  return 0;
}

