/* AXPY */
#define BLOCKSIZE 64
#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"

__global__ void
gddaxpy_kernel( int n, double *alphah, double *alphal, double *Xh,
                double *Xl, int incx,
                double *Yh, double *Yl,
                int incy )
{
    int x = threadIdx.x+__mul24( BLOCKSIZE, blockIdx.x );
    int offset1 = __mul24( x, incx );
    int offset2 = __mul24( x, incy );

    if( x < n )
    {

        double lyh = *( Yh+offset2 );
        double lyl = *( Yl+offset2 );
        double lxh = *( Xh+offset1 );
        double lxl = *( Xl+offset1 );
        /* y = a*y */

        RMUL( alphah, alphal, &lxh, &lxl, &lxh, &lxl );
        RADD( &lyh, &lyl, &lxh, &lxl, &lyh, &lyl );
        *( Yh+offset2 ) = lyh;
        *( Yl+offset2 ) = lyl;
    }
}

extern "C" void
gddaxpy( int *n, double *alphah, double *alphal, double *Xh,
              double *Xl, int* incx,
              double *Yh, double *Yl,
              int* incy )
{
/*     int num_gpus;*/
/*    CUDA_SAFE_CALL( cudaGetDeviceCount( &num_gpus ) );*/
/*    printf("numgpu = %d\n",num_gpus);*/


    dim3 blocks( ( *n+BLOCKSIZE-1 )/BLOCKSIZE, 1 );

    dim3 threads( BLOCKSIZE, 1 );


    gddaxpy_kernel<<< blocks, threads, 0, gddstream>>>
    ( *n,  alphah, alphal, Xh, Xl, *incx, Yh, Yl, *incy );
}
