#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"


#define BLOCKSIZE ( 64 )
#define BUFFERSIZE ( BLOCKSIZE )
#define TRANSBLOCK ( 256 )

extern "C" __global__ void
gddgbmv_kernel_N( int m, int n, int kl, int ku, double *alphah,
                  double*alphal,
                  double *Ah, double *Al, int lda, double * xh, double *xl,
                  int incx,
                  double *betah, double*betal, double*yh, double*yl,
                  int incy )
{
    int idx = threadIdx.x+__mul24( BLOCKSIZE, blockIdx.x );
 

    {
        unsigned int offset = idx+ku;
        Ah += offset;
        Al += offset;
        offset = threadIdx.x*incx;
        xh += offset;
        xl += offset;
    }
    lda -= 1;


    double resulth = 0.0;
    double resultl = 0.0;

    {
        __shared__ double buffh[BUFFERSIZE];
        __shared__ double buffl[BUFFERSIZE];
        int               n1 = ( n/BUFFERSIZE )*BUFFERSIZE;



	int startPosition =  (((int) blockIdx.x)*BUFFERSIZE-kl)/BUFFERSIZE;
	if(startPosition < 0) 
		startPosition = 0;
	startPosition *= BUFFERSIZE;
	Ah += startPosition * lda;
	Al += startPosition * lda;

	int endPosition = (((int) blockIdx.x)+1+(ku-1)/BLOCKSIZE+1)*BLOCKSIZE;
/*        if(threadIdx.x == 0) printf("bid:%d,n1:%d,spos:%d,epos:%d\n",blockIdx.x,n1,startPosition,endPosition);*/
	n1 = min(n1,endPosition);

        


        for( int i = startPosition; i < n1; i += BUFFERSIZE )
        {
            __syncthreads( );
            buffh[threadIdx.x] = xh[i*incx];
            buffl[threadIdx.x] = xl[i*incx];

            __syncthreads( );

#pragma unroll
            for( int j = 0; j < BUFFERSIZE; j++ )
            {
                double ph, pl;
                ph = *Ah;
                pl = *Al;
                int    b =
                    !( ( idx > kl+ i+j ) || ( idx < i+j-ku ) );
                ph = ph*b;
                pl = pl*b;
                RMUL( &ph, &pl, buffh+j, buffl+j, &ph, &pl );
                RADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
                Ah += lda;
                Al += lda;
            }
        }

        __syncthreads( );

        if( n > n1 && n1!=endPosition  )
        {
            buffh[threadIdx.x] = xh[n1*incx];
            buffl[threadIdx.x] = xl[n1*incx];

            __syncthreads( );
            for( int i = 0; i < ( n-n1 ); i++ )
            {
                double ph, pl;
                ph = *Ah;
                pl = *Al;

                int    b =
                    !( ( idx > kl+ i+n1 ) || ( idx < ( i+n1-ku ) ) );
                ph = ph*b;
                pl = pl*b;


                RMUL( &ph, &pl, buffh+i, buffl+i, &ph, &pl );
                RADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
                Ah += lda;
                Al += lda;
            }
        }
    }

    if( idx < m )
    {
        double ph, pl, qh, ql;
        RMUL( betah, betal, yh+idx*incy, yl+idx*incy, &ph, &pl );
        RMUL( alphah, alphal, &resulth, &resultl, &qh, &ql );
        RADD( &ph, &pl, &qh, &ql, yh+idx*incy, yl+idx*incy );
    }
}

extern "C" __global__ void
gddgbmv_kernel_T( int m, int n, int kl, int ku, double *alphah,
                  double*alphal,
                  double *Ah, double *Al, int lda, double * xh, double *xl,
                  int incx,
                  double *betah, double*betal, double*yh, double*yl,
                  int incy )
{
	int tx = threadIdx.x;
	int bid = blockIdx.y;

	__shared__ double sdatah[TRANSBLOCK];
	__shared__ double sdatal[TRANSBLOCK];

	double resh, resl;
	resh = resl = 0.0;
	int n1 = (m/TRANSBLOCK)*TRANSBLOCK;

	int offset = ku -blockIdx.y;
	Ah += offset;
	Al += offset;


	for (int i = 0; i < n1; i+=TRANSBLOCK) 
	{
		double ph = *(Ah+tx+i+lda*bid);
		double pl = *(Al+tx+i+lda*bid);
		int b = !((i+tx >bid + kl) || (i + tx < bid -ku));
		ph = ph * b;
		pl = pl * b;
		RMUL(&ph,&pl,xh+(tx+i)*incx,xl+(tx+i)*incx,&ph,&pl);
		RADD(&resh,&resl,&ph,&pl,&resh,&resl);


	}

	if(tx + n1+offset < m)
	{
		double ph = *(Ah+tx+n1+lda*bid);
		double pl = *(Al+tx+n1+lda*bid);
		int b = !((n1+tx> bid+kl)||(n1+tx < bid - ku));
		ph = ph*b;
		pl = pl*b;
		RMUL(&ph,&pl,xh+(tx+n1)*incx,xl+(tx+n1)*incx,&ph,&pl);
		RADD(&resh,&resl,&ph,&pl,&resh,&resl);
	}

	sdatah[tx] = resh;
	sdatal[tx] = resl;

	__syncthreads();

    if( TRANSBLOCK >= 512 )
    {
        if( tx < 256 )
        {
/*            sdata[tx] += sdata[tx+256];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+256, sdatal+tx+256,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }
    if( TRANSBLOCK >=        256 )
    {
        if( tx < 128 )
        {
/*            sdata[tx] += sdata[tx+128];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+128, sdatal+tx+128,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }
    if( TRANSBLOCK >=        128 )
    {
        if( tx <  64 )
        {
/*            sdata[tx] += sdata[tx+64];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+64, sdatal+tx+64,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }

    if( tx < 32 )
    {
        if( TRANSBLOCK >=  64 )
        {
/*            sdata[tx] += sdata[tx+32];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+32, sdatal+tx+32,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=  32 )
        {
/*                sdata[tx] += sdata[tx+16]; */
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+16, sdatal+tx+16,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=  16 )
        {
/*                sdata[tx] += sdata[tx+8];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+8, sdatal+tx+8,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   8 )
        {
/*                       sdata[tx] += sdata[tx+4];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+4, sdatal+tx+4,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   4 )
        {
/*                       sdata[tx] += sdata[tx+2];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+2, sdatal+tx+2,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   2 )
        {
/*                sdata[tx] += sdata[tx+1]; */
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+1, sdatal+tx+1,
                  sdatah+tx, sdatal+tx );
        }
    }

    if(tx == 0)
    {
/*            y[blockIdx.y*incy] = sdata[0]*(*alpha)+(*beta)*y[blockIdx.y*incy];*/

            double ph, pl, qh, ql;
            RMUL( betah, betal, yh+blockIdx.y*incy, yl+blockIdx.y*incy, &ph, &pl );
            RMUL( alphah, alphal, sdatah, sdatal, &qh, &ql );
            RADD( &ph, &pl, &qh, &ql, yh+blockIdx.y*incy, yl+blockIdx.y*incy );
    }



}

extern "C" void
gddgbmv( char*trans, int *m, int *n, int*kl, int*ku, double *alphah,
              double*alphal,
              double *Ah, double *Al, int *lda, double * xh, double *xl,
              int *incx,
              double *betah, double*betal, double*yh, double*yl,
              int *incy )
{
    int btrans = 1;

    if( *trans == 'N' || *trans == 'n' ) btrans = 0;
    if( !btrans )
    {
        dim3 blocks( ( *m+BLOCKSIZE-1 )/BLOCKSIZE, 1 );


        dim3 threads( BLOCKSIZE, 1 );

        gddgbmv_kernel_N<<< blocks, threads, 0, gddstream>>>
        ( *m, *n, *kl, *ku, alphah, alphal, Ah, Al, *lda,  xh, xl, *incx,
          betah,
          betal, yh, yl, *incy );
    }
    else
    {
        dim3 blocks( 1, *n );

        dim3 threads( TRANSBLOCK, 1 );

        gddgbmv_kernel_T<<< blocks, threads, 0, gddstream>>>
        ( *m, *n, *kl, *ku, alphah, alphal, Ah, Al, *lda,  xh, xl, *incx,
          betah,
          betal, yh, yl, *incy );
    }
}

