#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"


#define BLOCKSIZE ( 64 )
#define BUFFERSIZE ( BLOCKSIZE )
#define TRANSBLOCK (256)

extern "C" __global__ void
gddgemv_kernel_N( int m, int n, double *alphah, double*alphal,
                  double *Ah, double *Al, int lda, double * xh, double *xl,
                  int incx,
                  double *betah, double*betal, double*yh, double*yl,
                  int incy )
{
    unsigned int idx = threadIdx.x+__mul24( BLOCKSIZE, blockIdx.x );


    Ah += idx;
    Al += idx;
    {
        unsigned int offset = threadIdx.x*incx;
        xh += offset;
        xl += offset;
    }

    double resulth = 0.0;
    double resultl = 0.0;

    {
        __shared__ double buffh[BUFFERSIZE];
        __shared__ double buffl[BUFFERSIZE];

        int               n1 = ( n/BUFFERSIZE )*BUFFERSIZE;
        for( int i = 0; i < n1; i += BUFFERSIZE )
        {
            __syncthreads( );
            buffh[threadIdx.x] = xh[i*incx];
            buffl[threadIdx.x] = xl[i*incx];

            __syncthreads( );

#pragma unroll
            for( int j = 0; j < BUFFERSIZE; j++ )
            {
                double ph, pl;
/*                result += *(A) *buff[j];*/
                RMUL( Ah, Al, buffh+j, buffl+j, &ph, &pl );
                RADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
                Ah += lda;
                Al += lda;
            }
        }



        __syncthreads( );

        if( n > n1 )
        {
            buffh[threadIdx.x] = xh[n1*incx];
            buffl[threadIdx.x] = xl[n1*incx];

            __syncthreads( );
            for( int i = 0; i < ( n-n1 ); i++ )
            {
                double ph, pl;
/*                result += *(A)*buff[j];*/
                RMUL( Ah, Al, buffh+i, buffl+i, &ph, &pl );
                RADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
/*                if(idx < m){*/
/*                printf("i:%d,idx:%d,Ah:%e,xh:%e,resulth:%e\n",i,idx,*Ah,*(buffh+i),resulth);*/
/*                }*/
                Ah += lda;
                Al += lda;
            }
        }

        if( idx < m )
        {
            double ph, pl, qh, ql;
            RMUL( betah, betal, yh+idx*incy, yl+idx*incy, &ph, &pl );
            RMUL( alphah, alphal, &resulth, &resultl, &qh, &ql );
            RADD( &ph, &pl, &qh, &ql, yh+idx*incy, yl+idx*incy );
/*                printf("idx:%d,ph:%e,resulth:%e,yh:%e,qh:%e\n",idx,ph,resulth,*(yh+idx*incy),qh);*/
        }
    }
}


extern "C" __global__ void
gddgemv_kernel_T(int m, int n, double *alphah, double*alphal,
                  double *Ah, double *Al, int lda, double * xh, double *xl,
                  int incx,
                  double *betah, double*betal, double*yh, double*yl,
                  int incy)
{
	unsigned int tx = threadIdx.x;

	__shared__ double sdatah[TRANSBLOCK];
	__shared__ double sdatal[TRANSBLOCK];

	double  resh,resl;
	resh = resl = 0.0;
	unsigned int n1 = (m/TRANSBLOCK)*TRANSBLOCK;

	for (int i = 0; i <n1 ; i+=TRANSBLOCK) 
	{
		double ph,pl;

/*                res += Ah[tx + i + lda*blockIdx.y] * x[tx+i];*/
		RMUL(Ah+tx+i+lda*blockIdx.y,Al+tx+i+lda*blockIdx.y,
				xh+(tx+i)*incx,xl+(tx+i)*incx,
				&ph,&pl);
		RADD(&resh,&resl,&ph,&pl,&resh,&resl);
	}

	if(tx + n1 < m)
	{
		double ph,pl;

/*                resh +=  A[tx + n1 + lda*blockIdx.y] * x[tx + n1];*/
		RMUL(Ah+tx+n1+lda*blockIdx.y,Al+tx+n1+lda*blockIdx.y,
				xh+(tx+n1)*incx,xl+(tx+n1)*incx,
				&ph,&pl);
		RADD(&resh,&resl,&ph,&pl,&resh,&resl);
	}

	sdatah[tx] = resh;
	sdatal[tx] = resl;

	__syncthreads();

    if( TRANSBLOCK >= 512 )
    {
        if( tx < 256 )
        {
/*            sdata[tx] += sdata[tx+256];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+256, sdatal+tx+256,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }
    if( TRANSBLOCK >=        256 )
    {
        if( tx < 128 )
        {
/*            sdata[tx] += sdata[tx+128];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+128, sdatal+tx+128,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }
    if( TRANSBLOCK >=        128 )
    {
        if( tx <  64 )
        {
/*            sdata[tx] += sdata[tx+64];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+64, sdatal+tx+64,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }

    if( tx < 32 )
    {
        if( TRANSBLOCK >=  64 )
        {
/*            sdata[tx] += sdata[tx+32];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+32, sdatal+tx+32,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=  32 )
        {
/*                sdata[tx] += sdata[tx+16]; */
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+16, sdatal+tx+16,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=  16 )
        {
/*                sdata[tx] += sdata[tx+8];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+8, sdatal+tx+8,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   8 )
        {
/*                       sdata[tx] += sdata[tx+4];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+4, sdatal+tx+4,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   4 )
        {
/*                       sdata[tx] += sdata[tx+2];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+2, sdatal+tx+2,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   2 )
        {
/*                sdata[tx] += sdata[tx+1]; */
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+1, sdatal+tx+1,
                  sdatah+tx, sdatal+tx );
        }
    }

    if(tx == 0)
    {
/*            y[blockIdx.y*incy] = sdata[0]*(*alpha)+(*beta)*y[blockIdx.y*incy];*/

            double ph, pl, qh, ql;
            RMUL( betah, betal, yh+blockIdx.y*incy, yl+blockIdx.y*incy, &ph, &pl );
            RMUL( alphah, alphal, sdatah, sdatal, &qh, &ql );
            RADD( &ph, &pl, &qh, &ql, yh+blockIdx.y*incy, yl+blockIdx.y*incy );
    }


}


extern "C" void
gddgemv( char*trans, int *m, int *n, double *alphah, double*alphal,
              double *Ah, double *Al, int *lda, double * xh, double *xl,
              int *incx,
              double *betah, double*betal, double*yh, double*yl,
              int *incy )
{
    int btrans = 1;

    if( *trans == 'N' || *trans == 'n' ) btrans = 0;
    if( !btrans )
    {
        dim3 blocks( ( *m+BLOCKSIZE-1 )/BLOCKSIZE, 1 );
        dim3 threads( BLOCKSIZE, 1 );

        gddgemv_kernel_N<<< blocks, threads, 0, gddstream>>>
        ( *m, *n, alphah, alphal, Ah, Al, *lda,  xh, xl, *incx, betah,
          betal, yh, yl, *incy );
    }
    else
    {
        dim3 blocks( 1,*n );
        dim3 threads( TRANSBLOCK, 1 );
        gddgemv_kernel_T<<< blocks, threads, 0, gddstream>>>
        ( *m, *n, alphah, alphal, Ah, Al, *lda,  xh, xl, *incx, betah,
          betal, yh, yl, *incy );
    }
}

