#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"


#define BLOCKSIZE ( 64 )

extern "C" __global__ void
gddger_kernel( int m, int n, double *alphah, double*alphal,
               double * xh, double *xl,
               int incx,
               double*yh, double*yl,
               int incy,
               double* Ah, double *Al, int lda )
{
    unsigned int tid = threadIdx.x;
    unsigned int idx = tid+__mul24( BLOCKSIZE, blockIdx.x );


    Ah += idx;
    Al += idx;


    {
        int offset = threadIdx.x*incy;
        yh += offset;
        yl += offset;
    }


    {
        __shared__ double xhs[BLOCKSIZE];
        __shared__ double xls[BLOCKSIZE];
        __shared__ double yhs[BLOCKSIZE];
        __shared__ double yls[BLOCKSIZE];

        xhs[tid] = xh[idx*incx];
        xls[tid] = xl[idx*incx];


        int               n1 = ( n/BLOCKSIZE )*BLOCKSIZE;

        double            localalphah = *alphah;
        double            localalphal = *alphal;

        for( int i = 0; i < n1; i += BLOCKSIZE )
        {
            __syncthreads( );
            yhs[tid] = yh[i*incy];
            yls[tid] = yl[i*incy];

            __syncthreads( );

#pragma unroll
            for( int j = 0; j < BLOCKSIZE; j++ )
            {
                double ph, pl, lAh, lAl;

                lAh = *Ah;
                lAl = *Al;

                RMUL( xhs+tid, xls+tid, yhs+j, yls+j, &ph, &pl );
                RMUL( &ph, &pl, &localalphah, &localalphal, &ph, &pl );
                RADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );

                if( idx < m )
                {
                    *Ah = lAh;
                    *Al = lAl;

                }

                Ah += lda;
                Al += lda;
            }
        }


        __syncthreads( );

        if( n > n1 )
        {
            yhs[threadIdx.x] = yh[n1*incy];
            yls[threadIdx.x] = yl[n1*incy];

            __syncthreads( );
            for( int j = 0; j < ( n-n1 ); j++ )
            {
                double ph, pl, lAh, lAl;

                lAh = *Ah;
                lAl = *Al;

                RMUL( xhs+tid, xls+tid, yhs+j, yls+j, &ph, &pl );
                RMUL( &ph, &pl, &localalphah, &localalphal, &ph, &pl );
                RADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );


                if( idx < m )
                {
                    *Ah = lAh;
                    *Al = lAl;
                }

                Ah += lda;
                Al += lda;
            }
        }
    }
}

extern "C" void
gddger( int* m, int* n, double *alphah, double*alphal,
               double * xh, double *xl,
               int* incx,
               double*yh, double*yl,
               int* incy,
               double* Ah, double *Al, int* lda  )
{
        dim3 blocks( ( *m+BLOCKSIZE-1 )/BLOCKSIZE, 1 );

        dim3 threads( BLOCKSIZE, 1 );

        gddger_kernel<<< blocks, threads, 0, gddstream>>>
        ( *m, *n, alphah, alphal,  xh, xl, *incx,
          yh, yl, *incy , Ah, Al, *lda);
}

