#define BLOCKSIZE ( 2 )
#define TILEWIDTH ( 2*BLOCKSIZE )
#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "basic_algebra.h"
#include "cutil.h"


__global__ void
gddidamax_kernel( int n, double *Xh, double *Xl, int incx, double*dworkh,
                  double*dworkl ,int *dworkidx)
{
    __shared__ double sdatah[BLOCKSIZE];
    __shared__ double sdatal[BLOCKSIZE];
    __shared__ int    idx[BLOCKSIZE];
    unsigned int      tid = threadIdx.x;
    unsigned int      i = blockIdx.x*( BLOCKSIZE*2 )+tid;
    unsigned int      gridSize = BLOCKSIZE*2*gridDim.x;


    sdatah[tid] = 0.0;
    sdatal[tid] = 0.0;
    idx[tid] = tid;

    while( i < n )
    {
        double tmph, tmpl;
        tmph = *( Xh+i*incx );
        tmpl = *( Xl+i*incx );

        gdd_abs( &tmph, &tmpl );
        {
            bool flag;
            gdd_1stvalue_is_larger( sdatah+tid, sdatal+tid, &tmph, &tmpl,&flag );
            if( flag == false ) *( idx+tid ) = i;
        }


        if( i+BLOCKSIZE < n )
        {
            tmph = *( Xh+( i+BLOCKSIZE )*incx );
            tmpl = *( Xl+( i+BLOCKSIZE )*incx );

            gdd_abs( &tmph, &tmpl );
	    {
		    bool flag;
		    gdd_1stvalue_is_larger(sdatah+tid,sdatal+tid, &tmph, &tmpl,&flag );
            if( flag == false ) *( idx+tid ) = i+BLOCKSIZE;
	    }
        }

        i += gridSize;
    }

    __syncthreads( );

    if( BLOCKSIZE >= 512 )
    {
        if( tid < 256 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+256, sdatal+tid+256,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 256);
        }
        __syncthreads( );
    }
    if( BLOCKSIZE >= 256 )
    {
        if( tid < 128 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+128, sdatal+tid+128,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 128);
        }
        __syncthreads( );
    }
    if( BLOCKSIZE >= 128 )
    {
        if( tid <  64 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+64, sdatal+tid+64,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 64);
        }
        __syncthreads( );
    }

    if( tid < 32 )
    {
        if( BLOCKSIZE >=  64 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+32, sdatal+tid+32,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 32);
        }
        if( BLOCKSIZE >=  32 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+16, sdatal+tid+16,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 16);
        }
        if( BLOCKSIZE >=  16 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+8, sdatal+tid+8,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 8);
        }
        if( BLOCKSIZE >=   8 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+4, sdatal+tid+4,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 4);
        }
        if( BLOCKSIZE >=   4 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+2, sdatal+tid+2,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 2);
        }
        if( BLOCKSIZE >=   2 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+1, sdatal+tid+1,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 1);
        }
    }


    if( tid == 0 )
    {
        dworkh[blockIdx.x] = sdatah[0];
        dworkl[blockIdx.x] = sdatal[0];
	dworkidx[blockIdx.x] = idx[0];

    }
}



__global__ void
gddidamax_update_kernel( int n,  double*dworkh,double*dworkl,int *dworkidx,double*dwork2h,double*dwork2l,int*dwork2idx)
{
    __shared__ double sdatah[BLOCKSIZE];
    __shared__ double sdatal[BLOCKSIZE];
    __shared__ int    idx[BLOCKSIZE];
    unsigned int      tid = threadIdx.x;
    unsigned int      i = blockIdx.x*( BLOCKSIZE*2 )+tid;
    unsigned int      gridSize = BLOCKSIZE*2*gridDim.x;


    sdatah[tid] = 0.0;
    sdatal[tid] = 0.0;
    idx[tid] = 0;

    while( i < n )
    {
        double tmph, tmpl;
        tmph = *( dworkh+i );
        tmpl = *( dworkl+i );

        {
            bool flag;
            gdd_1stvalue_is_larger( sdatah+tid, sdatal+tid, &tmph, &tmpl,&flag );
            if( flag == false ) *( idx+tid ) = *(dworkidx+i);
        }

        if( i+BLOCKSIZE < n )
        {
            tmph = *( dworkh+ i+BLOCKSIZE  );
            tmpl = *( dworkl+ i+BLOCKSIZE  );

	    {
		    bool flag;
		    gdd_1stvalue_is_larger(sdatah+tid,sdatal+tid, &tmph, &tmpl,&flag );
            if( flag == false ) *( idx+tid ) = *(dworkidx+i+BLOCKSIZE);
	    }
        }

        i += gridSize;
    }

    __syncthreads( );

    if( BLOCKSIZE >= 512 )
    {
        if( tid < 256 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+256, sdatal+tid+256,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 256);
        }
        __syncthreads( );
    }
    if( BLOCKSIZE >= 256 )
    {
        if( tid < 128 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+128, sdatal+tid+128,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 128);
        }
        __syncthreads( );
    }
    if( BLOCKSIZE >= 128 )
    {
        if( tid <  64 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+64, sdatal+tid+64,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 64);
        }
        __syncthreads( );
    }

    if( tid < 32 )
    {
        if( BLOCKSIZE >=  64 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+32, sdatal+tid+32,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 32);
        }
        if( BLOCKSIZE >=  32 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+16, sdatal+tid+16,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 16);
        }
        if( BLOCKSIZE >=  16 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+8, sdatal+tid+8,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 8);
        }
        if( BLOCKSIZE >=   8 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+4, sdatal+tid+4,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 4);
        }
        if( BLOCKSIZE >=   4 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+2, sdatal+tid+2,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 2);
        }
        if( BLOCKSIZE >=   2 )
        {
	    bool flag;
	    gdd_1stvalue_is_larger(sdatah+tid, sdatal+tid, sdatah+tid+1, sdatal+tid+1,&flag);
	    if(flag == false) *(idx + tid) = *(idx + tid + 1);
        }
    }


    if( tid == 0 )
    {
        dwork2h[blockIdx.x] = sdatah[0];
        dwork2l[blockIdx.x] = sdatal[0];
	dwork2idx[blockIdx.x] = idx[0];

    }
}




extern "C" int
gddidamax( int *n,   double *Xh, double *Xl, int* incx )
{
	int ret;

    unsigned int gridSize = ( *n+TILEWIDTH-1 )/TILEWIDTH;

    dim3         blocks( gridSize, 1, 1 );

    dim3         threads( BLOCKSIZE, 1, 1 );


    double *     dworkh, *dworkl;
    int * dworkidx;
    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( gridSize, sizeof( double ), (void**) &dworkh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( gridSize, sizeof( double ), (void**) &dworkl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( gridSize, sizeof( int ), (void**) &dworkidx )
      )
    {
        cublasFree( dworkh );
        cublasFree( dworkl );
        cublasFree( dworkidx );
        printf( "Device Malloc Error!\n" );
        exit( -1 );
    }

    gddidamax_kernel<<< blocks, threads, 0, gddstream>>>
    ( *n, Xh, Xl, *incx, dworkh, dworkl,dworkidx );

    if( gridSize > 1 )
    {
        double *dwork2h, *dwork2l;
	int *dwork2idx;
        if(

            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( ( gridSize+TILEWIDTH-1 )/TILEWIDTH, sizeof( double ),
                        (void**) &dwork2h )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( ( gridSize+TILEWIDTH-1 )/TILEWIDTH, sizeof( double ),
                        (void**) &dwork2l )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( ( gridSize+TILEWIDTH-1 )/TILEWIDTH, sizeof( int ),
                        (void**) &dwork2idx )
          )
        {
            cublasFree( dwork2h );
            cublasFree( dwork2l );
            cublasFree( dwork2idx );
            printf( "Device Malloc Error!\n" );
            exit( -1 );
        }

        int count = 0;
        while( gridSize > 1 )
        {
            unsigned int vectorSize = gridSize;
            gridSize = ( gridSize+TILEWIDTH-1 )/TILEWIDTH;
            dim3         localblock( gridSize, 1 );

            /* reduce_kernel( */
            gddidamax_update_kernel
            <<< localblock, threads, 0, gddstream>>>
            ( vectorSize,dworkh, dworkl,dworkidx, dwork2h, dwork2l,dwork2idx );
            swapPointer( &dworkh, &dwork2h );
            swapPointer( &dworkl, &dwork2l );
            swapPointer( &dworkidx, &dwork2idx );
            count++;
        }


	cublasGetVector(1,sizeof(int),dworkidx,1,&ret,1);

        if( count%2 == 1 )
        {
            swapPointer( &dworkh, &dwork2h );
            swapPointer( &dworkl, &dwork2l );
            swapPointer( &dworkidx, &dwork2idx );
        }


        cublasFree( dwork2h );
        cublasFree( dwork2l );
        cublasFree( dwork2idx );
    }
    else
    {
	cublasGetVector(1,sizeof(int),dworkidx,1,&ret,1);
    
    }
    
    cublasFree( dworkh );
    cublasFree( dworkl );


    return ret+1;
}

