/* AXPY */
#define BLOCKSIZE ( 64 )
#define TILEWIDTH ( 2*BLOCKSIZE )
#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "basic_algebra.h"
#include "cutil.h"

__global__ void
gddsqrt_kernel( double* reth, double *retl )
{
    RSQRT( reth, retl );
}

__global__ void
gddnrm2_kernel( int n, double *Xh, double *Xl, int incx, double*dworkh,
                double*dworkl )
{
    __shared__ double sdatah[BLOCKSIZE];
    __shared__ double sdatal[BLOCKSIZE];
    unsigned int      tid = threadIdx.x;
    unsigned int      i = blockIdx.x*( BLOCKSIZE*2 )+tid;
    unsigned int      gridSize = BLOCKSIZE*2*gridDim.x;

/*    if(incx < 0){ */
/*            Xh += (-n+1)*incx;*/
/*            Xl += (-n+1)*incx;*/
/*    }*/


    sdatah[tid] = 0.0;
    sdatal[tid] = 0.0;

    while( i < n )
    {
        double tmph, tmpl;
        /*        sdata[tid] += X[i]*X[i];*/
        tmph = *( Xh+i*incx );
        tmpl = *( Xl+i*incx );
        RMUL( &tmph, &tmpl, &tmph, &tmpl, &tmph, &tmpl );
        RADD( sdatah+tid, sdatal+tid, &tmph, &tmpl, sdatah+tid, sdatal+tid );

        if( i+BLOCKSIZE < n )
        {
            tmph = *( Xh+( i+BLOCKSIZE )*incx );
            tmpl = *( Xl+( i+BLOCKSIZE )*incx );
            RMUL( &tmph, &tmpl, &tmph, &tmpl, &tmph, &tmpl );
            RADD( sdatah+tid, sdatal+tid, &tmph, &tmpl, sdatah+tid,
                  sdatal+tid );
        }

        i += gridSize;
    }

    __syncthreads( );

    if( BLOCKSIZE >= 512 )
    {
        if( tid < 256 )
        {
            /*            sdata[tid] += sdata[tid+256];*/
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+256, sdatal+tid+256,
                  sdatah+tid, sdatal+tid );
        }
        __syncthreads( );
    }
    if( BLOCKSIZE >= 256 )
    {
        if( tid < 128 )
        {
            /*            sdata[tid] += sdata[tid+128];*/
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+128, sdatal+tid+128,
                  sdatah+tid, sdatal+tid );
        }
        __syncthreads( );
    }
    if( BLOCKSIZE >= 128 )
    {
        if( tid <  64 )
        {
            /*            sdata[tid] += sdata[tid+64];*/
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+64, sdatal+tid+64,
                  sdatah+tid, sdatal+tid );
        }
        __syncthreads( );
    }

    if( tid < 32 )
    {
        if( BLOCKSIZE >=  64 )
        {
            /*            sdata[tid] += sdata[tid+32];*/
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+32, sdatal+tid+32,
                  sdatah+tid, sdatal+tid );
        }
        if( BLOCKSIZE >=  32 )
        {
            /*                sdata[tid] += sdata[tid+16]; */
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+16, sdatal+tid+16,
                  sdatah+tid, sdatal+tid );
        }
        if( BLOCKSIZE >=  16 )
        {
            /*                sdata[tid] += sdata[tid+8];*/
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+8, sdatal+tid+8,
                  sdatah+tid, sdatal+tid );
        }
        if( BLOCKSIZE >=   8 )
        {
            /*                       sdata[tid] += sdata[tid+4];*/
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+4, sdatal+tid+4,
                  sdatah+tid, sdatal+tid );
        }
        if( BLOCKSIZE >=   4 )
        {
            /*                       sdata[tid] += sdata[tid+2];*/
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+2, sdatal+tid+2,
                  sdatah+tid, sdatal+tid );
        }
        if( BLOCKSIZE >=   2 )
        {
            /*                sdata[tid] += sdata[tid+1]; */
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+1, sdatal+tid+1,
                  sdatah+tid, sdatal+tid );
        }
    }


    if( tid == 0 )
    {
        dworkh[blockIdx.x] = sdatah[0];
        dworkl[blockIdx.x] = sdatal[0];
    }
}

extern "C" void
gddsqsum( int *n,   double *Xh, double *Xl, int* incx,  double * reth, double*retl )
{
    unsigned int gridSize = ( *n+TILEWIDTH-1 )/TILEWIDTH;
    dim3         blocks( gridSize, 1, 1 );
    dim3         threads( BLOCKSIZE, 1, 1 );


    double *     dworkh, *dworkl;
    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( gridSize, sizeof( double ), (void**) &dworkh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( gridSize, sizeof( double ), (void**) &dworkl )
      )
    {
        cublasFree( dworkh );
        cublasFree( dworkl );
        printf( "Device Malloc Error!\n" );
        exit( -1 );
    }

    gddnrm2_kernel<<< blocks, threads, 0, gddstream>>>
    ( *n, Xh, Xl, *incx, dworkh, dworkl );

    if( gridSize > 1 )
    {
        double *dwork2h, *dwork2l;
        if(

            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( ( gridSize+TILEWIDTH-1 )/TILEWIDTH, sizeof( double ),
                        (void**) &dwork2h )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( ( gridSize+TILEWIDTH-1 )/TILEWIDTH, sizeof( double ),
                        (void**) &dwork2l )
          )
        {
            cublasFree( dwork2h );
            cublasFree( dwork2l );
            printf( "Device Malloc Error!\n" );
            exit( -1 );
        }

        int count = 0;
        while( gridSize > 1 )
        {
            unsigned int vectorSize = gridSize;
            gridSize = ( gridSize+TILEWIDTH-1 )/TILEWIDTH;
            dim3         localblock( gridSize, 1 );

            /* reduce_kernel( */
            reduce_kernel<BLOCKSIZE>
            <<< localblock, threads, 0, gddstream>>>
            ( dworkh, dworkl, dwork2h, dwork2l, vectorSize );
            swapPointer( &dworkh, &dwork2h );
            swapPointer( &dworkl, &dwork2l );
            count++;
        }


        elemcopy<double>( dworkh, reth );
        elemcopy<double>( dworkl, retl );

        if( count%2 == 1 )
        {
            swapPointer( &dworkh, &dwork2h );
            swapPointer( &dworkl, &dwork2l );
        }



        cublasFree( dwork2h );
        cublasFree( dwork2l );
    }
    else
    {
        elemcopy<double>( dworkh, reth );
        elemcopy<double>( dworkl, retl );
    }
    cublasFree( dworkh );
    cublasFree( dworkl );
}

extern "C" void
gddnrm2( int *n,   double *Xh, double *Xl, int* incx,double * reth, double*retl )
{
	if(*n==0)
	{
	       	*reth = 0.0;
	        *retl = 0.0;
	}else{
    gddsqsum( n,   Xh, Xl, incx, reth, retl );

    dim3 blocks( 1, 1, 1 );
    dim3 threads( 1, 1, 1 );
    gddsqrt_kernel<<< blocks, threads, 0, gddstream>>> ( reth, retl );
	}
}

