#define BLOCKSIZE 64
#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"

__global__ void
gddrot_kernel( int n,
               double *Xh, double *Xl,
               int incx,
               double *Yh, double *Yl,
               int incy,
               double *ch, double *cl,
               double *sh, double *sl )
{
    double       lch = *ch;
    double       lcl = *cl;
    double       lsh = *sh;
    double       lsl = *sl;
    unsigned int x = threadIdx.x+__mul24( BLOCKSIZE, blockIdx.x );
    unsigned int offset1 = __mul24( x, incx );
    unsigned int offset2 = __mul24( x, incy );

/*    if(incx < 0){ */
/*            Xh += (-n+1)*incx;*/
/*            Xl += (-n+1)*incx;*/
/*    }*/
/*    if(incy < 0){ */
/*            Yh += (-n+1)*incy;*/
/*            Yl += (-n+1)*incy;*/
/*    }*/

    if( x < n )
    {
        double lxh = *( Xh+offset1 );
        double lxl = *( Xl+offset1 );
        double lyh = *( Yh+offset2 );
        double lyl = *( Yl+offset2 );
        double tmp1h, tmp2h, tmp1l, tmp2l;
        /* y = a*y */

        RMUL( &lch, &lcl, &lxh, &lxl, &tmp1h, &tmp1l );
        RMUL( &lsh, &lsl, &lyh, &lyl, &tmp2h, &tmp2l );
        RADD( &tmp1h, &tmp1l, &tmp2h, &tmp2l, &tmp1h, &tmp1l );
        *( Xh+offset1 ) = tmp1h;
        *( Xl+offset1 ) = tmp1l;


        lxh = -lxh;
        lxl = -lxl;
        RMUL( &lsh, &lsl, &lxh, &lxl, &tmp1h, &tmp1l );
        RMUL( &lch, &lcl, &lyh, &lyl, &tmp2h, &tmp2l );
        RADD( &tmp1h, &tmp1l, &tmp2h, &tmp2l, &tmp1h, &tmp1l );
        *( Yh+offset2 ) = tmp1h;
        *( Yl+offset2 ) = tmp1l;
    }
}

extern "C" void
gddrot( int *n, double *Xh,
             double *Xl, int* incx,
             double *Yh, double *Yl,
             int* incy, double *ch, double*cl,
             double*sh, double*sl )
{
    dim3 blocks( ( *n+BLOCKSIZE-1 )/BLOCKSIZE, 1 );

    dim3 threads( BLOCKSIZE, 1 );


    gddrot_kernel<<< blocks, threads, 0, gddstream>>>
    ( *n,  Xh, Xl, *incx, Yh, Yl, *incy, ch, cl, sh, sl );
}

