#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"


extern "C" __global__
void
gddrotg_kernel( double*dah, double* dal, double*dbh, double*dbl,
              double*ch, double*cl, double*sh, double*sl )
{
    double roeh, roel;
    double scaleh, scalel;
    double zh, zl;
    double rh, rl;
    int    flag;

    double daabsh=*dah,daabsl=*dal,dbabsh=*dbh,dbabsl=*dbl;
    gdd_abs(&daabsh,&daabsl);
    gdd_abs(&dbabsh,&dbabsl);


    comp( dah, dal, dbh, dbl, &flag );

    if( flag )
    {
        /* da > db */
        roeh = *dah;
        roel = *dal;
    }
    else
    {
        roeh = *dbh;
        roel = *dbl;
    }

    RADD(&daabsh,&daabsl,&dbabsh,&dbabsl,&scaleh,&scalel);
    if(scaleh == 0.0 && scalel == 0.0)
    {
	    *ch = 1.0;
	    *cl = 0.0;
	    *sh = 0.0;
	    *sl = 0.0;
	    rh = 0.0;
	    rl = 0.0;
	    zh = 0.0;
	    zl = 0.0;
    }
    else
    {
	    double tmp1h,tmp1l;
	    RDIV(dah,dal,&scaleh,&scalel,&tmp1h,&tmp1l);
	    RMUL(&tmp1h,&tmp1l,&tmp1h,&tmp1l,&tmp1h,&tmp1l);

	    double tmp2h,tmp2l;
	    RDIV(dbh,dbl,&scaleh,&scalel,&tmp2h,&tmp2l);
	    RMUL(&tmp2h,&tmp2l,&tmp2h,&tmp2l,&tmp2h,&tmp2l);

	    RADD(&tmp1h,&tmp1l,&tmp2h,&tmp2l,&tmp1h,&tmp1l);
	    RSQRT(&tmp1h,&tmp1l);
	    RMUL(&scaleh,&scalel,&tmp1h,&tmp1l,&rh,&rl);

	    double oneh = 1.0,onel = 0.0;
	    gdd_sign(&oneh,&onel,&roeh,&roel,&tmp1h,&tmp1l);
	    RMUL(&tmp1h,&tmp1l,&rh,&rl,&rh,&rl);

	    RDIV(dah,dal,&rh,&rl,ch,cl);
	    RDIV(dbh,dbl,&rh,&rl,sh,sl);
	    zh = 1.0;
	    zl = 0.0;

    printf("flag:%d\n",flag);
	    if(flag)
	    {
		    zh = *sh;
		    zl = *sl;
	    }
	    else if (*ch != 0.0 && *cl != 0.0) {
		    RDIV(&oneh,&onel,ch,cl,&zh,&zl);
	    }
    }
    *dah = rh;
    *dal = rl;
    *dbh = zh;
    *dbl = zl;
}

extern "C" void gddrotg
( double*dah, double* dal, double*dbh, double*dbl,
  double*ch, double*cl, double*sh, double*sl )
{
	dim3 threads(1);
	dim3  blocks(1);

    gddrotg_kernel<<< blocks, threads, 0, gddstream>>>(dah,dal,dbh,dbl,ch,cl,sh,sl);
}


