#define BLOCKSIZE 64
#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"

__global__ void
gddrotm_00_kernel( int n,
                double *Xh, double *Xl,
                int incx,
                double *Yh, double *Yl,
                int incy,
                double *h12h, double *h12l,
	        double *h21h,double * h21l)
{
    unsigned int x = threadIdx.x+__mul24( BLOCKSIZE, blockIdx.x );
    unsigned int offset1 = __mul24( x, incx );
    unsigned int offset2 = __mul24( x, incy );

    if( x < n )
    {
        double lxh = *( Xh+offset1 );
        double lxl = *( Xl+offset1 );
        double lyh = *( Yh+offset2 );
        double lyl = *( Yl+offset2 );
        double tmph, tmpl;

        RMUL( h12h,  h12l,  &lyh, &lyl, &tmph, &tmpl );
        RADD( &tmph, &tmpl, &lxh, &lxl, &tmph, &tmpl );
        *( Xh+offset1 ) = tmph;
        *( Xl+offset1 ) = tmpl;


        RMUL( h21h, h21l, &lxh, &lxl, &tmph, &tmpl );
        RADD( &tmph, &tmpl, &lyh, &lyl, &tmph, &tmpl );
        *( Yh+offset2 ) = tmph;
        *( Yl+offset2 ) = tmpl;
    }
}



__global__ void
gddrotm_plus_kernel( int n,
                double *Xh, double *Xl,
                int incx,
                double *Yh, double *Yl,
                int incy,
                double *h11h, double *h11l,
	        double *h22h,double * h22l)
{
    unsigned int x = threadIdx.x+__mul24( BLOCKSIZE, blockIdx.x );
    unsigned int offset1 = __mul24( x, incx );
    unsigned int offset2 = __mul24( x, incy );

    if( x < n )
    {
        double lxh = *( Xh+offset1 );
        double lxl = *( Xl+offset1 );
        double lyh = *( Yh+offset2 );
        double lyl = *( Yl+offset2 );
        double tmph, tmpl;

        RMUL( h11h,  h11l,  &lxh, &lxl, &tmph, &tmpl );
        RADD( &tmph, &tmpl, &lyh, &lyl, &tmph, &tmpl );
        *( Xh+offset1 ) = tmph;
        *( Xl+offset1 ) = tmpl;


        RMUL( h22h, h22l, &lyh, &lyl, &tmph, &tmpl );
        RDIFF( &tmph, &tmpl, &lxh, &lxl, &tmph, &tmpl );
        *( Yh+offset2 ) = tmph;
        *( Yl+offset2 ) = tmpl;
    }
}


__global__ void
gddrotm_minus_kernel( int n,
                double *Xh, double *Xl,
                int incx,
                double *Yh, double *Yl,
                int incy,
                double *h11h, double *h11l,
                double *h21h, double *h21l,
                double *h12h, double *h12l,
	        double *h22h,double * h22l)
{
    unsigned int x = threadIdx.x+__mul24( BLOCKSIZE, blockIdx.x );
    unsigned int offset1 = __mul24( x, incx );
    unsigned int offset2 = __mul24( x, incy );

    if( x < n )
    {
        double lxh = *( Xh+offset1 );
        double lxl = *( Xl+offset1 );
        double lyh = *( Yh+offset2 );
        double lyl = *( Yl+offset2 );
        double tmp1h, tmp1l,tmp2h, tmp2l;

        RMUL( h11h,  h11l,  &lxh, &lxl, &tmp1h, &tmp1l );
        RMUL( h12h, h12l, &lyh, &lyl, &tmp2h, &tmp2l );
        RADD( &tmp1h, &tmp1l, &tmp2h, &tmp2l, &tmp1h, &tmp1l );

        *( Xh+offset1 ) = tmp1h;
        *( Xl+offset1 ) = tmp1l;


        RMUL( h21h,  h21l, &lxh, &lxl, &tmp1h, &tmp1l );
        RMUL( h22h,  h22l, &lyh, &lyl, &tmp2h, &tmp2l );
        RADD( &tmp1h, &tmp1l, &tmp2h, &tmp2l, &tmp1h, &tmp1l );

        *( Yh+offset2 ) = tmp1h;
        *( Yl+offset2 ) = tmp1l;
    }
}






extern "C" void
gddrotm( int *n, double *Xh,
              double *Xl, int *incx,
              double *Yh, double *Yl,
              int *incy, double *dparamh, double*dparaml )
{
       dim3    blocks( ( *n+BLOCKSIZE-1 )/BLOCKSIZE, 1 );

    dim3    threads( BLOCKSIZE, 1 );



 double flagh, flagl;
    cublasGetVector( 1, sizeof( double ), dparamh, 1, &flagh, 1 );
    cublasGetVector( 1, sizeof( double ), dparaml, 1, &flagl, 1 );
    
    if(*n <= 0 || (flagh == -2.0 && flagl == 0.0))
    {
	    return;
    }
    else if (flagh == 0 && flagl == 0) 
    {
    gddrotm_00_kernel<<< blocks, threads, 0, gddstream>>>
    ( *n,  Xh, Xl, *incx, Yh, Yl, *incy, dparamh+3,dparaml+3, dparamh+2,dparaml+2 );

    }
    else if (flagh < 0 || (flagh == 0 && flagl < 0)) 
    {
    gddrotm_minus_kernel<<< blocks, threads, 0, gddstream>>>
    ( *n,  Xh, Xl, *incx, Yh, Yl, *incy, 
      dparamh+1,dparaml+1, dparamh+2,dparaml+2 ,
      dparamh+3,dparaml+3, dparamh+4,dparaml+4 
      );
	    
    }
    else
    {
    gddrotm_plus_kernel<<< blocks, threads, 0, gddstream>>>
    ( *n,  Xh, Xl, *incx, Yh, Yl, *incy, dparamh+1,dparaml+1, dparamh+4,dparaml+4 );
    
    }
}

