#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"

extern "C" __global__
void
gddrotmg_kernel(  
		double *dd1h, double*dd1l,
		double *dd2h, double*dd2l,
		double *dx1h, double*dx1l,
	       	double *dy1h, double*dy1l,
	       	double *dparamh,double*dparaml )
{
    /* Initialized data */

     double zeroh = 0.0;
     double zerol = 0.0;
     double oneh = 1.0;
     double minusone = - 1.0;
     double twoh = 2.0;
     double gamh = 4096.;
     double gamsqh = 16777216.;
     double rgamsqh = 5.9604645e-8;

    /* Format strings */
     char       fmt_120[] = "";
     char       fmt_150[] = "";
     char       fmt_180[] = "";
     char       fmt_210[] = "";

    /* System generated locals */
    double        d__1h;
    double        d__1l;

    /* Local variables */
    double        duh, dp1h, dp2h, dq1h, dq2h, dh11h, dh12h, dh21h, dh22h;
    double        dul, dp1l, dp2l, dq1l, dq2l, dh11l, dh12l, dh21l, dh22l;
    int           igo;
    double        dflagh, dtemph;
    double        dflagl, dtempl;

    /* Assigned format variables */
    char *     igo_fmt;

    /*     .. Scalar Arguments .. */
    /*     .. */
    /*     .. Array Arguments .. */
    /*     .. */

    /*  Purpose */
    /*  ======= */

    /*     CONSTRUCT THE MODIFIED GIVENS TRANSFORMATION MATRIX H WHICH ZEROS */
    /*     THE SECOND COMPONENT OF THE 2-VECTOR  (DSQRT(DD1)*DX1,DSQRT(DD2)* */
    /*     DY2)**T. */
    /*     WITH DPARAM(1)=DFLAG, H HAS ONE OF THE FOLLOWING FORMS.. */

    /*     DFLAG=-1.D0     DFLAG=0.D0        DFLAG=1.D0     DFLAG=-2.D0 */

    /*       (DH11  DH12)    (1.D0  DH12)    (DH11  1.D0)    (1.D0  0.D0) */
    /*     H=(          )    (          )    (          )    (          ) */
    /*       (DH21  DH22),   (DH21  1.D0),   (-1.D0 DH22),   (0.D0  1.D0). */
    /*     LOCATIONS 2-4 OF DPARAM CONTAIN DH11, DH21, DH12, AND DH22 */
    /*     RESPECTIVELY. (VALUES OF 1.D0, -1.D0, OR 0.D0 IMPLIED BY THE */
    /*     VALUE OF DPARAM(1) ARE NOT STORED IN DPARAM.) */

    /*     THE VALUES OF GAMSQ AND RGAMSQ SET IN THE DATA STATEMENT MAY BE */
    /*     INEXACT.  THIS IS OK AS THEY ARE ONLY USED FOR TESTING THE SIZE */
    /*     OF DD1 AND DD2.  ALL ACTUAL SCALING OF DATA IS DONE USING GAM. */


    /*  Arguments */
    /*  ========= */

    /*  DD1    (input/output) DOUBLE PRECISION */

    /*  DD2    (input/output) DOUBLE PRECISION */

    /*  DX1    (input/output) DOUBLE PRECISION */

    /*  DY1    (input) DOUBLE PRECISION */

    /*  DPARAM (input/output)  DOUBLE PRECISION array, dimension 5 */
    /*     DPARAM(1)=DFLAG */
    /*     DPARAM(2)=DH11 */
    /*     DPARAM(3)=DH21 */
    /*     DPARAM(4)=DH12 */
    /*     DPARAM(5)=DH22 */

    /*  ===================================================================== */

    /*     .. Local Scalars .. */
    /*     .. */
    /*     .. Intrinsic Functions .. */
    /*     .. */
    /*     .. Data statements .. */

    /* Parameter adjustments */
/*    --dparam;*/
    --dparamh;
    --dparaml;

    /* Function Body */
    /*     .. */


/*    if( !( *dd1 < zero ) )*/
    int flag;
    comp(dd1h,dd1l,&zeroh,&zerol,&flag);
    if( flag )
    {
        goto L10;
    }
    /*       GO ZERO-H-D-AND-DX1.. */
    goto L60;
L10:
    /*     CASE-DD1-NONNEGATIVE */

/*    dp2 = *dd2*(*dy1);*/
    HOSTRMUL(dd2h,dd2l,dy1h,dy1l,&dp2h,&dp2l);
/*    if( !( dp2 == zero ) )*/
    if( !( dp2h == zeroh && dp2l == zerol ) )
    {
        goto L20;
    }
    dflagh = -twoh;
    dflagl = zerol;
    goto L260;
    /*     REGULAR-CASE.. */
L20:
/*    dp1 = *dd1*(*dx1);*/
/*    dq2 = dp2*(*dy1);*/
/*    dq1 = dp1*(*dx1);*/

    HOSTRMUL(dd1h,dd1l,dx1h,dx1l,&dp1h,&dp1l);
    HOSTRMUL(&dp2h,&dp2l,dy1h,dy1l,&dq2h,&dq2l);
    HOSTRMUL(&dp1h,&dp1l,dx1h,dx1l,&dq1h,&dq1l);

    gdd_abs(&dq1h,&dq1l);
    gdd_abs(&dq2h,&dq2l);

    comp(&dq2h,&dq2l,&dq1h,&dq1l,&flag);

    if( flag )
    {
        goto L40;
    }

/*    dh21 = -( *dy1 )/(*dx1);*/
    HOSTRDIV(dy1h,dy1l,dx1h,dx1l,&dh21h,&dh21l);
    HOSTRMUL(&dh21h,&dh21l,&minusone,&zerol,&dh21h,&dh21l);

/*    dh12 = dp2/dp1;*/
    HOSTRDIV(&dp2h,&dp2l,&dp1h,&dp1l,&dh12h,&dh12l);


/*    du = one-dh12*dh21;*/
    HOSTRMUL(&dh12h,&dh12l,&dh21h,&dh21l,&duh,&dul);
    HOSTRDIFF(&oneh,&zerol,&duh,&dul,&duh,&dul);



/*    if( !( du <= zero ) )*/
    comp(&zeroh,&zerol,&duh,&dul,&flag);
    if( !flag )
    {
        goto L30;
    }
    /*         GO ZERO-H-D-AND-DX1.. */
    goto L60;
L30:


/*    dflag = zero;*/
    dflagh = zeroh;
    dflagl = zerol;

/*    *dd1 /= du;*/
    HOSTRDIV(dd1h,dd1l,&duh,&dul,dd1h,dd1l);

/*    *dd2 /= du;*/
    HOSTRDIV(dd2h,dd2l,&duh,&dul,dd2h,dd2l);

/*    *dx1 *= du;*/
    HOSTRMUL(dx1h,dx1l,&duh,&dul,dx1h,dx1l);

    /*         GO SCALE-CHECK.. */
    goto L100;
L40:

    comp(&dq2h,&dq2l,&zeroh,&zerol,&flag);

/*    if( !( dq2 < zero ) )*/
    if( flag )
    {
        goto L50;
    }
    /*         GO ZERO-H-D-AND-DX1.. */
    goto L60;
L50:
/*    dflag = one;*/
    dflagh = oneh;
    dflagl = zerol;

/*    dh11 = dp1/dp2;*/
    HOSTRDIV(&dp1h,&dp1l,&dp2h,&dp2l,&dh11h,&dh11l);

/*    dh22 = *dx1/(*dy1);*/
    HOSTRDIV(dx1h,dx1l,dy1h,dy1l,&dh22h,&dh22l);

/*    du = one+dh11*dh22;*/
    HOSTRMUL(&dh11h,&dh11l,&dh22h,&dh22l,&duh,&dul);
    HOSTRADD(&oneh,&zerol,&duh,&dul,&duh,&dul);

/*    dtemp = *dd2/du;*/
    HOSTRDIV(dd2h,dd2l,&duh,&dul,&dtemph,&dtempl);

/*    *dd2 = *dd1/du;*/
    HOSTRDIV(dd1h,dd1l,&duh,&dul,dd2h,dd2l);

/*    *dd1 = dtemp;*/
    *dd1h = dtemph;
    *dd1l = dtempl;

/*    *dx1 = *dy1*du;*/
    HOSTRMUL(dy1h,dy1l,&duh,&dul,dx1h,dx1l);

    /*         GO SCALE-CHECK */
    goto L100;
    /*     PROCEDURE..ZERO-H-D-AND-DX1.. */
L60:

/*    dflag = -one;*/
    dflagh = minusone;
    dflagl = zerol;

/*    dh11 = zero;*/
    dh11h = zeroh;
    dh11l = zerol;

/*    dh12 = zero;*/
    dh12h = zeroh;
    dh12l = zerol;

/*    dh21 = zero;*/
    dh21h = zeroh;
    dh21l = zerol;
/*    dh22 = zero;*/
    dh22h = zeroh;
    dh22l = zerol;

/*    *dd1 = zero;*/
    *dd1h = zeroh;
    *dd1l = zerol;

/*    *dd2 = zero;*/
    *dd2h = zeroh;
    *dd2l = zerol;

/*    *dx1 = zero;*/
    *dx1h = zeroh;
    *dx1l = zerol;

    /*         RETURN.. */
    goto L220;
    /*     PROCEDURE..FIX-H.. */
L70:

/*    if( !( dflag >= zero ) )*/
    comp(&dflagh,&dflagl,&zeroh,&zerol,&flag);
    if( !( flag ) )
    {
        goto L90;
    }

    if( !( dflagh == zeroh && dflagl == zerol ) )
    {
        goto L80;
    }

/*    dh11 = one;*/
    dh11h = oneh;
    dh11l = zerol;

/*    dh22 = one;*/
    dh22h = oneh;
    dh22l = zerol;

/*    dflag = -one;*/
    dflagh = minusone;
    dflagl = zerol;

    goto L90;
L80:
/*    dh21 = -one;*/
    dh21h = minusone;
    dh21l = zerol;
    
/*    dh12 = one;*/
    dh12h = oneh;
    dh12l = zerol;

/*    dflag = -one;*/
    dflagh = minusone;
    dflagl = zerol;

L90:
    switch( igo )
    {
        case 0:
            goto L120;

        case 1:
            goto L150;

        case 2:
            goto L180;

        case 3:
            goto L210;
    }
    /*     PROCEDURE..SCALE-CHECK */
L100:
L110:
/*    if( !( *dd1 <= rgamsq ) )*/
    comp(&rgamsqh,&zerol,dd1h,dd1l,&flag);
    if(!flag)
    {
        goto L130;
    }
    if( *dd1h == zeroh && *dd1l == zerol )
    {
        goto L160;
    }
    igo = 0;
    igo_fmt = fmt_120;
    /*              FIX-H.. */
    goto L70;
L120:
    /* Computing 2nd power */
/*    d__1 = gam;*/
    d__1h = gamh;
    d__1l = zerol;

/*    *dd1 *= d__1*d__1;*/
    {
	double tmp1h,tmp1l;
	HOSTRMUL(&d__1h,&d__1l,&d__1h,&d__1l,&tmp1h,&tmp1l);
	HOSTRMUL(&tmp1h,&tmp1l,dd1h,dd1l,dd1h,dd1l);
    }

/*    *dx1 /= gam;*/
    HOSTRDIV(dx1h,dx1l,&gamh,&zerol,dx1h,dx1l);

/*    dh11 /= gam;*/
    HOSTRDIV(&dh11h,&dh11l,&gamh,&zerol,&dh11h,&dh11l);

/*    dh12 /= gam;*/
    HOSTRDIV(&dh12h,&dh12l,&gamh,&zerol,&dh12h,&dh21l);

    goto L110;
L130:
L140:

/*    if( !( *dd1 >= gamsq ) )*/
    comp(dd1h,dd1l,&gamsqh,&zerol,&flag);
    if(!flag)
    {
        goto L160;
    }
    igo = 1;
    igo_fmt = fmt_150;
    /*              FIX-H.. */
    goto L70;
L150:
    /* Computing 2nd power */
/*    d__1 = gam;*/
    d__1h = gamh;
    d__1l = zerol;

/*    *dd1 /= d__1*d__1;*/
    {
	double tmp1h,tmp1l;
	HOSTRMUL(&d__1h,&d__1l,&d__1h,&d__1l,&tmp1h,&tmp1l);
	HOSTRDIV(dd1h,dd1l,&tmp1h,&tmp1l,dd1h,dd1l);
    }
/*    *dx1 *= gam;*/
    HOSTRMUL(dx1h,dx1l,&gamh,&zerol,dx1h,dx1l);

/*    dh11 *= gam;*/
    HOSTRMUL(&dh11h,&dh11l,&gamh,&zerol,&dh11h,&dh11l);

/*    dh12 *= gam;*/
    HOSTRMUL(&dh12h,&dh12l,&gamh,&zerol,&dh12h,&dh21l);

    goto L140;
L160:
L170:

    {
	    double dd2absh,dd2absl;
	    dd2absh = *dd2h;
	    dd2absl = *dd2l;
	    gdd_abs(&dd2absh,&dd2absl);
	    comp(&rgamsqh,&zerol,dd2h,dd2l,&flag);

    }
    if( !( flag ) )
/*    if( !( fabs( *dd2 ) <= rgamsq ) )*/
    {
        goto L190;
    }
/*    if( *dd2 == zero )*/
    if(*dd2h == zeroh && *dd2l == zerol)
    {
        goto L220;
    }
    igo = 2;
    igo_fmt = fmt_180;
    /*              FIX-H.. */
    goto L70;
L180:
    /* Computing 2nd power */

/*    d__1 = gam;*/
    d__1h = gamh;
    d__1l = zerol;

/*    *dd2 *= d__1*d__1;*/
    {
	double tmp1h,tmp1l;
	HOSTRMUL(&d__1h,&d__1l,&d__1h,&d__1l,&tmp1h,&tmp1l);
	HOSTRMUL(dd2h,dd2l,&tmp1h,&tmp1l,dd2h,dd2l);
    }
    
/*    dh21 /= gam;*/
    HOSTRDIV(&dh21h,&dh21l,&gamh,&zerol,&dh21h,&dh21l);

/*    dh22 /= gam;*/
    HOSTRDIV(&dh22h,&dh22l,&gamh,&zerol,&dh22h,&dh22l);

    goto L170;
L190:
L200:

    {
	    double dd2absh,dd2absl;
	    dd2absh = *dd2h;
	    dd2absl = *dd2l;
	    gdd_abs(&dd2absh,&dd2absl);
	    comp(dd2h,dd2l,&gamsqh,&zerol,&flag);
    }
    if( !( flag ) )
/*    if( !( fabs( *dd2 ) >= gamsq ) )*/
    {
        goto L220;
    }
    igo = 3;
    igo_fmt = fmt_210;
    /*              FIX-H.. */
    goto L70;
L210:
    /* Computing 2nd power */

/*    d__1 = gam;*/
    d__1h = gamh;
    d__1l = zerol;

/*    *dd2 /= d__1*d__1;*/
    {
	double tmp1h,tmp1l;
	HOSTRMUL(&d__1h,&d__1l,&d__1h,&d__1l,&tmp1h,&tmp1l);
	HOSTRMUL(dd2h,dd2l,&tmp1h,&tmp1l,dd2h,dd2l);
    }

/*    dh21 *= gam;*/
    HOSTRMUL(&dh21h,&dh21l,&gamh,&zerol,&dh21h,&dh21l);

/*    dh22 *= gam;*/
    HOSTRMUL(&dh22h,&dh22l,&gamh,&zerol,&dh22h,&dh22l);

    goto L200;
L220:

    comp(&dflagh,&dflagl,&zeroh,&zerol,&flag);
/*    if( dflag < 0. )*/
    if( !(flag) )
    {
        goto L250;
    }
    else if( dflagh == 0 && dflagl == 0.0 )
    {
        goto L230;
    }
    else
    {
        goto L240;
    }
L230:
    dparamh[3] = dh21h;
    dparamh[4] = dh12h;
    dparaml[3] = dh21l;
    dparaml[4] = dh12l;
    goto L260;
L240:
    dparamh[2] = dh11h;
    dparamh[5] = dh22h;
    dparaml[2] = dh11l;
    dparaml[5] = dh22l;
    goto L260;
L250:
    dparamh[2] = dh11h;
    dparamh[3] = dh21h;
    dparamh[4] = dh12h;
    dparamh[5] = dh22h;
    dparaml[2] = dh11l;
    dparaml[3] = dh21l;
    dparaml[4] = dh12l;
    dparaml[5] = dh22l;
L260:
    dparamh[1] = dflagh;
    dparaml[1] = dflagl;
}


extern "C" void gddrotmg(
		double *dd1h, double*dd1l,
		double *dd2h, double*dd2l,
		double *dx1h, double*dx1l,
	       	double *dy1h, double*dy1l,
	       	double *dparamh,double*dparaml )
{
	dim3 threads(1);
	dim3  blocks(1);

    gddrotmg_kernel<<< blocks, threads, 0, gddstream>>>(dd1h,dd1l,dd2h,dd2l,dx1h,dx1l,dy1h,dy1l,dparamh,dparaml);
}



