/* AXPY */
#define BLOCKSIZE 64
#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"

__global__ void
gddscal_kernel( int n, double *alphah, double *alphal, double *Xh,
                double *Xl, int incx)
{
    unsigned int x = threadIdx.x+__mul24( BLOCKSIZE, blockIdx.x );
    unsigned int offset = __mul24( x, incx );

    if( x < n )
    {

        double lxh = *( Xh+offset );
        double lxl = *( Xl+offset );
        /* y = a*y */
        RMUL( alphah, alphal, &lxh, &lxl, &lxh, &lxl );
        *( Xh+offset ) = lxh;
        *( Xl+offset ) = lxl;
    }
}

extern "C" void
gddscal( int *n, double *alphah, double *alphal, double *Xh,
              double *Xl, int* incx)
{

    dim3 blocks( ( *n+BLOCKSIZE-1 )/BLOCKSIZE, 1 );

    dim3 threads( BLOCKSIZE, 1 );


    gddscal_kernel<<< blocks, threads, 0, gddstream>>>
    ( *n,  alphah, alphal, Xh, Xl, *incx );
}
