/* */
/* SWAP BLAS */
/* */

#define BLOCKSIZE 64
#include "gddcommon.h"

__global__ void
gddswap_kernel( int n, double *dxh, double *dxl, int incx,
                double *dyh, double *dyl, int incy )
{
    unsigned int x = threadIdx.x+__mul24( BLOCKSIZE, blockIdx.x );
    unsigned int offset1 = __mul24( x, incx );
    unsigned int offset2 = __mul24( x, incy );

    if( x < n )
    {
        /* For Higher Part */
        double *A1 = dxh+offset1;
        double *A2 = dyh+offset2;
        double  temp = *A1;
        *A1 = *A2;
        *A2 = temp;

        /* For Lower Part */
        A1 = dxl+offset1;
        A2 = dyl+offset2;
        temp = *A1;
        *A1 = *A2;
        *A2 = temp;
    }
}

extern "C" void
gddswap( int *n, double *dxh, double*dxl, int *incx,
              double *dyh, double * dyl, int *incy )
{
    dim3 blocks( ( *n+BLOCKSIZE-1 )/BLOCKSIZE, 1 );

    dim3 threads( BLOCKSIZE, 1 );

    gddswap_kernel<<< blocks, threads, 0, gddstream>>>
    ( *n,  dxh, dxl, *incx, dyh, dyl, *incy );
}

