#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"


#define BLOCKSIZE ( 64 )

extern "C" __global__ void
gddsyr_kernel_L( int n, double *alphah, double*alphal,
                  double * xh, double *xl,
                  int incx,
                  double* Ah, double *Al, int lda )
{
    unsigned int tid = threadIdx.x;
    unsigned int idx = tid+__mul24( BLOCKSIZE, blockIdx.x );


    Ah += idx;
    Al += idx;


    {
        __shared__ double xhs[BLOCKSIZE];
        __shared__ double xls[BLOCKSIZE];

        xhs[tid] = xh[idx*incx];
        xls[tid] = xl[idx*incx];

        {
            int offset =  tid*incx;
            xh += offset;
            xl += offset;
        }

        int               n1 = ( blockIdx.x )*BLOCKSIZE;

        double            regalphah = *alphah;
        double            regalphal = *alphal;

        /* Off-Diagonal Block */


        __shared__ double tilexhs[BLOCKSIZE];
        __shared__ double tilexls[BLOCKSIZE];


        for( int i = 0; i < n1; i += BLOCKSIZE )
        {
            __syncthreads( );


            tilexhs[tid] = xh[i*incx];
            tilexls[tid] = xl[i*incx];

            __syncthreads( );

#pragma unroll
            for( int j = 0; j < BLOCKSIZE; j++ )
            {
                double ph, pl, lAh, lAl;

                lAh = *Ah;
                lAl = *Al;

                RMUL( xhs+tid, xls+tid, tilexhs+j, tilexls+j, &ph, &pl );
                RMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );
                RADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );

                if( idx < n )
                {
                    *Ah = lAh;
                    *Al = lAl;
                }

                Ah += lda;
                Al += lda;
            }
        }


        /* Diagonal Block */
            __syncthreads( );
        {
            int m;
            if( n1+BLOCKSIZE < n )
            {
                m = BLOCKSIZE;
            }
            else
            {
                m = n-n1;
            }
            for( int j = 0; j < m; j++ )
            {
                double ph, pl, lAh, lAl;

                lAh = *Ah;
                lAl = *Al;

                RMUL( xhs+tid, xls+tid, xhs+j, xls+j, &ph, &pl );
                RMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );

                ph = ( j <= tid )*ph;
                pl = ( j <= tid )*pl;


                RADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );


                if( idx < n )
                {
                    *Ah = lAh;
                    *Al = lAl;
                }

                Ah += lda;
                Al += lda;
            }
        }
    }
}

extern "C" __global__ void
gddsyr_kernel_U( int n, double *alphah, double*alphal,
                  double * xh, double *xl,
                  int incx,
                  double* Ah, double *Al, int lda )
{
    unsigned int tid = threadIdx.x;
    unsigned int idx = tid+__mul24( BLOCKSIZE, blockIdx.x );


    {
        int offset = idx+__mul24( BLOCKSIZE, blockIdx.x )*lda;
        Ah += offset;
        Al += offset;
    }


    {
        __shared__ double xhs[BLOCKSIZE];
        __shared__ double xls[BLOCKSIZE];

        {
            xh += idx*incx;
            xl += idx*incx;
            xhs[tid] = *xh;
            xls[tid] = *xl;
        }

        int    n1 = BLOCKSIZE*blockIdx.x;

        double regalphah = *alphah;
        double regalphal = *alphal;
        /* Diagonal Block */


        if( n1+BLOCKSIZE  < n )
        {
            n1 = BLOCKSIZE;
        }
        else
        {
            n1 = n-n1;
        }


        for( int j = 0; j < n1; j++ )
        {
            double ph, pl,  lAh, lAl;

            lAh = *Ah;
            lAl = *Al;

            RMUL( xhs+tid, xls+tid, xhs+j, xls+j, &ph, &pl );
            RMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );

            ph = ( j >= tid )*ph;
            pl = ( j >= tid )*pl;


            RADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );


            if( idx < n )
            {
                *Ah = lAh;
                *Al = lAl;
            }

            Ah += lda;
            Al += lda;
        }

        /* Off-Diagonal Block */

        {
            int offset = BLOCKSIZE*incx ;
            xh += offset;
            xl += offset;
        }

        __shared__ double tilexhs[BLOCKSIZE];
        __shared__ double tilexls[BLOCKSIZE];


        n1 = BLOCKSIZE*( n/BLOCKSIZE-( blockIdx.x+1 ) );

             for( int i = 0; i < n1; i += BLOCKSIZE )
        {
            __syncthreads( );


            tilexhs[tid] = xh[i*incx];
            tilexls[tid] = xl[i*incx];

            __syncthreads( );

#pragma unroll
            for( int j = 0; j < BLOCKSIZE; j++ )
            {
                double ph, pl, lAh, lAl;

                lAh = *Ah;
                lAl = *Al;

                RMUL( xhs+tid, xls+tid, tilexhs+j, tilexls+j, &ph, &pl );
                RMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );
                RADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );

                if( idx < n )
                {
                    *Ah = lAh;
                    *Al = lAl;
                }

                Ah += lda;
                Al += lda;
            }
        }


        /* Last Block */

        __syncthreads( );

        int m  = n-( n1+( blockIdx.x+1 )*BLOCKSIZE );

        if( m > 0 && blockIdx.x+1 < gridDim.x )
        {


            tilexhs[tid] = xh[n1*incx];
            tilexls[tid] = xl[n1*incx];

            __syncthreads( );

	   

#pragma unroll
            for( int j = 0; j < m; j++ )
            {
                double ph, pl, lAh, lAl;

                lAh = *Ah;
                lAl = *Al;

                RMUL( xhs+tid, xls+tid, tilexhs+j, tilexls+j, &ph, &pl );
                RMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );
                RADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );


                if( idx < n )
                {
                    *Ah = lAh;
                    *Al = lAl;
                }

                Ah += lda;
                Al += lda;
            }
        }
    }
}

extern "C" void
gddsyr( char* uplo, int* n, double *alphah, double*alphal,
              double * xh, double *xl,
              int* incx,
              double* Ah, double *Al, int* lda  )
{
    dim3 blocks( ( *n+BLOCKSIZE-1 )/BLOCKSIZE, 1 );

    dim3 threads( BLOCKSIZE, 1 );

    if( *uplo == 'L' || *uplo == 'l' )
    {
        gddsyr_kernel_L<<< blocks, threads, 0, gddstream>>>
        ( *n, alphah, alphal,  xh, xl, *incx,
           Ah, Al, *lda );
    }
    else if( *uplo == 'U' || *uplo == 'u' )
    {
        gddsyr_kernel_U<<< blocks, threads, 0, gddstream>>>
        ( *n, alphah, alphal,  xh, xl, *incx,
          Ah, Al, *lda );
    
    }
    else
    {
	    // ???
    
    }
}

