#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"


#define BLOCKSIZE ( 64 )
#define BUFFERSIZE ( BLOCKSIZE )
#define TRANSBLOCK ( 128 )

extern "C" __global__ void
gddtbmv_kernel_N( char diag,int n, int kl, int ku,
                  double *Ah, double *Al, int lda, double * xh, double *xl,
                  int incx, double*yh, double*yl)
{
    int idx = threadIdx.x+__mul24( BLOCKSIZE, blockIdx.x );

    {
        unsigned int offset = idx+ku;
        Ah += offset;
        Al += offset;
        offset = threadIdx.x*incx;
        xh += offset;
        xl += offset;
    }
    lda -= 1;


    double resulth = 0.0;
    double resultl = 0.0;

    {
        __shared__ double buffh[BUFFERSIZE];
        __shared__ double buffl[BUFFERSIZE];
        int               n1 = ( n/BUFFERSIZE )*BUFFERSIZE;
        for( int i = 0; i < n1; i += BUFFERSIZE )
        {
            __syncthreads( );
            buffh[threadIdx.x] = xh[i*incx];
            buffl[threadIdx.x] = xl[i*incx];

            __syncthreads( );

#pragma unroll
            for( int j = 0; j < BUFFERSIZE; j++ )
            {
                double ph, pl;
                ph = *Ah;
                pl = *Al;

                int    b =
                    !( ( idx > kl+ i+j ) || ( idx < i+j-ku ) );
                ph = ph*b;
                pl = pl*b;


		//diagonal element
		b = ((diag == 'u' || diag == 'U') && (idx == i+j));
		ph = b + (!b)*ph;
		pl = (!b)*pl;

                RMUL( &ph, &pl, buffh+j, buffl+j, &ph, &pl );
                RADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
                Ah += lda;
                Al += lda;
            }
        }

        __syncthreads( );


        if( n > n1 )
        {
            buffh[threadIdx.x] = xh[n1*incx];
            buffl[threadIdx.x] = xl[n1*incx];

            __syncthreads( );
            for( int i = 0; i < ( n-n1 ); i++ )
            {
                double ph, pl;
                ph = *Ah;
                pl = *Al;

                int    b =
                    !( ( idx > kl+ i+n1 ) || ( idx < ( i+n1-ku ) ) );
                ph = ph*b;
                pl = pl*b;

/*                if(idx < n){*/
/*                printf("idx:%d:i:%d,ph:%e\n",idx,i+n1,ph);*/
/*                }*/
		//diagonal element
		b = ((diag == 'u' || diag == 'U') && (idx == i+n1));
		ph = b + (!b)*ph;
		pl = (!b)*pl;



                RMUL( &ph, &pl, buffh+i, buffl+i, &ph, &pl );


                RADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
                Ah += lda;
                Al += lda;
            }
        }
    }

    if( idx < n )
    {
 *( yh+idx*incx) = resulth;
	 *(yl+idx*incx) = resultl;
    }
}

extern "C" __global__ void
gddtbmv_kernel_T( char diag, int n, int kl, int ku,
                  double *Ah, double *Al, int lda, double * xh, double *xl,
                  int incx, double*yh, double*yl)
{
	int tx = threadIdx.x;
	int bid = blockIdx.y;

	__shared__ double sdatah[TRANSBLOCK];
	__shared__ double sdatal[TRANSBLOCK];

	double resh, resl;
	resh = resl = 0.0;
	int n1 = (n/TRANSBLOCK)*TRANSBLOCK;

	int offset = ku -blockIdx.y;
	Ah += offset;
	Al += offset;


	for (int i = 0; i < n1; i+=TRANSBLOCK) 
	{
		double ph = *(Ah+tx+i+lda*bid);
		double pl = *(Al+tx+i+lda*bid);
		int b = !((i+tx >bid + kl) || (i + tx < bid -ku));
		ph = ph * b;
		pl = pl * b;
		b = ((diag == 'u' || diag == 'U') && (bid == i+tx));
		ph = b + (!b)*ph;
		pl = (!b)*pl;
		RMUL(&ph,&pl,xh+(tx+i)*incx,xl+(tx+i)*incx,&ph,&pl);
		RADD(&resh,&resl,&ph,&pl,&resh,&resl);


	}

	if(tx + n1+offset < n)
	{
		double ph = *(Ah+tx+n1+lda*bid);
		double pl = *(Al+tx+n1+lda*bid);
		int b = !((n1+tx> bid+kl)||(n1+tx < bid - ku));
		ph = ph*b;
		pl = pl*b;
		b = ((diag == 'u' || diag == 'U') && (bid == n1+tx));
		ph = b + (!b)*ph;
		pl = (!b)*pl;
		RMUL(&ph,&pl,xh+(tx+n1)*incx,xl+(tx+n1)*incx,&ph,&pl);
		RADD(&resh,&resl,&ph,&pl,&resh,&resl);
	}

	sdatah[tx] = resh;
	sdatal[tx] = resl;

	__syncthreads();

    if( TRANSBLOCK >= 512 )
    {
        if( tx < 256 )
        {
/*            sdata[tx] += sdata[tx+256];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+256, sdatal+tx+256,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }
    if( TRANSBLOCK >=        256 )
    {
        if( tx < 128 )
        {
/*            sdata[tx] += sdata[tx+128];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+128, sdatal+tx+128,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }
    if( TRANSBLOCK >=        128 )
    {
        if( tx <  64 )
        {
/*            sdata[tx] += sdata[tx+64];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+64, sdatal+tx+64,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }

    if( tx < 32 )
    {
        if( TRANSBLOCK >=  64 )
        {
/*            sdata[tx] += sdata[tx+32];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+32, sdatal+tx+32,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=  32 )
        {
/*                sdata[tx] += sdata[tx+16]; */
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+16, sdatal+tx+16,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=  16 )
        {
/*                sdata[tx] += sdata[tx+8];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+8, sdatal+tx+8,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   8 )
        {
/*                       sdata[tx] += sdata[tx+4];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+4, sdatal+tx+4,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   4 )
        {
/*                       sdata[tx] += sdata[tx+2];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+2, sdatal+tx+2,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   2 )
        {
/*                sdata[tx] += sdata[tx+1]; */
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+1, sdatal+tx+1,
                  sdatah+tx, sdatal+tx );
        }
    }

    if(tx == 0)
    {
/*            y[blockIdx.y*incy] = sdata[0]*(*alpha)+(*beta)*y[blockIdx.y*incy];*/

	    *(yh+blockIdx.y*incx) = *sdatah;
	    *(yl+blockIdx.y*incx) = *sdatal;
    }



}

extern "C" void
gddtbmv( char*uplo, char*trans, char*diag, int *n,int *k,
              double *Ah, double *Al, int *lda, double * xh, double *xl,
              int *incx )
{
    int btrans = 1;
    if( *trans == 'N' || *trans == 'n' ) btrans = 0;

    bool    bupflag = true;
    if( *uplo == 'L' || *uplo == 'l' ) bupflag = false;

    double *dworkh,*dworkl;
    int sizex = (*n) * (*incx);
    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizex, sizeof( double ), (void**) &dworkh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizex, sizeof( double ), (void**) &dworkl )
      )
    {
        cublasFree( dworkh );
        cublasFree( dworkl );
        printf( "Device Malloc Error\n" );
        exit( -1 );
    }

    CUDA_SAFE_CALL( cudaMemcpy( dworkh, xh, sizeof( double )*sizex,
                                cudaMemcpyDeviceToDevice ) );
    CUDA_SAFE_CALL( cudaMemcpy( dworkl, xl, sizeof( double )*sizex,
                                cudaMemcpyDeviceToDevice ) );
    if( !btrans )
    {
        dim3 blocks( ( *n+BLOCKSIZE-1 )/BLOCKSIZE, 1 );
        dim3 threads( BLOCKSIZE, 1 );
	int kl,ku;
	    if(!bupflag){
		    kl = *k;
		    ku = 0;
	    }
	    else
	    {
		    kl = 0;
		    ku = *k;
	    }

	gddtbmv_kernel_N<<< blocks, threads, 0, gddstream>>>
	( *diag, *n, kl, ku, Ah, Al, *lda,  dworkh, dworkl, *incx, xh, xl );

    }
    else
    {
        dim3 blocks( 1, *n );
        dim3 threads( TRANSBLOCK, 1 );
	int kl,ku;
	    if(!bupflag)
	    {
		    kl = *k;
		    ku = 0;
	    }
	    else
	    {
		    kl = 0;
		    ku = *k;
	    }
	gddtbmv_kernel_T<<< blocks, threads, 0, gddstream>>>
	( *diag, *n, kl, ku, Ah, Al, *lda,  dworkh, dworkl, *incx, xh, xl );

    }

    cublasFree(dworkh);
    cublasFree(dworkl);
}
;
