#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"

#define BLOCKSIZE ( 256 )

void __global__
bandMatrix_copy_kernel( int n, int kl, int ku, double*Ah, double*Al,
                        int lda, double*Bh,
                        double*Bl )
{
    lda -= 1;
    int idx = threadIdx.x+__mul24( BLOCKSIZE, blockIdx.x );
    int offset = idx+ku;
    Ah += offset;
    Al += offset;

    int startPosition =  ( ( (int) blockIdx.x )*BLOCKSIZE-kl )/BLOCKSIZE;
    if( startPosition < 0 )
        startPosition = 0;
    startPosition *= BLOCKSIZE;
    Ah += startPosition*lda;
    Al += startPosition*lda;


    int n1 = ( n/BLOCKSIZE )*BLOCKSIZE;
    int endPosition =
        ( ( (int) blockIdx.x )+1+( ku-1 )/BLOCKSIZE+1 )*BLOCKSIZE;
    n1 = min( n1, endPosition );
    for( int i = 0; i < n1; i += BLOCKSIZE )
    {
        for( int j = 0; j < BLOCKSIZE; j++ )
        {
            double ph, pl;
            ph = *Ah;
            pl = *Al;
            int    b =
                !( ( idx > kl+i+j ) || ( idx < i+j-ku ) );
            ph = ph*b;
            pl = pl*b;

            if( idx < n )
            {
                Bh[idx+n*( i+j )] = ph;
                Bl[idx+n*( i+j )] = pl;
            }
            Ah += lda;
            Al += lda;
        }
    }

    if( n > n1 )
    {
        for( int i = 0; i < ( n-n1 ); i++ )
        {
            double ph, pl;
            ph = *Ah;
            pl = *Al;
            int    b =
                !( ( idx > kl+i+n1 ) || ( idx < ( i+n1-ku ) ) );
            ph = ph*b;
            pl = pl*b;
            if( idx < n )
            {
                Bh[idx+n*( i+n1 )] = ph;
                Bl[idx+n*( i+n1 )] = pl;
            }
            Ah += lda;
            Al += lda;
        }
    }
}

void
gddtbsv( char*uplo, char*trans, char*diag, int *n, int *k,
              double *Ah, double *Al, int *lda, double * xh, double *xl,
              int *incx )
{
    double *dworkh, *dworkl;
    int     sizew = ( *n )*( *n );

    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizew, sizeof( double ), (void**) &dworkh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizew, sizeof( double ), (void**) &dworkl )
      )
    {
        cublasFree( dworkh );
        cublasFree( dworkl );
        printf( "Device Malloc Error!\n" );
        exit( -1 );
    }
    cudaMemset( dworkh, 0, sizew*sizeof( double ) );
    cudaMemset( dworkl, 0, sizew*sizeof( double ) );

    dim3 blocks( ( *n+BLOCKSIZE-1 )/BLOCKSIZE );

    dim3 threads( BLOCKSIZE, 1, 1 );

    int  ku, kl;

    if( *uplo == 'u' || *uplo == 'U' )
    {
        ku = *k;
        kl = 0;
    }
    else
    {
        ku = 0;
        kl = *k;
    }
    bandMatrix_copy_kernel<<< blocks, threads, 0, gddstream>>>
    ( *n, kl, ku, Ah, Al, *lda, dworkh, dworkl );

    gddtrsv( uplo, trans, diag, n, dworkh, dworkl, n, xh, xl, incx );
}

