#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"


#define BLOCKSIZE ( 64 )
#define BUFFERSIZE ( BLOCKSIZE )
#define TRANSBLOCK ( 256 )

extern "C" __global__ void
gddtrmv_kernel_LN( char diag,int n, double *Ah, double *Al, int lda, double * xinh,
                   double *xinl,
                   int incx,
                   double*xouth,
                   double*xoutl )
{
    unsigned int idx = threadIdx.x+__mul24( BLOCKSIZE, blockIdx.x );


    Ah += idx;
    Al += idx;
    {
        unsigned int offset = threadIdx.x*incx;
        xinh += offset;
        xinl += offset;
    }

    double resulth = 0.0;
    double resultl = 0.0;

    {
        __shared__ double buffh[BUFFERSIZE];
        __shared__ double buffl[BUFFERSIZE];

        int               n1 = blockIdx.x*BUFFERSIZE;

        /* Off-Diagonal Blocks */
        for( int i = 0; i < n1; i += BUFFERSIZE )
        {
            __syncthreads( );

            buffh[threadIdx.x] = xinh[i*incx];
            buffl[threadIdx.x] = xinl[i*incx];

            __syncthreads( );
#pragma unroll
            for( int j = 0; j < BUFFERSIZE; j++ )
            {
                double ph, pl;
/*                result += *(A) *buff[j];*/
                RMUL( Ah, Al, buffh+j, buffl+j, &ph, &pl );
                RADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
                Ah += lda;
                Al += lda;
            }
        }

        __syncthreads( );
        {
            int m;
            if( n1+BUFFERSIZE < n )
            {
                m = BUFFERSIZE;
            }
            else
            {
                m = n-n1;
            }
	    
            buffh[threadIdx.x] = xinh[n1*incx];
            buffl[threadIdx.x] = xinl[n1*incx];
            __syncthreads( );

	    int tid = threadIdx.x;
            for( int i = 0; i < m; i++ )
            {
                double ph, pl;
/*                result += *(A)*buff[j];*/

		int switcher = ((diag == 'u' || diag == 'U') && i == tid);
		ph = switcher + !(switcher)*(*Ah);
		pl =  !(switcher)*(*Al);

                RMUL( &ph, &pl, buffh+i, buffl+i, &ph, &pl );
                if( i <= tid )
                {
                    RADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
                }

                Ah += lda;
                Al += lda;
            }
        }

        if( idx < n )
        {
            *( xouth+idx*incx ) = resulth;
            *( xoutl+idx*incx ) = resultl;
        }
    }
}

extern "C" __global__ void
gddtrmv_kernel_UN(char diag, int n, double *Ah, double *Al, int lda, double * xinh,
                   double *xinl,
                   int incx,
                   double*xouth,
                   double*xoutl )
{
    unsigned int tid = threadIdx.x;
    unsigned int idx = tid+__mul24( BLOCKSIZE, blockIdx.x );

    {
        int offset = idx+__mul24( BLOCKSIZE, blockIdx.x )*lda;
        Ah += offset;
        Al += offset;
    }

    double       resulth = 0.0;
    double       resultl = 0.0;

    {
        __shared__ double buffh[BUFFERSIZE];
        __shared__ double buffl[BUFFERSIZE];


        /* Diagonal Block */
        xinh += idx*incx;
        xinl += idx*incx;

        buffh[tid] = *xinh;
        buffl[tid] = *xinl;

        int               n1 = BUFFERSIZE*blockIdx.x;
        __syncthreads( );


        if( n1+BUFFERSIZE < n )
        {
            n1 = BUFFERSIZE;
        }
        else
        {
            n1 = n-n1;
        }

        for( int i = 0; i < n1; i++ )
        {
            double ph, pl;
/*                result += *(A)*buff[j];*/

		int switcher = ((diag == 'u' || diag == 'U') && i == tid);
		ph = switcher + !(switcher)*(*Ah);
		pl =  !(switcher)*(*Al);
	    
            RMUL( &ph,&pl, buffh+i, buffl+i, &ph, &pl );
            if( i >= threadIdx.x )
            {
                RADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
            }
            Ah += lda;
            Al += lda;
        }


        {
            int offset = BUFFERSIZE*incx;
            xinh += offset;
            xinl += offset;
        }

        /* Off-Diagonal Block */

        n1 = BUFFERSIZE*( n/BUFFERSIZE-( blockIdx.x+1 ) );

        for( int i = 0; i < n1; i += BUFFERSIZE )
        {
            __syncthreads( );
            buffh[threadIdx.x] = xinh[i*incx];
            buffl[threadIdx.x] = xinl[i*incx];
            __syncthreads( );

#pragma unroll
            for( int j = 0; j < BUFFERSIZE; j++ )
            {
                double ph, pl;
/*                result += *(A) *buff[j];*/
                RMUL( Ah, Al, buffh+j, buffl+j, &ph, &pl );
                RADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
                Ah += lda;
                Al += lda;
            }
        }


        /* Last Block */

        __syncthreads( );

        int m = n-( n1+( blockIdx.x+1 )*BLOCKSIZE );
        if( m > 0 && blockIdx.x+1 < gridDim.x )
        {
            buffh[threadIdx.x] = xinh[n1*incx];
            buffl[threadIdx.x] = xinl[n1*incx];
            __syncthreads( );

            for( int i = 0; i < m; i++ )
            {
                double ph, pl;
/*                result += *(A)*buff[j];*/
                RMUL( Ah, Al, buffh+i, buffl+i, &ph, &pl );
                RADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
                Ah += lda;
                Al += lda;
            }
        }


        if( idx < n )
        {
            *( xouth+idx*incx ) = resulth;
            *( xoutl+idx*incx ) = resultl;
        }
    }
}

extern "C" __global__ void
gddtrmv_kernel_UT(char diag, int n, double *Ah, double *Al, int lda, double * xinh,
                   double *xinl,
                   int incx,
                   double*xouth,
                   double*xoutl )
{
    unsigned int      tx = threadIdx.x;

    __shared__ double sdatah[TRANSBLOCK];
    __shared__ double sdatal[TRANSBLOCK];

    double            resh, resl;

    resh = resl = 0.0;
    unsigned int      n1 = ( blockIdx.y/TRANSBLOCK )*TRANSBLOCK;


    //Off-Diagonal Blocks

    for( int i = 0; i < n1; i += TRANSBLOCK )
    {
        double ph, pl;
        RMUL( Ah+tx+i+lda*blockIdx.y, Al+tx+i+lda*blockIdx.y,
              xinh+( tx+i )*incx, xinl+( tx+i )*incx,
              &ph, &pl );
        RADD( &resh, &resl, &ph, &pl, &resh, &resl );
    }

    //Diagonal Block

    if( tx+n1 < blockIdx.y+1 )
    {
        double ph, pl;


/*                resh +=  A[tx + n1 + lda*blockIdx.y] * x[tx + n1];*/

		int switcher = ((diag == 'u' || diag == 'U') && tx + n1 == blockIdx.y);
		ph = switcher + !(switcher)*(*(Ah+tx+n1+lda*blockIdx.y));
		pl =  !(switcher)*(*(Al+tx+n1+lda*blockIdx.y));


        RMUL( &ph, &pl,
              xinh+( tx+n1 )*incx, xinl+( tx+n1 )*incx,
              &ph, &pl );
        RADD( &resh, &resl, &ph, &pl, &resh, &resl );
    }

    sdatah[tx] = resh;
    sdatal[tx] = resl;



    __syncthreads( );


    if( TRANSBLOCK >= 512 )
    {
        if( tx < 256 )
        {
/*            sdata[tx] += sdata[tx+256];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+256, sdatal+tx+256,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }
    if( TRANSBLOCK >=        256 )
    {
        if( tx < 128 )
        {
/*            sdata[tx] += sdata[tx+128];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+128, sdatal+tx+128,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }
    if( TRANSBLOCK >=        128 )
    {
        if( tx <  64 )
        {
/*            sdata[tx] += sdata[tx+64];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+64, sdatal+tx+64,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }

    if( tx < 32 )
    {
        if( TRANSBLOCK >=  64 )
        {
/*            sdata[tx] += sdata[tx+32];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+32, sdatal+tx+32,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=  32 )
        {
/*                sdata[tx] += sdata[tx+16]; */
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+16, sdatal+tx+16,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=  16 )
        {
/*                sdata[tx] += sdata[tx+8];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+8, sdatal+tx+8,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   8 )
        {
/*                       sdata[tx] += sdata[tx+4];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+4, sdatal+tx+4,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   4 )
        {
/*                       sdata[tx] += sdata[tx+2];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+2, sdatal+tx+2,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   2 )
        {
/*                sdata[tx] += sdata[tx+1]; */
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+1, sdatal+tx+1,
                  sdatah+tx, sdatal+tx );
        }
    }

    if( tx == 0 )
    {
/*            y[blockIdx.y*incy] =
  sdata[0]*(*alpha)+(*beta)*y[blockIdx.y*incy];*/

        *( xouth+blockIdx.y*incx ) = sdatah[0];
        *( xoutl+blockIdx.y*incx ) = sdatal[0];
    }
}

extern "C" __global__ void
gddtrmv_kernel_LT( char  diag,int n, double *Ah, double *Al, int lda, double * xinh,
                   double *xinl,
                   int incx,
                   double*xouth,
                   double*xoutl )
{
    int      tx = threadIdx.x;

    __shared__ double sdatah[TRANSBLOCK];
    __shared__ double sdatal[TRANSBLOCK];


    double            resh, resl;

    resh = resl = 0.0;

    int      n1 = ( blockIdx.y/TRANSBLOCK )*TRANSBLOCK+tx;

    Ah += n1;
    Al += n1;

    xinh += n1*incx;
    xinl += n1*incx;

    /* INCLUDING DIAGONAL ELEMENT */
    if( n1 >= blockIdx.y && n1 <n)
    {
        double ph, pl;

		int switcher = ((diag == 'u' || diag == 'U') && n1 == blockIdx.y);
		ph = switcher + !(switcher)*(*(Ah+lda*blockIdx.y));
		pl =  !(switcher)*(*(Al+lda*blockIdx.y));

/*                resh +=  A[tx + n1 + lda*blockIdx.y] * x[tx + n1];*/
        RMUL( &ph,&pl,
              xinh, xinl,
              &ph, &pl );
        RADD( &resh, &resl, &ph, &pl, &resh, &resl );

    }

    Ah += TRANSBLOCK;
    Al += TRANSBLOCK;

    xinh += TRANSBLOCK*incx;
    xinl += TRANSBLOCK*incx;




    n1 = TRANSBLOCK*( n/TRANSBLOCK-(blockIdx.y/TRANSBLOCK) - 1);


    for( int i = 0; i < n1; i += TRANSBLOCK )
    {
        double ph, pl;


/*                res += Ah[tx + i + lda*blockIdx.y] * x[tx+i];*/
        RMUL( Ah+i+lda*blockIdx.y, Al+i+lda*blockIdx.y,
              xinh+( i )*incx, xinl+( i )*incx,
              &ph, &pl );
        RADD( &resh, &resl, &ph, &pl, &resh, &resl );
    }

    int m = n%TRANSBLOCK;
    /* Last BUFFER */

    if( tx < m && blockIdx.y/TRANSBLOCK < (gridDim.y-1)/TRANSBLOCK )
    {


        double ph, pl;

/*                resh +=  A[tx + n1 + lda*blockIdx.y] * x[tx + n1];*/
        RMUL( Ah+n1+lda*blockIdx.y, Al+n1+lda*blockIdx.y,
              xinh+( n1 )*incx, xinl+( n1 )*incx,
              &ph, &pl );
        RADD( &resh, &resl, &ph, &pl, &resh, &resl );
    }



    sdatah[tx] = resh;
    sdatal[tx] = resl;


    __syncthreads( );




    if( TRANSBLOCK >= 512 )
    {
        if( tx < 256 )
        {
/*            sdata[tx] += sdata[tx+256];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+256, sdatal+tx+256,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }
    if( TRANSBLOCK >=        256 )
    {
        if( tx < 128 )
        {
/*            sdata[tx] += sdata[tx+128];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+128, sdatal+tx+128,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }
    if( TRANSBLOCK >=        128 )
    {
        if( tx <  64 )
        {
/*            sdata[tx] += sdata[tx+64];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+64, sdatal+tx+64,
                  sdatah+tx, sdatal+tx );
        }
        __syncthreads( );
    }

   if( tx < 32 )
   {
        if( TRANSBLOCK >=  64 )
        {
/*            sdata[tx] += sdata[tx+32];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+32, sdatal+tx+32,
                  sdatah+tx, sdatal+tx );
	    
        }
        if( TRANSBLOCK >=  32 )
        {
/*                sdata[tx] += sdata[tx+16]; */
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+16, sdatal+tx+16,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=  16 )
        {
/*                sdata[tx] += sdata[tx+8];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+8, sdatal+tx+8,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   8 )
        {
/*                       sdata[tx] += sdata[tx+4];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+4, sdatal+tx+4,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   4 )
        {
/*                       sdata[tx] += sdata[tx+2];*/
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+2, sdatal+tx+2,
                  sdatah+tx, sdatal+tx );
        }
        if( TRANSBLOCK >=   2 )
        {
/*                sdata[tx] += sdata[tx+1]; */
            RADD( sdatah+tx, sdatal+tx, sdatah+tx+1, sdatal+tx+1,
                  sdatah+tx, sdatal+tx );

        }
    }

    if( tx == 0 )
    {
/*            y[blockIdx.y*incy] =
  sdata[0]*(*alpha)+(*beta)*y[blockIdx.y*incy];*/

        *( xouth+blockIdx.y*incx ) = sdatah[0];
        *( xoutl+blockIdx.y*incx ) = sdatal[0];
    }
}

extern "C" void
gddtrmv( char*uplo, char*trans, char*diag, int *n,
              double *Ah, double *Al, int *lda, double * xh, double *xl,
              int *incx )
{
    bool    btrans = true;
    if( *trans == 'N' || *trans == 'n' ) btrans = false;

    bool    bupflag = true;
    if( *uplo == 'L' || *uplo == 'l' ) bupflag = false;

    double *dworkh, *dworkl;
    int     sizex = ( *n )*( *incx );
    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizex, sizeof( double ), (void**) &dworkh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizex, sizeof( double ), (void**) &dworkl )
      )
    {
        cublasFree( dworkh );
        cublasFree( dworkl );
        printf( "Device Malloc Error\n" );
        exit( -1 );
    }

    CUDA_SAFE_CALL( cudaMemcpy( dworkh, xh, sizeof( double )*sizex,
                                cudaMemcpyDeviceToDevice ) );
    CUDA_SAFE_CALL( cudaMemcpy( dworkl, xl, sizeof( double )*sizex,
                                cudaMemcpyDeviceToDevice ) );


    if( !btrans && !bupflag )
    {
        dim3 blocks( ( *n+BLOCKSIZE-1 )/BLOCKSIZE, 1 );

        dim3 threads( BLOCKSIZE, 1 );

        gddtrmv_kernel_LN<<< blocks, threads, 0, gddstream>>>
        ( *diag,*n, Ah, Al, *lda,  dworkh, dworkl, *incx, xh, xl );
    }
    else if( !btrans && bupflag )
    {
        dim3 blocks( ( *n+BLOCKSIZE-1 )/BLOCKSIZE, 1 );

        dim3 threads( BLOCKSIZE, 1 );

        gddtrmv_kernel_UN<<< blocks, threads, 0, gddstream>>>
        ( *diag,*n, Ah, Al, *lda,  dworkh, dworkl, *incx, xh, xl );
    }
    else if( btrans && !bupflag )
    {
        dim3 blocks( 1, *n );

        dim3 threads( TRANSBLOCK, 1 );

        gddtrmv_kernel_LT<<< blocks, threads, 0, gddstream>>>
        ( *diag, *n, Ah, Al, *lda,  dworkh, dworkl, *incx, xh, xl );

    }
    else
    {
        dim3 blocks( 1, *n );

        dim3 threads( TRANSBLOCK, 1 );

        gddtrmv_kernel_UT<<< blocks, threads, 0, gddstream>>>
        ( *diag,*n, Ah, Al, *lda,  dworkh, dworkl, *incx, xh, xl );


    }
    cublasFree( dworkh );
    cublasFree( dworkl );
}

