#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"

#define qmod(a,b) ((a)-(__mul24((b),(a)/(b))))

#define b_copy();        dim3 dimBlock((M>=MAX_THREAD_PER_BLOCK)?MAX_THREAD_PER_BLOCK:(WARP_SIZE*((M/WARP_SIZE)+(M%WARP_SIZE!=0))), 1);	\
  dim3 dimGrid(M/dimBlock.x+(M%dimBlock.x!=0), N);			\
  b_copy_kernel<<< dimGrid, dimBlock, 0, gddstream >>>(M, N, bh, ldb, d_xh, M); \
  b_copy_kernel<<< dimGrid, dimBlock, 0, gddstream >>>(M, N, bl, ldb, d_xl, M); \
  cudaThreadSynchronize();

#define MAX_THREAD_PER_BLOCK 512
#define WARP_SIZE 32

#define BLOCK_SIZE 16 // inner blocking size, <=32
#define NB 128// outer blocking size, >BLOCK_SIZE 

static void gddgemm
( const char& transA, const char& transB,
  const int& m, const int& n, const int& k,
  double* alphah, double* alphal,
  double Ah[], double Al[], const int& lda,
  double Bh[], double Bl[], const int& ldb,
  double* betah, double* betal, 
  double Ch[], double Cl[], const int& ldc )
{
  gddgemm
    ( (char*)&transA, (char*)&transB,
      (int*)&m, (int*)&n, (int*)&k,
      (double*)alphah, (double*)alphal,
      (double*)Ah, (double*)Al, (int*)&lda,
      (double*)Bh, (double*)Bl, (int*)&ldb,
      (double*)betah, (double*)betal, 
      (double*)Ch, (double*)Cl, (int*)&ldc );
}


__global__ void
diag_strtri_kernel_upper (char diag, double *Ah, double *Al, double *d_dinvAh, double *d_dinvAl, int lda)
{
  int i,j;
  double Ystxh=0;
  double Ystxl=0;
  double *yh=NULL, *Aoffh=NULL;
  double *yl=NULL, *Aoffl=NULL;
  int switcher=0;
  double done = 1.0,dzero = 0.0;

  // Thread index
  int tx = threadIdx.x;

  // Block index
  int bx = blockIdx.x;
                
  Aoffh = Ah+bx*lda*BLOCK_SIZE+bx*BLOCK_SIZE;
  Aoffl = Al+bx*lda*BLOCK_SIZE+bx*BLOCK_SIZE;
  int NumBLperNB = NB/BLOCK_SIZE;
  d_dinvAh += bx/NumBLperNB*NB*NB+(bx%NumBLperNB)*(NB*BLOCK_SIZE+BLOCK_SIZE);
  d_dinvAl += bx/NumBLperNB*NB*NB+(bx%NumBLperNB)*(NB*BLOCK_SIZE+BLOCK_SIZE);

  __shared__ double Bsh[BLOCK_SIZE*BLOCK_SIZE];
  __shared__ double Bsl[BLOCK_SIZE*BLOCK_SIZE];
  __shared__ double workspaceh[BLOCK_SIZE]; // workspace used to store the current working column
  __shared__ double workspacel[BLOCK_SIZE]; // workspace used to store the current working column

  // load A
#pragma unroll
  for (i=0; i<BLOCK_SIZE; i++){
    Bsh[i*BLOCK_SIZE+tx] = ((double)(tx<=i))*(*(Aoffh+i*lda+tx));        // read in the whole square block of my A and zero out the non data triangular
/*    printf("%e,\n",Bsh[j*BLOCK_SIZE+tx]);*/
    Bsl[i*BLOCK_SIZE+tx] = ((double)(tx<=i))*(*(Aoffl+i*lda+tx));        // read in the whole square block of my A and zero out the non data triangular
  }

  // Synchronize to make sure the matrices are loaded
  __syncthreads();

  switcher = (diag=='u' || diag=='U');
  int diagsw = (Bsh[tx*BLOCK_SIZE+tx]==0 && Bsl[tx*BLOCK_SIZE+tx]==0);

  if(switcher || diagsw){
  Bsh[tx*BLOCK_SIZE+tx] =1.0;
  Bsl[tx*BLOCK_SIZE+tx] =0.0;
  }
  else 
  {
	  RDIV(&done,&dzero,&Bsh[tx*BLOCK_SIZE+tx],&Bsl[tx*BLOCK_SIZE+tx],&Bsh[tx*BLOCK_SIZE+tx],&Bsl[tx*BLOCK_SIZE+tx]);
  
  }

/*  Bsh[tx*BLOCK_SIZE+tx] = switcher+!switcher*(1/(diagsw+(!diagsw)*Bsh[tx*BLOCK_SIZE+tx]));        // solve the diagonals*/
/*  Bsl[tx*BLOCK_SIZE+tx] = !switcher*(1/(diagsw+(!diagsw)*Bsl[tx*BLOCK_SIZE+tx]));        // solve the diagonals*/

  /* the upper case */
  for (i=0; i<BLOCK_SIZE; i++){
    Ystxh = 0;
    Ystxl = 0;
    switcher = (tx<i);

    //strmv
    workspaceh[tx] = *(Bsh+i*BLOCK_SIZE+tx);
    workspacel[tx] = *(Bsl+i*BLOCK_SIZE+tx);
    yh = Bsh+i*BLOCK_SIZE;
    yl = Bsl+i*BLOCK_SIZE;

#pragma unroll
    for (j=0; j<i; j++){
      double temph, templ;
      RMUL( &Bsh[j*BLOCK_SIZE+tx], &Bsl[j*BLOCK_SIZE+tx],
	    &workspaceh[j], &workspacel[j], &temph, &templ );
/*    printf("%e,%e,\n",Bsh[j*BLOCK_SIZE+tx],workspaceh[j]);*/
      temph = switcher*temph;
      templ = switcher*templ;

/*    printf("%e,\n",Ystxh);*/
      RADD( &temph, &templ, &Ystxh, &Ystxl, &Ystxh, &Ystxl );
    }

    //sscal
    switcher = (tx != i);

    {
      double temp1h, temp1l;
      double temp2h, temp2l;
      double temp3h, temp3l;

      temp1h = (-1.0)*Bsh[i*BLOCK_SIZE+i];
      temp1l = (-1.0)*Bsl[i*BLOCK_SIZE+i];

      RMUL( &temp1h, &temp1l, &Ystxh, &Ystxl, &temp2h, &temp2l );
/*    printf("%e,\n",Ystxh);*/

      temp2h = switcher*temp2h;
      temp2l = switcher*temp2l;

      temp3h = !switcher*yh[tx];
      temp3l = !switcher*yl[tx];

      RADD( &temp2h, &temp2l, &temp3h, &temp3l, &yh[tx], &yl[tx] );

      //      yh[tx] = switcher*Ystxh*(-Bsh[i*BLOCK_SIZE+i])+!switcher*yh[tx];
      //      yl[tx] = switcher*Ystxl*(-Bsl[i*BLOCK_SIZE+i])+!switcher*yl[tx];
    }

    __syncthreads();
  }

  // write back A
#pragma unroll
  for (i=0; i<BLOCK_SIZE; i++){
    *(d_dinvAh+i*NB+tx) = Bsh[i*BLOCK_SIZE+tx];
    *(d_dinvAl+i*NB+tx) = Bsl[i*BLOCK_SIZE+tx];
/*    printf("%e,\n",*(d_dinvAh+i*NB+tx));*/
  }

}
__global__ void
diag_strtri_kernel_lower (char diag, double *Ah, double *Al, double *d_dinvAh, double *d_dinvAl, int lda)
{
  int i,j;
  double Ystxh=0;
  double Ystxl=0;
  double *Bwh=NULL, *xh=NULL, *yh=NULL, *Aoffh=NULL;
  double *Bwl=NULL, *xl=NULL, *yl=NULL, *Aoffl=NULL;
  int switcher=0;

  // Thread index
  int tx = threadIdx.x;
  int txw;

  // Block index
  int bx = blockIdx.x;
                
  Aoffh = Ah+bx*lda*BLOCK_SIZE+bx*BLOCK_SIZE;
  Aoffl = Al+bx*lda*BLOCK_SIZE+bx*BLOCK_SIZE;

  int NumBLperNB = NB/BLOCK_SIZE;
  d_dinvAh += bx/NumBLperNB*NB*NB+(bx%NumBLperNB)*(NB*BLOCK_SIZE+BLOCK_SIZE);
  d_dinvAl += bx/NumBLperNB*NB*NB+(bx%NumBLperNB)*(NB*BLOCK_SIZE+BLOCK_SIZE);

  __shared__ double Bsh[BLOCK_SIZE*BLOCK_SIZE];
  __shared__ double Bsl[BLOCK_SIZE*BLOCK_SIZE];
  __shared__ double workspaceh[BLOCK_SIZE]; // workspace used to store the current working column
  __shared__ double workspacel[BLOCK_SIZE]; // workspace used to store the current working column

  // load A
#pragma unroll
  for (i=0; i<BLOCK_SIZE; i++){
    Bsh[i*BLOCK_SIZE+tx] = ((double)(tx>=i))*(*(Aoffh+i*lda+tx));        // read in the whole square block of my A and zero out the non data triangular
    Bsl[i*BLOCK_SIZE+tx] = ((double)(tx>=i))*(*(Aoffl+i*lda+tx));        // read in the whole square block of my A and zero out the non data triangular
  }

  // not the upper or lower diagonal
  // Synchronize to make sure the matrices are loaded
  __syncthreads();

  switcher = (diag=='u' || diag=='U');
  int diagsw = (Bsh[tx*BLOCK_SIZE+tx]==0 && Bsl[tx*BLOCK_SIZE+tx]==0);

  {
    double work1h, work1l;
    double work2h, work2l;
    double work3h, work3l;
    double work4h, work4l;
    double work5h, work5l;
    double work6h, work6l;

    work1h = (!diagsw)*Bsh[tx*BLOCK_SIZE+tx];
    work1l = (!diagsw)*Bsl[tx*BLOCK_SIZE+tx];

    work2h = diagsw ? 1.0 : 0.0;
    work2l = 0.0;

    RADD( &work2h, &work2l, &work1h, &work1l, &work3h, &work3l );
    {
    double done  = 1.0;
    double dzero = 0.0;
    RDIV( &done,&dzero,&work3h, &work3l, &work4h, &work4l );
    }

    work5h = !switcher*work4h;
    work5l = !switcher*work4l;

    work6h = switcher ? 1.0 : 0.0;
    work6l = 0.0;

    RADD( &work6h, &work6l, &work5h, &work5l, &Bsh[tx*BLOCK_SIZE+tx], &Bsl[tx*BLOCK_SIZE+tx] );

    //    Bsh[tx*BLOCK_SIZE+tx] = switcher+!switcher*(1/(diagsw+(!diagsw)*Bsh[tx*BLOCK_SIZE+tx]));        // solve the diagonals
    //    Bsl[tx*BLOCK_SIZE+tx] = switcher+!switcher*(1/(diagsw+(!diagsw)*Bsl[tx*BLOCK_SIZE+tx]));        // solve the diagonals
  }

  /*
   * the lower case
   */

  switcher = !(tx<BLOCK_SIZE-1);
  Bsh[(BLOCK_SIZE-1)*BLOCK_SIZE+tx] = (double)switcher*Bsh[(BLOCK_SIZE-1)*BLOCK_SIZE+tx];        //zero out the last column, except the diagonal element
  Bsl[(BLOCK_SIZE-1)*BLOCK_SIZE+tx] = (double)switcher*Bsl[(BLOCK_SIZE-1)*BLOCK_SIZE+tx];        //zero out the last column, except the diagonal element

  for (i=BLOCK_SIZE-2; i>=0; i--){
    Ystxh = 0;
    Ystxl = 0;
    switcher = (tx>i);

    //strmv
    Bwh = Bsh+(i+1)*BLOCK_SIZE+i+1;
    Bwl = Bsl+(i+1)*BLOCK_SIZE+i+1;
    workspaceh[tx] = *(Bsh+i*BLOCK_SIZE+tx);
    workspacel[tx] = *(Bsl+i*BLOCK_SIZE+tx);
    xh = workspaceh+i+1;
    xl = workspacel+i+1;
    yh = Bsh+i*BLOCK_SIZE;
    yl = Bsl+i*BLOCK_SIZE;

    txw = (tx-i-1);

#pragma unroll
    for (j=0; j<BLOCK_SIZE-i-1; j++){
      double temp1h, temp1l;
      RMUL( &Bwh[j*BLOCK_SIZE+txw], &Bwl[j*BLOCK_SIZE+txw], &xh[j], &xl[j], &temp1h, &temp1l );

      temp1h = switcher*temp1h;
      temp1l = switcher*temp1l;

      RADD( &temp1h, &temp1l, &Ystxh, &Ystxl, &Ystxh, &Ystxl );
      //      Ystxh += (double)switcher*(Bwh[j*BLOCK_SIZE+txw]*xh[j]);
      //      Ystxl += (double)switcher*(Bwl[j*BLOCK_SIZE+txw]*xl[j]);
    }

    //sscal
    switcher = (tx != i); 

    {
      double temp1h, temp1l;
      double temp2h, temp2l;
      double temp3h, temp3l;

      temp1h = (-1.0)*Bsh[i*BLOCK_SIZE+i];
      temp1l = (-1.0)*Bsl[i*BLOCK_SIZE+i];

      RMUL( &temp1h, &temp1l, &Ystxh, &Ystxl, &temp2h, &temp2l );

      temp2h = switcher*temp2h;
      temp2l = switcher*temp2l;

      temp3h = !switcher*yh[tx];
      temp3l = !switcher*yl[tx];

      RADD( &temp2h, &temp2l, &temp3h, &temp3l, &yh[tx], &yl[tx] );

      //      yh[tx] = switcher*Ystxh*(-Bsh[i*BLOCK_SIZE+i])+!switcher*yh[tx];
      //      yl[tx] = switcher*Ystxl*(-Bsl[i*BLOCK_SIZE+i])+!switcher*yl[tx];
    }

    __syncthreads();
  }

  // write back A
#pragma unroll
  for (i=0; i<BLOCK_SIZE; i++){
    *(d_dinvAh+i*NB+tx) = Bsh[i*BLOCK_SIZE+tx];
    *(d_dinvAl+i*NB+tx) = Bsl[i*BLOCK_SIZE+tx];
  }
}

__device__ void ssaxpy( double ah, double al, double *bh, double *bl, double *ch, double *cl )
{
  double workh, workl;

  for( int i=0; i<16; i++ ){
    RMUL( &ah, &al, &bh[i], &bl[i], &workh, &workl );
    RADD( &workh, &workl, &ch[i], &cl[i], &ch[i], &cl[i] );
  }
}

__device__ void sgemm_kernel_16
( double *Ah, double *Al, int lda,
  double *Bh, double *Bl, int ldb,
  double *Ch, double *Cl, int ldc,
  double* alphah, double* alphal, int blk, int inx, int iny,
  double *ch, double *cl )
{

  const double *Blasth = Bh + blk;
  const double *Blastl = Bl + blk;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  do{
    double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
    double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

    bsh[inx][iny]    = Bh[0*ldb];
    bsh[inx][iny+4]  = Bh[4*ldb];
    bsh[inx][iny+8]  = Bh[8*ldb];
    bsh[inx][iny+12] = Bh[12*ldb];
    bsh[inx+4][iny]    = Bh[4+0*ldb];
    bsh[inx+4][iny+4]  = Bh[4+4*ldb];
    bsh[inx+4][iny+8]  = Bh[4+8*ldb];
    bsh[inx+4][iny+12] = Bh[4+12*ldb];
    bsh[inx+8][iny]    = Bh[8+0*ldb];
    bsh[inx+8][iny+4]  = Bh[8+4*ldb];
    bsh[inx+8][iny+8]  = Bh[8+8*ldb];
    bsh[inx+8][iny+12] = Bh[8+12*ldb];
    bsh[inx+12][iny]    = Bh[12+0*ldb];
    bsh[inx+12][iny+4]  = Bh[12+4*ldb];
    bsh[inx+12][iny+8]  = Bh[12+8*ldb];
    bsh[inx+12][iny+12] = Bh[12+12*ldb];

    bsl[inx][iny]    = Bl[0*ldb];
    bsl[inx][iny+4]  = Bl[4*ldb];
    bsl[inx][iny+8]  = Bl[8*ldb];
    bsl[inx][iny+12] = Bl[12*ldb];
    bsl[inx+4][iny]    = Bl[4+0*ldb];
    bsl[inx+4][iny+4]  = Bl[4+4*ldb];
    bsl[inx+4][iny+8]  = Bl[4+8*ldb];
    bsl[inx+4][iny+12] = Bl[4+12*ldb];
    bsl[inx+8][iny]    = Bl[8+0*ldb];
    bsl[inx+8][iny+4]  = Bl[8+4*ldb];
    bsl[inx+8][iny+8]  = Bl[8+8*ldb];
    bsl[inx+8][iny+12] = Bl[8+12*ldb];
    bsl[inx+12][iny]    = Bl[12+0*ldb];
    bsl[inx+12][iny+4]  = Bl[12+4*ldb];
    bsl[inx+12][iny+8]  = Bl[12+8*ldb];
    bsl[inx+12][iny+12] = Bl[12+12*ldb];
    __syncthreads();

    Ah += 4*lda;
    Al += 4*lda;
    ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
    ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
    ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
    ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

    Ah += 4*lda;
    Al += 4*lda;
    ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
    ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
    ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
    ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

    Ah += 4*lda;
    Al += 4*lda;
    ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
    ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
    ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
    ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

    Ah += 4*lda;
    Al += 4*lda;
    ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
    ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
    ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
    ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

    Bh += 16;
    Bl += 16;

    __syncthreads();
  } while( Bh < Blasth && Bl < Blastl );

  for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
    RMUL( alphah, alphal, &ch[i], &cl[i], &Ch[0], &Cl[0] );
  }
}

/*
 * B21 = -inv(A11)*A12*inv(A22)
 */
#define qmod(a,b) ((a)-(__mul24((b),(a)/(b))))
__global__ void
triple_sgemm_update_16_R (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  //        const int page = (blockIdx.y)%(npages);
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x * (blockDim.x*blockDim.y);
  const int iby = bIdy * 16;
  const int id = inx + iny*blockDim.x;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  int PagesPerNB = NB/(blk*2);
  //--------------------------part one---------------------------//
  {
    // A12*inv(A22) -> A12
    // A=A12, B=inv(A22), C=A12(d_dinvA)
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int ldb = NB;
    int ldc = NB;

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = Ahin + page*lda*blk*2 + blk*lda + page*blk*2;  
    Bh = d_dinvAh + blk*NB + blk;
    Ch = d_dinvAh + blk*NB;
    Al = Alin + page*lda*blk*2 + blk*lda + page*blk*2;  
    Bl = d_dinvAl + blk*NB + blk;
    Cl = d_dinvAl + blk*NB;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );
    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];
      bsh[inx+4][iny]    = Bh[4+0*ldb];
      bsh[inx+4][iny+4]  = Bh[4+4*ldb];
      bsh[inx+4][iny+8]  = Bh[4+8*ldb];
      bsh[inx+4][iny+12] = Bh[4+12*ldb];
      bsh[inx+8][iny]    = Bh[8+0*ldb];
      bsh[inx+8][iny+4]  = Bh[8+4*ldb];
      bsh[inx+8][iny+8]  = Bh[8+8*ldb];
      bsh[inx+8][iny+12] = Bh[8+12*ldb];
      bsh[inx+12][iny]    = Bh[12+0*ldb];
      bsh[inx+12][iny+4]  = Bh[12+4*ldb];
      bsh[inx+12][iny+8]  = Bh[12+8*ldb];
      bsh[inx+12][iny+12] = Bh[12+12*ldb];

      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];
      bsl[inx+4][iny]    = Bl[4+0*ldb];
      bsl[inx+4][iny+4]  = Bl[4+4*ldb];
      bsl[inx+4][iny+8]  = Bl[4+8*ldb];
      bsl[inx+4][iny+12] = Bl[4+12*ldb];
      bsl[inx+8][iny]    = Bl[8+0*ldb];
      bsl[inx+8][iny+4]  = Bl[8+4*ldb];
      bsl[inx+8][iny+8]  = Bl[8+8*ldb];
      bsl[inx+8][iny+12] = Bl[8+12*ldb];
      bsl[inx+12][iny]    = Bl[12+0*ldb];
      bsl[inx+12][iny+4]  = Bl[12+4*ldb];
      bsl[inx+12][iny+8]  = Bl[12+8*ldb];
      bsl[inx+12][iny+12] = Bl[12+12*ldb];

      __syncthreads();


      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = ch[i];
      Cl[0] = cl[i];
    }
  }
                        
  __syncthreads();
        
  //--------------------------part two---------------------------//
  {
    // -inv(A11)*A12 -> A12
    // A=inv(A11), B=A12, C=A12
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int lda = NB;
    int ldb = NB;
    int ldc = NB;

    Ah = d_dinvAh;
    Bh = Ch = d_dinvAh + blk*NB;
    Al = d_dinvAl;
    Bl = Cl = d_dinvAl + blk*NB;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );

    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];
      bsh[inx+4][iny]    = Bh[4+0*ldb];
      bsh[inx+4][iny+4]  = Bh[4+4*ldb];
      bsh[inx+4][iny+8]  = Bh[4+8*ldb];
      bsh[inx+4][iny+12] = Bh[4+12*ldb];
      bsh[inx+8][iny]    = Bh[8+0*ldb];
      bsh[inx+8][iny+4]  = Bh[8+4*ldb];
      bsh[inx+8][iny+8]  = Bh[8+8*ldb];
      bsh[inx+8][iny+12] = Bh[8+12*ldb];
      bsh[inx+12][iny]    = Bh[12+0*ldb];
      bsh[inx+12][iny+4]  = Bh[12+4*ldb];
      bsh[inx+12][iny+8]  = Bh[12+8*ldb];
      bsh[inx+12][iny+12] = Bh[12+12*ldb];

      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];
      bsl[inx+4][iny]    = Bl[4+0*ldb];
      bsl[inx+4][iny+4]  = Bl[4+4*ldb];
      bsl[inx+4][iny+8]  = Bl[4+8*ldb];
      bsl[inx+4][iny+12] = Bl[4+12*ldb];
      bsl[inx+8][iny]    = Bl[8+0*ldb];
      bsl[inx+8][iny+4]  = Bl[8+4*ldb];
      bsl[inx+8][iny+8]  = Bl[8+8*ldb];
      bsl[inx+8][iny+12] = Bl[8+12*ldb];
      bsl[inx+12][iny]    = Bl[12+0*ldb];
      bsl[inx+12][iny+4]  = Bl[12+4*ldb];
      bsl[inx+12][iny+8]  = Bl[12+8*ldb];
      bsl[inx+12][iny+12] = Bl[12+12*ldb];

      __syncthreads();


      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = (-1.0)*ch[i];
      Cl[0] = (-1.0)*cl[i];
    }
  }
}

/*
 * B21 = -inv(A22)*A21*inv(A11)
 */
#define qmod(a,b) ((a)-(__mul24((b),(a)/(b))))
__global__ void
triple_sgemm_update_16_part1_L (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  //        const int page = (blockIdx.y)%(npages);
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x * (blockDim.x*blockDim.y);
  const int iby = bIdy * 16;
  const int id = inx + iny*blockDim.x;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  //--------------------------part one---------------------------//
  {
    // A21*inv(A11) -> A21
    // A=A21, B=inv(A11), C=A21
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int ldb = NB;
    int ldc = NB;

    int PagesPerNB = NB/(blk*2);

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = Ahin + page*lda*blk*2 + page*blk*2 + blk;  
    Bh = d_dinvAh; 
    Ch = d_dinvAh + blk;
    Al = Alin + page*lda*blk*2 + page*blk*2 + blk;  
    Bl = d_dinvAl; 
    Cl = d_dinvAl + blk;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );

    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];
      bsh[inx+4][iny]    = Bh[4+0*ldb];
      bsh[inx+4][iny+4]  = Bh[4+4*ldb];
      bsh[inx+4][iny+8]  = Bh[4+8*ldb];
      bsh[inx+4][iny+12] = Bh[4+12*ldb];
      bsh[inx+8][iny]    = Bh[8+0*ldb];
      bsh[inx+8][iny+4]  = Bh[8+4*ldb];
      bsh[inx+8][iny+8]  = Bh[8+8*ldb];
      bsh[inx+8][iny+12] = Bh[8+12*ldb];
      bsh[inx+12][iny]    = Bh[12+0*ldb];
      bsh[inx+12][iny+4]  = Bh[12+4*ldb];
      bsh[inx+12][iny+8]  = Bh[12+8*ldb];
      bsh[inx+12][iny+12] = Bh[12+12*ldb];

      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];
      bsl[inx+4][iny]    = Bl[4+0*ldb];
      bsl[inx+4][iny+4]  = Bl[4+4*ldb];
      bsl[inx+4][iny+8]  = Bl[4+8*ldb];
      bsl[inx+4][iny+12] = Bl[4+12*ldb];
      bsl[inx+8][iny]    = Bl[8+0*ldb];
      bsl[inx+8][iny+4]  = Bl[8+4*ldb];
      bsl[inx+8][iny+8]  = Bl[8+8*ldb];
      bsl[inx+8][iny+12] = Bl[8+12*ldb];
      bsl[inx+12][iny]    = Bl[12+0*ldb];
      bsl[inx+12][iny+4]  = Bl[12+4*ldb];
      bsl[inx+12][iny+8]  = Bl[12+8*ldb];
      bsl[inx+12][iny+12] = Bl[12+12*ldb];

      __syncthreads();


      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = ch[i];
      Cl[0] = cl[i];
    }
  }
                        
  __syncthreads();
        
}

/*
 * B21 = -inv(A22)*A21*inv(A11)
 */
#define qmod(a,b) ((a)-(__mul24((b),(a)/(b))))
__global__ void
triple_sgemm_update_16_part2_L (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x * (blockDim.x*blockDim.y);
  const int iby = bIdy * 16;
  const int id = inx + iny*blockDim.x;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];
        
  //--------------------------part two---------------------------//
  {
    // -inv(A22)*A21 -> A21
    // A=inv(A22), B=A21, C=A21
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int lda = NB;
    int ldb = NB;
    int ldc = NB;

    int PagesPerNB = NB/(blk*2);
    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = d_dinvAh + blk*NB + blk;
    Bh = Ch = d_dinvAh + blk;
    Al = d_dinvAl + blk*NB + blk;
    Bl = Cl = d_dinvAl + blk;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );

    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];
      bsh[inx+4][iny]    = Bh[4+0*ldb];
      bsh[inx+4][iny+4]  = Bh[4+4*ldb];
      bsh[inx+4][iny+8]  = Bh[4+8*ldb];
      bsh[inx+4][iny+12] = Bh[4+12*ldb];
      bsh[inx+8][iny]    = Bh[8+0*ldb];
      bsh[inx+8][iny+4]  = Bh[8+4*ldb];
      bsh[inx+8][iny+8]  = Bh[8+8*ldb];
      bsh[inx+8][iny+12] = Bh[8+12*ldb];
      bsh[inx+12][iny]    = Bh[12+0*ldb];
      bsh[inx+12][iny+4]  = Bh[12+4*ldb];
      bsh[inx+12][iny+8]  = Bh[12+8*ldb];
      bsh[inx+12][iny+12] = Bh[12+12*ldb];


      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];
      bsl[inx+4][iny]    = Bl[4+0*ldb];
      bsl[inx+4][iny+4]  = Bl[4+4*ldb];
      bsl[inx+4][iny+8]  = Bl[4+8*ldb];
      bsl[inx+4][iny+12] = Bl[4+12*ldb];
      bsl[inx+8][iny]    = Bl[8+0*ldb];
      bsl[inx+8][iny+4]  = Bl[8+4*ldb];
      bsl[inx+8][iny+8]  = Bl[8+8*ldb];
      bsl[inx+8][iny+12] = Bl[8+12*ldb];
      bsl[inx+12][iny]    = Bl[12+0*ldb];
      bsl[inx+12][iny+4]  = Bl[12+4*ldb];
      bsl[inx+12][iny+8]  = Bl[12+8*ldb];
      bsl[inx+12][iny+12] = Bl[12+12*ldb];

      __syncthreads();


      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = (-1)*ch[i];
      Cl[0] = (-1)*cl[i];
    }
  }
  __syncthreads();
}

/*
 * B21 = -inv(A11)*A12*inv(A22)
 */
__global__ void
triple_sgemm_update_32_part1_R (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x * (blockDim.x*blockDim.y);
  const int iby = bIdy * 16;
  const int id = inx + iny*blockDim.x;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  int PagesPerNB = NB/(blk*2);
  //--------------------------part one---------------------------//
  {
    // A12*inv(A22) -> A21
    // A=A12, B=inv(A22), C=A12(d_dinvA)
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int ldb = NB;
    int ldc = NB;

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = Ahin + page*lda*blk*2 + blk*lda + page*blk*2;  
    Bh = d_dinvAh + blk*NB + blk;
    Ch = d_dinvAh + blk*NB;
    Al = Alin + page*lda*blk*2 + blk*lda + page*blk*2;  
    Bl = d_dinvAl + blk*NB + blk;
    Cl = d_dinvAl + blk*NB;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );
    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];
      bsh[inx+8][iny]    = Bh[8+0*ldb];
      bsh[inx+8][iny+4]  = Bh[8+4*ldb];
      bsh[inx+8][iny+8]  = Bh[8+8*ldb];
      bsh[inx+8][iny+12] = Bh[8+12*ldb];

      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];
      bsl[inx+8][iny]    = Bl[8+0*ldb];
      bsl[inx+8][iny+4]  = Bl[8+4*ldb];
      bsl[inx+8][iny+8]  = Bl[8+8*ldb];
      bsl[inx+8][iny+12] = Bl[8+12*ldb];

      __syncthreads();


      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();

    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = ch[i];
      Cl[0] = cl[i];
    }
  }
                        
  __syncthreads();
}
/*
 * B21 = -inv(A11)*A12*inv(A22)
 */
__global__ void
triple_sgemm_update_32_part2_R (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x * (blockDim.x*blockDim.y);
  const int iby = bIdy * 16;
  const int id = inx + iny*blockDim.x;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  int PagesPerNB = NB/(blk*2);
        
  //--------------------------part two---------------------------//
  {
    // -inv(A11)*A12 -> A12
    // A=inv(A11), B=A12, C=A12
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int lda = NB;
    int ldb = NB;
    int ldc = NB;

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = d_dinvAh;
    Bh = Ch = d_dinvAh + blk*NB;
    Al = d_dinvAl;
    Bl = Cl = d_dinvAl + blk*NB;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );
    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];
      bsh[inx+8][iny]    = Bh[8+0*ldb];
      bsh[inx+8][iny+4]  = Bh[8+4*ldb];
      bsh[inx+8][iny+8]  = Bh[8+8*ldb];
      bsh[inx+8][iny+12] = Bh[8+12*ldb];

      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];
      bsl[inx+8][iny]    = Bl[8+0*ldb];
      bsl[inx+8][iny+4]  = Bl[8+4*ldb];
      bsl[inx+8][iny+8]  = Bl[8+8*ldb];
      bsl[inx+8][iny+12] = Bl[8+12*ldb];
      __syncthreads();


      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = (-1)*ch[i];
      Cl[0] = (-1)*cl[i];
    }
  }
}
/*
 * B21 = -inv(A22)*A21*inv(A11)
 */
__global__ void
triple_sgemm_update_32_part1_L (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x * (blockDim.x*blockDim.y);
  const int iby = bIdy * 16;
  const int id = inx + iny*blockDim.x;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  int PagesPerNB = NB/(blk*2);
  //--------------------------part one---------------------------//
  {
    // A21*inv(A11) -> A21
    // A=A21, B=inv(A11), C=A21
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int ldb = NB;
    int ldc = NB;

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = Ahin + page*lda*blk*2 + page*blk*2 + blk;  
    Bh = d_dinvAh;
    Ch = d_dinvAh + blk;
    Al = Alin + page*lda*blk*2 + page*blk*2 + blk;  
    Bl = d_dinvAl;
    Cl = d_dinvAl + blk;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );
    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];
      bsh[inx+8][iny]    = Bh[8+0*ldb];
      bsh[inx+8][iny+4]  = Bh[8+4*ldb];
      bsh[inx+8][iny+8]  = Bh[8+8*ldb];
      bsh[inx+8][iny+12] = Bh[8+12*ldb];

      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];
      bsl[inx+8][iny]    = Bl[8+0*ldb];
      bsl[inx+8][iny+4]  = Bl[8+4*ldb];
      bsl[inx+8][iny+8]  = Bl[8+8*ldb];
      bsl[inx+8][iny+12] = Bl[8+12*ldb];

      __syncthreads();

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = ch[i];
      Cl[0] = cl[i];
    }
  }
                        
  __syncthreads();
}
/*
 * B21 = -inv(A22)*A21*inv(A11)
 */
__global__ void
triple_sgemm_update_32_part2_L (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x * (blockDim.x*blockDim.y);
  const int iby = bIdy * 16;
  const int id = inx + iny*blockDim.x;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  int PagesPerNB = NB/(blk*2);
  //--------------------------part two---------------------------//
  {
    // -inv(A22)*A21 -> A21
    // A=inv(A22), B=A21, C=A21
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int lda = NB;
    int ldb = NB;
    int ldc = NB;

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = d_dinvAh + blk*NB + blk;
    Bh = Ch = d_dinvAh + blk;
    Al = d_dinvAl + blk*NB + blk;
    Bl = Cl = d_dinvAl + blk;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );
    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];
      bsh[inx+8][iny]    = Bh[8+0*ldb];
      bsh[inx+8][iny+4]  = Bh[8+4*ldb];
      bsh[inx+8][iny+8]  = Bh[8+8*ldb];
      bsh[inx+8][iny+12] = Bh[8+12*ldb];

      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];
      bsl[inx+8][iny]    = Bl[8+0*ldb];
      bsl[inx+8][iny+4]  = Bl[8+4*ldb];
      bsl[inx+8][iny+8]  = Bl[8+8*ldb];
      bsl[inx+8][iny+12] = Bl[8+12*ldb];

      __syncthreads();

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = (-1)*ch[i];
      Cl[0] = (-1)*cl[i];
    }
  }
}

/*
 * B21 = -inv(A11)*A12*inv(A22)
 */
__global__ void
triple_sgemm_update_64_part1_R (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x*64;
  const int iby = bIdy*16;
  const int id = inx + iny*16;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  int PagesPerNB = NB/(blk*2);
  //--------------------------part one---------------------------//
  {
    // A12*inv(A22) -> A12(d_dinvA)
    // A=A12, B=inv(A22), C=A12
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int ldb = NB;
    int ldc = NB;

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = Ahin + page*lda*blk*2 + blk*lda + page*blk*2;  
    Bh = d_dinvAh + blk*NB + blk;
    Ch = d_dinvAh + blk*NB;
    Al = Alin + page*lda*blk*2 + blk*lda + page*blk*2;  
    Bl = d_dinvAl + blk*NB + blk;
    Cl = d_dinvAl + blk*NB;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );
    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];
      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];

      __syncthreads();

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = ch[i];
      Cl[0] = cl[i];
    }
  }
                        
}
/*
 * B21 = -inv(A11)*A12*inv(A22)
 */
__global__ void
triple_sgemm_update_64_part2_R (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x*64;
  const int iby = bIdy*16;
  const int id = inx + iny*16;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  int PagesPerNB = NB/(blk*2);
                        
  //--------------------------part two---------------------------//
  {
    // -inv(A11)*A12 -> A12
    // A=inv(A11), B=A12, C=A12
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int lda = NB;
    int ldb = NB;
    int ldc = NB;

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = d_dinvAh;
    Bh = Ch = d_dinvAh + blk*NB;
    Al = d_dinvAl;
    Bl = Cl = d_dinvAl + blk*NB;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );
    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];

      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];

      __syncthreads();

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = (-1)*ch[i];
      Cl[0] = (-1)*cl[i];
    }
  }
}
/*
 * B21 = -inv(A22)*A21*inv(A11)
 */
__global__ void
triple_sgemm_update_64_part1_L (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x*64;
  const int iby = bIdy*16;
  const int id = inx + iny*16;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  int PagesPerNB = NB/(blk*2);
  //--------------------------part one---------------------------//
  {
    // A21*inv(A11) -> A21
    // A=A21, B=inv(A11), C=A21
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int ldb = NB;
    int ldc = NB;

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = Ahin + page*lda*blk*2 + page*blk*2 + blk;  
    Bh = d_dinvAh;
    Ch = d_dinvAh + blk;
    Al = Alin + page*lda*blk*2 + page*blk*2 + blk;  
    Bl = d_dinvAl;
    Cl = d_dinvAl + blk;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );

    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];
      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];

      __syncthreads();

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = ch[i];
      Cl[0] = cl[i];
    }
  }
}
/*
 * B21 = -inv(A22)*A21*inv(A11)
 */
__global__ void
triple_sgemm_update_64_part2_L (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x*64;
  const int iby = bIdy*16;
  const int id = inx + iny*16;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  int PagesPerNB = NB/(blk*2);
                        
  //--------------------------part two---------------------------//
  {
    // -inv(A22)*A21 -> A21
    // A=inv(A22), B=A21, C=A21
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int lda = NB;
    int ldb = NB;
    int ldc = NB;

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = d_dinvAh + blk*NB + blk;
    Bh = Ch = d_dinvAh + blk;
    Al = d_dinvAl + blk*NB + blk;
    Bl = Cl = d_dinvAl + blk;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );
    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];

      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];

      __syncthreads();

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = (-1)*ch[i];
      Cl[0] = (-1)*cl[i];
    }
  }
}

/*
 * B21 = -inv(A11)*A12*inv(A22)
 */
__global__ void
triple_sgemm_update_above64_part1_R (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x*64;
  const int iby = bIdy*16;
  const int id = inx + iny*16;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  int PagesPerNB = NB/(blk*2);
  //--------------------------part one---------------------------//
  {
    // A12*inv(A22) -> A12(d_dinvA)
    // A=A12, B=inv(A22), C=A12
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int ldb = NB;
    int ldc = NB;

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = Ahin + page*lda*blk*2 + blk*lda + page*blk*2;  
    Bh = d_dinvAh + blk*NB + blk;
    Ch = d_dinvAh + blk*NB;
    Al = Alin + page*lda*blk*2 + blk*lda + page*blk*2;  
    Bl = d_dinvAl + blk*NB + blk;
    Cl = d_dinvAl + blk*NB;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );
    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];
      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];

      __syncthreads();

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = ch[i];
      Cl[0] = cl[i];
    }
  }
                        
}
/*
 * B21 = -inv(A22)*A21*inv(A11)
 */
__global__ void
triple_sgemm_update_above64_part1_L (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x*64;
  const int iby = bIdy*16;
  const int id = inx + iny*16;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  int PagesPerNB = NB/(blk*2);
  //--------------------------part one---------------------------//
  {
    // A21*inv(A11) -> A21
    // A=A21, B=inv(A11), C=A21
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int ldb = NB;
    int ldc = NB;

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = Ahin + page*lda*blk*2 + page*blk*2 + blk;  
    Bh = d_dinvAh;
    Ch = d_dinvAh + blk;
    Al = Alin + page*lda*blk*2 + page*blk*2 + blk;  
    Bl = d_dinvAl;
    Cl = d_dinvAl + blk;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );
    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];

      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];
      __syncthreads();

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = ch[i];
      Cl[0] = cl[i];
    }
  }
                        
}

/*
 * B21 = -inv(A11)*A12*inv(A22)
 */
__global__ void
triple_sgemm_update_above64_part2_R (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x*64;
  const int iby = bIdy*16;
  const int id = inx + iny*16;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  int PagesPerNB = NB/(blk*2);
        
  //--------------------------part two---------------------------//
  {
    // -inv(A11)*A12 -> A12
    // A=inv(A11), B=A12, C=A12
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int lda = NB;
    int ldb = NB;
    int ldc = NB;

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = d_dinvAh;
    Bh = d_dinvAh + blk*NB;
    Ch = d_dinvAh + blk;
    Al = d_dinvAl;
    Bl = d_dinvAl + blk*NB;
    Cl = d_dinvAl + blk;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );
    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];
      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];

      __syncthreads();

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = (-1.0)*ch[i];
      Cl[0] = (-1.0)*cl[i];
    }
  }
}
/*
 * part 3, copy data into position 
 */
__global__ void
triple_sgemm_update_above64_part3_R (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x*64;
  const int iby = bIdy*16;
  const int id = inx + iny*16;

  int PagesPerNB = NB/(blk*2);
        
  //--------------------------part two---------------------------//
  {
    // -inv(A11)*A12 -> A12
    // A=inv(A11), B=A12, C=A12
    double *C_temph, *C_realh;
    double *C_templ, *C_reall;
    int ldc = NB;

    C_temph = d_dinvAh + NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2) + blk;
    C_templ = d_dinvAl + NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2) + blk;

    C_realh = d_dinvAh + NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB + blk*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    C_reall = d_dinvAl + NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB + blk*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    C_temph += ibx + id  + __mul24( iby, ldc );
    C_templ += ibx + id  + __mul24( iby, ldc );

    C_realh += ibx + id  + __mul24( iby, ldc );
    C_reall += ibx + id  + __mul24( iby, ldc );


    for( int i = 0; i < 16; i++, C_temph+=ldc, C_realh+=ldc, C_templ+=ldc, C_reall+=ldc ){
      C_realh[0] = C_temph[0];
      C_reall[0] = C_templ[0];
      C_temph[0] = 0;
      C_templ[0] = 0;
    }
  }
}

/*
 * part 3: copy data back to position 
 */
__global__ void
triple_sgemm_update_above64_part3_L (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x*64;
  const int iby = bIdy*16;
  const int id = inx + iny*16;

  int PagesPerNB = NB/(blk*2);
        
  //--------------------------part three---------------------------//
  {
    // -inv(A22)*A21 -> A21
    // A=inv(A22), B=A21, C=A21
    double *C_temph, *C_realh;
    double *C_templ, *C_reall;
    int ldc = NB;

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    C_realh = d_dinvAh + blk;
    C_reall = d_dinvAl + blk;
                
    C_temph = d_dinvAh + blk*NB;
    C_templ = d_dinvAl + blk*NB;

    C_temph += ibx + id  + __mul24( iby, ldc );
    C_templ += ibx + id  + __mul24( iby, ldc );
    C_realh += ibx + id  + __mul24( iby, ldc );
    C_reall += ibx + id  + __mul24( iby, ldc );

    for( int i = 0; i < 16; i++, C_realh+=ldc, C_temph+=ldc, C_reall+=ldc, C_templ+=ldc ){
      C_realh[0] = C_temph[0];
      C_temph[0] = 0;
      C_reall[0] = C_templ[0];
      C_templ[0] = 0;
    }
  }
  __syncthreads();
}
/*
 * B21 = -inv(A22)*A21*inv(A11)
 */
__global__ void
triple_sgemm_update_above64_part2_L (double* Ahin, double* Alin, double *d_dinvAh, double *d_dinvAl, int blk, int lda, int npages)
{
  const int bIdy = blockIdx.y/npages;
  const int page = qmod(blockIdx.y, npages);
  const int inx = threadIdx.x;
  const int iny = threadIdx.y;
  const int ibx = blockIdx.x*64;
  const int iby = bIdy*16;
  const int id = inx + iny*16;
  __shared__ double bsh[16][17];
  __shared__ double bsl[16][17];

  int PagesPerNB = NB/(blk*2);
        
  //--------------------------part two---------------------------//
  {
    // -inv(A22)*A21 -> A21
    // A=inv(A22), B=A21, C=A21
    double *Ah, *Bh, *Ch;
    double *Al, *Bl, *Cl;
    int lda = NB;
    int ldb = NB;
    int ldc = NB;

    d_dinvAh += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);
    d_dinvAl += NB*NB*(page/PagesPerNB) + 
      (qmod(page,PagesPerNB))*(blk*2)*NB +
      (qmod(page,PagesPerNB))*(blk*2);

    Ah = d_dinvAh + blk*NB + blk;
    Bh = d_dinvAh + blk;
    Al = d_dinvAl + blk*NB + blk;
    Bl = d_dinvAl + blk;
                
    Ch = d_dinvAh + blk*NB;
    Cl = d_dinvAl + blk*NB;

    Ah += ibx + id;
    Bh += inx + __mul24( iby + iny, ldb );
    Ch += ibx + id  + __mul24( iby, ldc );
    Al += ibx + id;
    Bl += inx + __mul24( iby + iny, ldb );
    Cl += ibx + id  + __mul24( iby, ldc );

    const double *Blasth = Bh + blk;
    const double *Blastl = Bl + blk;

    double ch[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    double cl[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

    do{
      double ah[4] = { Ah[0*lda], Ah[1*lda], Ah[2*lda], Ah[3*lda] };
      double al[4] = { Al[0*lda], Al[1*lda], Al[2*lda], Al[3*lda] };

      bsh[inx][iny]    = Bh[0*ldb];
      bsh[inx][iny+4]  = Bh[4*ldb];
      bsh[inx][iny+8]  = Bh[8*ldb];
      bsh[inx][iny+12] = Bh[12*ldb];
      bsl[inx][iny]    = Bl[0*ldb];
      bsl[inx][iny+4]  = Bl[4*ldb];
      bsl[inx][iny+8]  = Bl[8*ldb];
      bsl[inx][iny+12] = Bl[12*ldb];

      __syncthreads();

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[0][0], &bsl[0][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[1][0], &bsl[1][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[2][0], &bsl[2][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[3][0], &bsl[3][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[4][0], &bsl[4][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[5][0], &bsl[5][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[6][0], &bsl[6][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[7][0], &bsl[7][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[ 8][0], &bsl[ 8][0], ch, cl ); ah[0] = Ah[0*lda]; al[0] = Al[0*lda];
      ssaxpy( ah[1], al[1], &bsh[ 9][0], &bsl[ 9][0], ch, cl ); ah[1] = Ah[1*lda]; al[1] = Al[1*lda];
      ssaxpy( ah[2], al[2], &bsh[10][0], &bsl[10][0], ch, cl ); ah[2] = Ah[2*lda]; al[2] = Al[2*lda];
      ssaxpy( ah[3], al[3], &bsh[11][0], &bsl[11][0], ch, cl ); ah[3] = Ah[3*lda]; al[3] = Al[3*lda];

      Ah += 4*lda;
      Al += 4*lda;
      ssaxpy( ah[0], al[0], &bsh[12][0], &bsl[12][0], ch, cl );
      ssaxpy( ah[1], al[1], &bsh[13][0], &bsl[13][0], ch, cl );
      ssaxpy( ah[2], al[2], &bsh[14][0], &bsl[14][0], ch, cl );
      ssaxpy( ah[3], al[3], &bsh[15][0], &bsl[15][0], ch, cl );

      Bh += 16;
      Bl += 16;

      __syncthreads();
    } while( Bh < Blasth && Bl < Blastl );

    for( int i = 0; i < 16; i++, Ch += ldc, Cl += ldc ){
      Ch[0] = (-1)*ch[i];
      Cl[0] = (-1)*cl[i];
    }
  }
}

__global__ void
b_copy_kernel (int M, int N, double *b, int ldb, double *d_x, int ldx)
{
  int by = blockIdx.y;

  int gx = blockIdx.x*blockDim.x+threadIdx.x;

  if (gx < M)
    b[by*ldb+gx] = d_x[by*ldx+gx];
}


extern "C"
void diag_strtri (int M, char uplo, char diag, double *Ah, double *Al, double *d_dinvAh, double *d_dinvAl, int lda)
{
  int nblocks = M/BLOCK_SIZE+(M%BLOCK_SIZE!=0);

  if (uplo == 'l' || uplo == 'L'){
    // solve the diagonal blocks
    diag_strtri_kernel_lower<<< nblocks, BLOCK_SIZE, 0, gddstream >>>(diag, Ah, Al, d_dinvAh, d_dinvAl, lda);

    // update the inverse up to the size of BLOCK_SIZE
    for (int i=BLOCK_SIZE; i<NB; i*=2){
      int npages = M/(i*2)+(M%(i*2)!=0);
      dim3 dimBlock((i<=32)?(i/4):16, 4);
      dim3 dimGrid(i/(dimBlock.x*dimBlock.y), npages*(i/16));        //emulated 3D grid, see 3d_grid.txt 
                        
      switch (i){
      case 16:
	triple_sgemm_update_16_part1_L<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	triple_sgemm_update_16_part2_L<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	break;
      case 32:
	triple_sgemm_update_32_part1_L<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	triple_sgemm_update_32_part2_L<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	break;
      case 64:
	triple_sgemm_update_64_part1_L<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	triple_sgemm_update_64_part2_L<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	break;
      default:
	triple_sgemm_update_above64_part1_L<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	triple_sgemm_update_above64_part2_L<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	triple_sgemm_update_above64_part3_L<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	break;
      }
      if (i*2>=M) break;
    }
  }
  else{
    diag_strtri_kernel_upper<<< nblocks, BLOCK_SIZE, 0, gddstream >>>(diag, Ah, Al, d_dinvAh, d_dinvAl, lda);

    // update the inverse up to the size of BLOCK_SIZE
    for (int i=BLOCK_SIZE; i<NB; i*=2){
      int npages = M/(i*2)+(M%(i*2)!=0);
      dim3 dimBlock((i<=32)?(i/4):16, 4);
      dim3 dimGrid(i/(dimBlock.x*dimBlock.y), npages*(i/16));        //emulated 3D grid, see 3d_grid.txt 
                        
      switch (i){
      case 16:
	triple_sgemm_update_16_R<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	break;
      case 32:
	triple_sgemm_update_32_part1_R<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	triple_sgemm_update_32_part2_R<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	break;
      case 64:
	triple_sgemm_update_64_part1_R<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	triple_sgemm_update_64_part2_R<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	break;
      default:
	triple_sgemm_update_above64_part1_R<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	triple_sgemm_update_above64_part2_R<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	triple_sgemm_update_above64_part3_R<<< dimGrid, dimBlock, 0, gddstream >>>(Ah, Al, d_dinvAh, d_dinvAl, i, lda, npages);
	break;
      }
      if (i*2>=M) break;
    }
  }
}

extern "C"
void gddtrsminner_
( char side, char uplo, char tran, char diag, int M, int N, 
  double* alphah, double* alphal,
  double Ah[],double Al[],
  int lda,
  double bh[], double bl[],
  int ldb)
{
  int i;
  double *d_dinvAh, *d_xh;
  double *d_dinvAl, *d_xl;

  /* quick return on wrong size */
  if (M<=0 || N<=0)
    return;

  double *zeroh, *zerol;
  double *oneh, *onel;
  double *minus_oneh, *minus_onel;

  {
    double zerohh=0.0, zerohl=0.0;
    cublasAlloc( 1, sizeof(double), (void**)&zeroh );
    cublasAlloc( 1, sizeof(double), (void**)&zerol );
    cublasSetVector( 1, sizeof(double), &zerohh, 1, zeroh, 1 );
    cublasSetVector( 1, sizeof(double), &zerohl, 1, zerol, 1 );
  }
  {
    double onehh=1.0, onehl=0.0;
    cublasAlloc( 1, sizeof(double), (void**)&oneh );
    cublasAlloc( 1, sizeof(double), (void**)&onel );
    cublasSetVector( 1, sizeof(double), &onehh, 1, oneh, 1 );
    cublasSetVector( 1, sizeof(double), &onehl, 1, onel, 1 );
  }
  {
    double minus_onehh=-1.0, minus_onehl=0.0;
    cublasAlloc( 1, sizeof(double), (void**)&minus_oneh );
    cublasAlloc( 1, sizeof(double), (void**)&minus_onel );
    cublasSetVector( 1, sizeof(double), &minus_onehh, 1, minus_oneh, 1 );
    cublasSetVector( 1, sizeof(double), &minus_onehl, 1, minus_onel, 1 );
  }
  

  if (side == 'l' || side == 'L'){
    cudaMalloc((void**)&d_dinvAh, NB*((M/NB)+(M%NB!=0))*NB*sizeof(double));
    cudaMalloc((void**)&d_dinvAl, NB*((M/NB)+(M%NB!=0))*NB*sizeof(double));

    cudaMalloc((void**)&d_xh, N*M*sizeof(double));
    cudaMalloc((void**)&d_xl, N*M*sizeof(double));

    cudaMemset (d_xh, 0, N*M*sizeof(double));
    cudaMemset (d_xl, 0, N*M*sizeof(double));

    cudaMemset (d_dinvAh, 0, NB*((M/NB)+(M%NB!=0))*NB*sizeof(double));
    cudaMemset (d_dinvAl, 0, NB*((M/NB)+(M%NB!=0))*NB*sizeof(double));

    diag_strtri (M, uplo, diag, Ah, Al, d_dinvAh, d_dinvAl, lda);

    if (tran == 'N' || tran == 'n')
      /* the non-transpose case */
      {
	if (uplo == 'L' || uplo == 'l')
	  {
	    /* the lower case */
                                
	    /* handle the first block seperately with alpha */

	    int MM = min (NB, M);
	    gddgemm ('N', 'N', MM, N, MM, alphah, alphal, d_dinvAh, d_dinvAl, NB, bh, bl, ldb, zeroh, zerol, d_xh, d_xl, M);  

	    if (NB>=M)
	      {
		b_copy();
		cudaFree(d_dinvAh);
		cudaFree(d_dinvAl);
		cudaFree(d_xh);
		cudaFree(d_xl);
		return;
	      }

	    gddgemm ('N', 'N', M-NB, N, NB, minus_oneh, minus_onel, Ah+NB, Al+NB, lda, d_xh, d_xl, M, alphah, alphal, bh+NB, bl+NB, ldb);

	    /* the rest blocks */
	    for (i=NB; i<M; i+=NB)
	      {
		MM = min (M-i, NB);
		gddgemm ('N', 'N', MM, N, MM, oneh, onel, d_dinvAh+i*NB, d_dinvAl+i*NB, NB, bh+i, bl+i, ldb, zeroh, zerol, d_xh+i, d_xl+i, M);  
                                        
		if (i+NB>=M)
		  break;

		gddgemm ('N', 'N', M-i-NB, N, NB, minus_oneh, minus_onel, Ah+i*lda+i+NB, Al+i*lda+i+NB, lda, d_xh+i, d_xl+i, M, oneh, onel, bh+i+NB, bl+i+NB, ldb);
	      }
	  }
	else
	  {
	    /* the upper case */

	    /* handle the first block seperately with alpha */
	    int MM = (M%NB==0)?NB:(M%NB); 
	    i = M-MM;
	    gddgemm ('N', 'N', MM, N, MM, alphah, alphal, d_dinvAh+i*NB, d_dinvAl+i*NB, NB, bh+i, bl+i, ldb, zeroh, zerol, d_xh+i, d_xl+i, M); 
                                        
	    if (i-NB<0)
	      {
		b_copy();
		cudaFree(d_dinvAh);
		cudaFree(d_dinvAl);
		cudaFree(d_xh);
		cudaFree(d_xl);
		return;
	      }

	    gddgemm ('N', 'N', i, N, MM, minus_oneh, minus_onel, Ah+i*lda, Al+i*lda, lda, d_xh+i, d_xl+i, M, alphah, alphal, bh, bl, ldb);

	    /* the rest blocks */
	    for (i=M-MM-NB; i>=0; i-=NB)
	      {
		gddgemm ('N', 'N', NB, N, NB, oneh, onel, d_dinvAh+i*NB, d_dinvAl+i*NB, NB, bh+i, bl+i, ldb, zeroh, zerol, d_xh+i, d_xl+i, M);

		if (i-NB<0)
		  break;

		gddgemm ('N', 'N', i, N, NB, minus_oneh, minus_onel, Ah+i*lda, Al+i*lda, lda, d_xh+i, d_xl+i, M, oneh, onel, bh, bl, ldb);
	      }

	  }
      }
    else
      /* the transpose case */
      {
	if (uplo == 'L' || uplo == 'l')
	  {
	    /* the lower case */
                                
	    /* handle the first block seperately with alpha */
	    int MM = (M%NB==0)?NB:(M%NB); 
	    i=M-MM; 
	    gddgemm ('T', 'N', MM, N, MM, alphah, alphal, d_dinvAh+i*NB, d_dinvAl+i*NB, NB, bh+i, bl+i, ldb, zeroh, zerol, d_xh+i, d_xl+i, M);  

	    if (i-NB<0)
	      {
		b_copy();
		cudaFree(d_dinvAh);
		cudaFree(d_dinvAl);
		cudaFree(d_xh);
		cudaFree(d_xl);
		return;
	      }

	    gddgemm ('T', 'N', i, N, MM, minus_oneh, minus_onel, Ah+i, Al+i, lda, d_xh+i, d_xl+i, M, alphah, alphal, bh, bl, ldb);

	    /* the rest blocks */
	    for (i=M-MM-NB; i>=0; i-=NB)
	      {
		gddgemm ('T', 'N', NB, N, NB, oneh, onel, d_dinvAh+i*NB, d_dinvAl+i*NB, NB, bh+i, bl+i, ldb, zeroh, zerol,d_xh+i, d_xl+i, M);  

		if (i-NB<0)
		  break;

		gddgemm ('T', 'N', i, N, NB, minus_oneh, minus_onel, Ah+i, Al+i, lda, d_xh+i, d_xl+i, M, oneh, onel, bh, bl, ldb);
	      }
	  }
	else
	  {
	    /* the upper case */
                                        
	    /* handle the first block seperately with alpha */
	    int MM = min (NB, M); 
	    gddgemm ('T', 'N', MM, N, MM, alphah, alphal, d_dinvAh, d_dinvAl, NB, bh, bl, ldb, zeroh, zerol, d_xh, d_xl, M);  

	    if (NB>=M)
	      {
		b_copy();
		cudaFree(d_dinvAh);
		cudaFree(d_dinvAl);
		cudaFree(d_xh);
		cudaFree(d_xl);
		return;
	      }

	    gddgemm ('T', 'N', M-NB, N, NB, minus_oneh, minus_onel, Ah+(NB)*lda, Al+(NB)*lda, lda, d_xh, d_xl, M, alphah, alphal, bh+NB, bl+NB, ldb);

	    /* the rest blocks */
	    for (i=NB; i<M; i+=NB)
	      {
		MM = min (M-i, NB);
		gddgemm ('T', 'N', MM, N, MM, oneh, onel, d_dinvAh+i*NB, d_dinvAl+i*NB, NB, bh+i, bl+i, ldb, zeroh, zerol, d_xh+i, d_xl+i, M);  
                                        
		if (i+NB>=M)
		  break;

		gddgemm ('T', 'N', M-i-NB, N, NB, minus_oneh, minus_onel, Ah+(i+NB)*lda+i, Al+(i+NB)*lda+i, lda, d_xh+i, d_xl+i, M, oneh, onel, bh+i+NB, bl+i+NB, ldb);
	      }
	  }
      }
  }
  else
    {        // side=R

      /* inverse the diagonals
       * Allocate device memory for the inversed diagonal blocks, size=N*BLOCK_SIZE 
       */
      cudaMalloc((void**)&d_dinvAh, NB*((N/NB)+(N%NB!=0))*NB*sizeof(double));
      cudaMalloc((void**)&d_dinvAl, NB*((N/NB)+(N%NB!=0))*NB*sizeof(double));

      cudaMalloc((void**)&d_xh, N*M*sizeof(double));
      cudaMalloc((void**)&d_xl, N*M*sizeof(double));

      cudaMemset (d_xh, 0, N*M*sizeof(double));
      cudaMemset (d_xl, 0, N*M*sizeof(double));

      cudaMemset (d_dinvAh, 0, NB*((N/NB)+(N%NB!=0))*NB*sizeof(double));
      cudaMemset (d_dinvAl, 0, NB*((N/NB)+(N%NB!=0))*NB*sizeof(double));

      diag_strtri (N, uplo, diag, Ah, Al, d_dinvAh, d_dinvAl, lda);

      if (tran == 'N' || tran == 'n')
	/* the non-transpose case */
	{
	  if (uplo == 'L' || uplo == 'l')
	    {
	      /* the lower case */
                                
	      /* handle the first block seperately with alpha */
	      int NN = (N%NB==0)?NB:(N%NB);
	      i=N-NN;
	      gddgemm ('N', 'N', M, NN, NN, alphah, alphal, bh+ldb*i, bl+ldb*i, ldb, d_dinvAh+i*NB, d_dinvAl+i*NB, NB, zeroh, zerol, d_xh+i*M, d_xl+i*M, M); 

	      if (i-NB<0)
		{
		  b_copy();
		  cudaFree(d_xh);
		  cudaFree(d_xl);
		  cudaFree(d_dinvAh);
		  cudaFree(d_dinvAl);
		  return;
		}

	      gddgemm ('N', 'N', M, i, NN, minus_oneh, minus_onel, d_xh+i*M, d_xl+i*M, M, Ah+i, Al+i, lda, alphah, alphal, bh, bl, ldb);

	      /* the rest blocks */
	      for (i=N-NN-NB; i>=0; i-=NB)
		{
		  gddgemm ('N', 'N', M, NB, NB, oneh, onel, bh+ldb*i, bl+ldb*i, ldb, d_dinvAh+i*NB, d_dinvAl+i*NB, NB, zeroh, zerol, d_xh+i*M, d_xl+i*M, M); 
                                        
		  if (i-NB<0)
		    break;

		  gddgemm ('N', 'N', M, i, NB, minus_oneh, minus_onel, d_xh+i*M, d_xl+i*M, M, Ah+i, Al+i, lda, oneh, onel, bh, bl, ldb);
		}
	    }
	  else
	    {
	      /* the upper case */
                                
	      /* handle the first block seperately with alpha */
	      int NN = min(NB, N); 
	      gddgemm ('N', 'N', M, NN, NN, alphah, alphal, bh, bl, ldb, d_dinvAh, d_dinvAl, NB, zeroh, zerol, d_xh, d_xl, M);  

	      if (NB>=N)
		{
		  b_copy();
		  cudaFree(d_xh);
		  cudaFree(d_xl);
		  cudaFree(d_dinvAh);
		  cudaFree(d_dinvAl);
		  return;
		}

	      gddgemm ('N', 'N', M, N-NB, NB, minus_oneh, minus_onel, d_xh, d_xl, M, Ah+NB*lda, Al+NB*lda, lda, alphah, alphal, bh+NB*ldb, bl+NB*ldb, ldb);
                                
	      /* the rest blocks */
	      for (i=NB; i<N; i+=NB)
		{
		  NN = min(NB, N-i); 
		  gddgemm ('N', 'N', M, NN, NN, oneh, onel, bh+ldb*i, bl+ldb*i, ldb, d_dinvAh+i*NB, d_dinvAl+i*NB, NB, zeroh, zerol, d_xh+i*M, d_xl+i*M, M);  

		  if (i+NB>=N)
		    break;

		  gddgemm ('N', 'N', M, N-i-NB, NB, minus_oneh, minus_onel, d_xh+i*M, d_xl+i*M, M,  Ah+(i+NB)*lda+i, Al+(i+NB)*lda+i, lda, oneh, onel, bh+(i+NB)*ldb, bl+(i+NB)*ldb, ldb);
		}
	    }
	}
      else
	/* the transpose case */
	{
	  if (uplo == 'L' || uplo == 'l')
	    {
	      /* the lower case */
                                
	      /* handle the first block seperately with alpha */
	      int NN = min(NB, N); 
	      gddgemm ('N', 'T', M, NN, NN, alphah, alphal, bh, bl, ldb, d_dinvAh, d_dinvAl, NB, zeroh, zerol, d_xh, d_xl, M);  

	      if (NB>=N)
		{
		  b_copy();
		  cudaFree(d_xh);
		  cudaFree(d_xl);
		  cudaFree(d_dinvAh);
		  cudaFree(d_dinvAl);
		  return;
		}

	      gddgemm ('N', 'T', M, N-NB, NB, minus_oneh, minus_onel, d_xh, d_xl, M, Ah+NB, Al+NB, lda, alphah, alphal, bh+NB*ldb, bl+NB*ldb, ldb);

	      /* the rest blocks */
	      for (i=NB; i<N; i+=NB)
		{
		  NN = min(NB, N-i); 
		  gddgemm ('N', 'T', M, NN, NN, oneh, onel, bh+ldb*i, bl+ldb*i, ldb, d_dinvAh+i*NB, d_dinvAl+i*NB, NB, zeroh, zerol, d_xh+i*M, d_xl+i*M, M);  

		  if (i+NB>=N)
		    break;

		  gddgemm ('N', 'T', M, N-i-NB, NB, minus_oneh, minus_onel, d_xh+i*M, d_xl+i*M, M, Ah+i*lda+NB+i, Al+i*lda+NB+i, lda, oneh, onel, bh+(i+NB)*ldb, bl+(i+NB)*ldb, ldb);
		}
	    }
	  else
	    {
	      /* the upper case */
                                
	      /* handle the first block seperately with alpha */
	      int NN = (N%NB==0)?NB:(N%NB);
	      i=N-NN;
	      gddgemm ('N', 'T', M, NN, NN, alphah, alphal, bh+ldb*i, bl+ldb*i, ldb, d_dinvAh+i*NB, d_dinvAl+i*NB, NB, zeroh, zerol, d_xh+i*M, d_xl+i*M, M); 

	      if (i-NB<0)
		{
		  b_copy();
		  cudaFree(d_xh);
		  cudaFree(d_xl);
		  cudaFree(d_dinvAh);
		  cudaFree(d_dinvAl);
		  return;
		}

	      gddgemm ('N', 'T', M, i, NN, minus_oneh, minus_onel, d_xh+i*M, d_xl+i*M, M, Ah+i*lda, Al+i*lda, lda, alphah, alphal, bh, bl, ldb);
                                
	      /* the rest blocks */
	      for (i=N-NN-NB; i>=0; i-=NB)
		{
		  gddgemm ('N', 'T', M, NB, NB, oneh, onel, bh+ldb*i, bl+ldb*i, ldb, d_dinvAh+i*NB, d_dinvAl+i*NB, NB, zeroh, zerol, d_xh+i*M, d_xl+i*M, M); 

		  if (i-NB<0)
		    break;

		  gddgemm ('N', 'T', M, i, NB, minus_oneh, minus_onel, d_xh+i*M, d_xl+i*M, M, Ah+i*lda, Al+i*lda, lda, oneh, onel, bh, bl, ldb);
		}
	    }
	}
    }
                
  b_copy();
  cudaFree(d_dinvAh);
  cudaFree(d_dinvAl);
  cudaFree(d_xh);
  cudaFree(d_xl);
}

extern "C"
void gddtrsm
( char *side, char *uplo, char *tran, char *diag, int *M, int *N, 
  double* alphah, double* alphal,
  double Ah[],double Al[],
  int *lda,
  double bh[], double bl[],
  int *ldb)
{
	gddtrsminner_(*side,*uplo,*tran,*diag,*M,*N,alphah,alphal,Ah,Al,*lda,bh,bl,*ldb);

}
