#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"
#include <basic_algebra.h>
#define NB  32

__global__ void
vector_copy_kernel( int N, double *xinh, double *xinl, int incxin,
                    double*xouth,
                    double*xoutl, int incxout )
{
    int idx = threadIdx.x+NB*blockIdx.x;

    *( xouth+idx*incxout ) = *( xinh+idx*incxin );
    *( xoutl+idx*incxout ) = *( xinl+idx*incxin );
}

void
gddtrsv( char *uplo, char *tran, char *diag, int *N,
              double Ah[], double Al[],
              int *lda,
              double xh[], double xl[],
              int *incx )
{
    double *dworkh, *dworkl, *dalphah, *dalphal;

    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( *N, sizeof( double ), (void**) &dworkh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( *N, sizeof( double ), (void**) &dworkl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( double ), (void**) &dalphah )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( double ), (void**) &dalphal )
      )
    {
        cublasFree( dworkh );
        cublasFree( dworkl );
        cublasFree( dalphah );
        cublasFree( dalphal );
        printf( "Device Malloc Error!\n" );
        exit( -1 );
    }

    dim3 blocks( ( *N+NB-1 )/NB );
    dim3 threads( NB, 1, 1 );

    set_one_kernel<<< 1, 1, 0, gddstream>>>
    ( dalphah, dalphal );

    vector_copy_kernel<<< blocks, threads, 0, gddstream>>>
    ( *N, xh, xl, *incx, dworkh, dworkl, 1 );

    gddtrsminner_( 'L', *uplo, *tran, *diag, *N, 1,
                        dalphah, dalphal,
                        Ah, Al, *lda,
                        dworkh, dworkl, *N );
    vector_copy_kernel<<<blocks,threads,0,gddstream>>>
    ( *N, dworkh, dworkl, 1, xh, xl, *incx );



    cublasFree( dworkh );
    cublasFree( dworkl );
    cublasFree( dalphah );
    cublasFree( dalphal );
}

