#define BLOCKSIZE (256)
#define TILEWIDTH (2*BLOCKSIZE)
#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "basic_algebra.h"
#include "cutil.h"

__global__ void
gddzdotu_kernel( int n, cuDoubleComplex *Xh, cuDoubleComplex *Xl, int incx, cuDoubleComplex *Yh,
               cuDoubleComplex *Yl,
               int incy, cuDoubleComplex*dworkh, cuDoubleComplex*dworkl )
{

    __shared__ cuDoubleComplex sdatal[BLOCKSIZE];
    __shared__ cuDoubleComplex sdatah[BLOCKSIZE];
    unsigned int      tid = threadIdx.x;
    unsigned int      i = blockIdx.x*( BLOCKSIZE*2 )+tid;
    unsigned int      gridSize = BLOCKSIZE*2*gridDim.x;


    sdatal[tid] = make_cuDoubleComplex(0.0,0.0);
    sdatah[tid] = make_cuDoubleComplex(0.0,0.0);



    while( i < n )
    {
        cuDoubleComplex tmph, tmpl;
        /*        sdata[tid] += X[i]*Y[i]+X[i+BLOCKSIZE]*Y{i+BLOCKSIZE];*/
	tmph = *(Xh+i*incx);
	tmpl = *(Xl+i*incx);
        CMUL( &tmph,&tmpl, Yh+i*incy, Yl+i*incy, &tmph, &tmpl );
        CADD( sdatah+tid, sdatal+tid, &tmph, &tmpl, sdatah+tid, sdatal+tid );

        if( i+BLOCKSIZE < n )
        {
		tmph = *(Xh+( i+BLOCKSIZE )*incx);
		tmpl = *(Xl+( i+BLOCKSIZE )*incx);
            CMUL( &tmph, &tmpl,
                  Yh+( i+BLOCKSIZE )*incy, Yl+( i+BLOCKSIZE )*incy,
                  &tmph, &tmpl );
            CADD( sdatah+tid, sdatal+tid, &tmph, &tmpl, sdatah+tid,
                  sdatal+tid );
        }

        i += gridSize;
    }

    __syncthreads( );

    if( BLOCKSIZE >= 512 )
    {
        if( tid < 256 )
        {
            /*            sdata[tid] += sdata[tid+256];*/
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+256, sdatal+tid+256,
                  sdatah+tid, sdatal+tid );
        }
        __syncthreads( );
    }
    if( BLOCKSIZE >= 256 )
    {
        if( tid < 128 )
        {
            /*            sdata[tid] += sdata[tid+128];*/
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+128, sdatal+tid+128,
                  sdatah+tid, sdatal+tid );
        }
        __syncthreads( );
    }
    if( BLOCKSIZE >= 128 )
    {
        if( tid <  64 )
        {
            /*            sdata[tid] += sdata[tid+64];*/
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+64, sdatal+tid+64,
                  sdatah+tid, sdatal+tid );
        }
        __syncthreads( );
    }

    if( tid < 32 )
    {
        if( BLOCKSIZE >=  64 )
        {
            /*            sdata[tid] += sdata[tid+32];*/
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+32, sdatal+tid+32,
                  sdatah+tid, sdatal+tid );
        }
        if( BLOCKSIZE >=  32 )
        {
            /*                sdata[tid] += sdata[tid+16]; */
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+16, sdatal+tid+16,
                  sdatah+tid, sdatal+tid );
        }
        if( BLOCKSIZE >=  16 )
        {
            /*                sdata[tid] += sdata[tid+8];*/
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+8, sdatal+tid+8,
                  sdatah+tid, sdatal+tid );
        }
        if( BLOCKSIZE >=   8 )
        {
            /*                       sdata[tid] += sdata[tid+4];*/
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+4, sdatal+tid+4,
                  sdatah+tid, sdatal+tid );
        }
        if( BLOCKSIZE >=   4 )
        {
            /*                       sdata[tid] += sdata[tid+2];*/
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+2, sdatal+tid+2,
                  sdatah+tid, sdatal+tid );
        }
        if( BLOCKSIZE >=   2 )
        {
            /*                sdata[tid] += sdata[tid+1]; */
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+1, sdatal+tid+1,
                  sdatah+tid, sdatal+tid );
        }
    }


    if( tid == 0 )
    {
        dworkh[blockIdx.x] = sdatah[0];
        dworkl[blockIdx.x] = sdatal[0];
    }
}

extern "C" void
gddzdotu( int *n,   cuDoubleComplex *Xh, cuDoubleComplex *Xl, int* incx, cuDoubleComplex *Yh,
             cuDoubleComplex *Yl, int*incy, cuDoubleComplex * reth, cuDoubleComplex*retl )
{
    unsigned int gridSize = ( *n+TILEWIDTH-1 )/TILEWIDTH;

    dim3         blocks( gridSize, 1 ,1);

    dim3         threads( BLOCKSIZE, 1 ,1);


    cuDoubleComplex *     dworkh, *dworkl;
    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( gridSize, sizeof( cuDoubleComplex ), (void**) &dworkh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( gridSize, sizeof( cuDoubleComplex ), (void**) &dworkl )
      )
    {
        cublasFree( dworkh );
        cublasFree( dworkl );
        printf( "Device Malloc Error!\n" );
        exit( -1 );
    }

    gddzdotu_kernel<<< blocks, threads, 0, gddstream>>>
    ( *n, Xh, Xl, *incx, Yh, Yl, *incy, dworkh, dworkl );

    if( gridSize > 1 )
    {
       cuDoubleComplex *dwork2h, *dwork2l;
       if(

           CUBLAS_STATUS_SUCCESS !=
           cublasAlloc( ( gridSize+TILEWIDTH-1 )/TILEWIDTH, sizeof( cuDoubleComplex ),
                       (void**) &dwork2h )
         || CUBLAS_STATUS_SUCCESS !=
           cublasAlloc( ( gridSize+TILEWIDTH-1 )/TILEWIDTH, sizeof( cuDoubleComplex ),
                       (void**) &dwork2l )
         )
       {
           cublasFree( dwork2h );
           cublasFree( dwork2l );
           printf( "Device Malloc Error!\n" );
           exit( -1 );
       }

       int count = 0;
        while( gridSize > 1 )
        {
            unsigned int vectorSize = gridSize;
            gridSize = ( gridSize+TILEWIDTH-1 )/TILEWIDTH;
            dim3         localblock( gridSize, 1 );

            /* reduce_kernel( */
	    reduce_kernel_complex<BLOCKSIZE>
	    <<< localblock, threads, 0, gddstream>>>
	    ( dworkh, dworkl, dwork2h, dwork2l, vectorSize );
            swapPointer( &dworkh, &dwork2h );
            swapPointer( &dworkl, &dwork2l );
            count++;
        }


	elemcopy<cuDoubleComplex>( dworkh, reth );
	elemcopy<cuDoubleComplex>( dworkl, retl );

        if( count%2 == 1 )
        {
            swapPointer( &dworkh, &dwork2h );
            swapPointer( &dworkl, &dwork2l );
        }

        cublasFree( dwork2h );
        cublasFree( dwork2l );
    }
    else
    {
	elemcopy<cuDoubleComplex>( dworkh, reth );
	elemcopy<cuDoubleComplex>( dworkl, retl );
    }
    cublasFree( dworkh );
    cublasFree( dworkl );
}

