#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"


#define BLOCKSIZE ( 16 )
#define BUFFERSIZE ( BLOCKSIZE )
#define TRANSBLOCK ( 256 )

extern "C" __global__ void
gddzhbmv_kernel_L( int n, int k, cuDoubleComplex *alphah, cuDoubleComplex*alphal,
                  cuDoubleComplex *Ah, cuDoubleComplex *Al, int lda, cuDoubleComplex * xh, cuDoubleComplex *xl,
                  int incx,
                  cuDoubleComplex *betah, cuDoubleComplex*betal, cuDoubleComplex*yh, cuDoubleComplex*yl,
                  int incy  )
{
    int tid = threadIdx.x;
    int idx = tid+__mul24( BLOCKSIZE, blockIdx.x );

    {
        int offset = idx;
        Ah += offset;
        Al += offset;
        offset = tid*incx;
        xh += offset;
        xl += offset;
    }
    lda -= 1;

    cuDoubleComplex resulth = make_cuDoubleComplex(0.0,0.0);
    cuDoubleComplex resultl = make_cuDoubleComplex(0.0,0.0);
    {
        __shared__ cuDoubleComplex buffh[BUFFERSIZE];
        __shared__ cuDoubleComplex buffl[BUFFERSIZE];

        int               n1 = blockIdx.x*BUFFERSIZE;

	int position =  (n1 -k)/BUFFERSIZE;
	if(position < 0) 
		position = 0;
	position *= BUFFERSIZE;
	Ah += position * lda;
	Al += position * lda;
        /* Lower Blocks */
        for( int i = position; i < n1; i += BUFFERSIZE )
        {
            __syncthreads( );
            buffh[tid] = xh[i*incx];
            buffl[tid] = xl[i*incx];

            __syncthreads( );
#pragma unroll
            for( int j = 0; j < BUFFERSIZE; j++ )
            {
                cuDoubleComplex ph, pl;
                ph = *Ah;
                pl = *Al;
/*                    int b = !((idx > k+ i+j) || (idx < i+j -k));*/
                int    b = !( idx > k+i+j );
                ph = cuCmul(ph,make_cuDoubleComplex((double)b,0.0));
                pl = cuCmul(pl,make_cuDoubleComplex((double)b,0.0));

                CMUL( &ph, &pl, buffh+j, buffl+j, &ph, &pl );

                CADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );

                Ah += lda;
                Al += lda;
            }
        }


        __syncthreads( );

        __shared__ cuDoubleComplex buffAh[BLOCKSIZE][BLOCKSIZE+1];
        __shared__ cuDoubleComplex buffAl[BLOCKSIZE][BLOCKSIZE+1];
        /* Diagonal Block */
        {
            int offset = n1*incx;
            xh += offset;
            xl += offset;
        }

        {
            int m;
            if( n1+BUFFERSIZE < n )
            {
                m = BUFFERSIZE;
            }
            else
            {
                m = n-n1;
            }

            buffh[tid] = *xh;
            buffl[tid] = *xl;


            for( int i = 0; i < m; i++ )
            {
                int offset  = i*lda;
                buffAh[tid][i] = *( Ah+offset );
                buffAl[tid][i] = *( Al+offset );
            }
            __syncthreads( );

            for( int i = 0; i < m; i++ )
            {
                cuDoubleComplex ph, pl;
                int    x, y;
                x = ( tid > i ) ? tid : i;
                y = ( tid > i ) ? i : tid;
                ph = buffAh[x][y];
                pl = buffAl[x][y];
                int    b = !( ( idx > k+i+n1 ) || ( idx < i+n1-k ) );
                ph = cuCmul(ph,make_cuDoubleComplex((double)b,0.0));
                pl = cuCmul(pl,make_cuDoubleComplex((double)b,0.0));

		if(tid < i)
		{
			ph = cuConj(ph);
			pl = cuConj(pl);
		}
		else if(tid == i)
		{
			ph = make_cuDoubleComplex(cuCreal(ph),0.0);
			pl = make_cuDoubleComplex(cuCreal(pl),0.0);
		}


                CMUL( &ph, &pl, buffh+i, buffl+i, &ph, &pl );
                CADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
            }

            Ah += BLOCKSIZE;
            Al += BLOCKSIZE;

            xh += m*incx;
            xl += m*incx;
        }

        /* Upper Blocks */

        int no = n1+BLOCKSIZE;


        n1 = ( n/BLOCKSIZE-( blockIdx.x+1 ) )*BLOCKSIZE;
	position = ((k-1)/BLOCKSIZE+1)*BLOCKSIZE;
/*        if(tid==0) printf("bid:%d,pos:%d,n1:%d\n",blockIdx.x,position,n1);*/
	n1 = min(n1,position);




        for( int i = 0; i < n1; i += BLOCKSIZE )
        {
            __syncthreads( );

            buffh[tid] = xh[i*incx];
            buffl[tid] = xl[i*incx];
            for( int j = 0; j < BLOCKSIZE; j++ )
            {
                int offset = j*lda;
                buffAh[j][tid] = cuConj(*( Ah+offset ));
                buffAl[j][tid] = cuConj(*( Al+offset ));
            }
            __syncthreads( );

#pragma unroll
            for( int j = 0; j < BUFFERSIZE; j++ )
            {
                cuDoubleComplex ph, pl;
                ph = buffAh[tid][j];
                pl = buffAl[tid][j];
/*                    int b = !((idx > k+ i+j) || (idx < i+j -k));*/
                int    b = !( ( idx < no+i+j-k ) );
                ph = cuCmul(ph,make_cuDoubleComplex((double)b,0.0));
                pl = cuCmul(pl,make_cuDoubleComplex((double)b,0.0));
                CMUL( &ph, &pl, buffh+j, buffl+j, &ph, &pl );
                CADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
            }

            Ah += BLOCKSIZE;
            Al += BLOCKSIZE;
        }

        /* Last Block */
        no += n1;

        __syncthreads( );
        int m = n-( n1+( blockIdx.x+1 )*BLOCKSIZE );
/*        if(tid==0) printf("bid:%d,m:%d,n1:%d,position:%d\n",blockIdx.x,m,n1,position);*/

        if( m > 0 && n1 !=position && blockIdx.x+1 < gridDim.x )
        {
            buffh[tid] = xh[n1*incx];
            buffl[tid] = xl[n1*incx];
            for( int j = 0; j < BLOCKSIZE; j++ )
            {
                int offset = j*lda;
                buffAh[j][tid] = cuConj(*( Ah+offset ));
                buffAl[j][tid] = cuConj(*( Al+offset ));
            }
            __syncthreads( );

            for( int i = 0; i < m; i++ )
            {
                cuDoubleComplex ph, pl;
/*                result += *(A)*buff[j];*/
                ph = buffAh[tid][i];
                pl = buffAl[tid][i];
/*                    int b = !((idx > k+ i+j) || (idx < i+j -k));*/
                int    b = !( ( idx < no+i-k ) );
                ph = cuCmul(ph,make_cuDoubleComplex((double)b,0.0));
                pl = cuCmul(pl,make_cuDoubleComplex((double)b,0.0));
                CMUL( &ph, &pl, buffh+i, buffl+i, &ph, &pl );
                CADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
                Ah += lda;
                Al += lda;
            }
        }

        if( idx < n )
        {
            cuDoubleComplex ph, pl, qh, ql;
            CMUL( betah, betal, yh+idx*incy, yl+idx*incy, &ph, &pl );
            CMUL( alphah, alphal, &resulth, &resultl, &qh, &ql );
            CADD( &ph, &pl, &qh, &ql, yh+idx*incy, yl+idx*incy );
        }
    }
}

extern "C" __global__ void
gddzhbmv_kernel_U( int n, int k, cuDoubleComplex *alphah, cuDoubleComplex*alphal,
                  cuDoubleComplex *Ah, cuDoubleComplex *Al, int lda, cuDoubleComplex * xh, cuDoubleComplex *xl,
                  int incx,
                  cuDoubleComplex *betah, cuDoubleComplex*betal, cuDoubleComplex*yh, cuDoubleComplex*yl,
                  int incy  )
{
    int          tid = threadIdx.x;
    int idx = tid+__mul24( BLOCKSIZE, blockIdx.x );

    lda-=1;
    Ah += k+tid+__mul24( BLOCKSIZE*lda, blockIdx.x );
    Al += k+tid+__mul24( BLOCKSIZE*lda, blockIdx.x );
    {
        unsigned int offset = tid*incx;
        xh += offset;
        xl += offset;
    }

    cuDoubleComplex       resulth = make_cuDoubleComplex(0.0,0.0);
    cuDoubleComplex       resultl = make_cuDoubleComplex(0.0,0.0);
    {
        __shared__ cuDoubleComplex buffh[BUFFERSIZE];
        __shared__ cuDoubleComplex buffl[BUFFERSIZE];

        int               n1 = blockIdx.x*BUFFERSIZE;

        __shared__ cuDoubleComplex buffAh[BLOCKSIZE][BLOCKSIZE+1];
        __shared__ cuDoubleComplex buffAl[BLOCKSIZE][BLOCKSIZE+1];

	int position =  (n1 -k)/BUFFERSIZE;
	if(position < 0) 
		position = 0;
	position *= BUFFERSIZE;
	Ah += position ;
	Al += position ;


        /* Lower Blocks */

        for( int i = position; i < n1; i += BUFFERSIZE )
        {
            __syncthreads( );

            buffh[tid] = xh[i*incx];
            buffl[tid] = xl[i*incx];
            for( int j = 0; j < BLOCKSIZE; j++ )
            {
                int offset = j*lda;
                buffAh[j][tid] = cuConj(*( Ah+offset ));
                buffAl[j][tid] = cuConj(*( Al+offset ));
            }
            __syncthreads( );

#pragma unroll
            for( int j = 0; j < BUFFERSIZE; j++ )
            {
                cuDoubleComplex ph, pl;
                ph = buffAh[tid][j];
                pl = buffAl[tid][j];
                int    b = !( idx > k+i+j );
                ph = cuCmul(ph,make_cuDoubleComplex((double)b,0.0));
                pl = cuCmul(pl,make_cuDoubleComplex((double)b,0.0));
                CMUL( &ph, &pl, buffh+j, buffl+j, &ph, &pl );
                CADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
            }

            Ah += BLOCKSIZE;
            Al += BLOCKSIZE;
        }


        /* Diagonal Block */
        {
            int offset = n1*incx;
            xh += offset;
            xl += offset;
        }

        {
            int m;
            if( n1+BUFFERSIZE < n )
            {
                m = BUFFERSIZE;
            }
            else
            {
                m = n-n1;
            }

            buffh[tid] = *xh;
            buffl[tid] = *xl;
            for( int i = 0; i < m; i++ )
            {
                int offset  = i*lda;
                buffAh[tid][i] = *( Ah+offset );
                buffAl[tid][i] = *( Al+offset );

            }
            __syncthreads( );

            for( int i = 0; i < m; i++ )
            {
                cuDoubleComplex ph, pl;
                int x, y;
                x = ( tid > i ) ? i : tid;
                y = ( tid > i ) ? tid : i;
		ph = buffAh[x][y];
		pl = buffAl[x][y];
                int    b = !( ( idx > k+i+n1 ) || ( idx < i+n1-k ) );
                ph = cuCmul(ph,make_cuDoubleComplex((double)b,0.0));
                pl = cuCmul(pl,make_cuDoubleComplex((double)b,0.0));

		if(tid > i)
		{
			ph = cuConj(ph);
			pl = cuConj(pl);
		}else if (tid == i) {
			ph = make_cuDoubleComplex(cuCreal(ph),0.0);
			pl = make_cuDoubleComplex(cuCreal(pl),0.0);
		}

                CMUL( &ph, &pl, buffh+i, buffl+i, &ph, &pl );
                CADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );

            }

            Ah += lda*BLOCKSIZE;
            Al += lda*BLOCKSIZE;


            xh += m*incx;
            xl += m*incx;
        }
        /* Upper Blocks */
        int no = n1+BLOCKSIZE;

        n1 = ( n/BLOCKSIZE-( blockIdx.x+1 ) )*BLOCKSIZE;
	position = ((k-1)/BLOCKSIZE+1)*BLOCKSIZE;
/*        if(tid==0) printf("bid:%d,pos:%d,n1:%d\n",blockIdx.x,position,n1);*/
	n1 = min(n1,position);

        for( int i = 0; i < n1; i += BLOCKSIZE )
        {
            __syncthreads( );

            buffh[tid] = xh[i*incx];
            buffl[tid] = xl[i*incx];

            __syncthreads( );

#pragma unroll
            for( int j = 0; j < BUFFERSIZE; j++ )
            {
                cuDoubleComplex ph, pl;
		ph = *Ah;
		pl = *Al;
                int    b = !( ( idx < no+i+j-k ) );
                ph = cuCmul(ph,make_cuDoubleComplex((double)b,0.0));
                pl = cuCmul(pl,make_cuDoubleComplex((double)b,0.0));
                CMUL( &ph, &pl, buffh+j, buffl+j, &ph, &pl );
                CADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );

                Ah += lda;
                Al += lda;
            }
        }

	//Last Block
        no += n1;
        __syncthreads( );
        int m = n-( n1+( blockIdx.x+1 )*BLOCKSIZE );
        if( m > 0 && n1 != position && blockIdx.x+1 < gridDim.x )
        {
            buffh[tid] = xh[n1*incx];
            buffl[tid] = xl[n1*incx];
        __syncthreads( );

            for( int i = 0; i < m; i++ )
            {
                cuDoubleComplex ph, pl;
		ph = *Ah;
		pl = *Al;
/*                result += *(A)*buff[j];*/
                int    b = !( ( idx < no+i-k ) );
                ph = cuCmul(ph,make_cuDoubleComplex((double)b,0.0));
                pl = cuCmul(pl,make_cuDoubleComplex((double)b,0.0));
                CMUL( &ph, &pl, buffh+i, buffl+i, &ph, &pl );
                CADD( &resulth, &resultl, &ph, &pl, &resulth, &resultl );
                Ah += lda;
                Al += lda;
            }


	}

        if( idx < n )
        {
            cuDoubleComplex ph, pl, qh, ql;
            CMUL( betah, betal, yh+idx*incy, yl+idx*incy, &ph, &pl );
            CMUL( alphah, alphal, &resulth, &resultl, &qh, &ql );
            CADD( &ph, &pl, &qh, &ql, yh+idx*incy, yl+idx*incy );
        }
    }

}

extern "C" void
gddzhbmv( char*uplo, int *n, int *k, cuDoubleComplex *alphah, cuDoubleComplex*alphal,
              cuDoubleComplex *Ah, cuDoubleComplex *Al, int *lda, cuDoubleComplex * xh, cuDoubleComplex *xl,
              int *incx,
              cuDoubleComplex *betah, cuDoubleComplex*betal, cuDoubleComplex*yh, cuDoubleComplex*yl,
              int *incy )
{
    int  bupflag = 1;

    if( *uplo == 'L' || *uplo == 'l' ) bupflag = 0;
    dim3 blocks( ( *n+BLOCKSIZE-1 )/BLOCKSIZE, 1 );

    dim3 threads( BLOCKSIZE, 1 );

    if( !bupflag )
    {
        gddzhbmv_kernel_L<<< blocks, threads, 0, gddstream>>>
        ( *n, *k, alphah, alphal, Ah, Al, *lda,  xh, xl, *incx, betah,
          betal, yh, yl, *incy );
    }
    else
    {
        gddzhbmv_kernel_U<<< blocks, threads, 0, gddstream>>>
        ( *n, *k, alphah, alphal, Ah, Al, *lda,  xh, xl, *incx, betah,
          betal, yh, yl, *incy );
    }
}

