#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"


#define BLOCKSIZE ( 64 )

extern "C" __global__ void
gddzher_kernel_L( int n, double *alphah, double*alphal,
                  cuDoubleComplex * xh, cuDoubleComplex *xl,
                  int incx,
                  cuDoubleComplex* Ah, cuDoubleComplex *Al, int lda )
{
    unsigned int tid = threadIdx.x;
    unsigned int idx = tid+__mul24( BLOCKSIZE, blockIdx.x );


    Ah += idx;
    Al += idx;


    {
        __shared__ cuDoubleComplex xhs[BLOCKSIZE];
        __shared__ cuDoubleComplex xls[BLOCKSIZE];

        xhs[tid] = xh[idx*incx];
        xls[tid] = xl[idx*incx];

        {
            int offset =  tid*incx;
            xh += offset;
            xl += offset;
        }

        int                        n1 = ( blockIdx.x )*BLOCKSIZE;

        cuDoubleComplex            regalphah = make_cuDoubleComplex(*alphah,0.0);
        cuDoubleComplex            regalphal = make_cuDoubleComplex(*alphal,0.0);

        /* Off-Diagonal Block */


        __shared__ cuDoubleComplex tilexhs[BLOCKSIZE];
        __shared__ cuDoubleComplex tilexls[BLOCKSIZE];


        for( int i = 0; i < n1; i += BLOCKSIZE )
        {
            __syncthreads( );


            tilexhs[tid] = cuConj( xh[i*incx] );
            tilexls[tid] = cuConj( xl[i*incx] );

            __syncthreads( );

#pragma unroll
            for( int j = 0; j < BLOCKSIZE; j++ )
            {
                cuDoubleComplex ph, pl, lAh, lAl;

                lAh = *Ah;
                lAl = *Al;

                CMUL( xhs+tid, xls+tid, tilexhs+j, tilexls+j, &ph, &pl );
                CMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );
                CADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );

                if( idx < n )
                {
                    *Ah = lAh;
                    *Al = lAl;
                }

                Ah += lda;
                Al += lda;
            }
        }


            __syncthreads( );
        /* Diagonal Block */
        {
            int m;
            if( n1+BLOCKSIZE < n )
            {
                m = BLOCKSIZE;
            }
            else
            {
                m = n-n1;
            }
            __syncthreads( );
            for( int j = 0; j < m; j++ )
            {
                cuDoubleComplex ph, pl, lAh, lAl;

                lAh = *Ah;
                lAl = *Al;

                ph = cuConj( xhs[j] );
                pl = cuConj( xls[j] );
                CMUL( xhs+tid, xls+tid, &ph, &pl, &ph, &pl );

                CMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );

                cuDoubleComplex b = make_cuDoubleComplex( ( j <= tid ), 0.0 );
                ph = cuCmul( b, ph );
                pl = cuCmul( b, pl );

                CADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );

		if(tid == j)
		{
			lAh = make_cuDoubleComplex(lAh.x,0.0);
			lAl = make_cuDoubleComplex(lAl.x,0.0);
		}


                if( idx < n )
                {
                    *Ah = lAh;
                    *Al = lAl;
                }

                Ah += lda;
                Al += lda;
            }
        }
    }
}

extern "C" __global__ void
gddzher_kernel_U( int n, double *alphah, double*alphal,
                  cuDoubleComplex * xh, cuDoubleComplex *xl,
                  int incx,
                  cuDoubleComplex* Ah, cuDoubleComplex *Al, int lda )
{
    unsigned int tid = threadIdx.x;
    unsigned int idx = tid+__mul24( BLOCKSIZE, blockIdx.x );


    {
        int offset = idx+__mul24( BLOCKSIZE, blockIdx.x )*lda;
        Ah += offset;
        Al += offset;
    }


    {
        __shared__ cuDoubleComplex xhs[BLOCKSIZE];
        __shared__ cuDoubleComplex xls[BLOCKSIZE];

        {
            xh += idx*incx;
            xl += idx*incx;
            xhs[tid] = *xh;
            xls[tid] = *xl;
        }

        int                        n1 = BLOCKSIZE*blockIdx.x;

        cuDoubleComplex            regalphah = make_cuDoubleComplex(*alphah,0.0);
        cuDoubleComplex            regalphal = make_cuDoubleComplex(*alphal,0.0);
        /* Diagonal Block */


            __syncthreads( );
        if( n1+BLOCKSIZE  < n )
        {
            n1 = BLOCKSIZE;
        }
        else
        {
            n1 = n-n1;
        }


        for( int j = 0; j < n1; j++ )
        {
            cuDoubleComplex ph, pl,  lAh, lAl;

            lAh = *Ah;
            lAl = *Al;

            ph = cuConj( xhs[j] );
            pl = cuConj( xls[j] );
            CMUL( xhs+tid, xls+tid, &ph, &pl, &ph, &pl );

            CMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );


            cuDoubleComplex b = make_cuDoubleComplex( ( j >= tid ), 0.0 );
            ph = cuCmul( b, ph );
            pl = cuCmul( b, pl );


            CADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );


            if( idx < n )
            {
                *Ah = lAh;
                *Al = lAl;
            }

            Ah += lda;
            Al += lda;
        }

        /* Off-Diagonal Block */

        {
            int offset = BLOCKSIZE*incx;
            xh += offset;
            xl += offset;
        }

        __shared__ cuDoubleComplex tilexhs[BLOCKSIZE];
        __shared__ cuDoubleComplex tilexls[BLOCKSIZE];


        n1 = BLOCKSIZE*( n/BLOCKSIZE-( blockIdx.x+1 ) );

        for( int i = 0; i < n1; i += BLOCKSIZE )
        {
            __syncthreads( );


            tilexhs[tid] = cuConj( xh[i*incx] );
            tilexls[tid] = cuConj( xl[i*incx] );

            __syncthreads( );

#pragma unroll
            for( int j = 0; j < BLOCKSIZE; j++ )
            {
                cuDoubleComplex ph, pl, lAh, lAl;

                lAh = *Ah;
                lAl = *Al;

                CMUL( xhs+tid, xls+tid, tilexhs+j, tilexls+j, &ph, &pl );
                CMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );
                CADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );

                if( idx < n )
                {
                    *Ah = lAh;
                    *Al = lAl;
                }

                Ah += lda;
                Al += lda;
            }
        }


        /* Last Block */

        __syncthreads( );

        int m  = n-( n1+( blockIdx.x+1 )*BLOCKSIZE );

        if( m > 0 && blockIdx.x+1 < gridDim.x )
        {
            tilexhs[tid] = cuConj( xh[n1*incx] );
            tilexls[tid] = cuConj( xl[n1*incx] );

            __syncthreads( );


#pragma unroll
            for( int j = 0; j < m; j++ )
            {
                cuDoubleComplex ph, pl, lAh, lAl;

                lAh = *Ah;
                lAl = *Al;

                CMUL( xhs+tid, xls+tid, tilexhs+j, tilexls+j, &ph, &pl );
                CMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );
                CADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );


                if( idx < n )
                {
                    *Ah = lAh;
                    *Al = lAl;
                }

                Ah += lda;
                Al += lda;
            }
        }
    }
}

extern "C" void
gddzher( char* uplo, int* n, double *alphah,
              double*alphal,
              cuDoubleComplex * xh, cuDoubleComplex *xl,
              int* incx,
              cuDoubleComplex* Ah, cuDoubleComplex *Al,
              int* lda  )
{
    dim3 blocks( ( *n+BLOCKSIZE-1 )/BLOCKSIZE, 1 );

    dim3 threads( BLOCKSIZE, 1 );

    if( *uplo == 'L' || *uplo == 'l' )
    {
        gddzher_kernel_L<<< blocks, threads, 0, gddstream>>>
        ( *n, alphah, alphal,  xh, xl, *incx,
          Ah, Al, *lda );
    }
    else if( *uplo == 'U' || *uplo == 'u' )
    {
        gddzher_kernel_U<<< blocks, threads, 0, gddstream>>>
        ( *n, alphah, alphal,  xh, xl, *incx,
          Ah, Al, *lda );
    }
    else
    {
        /* ??? */
    }
}

