#include "gddcommon.h"
#include "gdd_basic_inline.h"
#include "cutil.h"


#define BLOCKSIZE ( 64 )

extern "C" __global__ void
gddzher2_kernel_L( int n, cuDoubleComplex *alphah, cuDoubleComplex*alphal,
                  cuDoubleComplex * xh, cuDoubleComplex *xl,
                  int incx,
                  cuDoubleComplex*yh, cuDoubleComplex*yl,
                  int incy,
                  cuDoubleComplex* Ah, cuDoubleComplex *Al, int lda )
{
    unsigned int tid = threadIdx.x;
    unsigned int idx = tid+__mul24( BLOCKSIZE, blockIdx.x );


    Ah += idx;
    Al += idx;


    {
        __shared__ cuDoubleComplex xhs[BLOCKSIZE];
        __shared__ cuDoubleComplex xls[BLOCKSIZE];
        __shared__ cuDoubleComplex yhs[BLOCKSIZE];
        __shared__ cuDoubleComplex yls[BLOCKSIZE];

        xhs[tid] = xh[idx*incx];
        xls[tid] = xl[idx*incx];
        yhs[tid] = yh[idx*incy];
        yls[tid] = yl[idx*incy];


        {
            int offset = tid*incy;
            yh += offset;
            yl += offset;
            offset =  tid*incx;
            xh += offset;
            xl += offset;
        }

        int               n1 = ( blockIdx.x )*BLOCKSIZE;

        cuDoubleComplex            regalphah = *alphah;
        cuDoubleComplex            regalphal = *alphal;
	cuDoubleComplex conjalphah = cuConj(regalphah);
	cuDoubleComplex conjalphal = cuConj(regalphal);

        /* Off-Diagonal Block */


        __shared__ cuDoubleComplex tilexhs[BLOCKSIZE];
        __shared__ cuDoubleComplex tilexls[BLOCKSIZE];
        __shared__ cuDoubleComplex tileyhs[BLOCKSIZE];
        __shared__ cuDoubleComplex tileyls[BLOCKSIZE];


        for( int i = 0; i < n1; i += BLOCKSIZE )
        {
            __syncthreads( );


            tilexhs[tid] = cuConj(xh[i*incx]);
            tilexls[tid] = cuConj(xl[i*incx]);
            tileyhs[tid] = cuConj(yh[i*incy]);
            tileyls[tid] = cuConj(yl[i*incy]);

            __syncthreads( );

#pragma unroll
            for( int j = 0; j < BLOCKSIZE; j++ )
            {
                cuDoubleComplex ph, pl, lAh, lAl;

                lAh = *Ah;
                lAl = *Al;

                CMUL( xhs+tid, xls+tid, tileyhs+j, tileyls+j, &ph, &pl );
                CMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );
                CADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );


                CMUL(  yhs+tid, yls+tid, tilexhs+j, tilexls+j, &ph, &pl );
                CMUL( &ph, &pl, &conjalphah, &conjalphal, &ph, &pl );
                CADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );

                if( idx < n )
                {
                    *Ah = lAh;
                    *Al = lAl;
                }

                Ah += lda;
                Al += lda;
            }
        }


        /* Diagonal Block */
            __syncthreads( );
        {
            int m;
            if( n1+BLOCKSIZE < n )
            {
                m = BLOCKSIZE;
            }
            else
            {
                m = n-n1;
            }
            __syncthreads( );
            for( int j = 0; j < m; j++ )
            {
                cuDoubleComplex ph, pl, qh, ql, lAh, lAl;

                lAh = *Ah;
                lAl = *Al;

		ph = cuConj(yhs[j]);
		pl = cuConj(yls[j]);
                CMUL( xhs+tid, xls+tid, &ph, &pl, &ph, &pl );
                CMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );

		qh = cuConj(xhs[j]);
		ql = cuConj(xls[j]);
                CMUL( &qh, &ql, yhs+tid, yls+tid, &qh, &ql );
                CMUL( &qh, &ql, &conjalphah, &conjalphal, &qh, &ql );

                CADD( &ph, &pl, &qh, &ql, &ph, &pl );


		cuDoubleComplex b = make_cuDoubleComplex(( j <= tid ),0.0);
                ph = cuCmul(b,ph);
                pl = cuCmul(b,pl);




                CADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );

		if(idx == n1 + j)
		{
			lAh = make_cuDoubleComplex(lAh.x,0.0);
			lAl = make_cuDoubleComplex(lAl.x,0.0);
		}


                if( idx < n )
                {
                    *Ah = lAh;
                    *Al = lAl;
                }

                Ah += lda;
                Al += lda;
            }
        }
    }
}

extern "C" __global__ void
gddzher2_kernel_U( int n, cuDoubleComplex *alphah, cuDoubleComplex*alphal,
                  cuDoubleComplex * xh, cuDoubleComplex *xl,
                  int incx,
                  cuDoubleComplex*yh, cuDoubleComplex*yl,
                  int incy,
                  cuDoubleComplex* Ah, cuDoubleComplex *Al, int lda )
{
    unsigned int tid = threadIdx.x;
    unsigned int idx = tid+__mul24( BLOCKSIZE, blockIdx.x );


    {
        int offset = idx+__mul24( BLOCKSIZE, blockIdx.x )*lda;
        Ah += offset;
        Al += offset;
    }


    {
        __shared__ cuDoubleComplex xhs[BLOCKSIZE];
        __shared__ cuDoubleComplex xls[BLOCKSIZE];
        __shared__ cuDoubleComplex yhs[BLOCKSIZE];
        __shared__ cuDoubleComplex yls[BLOCKSIZE];

        {
            xh += idx*incx;
            xl += idx*incx;
            yh += idx*incy;
            yl += idx*incy;
            xhs[tid] = *xh;
            xls[tid] = *xl;
            yhs[tid] = *yh;
            yls[tid] = *yl;
        }

        int    n1 = BLOCKSIZE*blockIdx.x;

        cuDoubleComplex regalphah = *alphah;
        cuDoubleComplex regalphal = *alphal;
	cuDoubleComplex conjalphah = cuConj(regalphah);
	cuDoubleComplex conjalphal = cuConj(regalphal);
        /* Diagonal Block */


        if( n1+BLOCKSIZE  < n )
        {
            n1 = BLOCKSIZE;
        }
        else
        {
            n1 = n-n1;
        }

            __syncthreads( );

        for( int j = 0; j < n1; j++ )
        {
            cuDoubleComplex ph, pl, qh, ql, lAh, lAl;

            lAh = *Ah;
            lAl = *Al;

	    ph = cuConj(yhs[j]);
	    pl = cuConj(yls[j]);
            CMUL( xhs+tid, xls+tid, &ph,&pl, &ph, &pl );
            CMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );
	    qh = cuConj(xhs[j]);
	    ql = cuConj(xls[j]);
            CMUL( &qh, &ql, yhs+tid, yls+tid, &qh, &ql );
            CMUL( &qh, &ql, &conjalphah, &conjalphal, &qh, &ql );
            CADD( &ph, &pl, &qh, &ql, &ph, &pl );


		cuDoubleComplex b = make_cuDoubleComplex(( j >= tid ),0.0);

            ph = cuCmul(b,ph);
            pl = cuCmul(b,pl);


            CADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );

		if(tid ==  j)
		{
			lAh = make_cuDoubleComplex(lAh.x,0.0);
			lAl = make_cuDoubleComplex(lAl.x,0.0);
		}

            if( idx < n )
            {
                *Ah = lAh;
                *Al = lAl;
            }


            Ah += lda;
            Al += lda;
        }

        /* Off-Diagonal Block */

        {
            int offset = BLOCKSIZE *incy;
            yh += offset;
            yl += offset;
            offset = BLOCKSIZE*incx ;
            xh += offset;
            xl += offset;
        }

        __shared__ cuDoubleComplex tilexhs[BLOCKSIZE];
        __shared__ cuDoubleComplex tilexls[BLOCKSIZE];
        __shared__ cuDoubleComplex tileyhs[BLOCKSIZE];
        __shared__ cuDoubleComplex tileyls[BLOCKSIZE];


        n1 = BLOCKSIZE*( n/BLOCKSIZE-( blockIdx.x+1 ) );

             for( int i = 0; i < n1; i += BLOCKSIZE )
        {
            __syncthreads( );


            tilexhs[tid] = cuConj(xh[i*incx]);
            tilexls[tid] = cuConj(xl[i*incx]);
            tileyhs[tid] = cuConj(yh[i*incy]);
            tileyls[tid] = cuConj(yl[i*incy]);

            __syncthreads( );

#pragma unroll
            for( int j = 0; j < BLOCKSIZE; j++ )
            {
                cuDoubleComplex ph, pl, lAh, lAl;

                lAh = *Ah;
                lAl = *Al;

                CMUL( xhs+tid, xls+tid, tileyhs+j, tileyls+j, &ph, &pl );
                CMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );
                CADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );


                CMUL(  yhs+tid, yls+tid, tilexhs+j, tilexls+j, &ph, &pl );
                CMUL( &ph, &pl, &conjalphah, &conjalphal, &ph, &pl );
                CADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );

/*                if(j==tid)*/
/*                {*/
/*                        lAh = make_cuDoubleComplex(lAh.x,0.0);*/
/*                        lAl = make_cuDoubleComplex(lAl.x,0.0);*/
/*                }*/


                if( idx < n )
                {
                    *Ah = lAh;
                    *Al = lAl;
                }

                Ah += lda;
                Al += lda;
            }
        }


        /* Last Block */

        __syncthreads( );

        int m  = n-( n1+( blockIdx.x+1 )*BLOCKSIZE );

        if( m > 0 && blockIdx.x+1 < gridDim.x )
        {


            tilexhs[tid] = cuConj(xh[n1*incx]);
            tilexls[tid] = cuConj(xl[n1*incx]);
            tileyhs[tid] = cuConj(yh[n1*incy]);
            tileyls[tid] = cuConj(yl[n1*incy]);

            __syncthreads( );

	   

#pragma unroll
            for( int j = 0; j < m; j++ )
            {
                cuDoubleComplex ph, pl, lAh, lAl;

                lAh = *Ah;
                lAl = *Al;

                CMUL( xhs+tid, xls+tid, tileyhs+j, tileyls+j, &ph, &pl );
                CMUL( &ph, &pl, &regalphah, &regalphal, &ph, &pl );
                CADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );


                CMUL(  yhs+tid, yls+tid, tilexhs+j, tilexls+j, &ph, &pl );
                CMUL( &ph, &pl, &conjalphah, &conjalphal, &ph, &pl );
                CADD( &lAh, &lAl, &ph, &pl, &lAh, &lAl );

            if( idx < n )
            {
                *Ah = lAh;
                *Al = lAl;
            }

                Ah += lda;
                Al += lda;
            }
        }
    }
}

extern "C" void
gddzher2( char* uplo, int* n, cuDoubleComplex *alphah, cuDoubleComplex*alphal,
              cuDoubleComplex * xh, cuDoubleComplex *xl,
              int* incx,
              cuDoubleComplex*yh, cuDoubleComplex*yl,
              int* incy,
              cuDoubleComplex* Ah, cuDoubleComplex *Al, int* lda  )
{
    dim3 blocks( ( *n+BLOCKSIZE-1 )/BLOCKSIZE, 1 );

    dim3 threads( BLOCKSIZE, 1 );

    if( *uplo == 'L' || *uplo == 'l' )
    {
        gddzher2_kernel_L<<< blocks, threads, 0, gddstream>>>
        ( *n, alphah, alphal,  xh, xl, *incx,
          yh, yl, *incy, Ah, Al, *lda );
    }
    else if( *uplo == 'U' || *uplo == 'u' )
    {
        gddzher2_kernel_U<<< blocks, threads, 0, gddstream>>>
        ( *n, alphah, alphal,  xh, xl, *incx,
          yh, yl, *incy, Ah, Al, *lda );
    
    }
    else
    {
	    // ???
    
    }
}

