#include "gddcommon.h"
#include <omp.h>
#include <cutil.h>
#include "multiple_gpu.h"
#include "config.h"

namespace multiple_gpus
{
int num_gpus;
int num_omps;
#define _DEBUG_



extern "C"
void
bind_cpu_to_gpu_( int *ngpu )
{
    CUDA_SAFE_CALL( cudaGetDeviceCount( &num_gpus ) );

    *ngpu = num_gpus;
    num_omps = omp_get_max_threads( );
#ifdef _DEBUG_
    printf( "Number of Host OpenMP Threads : %d\n", num_omps );
    printf( "Number of CUDA Device         : %d\n", num_gpus );
#endif

    if( num_gpus < 1 )
    {
	printf( "There is no CUDA Device. Program Stops.\n" );
	exit( -1 );
    }

    if( num_gpus > num_omps )
    {
	printf(
	    "Number of Host OpenMP Threads exceeds the Number of GPUs.\n" );
	exit( -1 );
    }
//  bind gpu card <=> openmp thread
#pragma omp parallel for
  for (int igx=0;igx<num_gpus;igx++) {
    int omp_thread_num  = omp_get_thread_num();
    int cuda_device_num = igx % num_gpus;
#ifdef _DEBUG_
    printf("OpenMP Thread = %d  Cuda Device = %d\n",omp_thread_num,cuda_device_num);
#endif
    if (num_gpus > 1) CUDA_SAFE_CALL(cudaSetDevice(cuda_device_num));
  }  
#ifdef _DEBUG_
  fflush(stdout);
#endif
}




//1GPUでのテスト用
extern "C"
void
bind_cpu_to_gpu_test_( int *ngpu )
{
    CUDA_SAFE_CALL( cudaGetDeviceCount( &num_gpus ) );

    //test用
    num_gpus = 2;
    *ngpu = num_gpus;
    num_omps = omp_get_max_threads( );
#ifdef _DEBUG_
    printf( "Number of Host OpenMP Threads : %d\n", num_omps );
    printf( "Number of CUDA Device         : %d\n", num_gpus );
#endif

    if( num_gpus < 1 )
    {
        printf( "There is no CUDA Device. Program Stops.\n" );
        exit( -1 );
    }

    if( num_gpus > num_omps )
    {
        printf(
            "Number of Host OpenMP Threads exceeds the Number of GPUs.\n" );
        exit( -1 );
    }
//  bind gpu card <=> openmp thread
#pragma omp parallel for
  for (int igx=0;igx<num_gpus;igx++) {
    int omp_thread_num  = omp_get_thread_num();
    //テスト用
    int cuda_device_num = 0;
/*    int cuda_device_num = igx % num_gpus;*/
#ifdef _DEBUG_
    printf("OpenMP Thread = %d  Cuda Device = %d\n",omp_thread_num,cuda_device_num);
#endif
    if (num_gpus > 1) CUDA_SAFE_CALL(cudaSetDevice(cuda_device_num));
  }  
#ifdef _DEBUG_
  fflush(stdout);
#endif
}

extern "C" 
void get_num_gpu_(int *ngpu){
  *ngpu = num_gpus;
}

extern "C" void 
unbind_cpu_to_gpu_(){
#pragma omp parallel for
  for (int igx=0;igx<num_gpus;igx++) {
    if (num_gpus > 1) CUDA_SAFE_CALL(cudaThreadExit());
  }
#ifdef _DEBUG_
  fflush(stdout);
#endif
}



}

