#include "gdd_basic_inline.h"
#include "cutil.h"
#ifndef _BASICALGEBLA_
 #define _BASICALGEBLA_

template <int blockSize> __global__ static void
reduce_kernel( double* g_idatah, double *g_idatal, double* g_odatah,
               double*g_odatal,
               unsigned int n )
{

    __shared__ double sdatah[blockSize];
    __shared__ double sdatal[blockSize];
    unsigned int             tid = threadIdx.x;
    unsigned int             i = blockIdx.x*( blockSize*2 )+tid;
    unsigned int             gridSize = blockSize*2*gridDim.x;

    sdatah[tid] = 0.0;
    sdatal[tid] = 0.0;

    while( i < n )
    {
/*        sdata[tid] += g_idata[i]+g_idata[i+blockSize];*/
        RADD( g_idatah+i, g_idatal+i,
              sdatah+tid, sdatal+tid,
              sdatah+tid, sdatal+tid );
        if( i+blockSize < n )
        {
            RADD( g_idatah+i+blockSize, g_idatal+i+blockSize,
                  sdatah+tid, sdatal+tid,
                  sdatah+tid, sdatal+tid );
        }


        i += gridSize;
    }

    __syncthreads( );

    if( blockSize >= 512 )
    {
        if( tid < 256 )
        {
/*            sdata[tid] += sdata[tid+256];*/
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+256, sdatal+tid+256,
                  sdatah+tid, sdatal+tid );
        }
        __syncthreads( );
    }
    if( blockSize >=        256 )
    {
        if( tid < 128 )
        {
/*            sdata[tid] += sdata[tid+128];*/
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+128, sdatal+tid+128,
                  sdatah+tid, sdatal+tid );
        }
        __syncthreads( );
    }
    if( blockSize >=        128 )
    {
        if( tid <  64 )
        {
/*            sdata[tid] += sdata[tid+64];*/
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+64, sdatal+tid+64,
                  sdatah+tid, sdatal+tid );
        }
        __syncthreads( );
    }

    if( tid < 32 )
    {
        if( blockSize >=  64 )
        {
/*            sdata[tid] += sdata[tid+32];*/
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+32, sdatal+tid+32,
                  sdatah+tid, sdatal+tid );
        }
        if( blockSize >=  32 )
        {
/*                sdata[tid] += sdata[tid+16]; */
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+16, sdatal+tid+16,
                  sdatah+tid, sdatal+tid );
        }
        if( blockSize >=  16 )
        {
/*                sdata[tid] += sdata[tid+8];*/
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+8, sdatal+tid+8,
                  sdatah+tid, sdatal+tid );
        }
        if( blockSize >=   8 )
        {
/*                       sdata[tid] += sdata[tid+4];*/
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+4, sdatal+tid+4,
                  sdatah+tid, sdatal+tid );
        }
        if( blockSize >=   4 )
        {
/*                       sdata[tid] += sdata[tid+2];*/
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+2, sdatal+tid+2,
                  sdatah+tid, sdatal+tid );
        }
        if( blockSize >=   2 )
        {
/*                sdata[tid] += sdata[tid+1]; */
            RADD( sdatah+tid, sdatal+tid, sdatah+tid+1, sdatal+tid+1,
                  sdatah+tid, sdatal+tid );
        }
    }


    if( tid == 0 )
    {
        g_odatah[blockIdx.x] = sdatah[0];
        g_odatal[blockIdx.x] = sdatal[0];
    }
}




template <int blockSize> __global__ static void
reduce_kernel_complex( cuDoubleComplex* g_idatah, cuDoubleComplex *g_idatal, cuDoubleComplex* g_odatah,
               cuDoubleComplex*g_odatal,
               unsigned int n )
{

    __shared__ cuDoubleComplex sdatah[blockSize];
    __shared__ cuDoubleComplex sdatal[blockSize];
    unsigned int             tid = threadIdx.x;
    unsigned int             i = blockIdx.x*( blockSize*2 )+tid;
    unsigned int             gridSize = blockSize*2*gridDim.x;

    sdatah[tid] = make_cuDoubleComplex(0.0,0.0);
    sdatal[tid] = make_cuDoubleComplex(0.0,0.0);

    while( i < n )
    {
/*        sdata[tid] += g_idata[i]+g_idata[i+blockSize];*/
        CADD( g_idatah+i, g_idatal+i,
              sdatah+tid, sdatal+tid,
              sdatah+tid, sdatal+tid );
        if( i+blockSize < n )
        {
            CADD( g_idatah+i+blockSize, g_idatal+i+blockSize,
                  sdatah+tid, sdatal+tid,
                  sdatah+tid, sdatal+tid );
        }


        i += gridSize;
    }

    __syncthreads( );

    if( blockSize >= 512 )
    {
        if( tid < 256 )
        {
/*            sdata[tid] += sdata[tid+256];*/
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+256, sdatal+tid+256,
                  sdatah+tid, sdatal+tid );
        }
        __syncthreads( );
    }
    if( blockSize >=        256 )
    {
        if( tid < 128 )
        {
/*            sdata[tid] += sdata[tid+128];*/
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+128, sdatal+tid+128,
                  sdatah+tid, sdatal+tid );
        }
        __syncthreads( );
    }
    if( blockSize >=        128 )
    {
        if( tid <  64 )
        {
/*            sdata[tid] += sdata[tid+64];*/
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+64, sdatal+tid+64,
                  sdatah+tid, sdatal+tid );
        }
        __syncthreads( );
    }

    if( tid < 32 )
    {
        if( blockSize >=  64 )
        {
/*            sdata[tid] += sdata[tid+32];*/
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+32, sdatal+tid+32,
                  sdatah+tid, sdatal+tid );
        }
        if( blockSize >=  32 )
        {
/*                sdata[tid] += sdata[tid+16]; */
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+16, sdatal+tid+16,
                  sdatah+tid, sdatal+tid );
        }
        if( blockSize >=  16 )
        {
/*                sdata[tid] += sdata[tid+8];*/
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+8, sdatal+tid+8,
                  sdatah+tid, sdatal+tid );
        }
        if( blockSize >=   8 )
        {
/*                       sdata[tid] += sdata[tid+4];*/
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+4, sdatal+tid+4,
                  sdatah+tid, sdatal+tid );
        }
        if( blockSize >=   4 )
        {
/*                       sdata[tid] += sdata[tid+2];*/
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+2, sdatal+tid+2,
                  sdatah+tid, sdatal+tid );
        }
        if( blockSize >=   2 )
        {
/*                sdata[tid] += sdata[tid+1]; */
            CADD( sdatah+tid, sdatal+tid, sdatah+tid+1, sdatal+tid+1,
                  sdatah+tid, sdatal+tid );
        }
    }


    if( tid == 0 )
    {
        g_odatah[blockIdx.x] = sdatah[0];
        g_odatal[blockIdx.x] = sdatal[0];
    }
}





template<typename T> 
inline static void
elemcopy( T *src, T *dest )
{
    CUDA_SAFE_CALL( cudaMemcpy( dest, src, sizeof( T ),
                                cudaMemcpyDeviceToDevice ) );
}


__global__ void static
set_one_kernel( double *dalphah, double* dalphal )
{
    *dalphah = 1.0;
    *dalphal = 0.0;
}



#endif

