#include "config.h"
#ifndef _INLINE_BASIC_
 #define _INLINE_BASIC_
/* */
/* For Addition */
/* */


static __device__ __host__ inline void
gdd_mul_host( double *ahi, double*alo, double *bhi, double*blo, double*chi,
              double*clo )
{
    double p;
    double a_hh, b_hh;
    double a_hl, b_hl;
    double e;

    p = *ahi*( *bhi );
    a_hh = 134217729.0*( *ahi );
    a_hh = a_hh-( a_hh-*ahi );
    a_hl = *ahi-a_hh;
    b_hh = 134217729.0*( *bhi );
    b_hh = b_hh-( b_hh-( *bhi ) );
    b_hl = ( *bhi )-b_hh;
    e = ( ( a_hh*b_hh-p )+a_hh*b_hl+a_hl*b_hh )+a_hl*b_hl;
    e = e+( ( *ahi )*( *blo )+( *alo )*( *bhi ) );

    *chi = p+e;
    *clo = e-( *chi-p );
}



static __host__ __device__ inline void
two_sum( double *a, double *b, double *sum, double * err )
{
    *sum = *a+*b;
    double v = *sum-*a;
    *err = ( *a-( *sum-v ) )+( *b-v );
}

static __host__ __device__ inline void
ctwo_sum( cuDoubleComplex *a, cuDoubleComplex *b, cuDoubleComplex *sum,
          cuDoubleComplex * err )
{
    *sum = cuCadd( *a, *b );
    cuDoubleComplex v = cuCsub( *sum, *a );
    *err = cuCadd( cuCsub( *a, cuCsub( *sum, v ) ), cuCsub( *b, v ) );
}

static __host__ __device__  inline void
quick_two_sum( double *a, double *b, double*sum, double *err )
{
    *sum = *a+*b;
    *err = *b-( *sum-*a );
}

static __host__ __device__  inline void
cquick_two_sum( cuDoubleComplex *a, cuDoubleComplex *b, cuDoubleComplex*sum,
                cuDoubleComplex *err )
{
    *sum = cuCadd( *a, *b );
    *err = cuCsub( *b, cuCsub( *sum, *a ) );
}

static __host__ __device__ inline void
two_diff( double *a, double *b, double*sum, double*err )
{
    *sum = *a-*b;
    double v = *sum-*a;
    *err = ( *a-( *sum-v ) )-( *b+v );
}

static __host__ __device__ inline void
ctwo_diff( cuDoubleComplex *a, cuDoubleComplex *b, cuDoubleComplex*sum,
           cuDoubleComplex*err )
{
    *sum = cuCsub( *a, *b );
    cuDoubleComplex v = cuCsub( *sum, *a );
    *err = cuCsub( cuCsub( *a, cuCsub( *sum, v ) ), cuCadd( *b, v ) );
}

static __host__ __device__  inline void
gdd_add( double *ahi, double *alo, double *bhi, double*blo, double*chi,
         double *clo )
{
    double sum, err;

    two_sum( ahi, bhi, &sum, &err );
    err += *alo+*blo;
    quick_two_sum( &sum, &err, chi, clo );
}

static __host__ __device__  inline void
cgdd_add( cuDoubleComplex *ahi, cuDoubleComplex *alo, cuDoubleComplex *bhi,
          cuDoubleComplex*blo, cuDoubleComplex*chi,
          cuDoubleComplex *clo )
{
    cuDoubleComplex sum, err;

    ctwo_sum( ahi, bhi, &sum, &err );
    err = cuCadd( *alo, err ) ;
    err = cuCadd( *blo, err ) ;
    cquick_two_sum( &sum, &err, chi, clo );
}

static __host__ __device__ inline void
gdd_add_expand( double *ahi, double *alo, double*bhi, double *blo,
                double *chi,
                double *clo )
{
    double tmp;
    double v;

    /* two_sum */
    tmp = *ahi+*bhi;
    v = tmp-*ahi;
    v = ( ( *bhi-v )+( *ahi-( tmp-v ) ) );

    /* e += (alo + blo); */
    v = v+( *alo+*blo );

    /* quick_two_sum */
    *chi = tmp+v;
    *clo = v-( *chi-tmp );
}

static __host__ __device__  inline void
gdd_diff( double *ahi, double *alo, double *bhi, double*blo, double*chi,
          double *clo )
{
    double sum, err;

    two_diff( ahi, bhi, &sum, &err );
    err += *alo;
    err -= *blo;
    quick_two_sum( &sum, &err, chi, clo );
}

static __host__ __device__  inline void
cgdd_diff( cuDoubleComplex *ahi, cuDoubleComplex *alo, cuDoubleComplex *bhi,
           cuDoubleComplex*blo, cuDoubleComplex*chi,
           cuDoubleComplex *clo )
{
    cuDoubleComplex sum, err;

    ctwo_diff( ahi, bhi, &sum, &err );
    err = cuCsub( cuCadd( err, *alo ), *blo );
    cquick_two_sum( &sum, &err, chi, clo );
}

/* */
/* For multiplication */
/* */


 #ifdef __cplusplus

static __device__ inline void
gdd_mul_fma( double *ahi, double*alo, double*bhi, double*blo,
                    double*chi,
                    double*clo )
{
    double tmp1;
    double v;

    tmp1 = *ahi*( *bhi );

    v = __fma_rn( *ahi, *bhi, -1.0*tmp1 );

    v = v+( *ahi*( *blo )+*alo*( *bhi ) );

    *chi = tmp1+v;
    *clo = v-( *chi-tmp1 );
}

static __device__  inline void
gdd_mul_fma_sloppy( double *ahi, double *alo, double*bhi, double*blo, double*chi,
             double*clo )
{
    double tmp1, tmp2, v;

    tmp1  = *ahi*( *blo );
    tmp1 += *alo*( *bhi );

    v = __fma_rn( *ahi, *bhi, tmp1 );
    tmp2 = v;
    v = __fma_rn( *ahi, *bhi, -1.0*v );
    v = v+tmp1;
    *chi = tmp2;
    *clo = v;
}

static __device__ inline void
gdd_sqrt( double * ah, double * al )
{
    if( *ah == 0.0 && *al == 0.0 )
    {
        return;
    }
    else if( *ah < 0 )
    {
        printf( "gdd_sqrt:N must be positive.\n" );
    }

    double yh, yl, zh, zl;
    double zero = 0.0, half = 0.5;

    double x = 1.0/sqrt( *ah );
    double ax = ( *ah )*x;
    RMUL( &x, &zero, &half, &zero, &yh, &yl );
    RMUL( &ax, &zero, &ax, &zero, &zh, &zl );
    RDIFF( ah, al, &zh, &zl, &zh, &zl );
    RMUL( &zh, &zl, &yh, &yl, &zh, &zl );
    RADD( &ax, &zero, &zh, &zl, ah, al );
}



static __device__ inline void
gdd_div_fma_sloppy( double *ahi, double*alo, double*bhi, double*blo,
                    double*chi,double*clo)
{
	double rh,rl;
	double s1,s2;
	double zero = 0.0;
	double q1,q2;
       q1	= (*ahi)/(*bhi);

	RMUL(&q1,&zero,bhi,blo,&rh,&rl);
	two_diff(ahi,&rh,&s1,&s2);
	s2 -= rl;
	s2 += *alo;

	q2 = (s1+s2)/(*bhi);

	quick_two_sum(&q1,&q2,chi,clo);

}


/*static __device__ inline void*/
/*gdd_sqrt( double * ah, double * al )*/
/*{*/
/*    if( *ah == 0.0 && *al == 0.0 )*/
/*    {*/
/*        return;*/
/*    }*/
/*    else if( *ah < 0 )*/
/*    {*/
/*        printf( "gdd_sqrt:N must be positive.\n" );*/
/*    }*/

/*    double yh,yl,zh,zl,th,tl,ch,cl;*/
/*    double zero = 0.0,half = 0.5;*/

/*    double x = 1.0/sqrt( *ah );*/
/*    double ax = ( *ah )*x;*/
/*    RMUL(&x ,&zero,&half,&zero,&yh,&yl);*/
/*    RMUL(&ax,&zero,&ax,&zero,&zh,&zl);*/
/*    RDIFF(ah,al,&zh,&zl,&th,&tl);*/
/*    RMUL(&th,&tl,&yh,&yl,&ch,&cl);*/
/*    RADD(&ax,&zero,&ch,&cl,ah,al);*/
/*}*/


/* if( a < b ) */
/*  {flag = false; */
/*  a = b} */
/*  else */
/*  {flag=true;} */


static __device__ __host__ inline void
gdd_1stvalue_is_larger( double*ahi, double*alo, double*bhi, double*blo,
                        bool * flag )
{
    if( *ahi < *bhi || ( *ahi == *bhi && *alo < *blo ) )
    {
        *ahi = *bhi;
        *alo = *blo;
	*flag = false;
    }
    else
    {
	*flag = true;
    }
}




 #endif

static __host__ __device__ inline void
comp( double*ahi, double*alo, double*bhi, double*blo,
                        int * flag )
{
    if( *ahi < *bhi || ( *ahi == *bhi && *alo < *blo ) )
    {
        *flag = 0;
    }
    else
    {
        *flag = 1;
    }
}




static __device__ __host__ inline void
gdd_div_host( double *ahi, double*alo, double *bhi, double*blo, double*chi,
              double*clo )
{
	double rh,rl;
	double s1,s2;
	double zero = 0.0;
	double q1,q2;
       q1	= (*ahi)/(*bhi);

	HOSTRMUL(&q1,&zero,bhi,blo,&rh,&rl);
	two_diff(ahi,&rh,&s1,&s2);
	s2 -= rl;
	s2 += *alo;
/*        s2 += alo;*/

	q2 = (s1+s2)/(*bhi);

	quick_two_sum(&q1,&q2,chi,clo);
}


static __device__ __host__ inline void
cgdd_mul_host( cuDoubleComplex *ahi, cuDoubleComplex*alo,
               cuDoubleComplex *bhi, cuDoubleComplex*blo,
               cuDoubleComplex*chi,
               cuDoubleComplex*clo )
{
    cuDoubleComplex p;
    cuDoubleComplex a_hh, b_hh;
    cuDoubleComplex a_hl, b_hl;
    cuDoubleComplex e;

    p = cuCmul( *ahi, ( *bhi ) );
    a_hh = cuCmul( make_cuDoubleComplex( 134217729.0, 0.0 ), ( *ahi ) );
    a_hh = cuCsub( a_hh, cuCsub( a_hh, *ahi ) );
    a_hl = cuCsub( *ahi, a_hh );
    b_hh = cuCmul( make_cuDoubleComplex( 134217729.0, 0.0 ), ( *bhi ) );
    b_hh = cuCsub( b_hh, cuCsub( b_hh, ( *bhi ) ) );
    b_hl = cuCsub( ( *bhi ), b_hh );
    e =cuCadd( cuCadd( cuCsub( cuCmul( a_hh,b_hh ), p ),
                       cuCadd( cuCmul( a_hh, b_hl ), 
			       cuCmul( a_hl,b_hh ) ) ), 
		    cuCmul(a_hl,b_hl) );
    e = cuCadd(e,cuCadd( cuCmul(*ahi,*blo),cuCmul( *alo , *bhi ) ));

    *chi = cuCadd(p,e);
    *clo = cuCsub(e,cuCsub( *chi,p ));
}

/*static __device__ __host__ inline void*/
/*gdd_sqrt_host( double * ah, double * al )*/
/*{*/
/*    if( *ah == 0.0 && *al == 0.0 )*/
/*    {*/
/*        return;*/
/*    }*/
/*    else if( *ah < 0 )*/
/*    {*/
/*        printf( "gdd_sqrt:N must be positive.\n" );*/
/*    }*/

/*    double yh,yl,zh,zl,th,tl,ch,cl;*/
/*    double zero = 0.0,half = 0.5;*/

/*    double x = 1.0/sqrt( *ah );*/
/*    double ax = ( *ah )*x;*/
/*    HOSTRMUL(&x ,&zero,&half,&zero,&yh,&yl);*/
/*    HOSTRMUL(&ax,&zero,&ax,&zero,&zh,&zl);*/
/*    HOSTRDIFF(ah,al,&zh,&zl,&th,&tl);*/
/*    HOSTRMUL(&th,&tl,&yh,&yl,&ch,&cl);*/
/*    HOSTRADD(&ax,&zero,&ch,&cl,ah,al);*/
/*}*/

static __device__ __host__ inline void
gdd_sqrt_host( double * ah, double * al )
{
    if( *ah == 0.0 && *al == 0.0 )
    {
        return;
    }
    else if( *ah < 0 )
    {
        printf( "gdd_sqrt:N must be positive.\n" );
    }

    double yh, yl, zh, zl;
    double zero = 0.0, half = 0.5;

    double x = 1.0/sqrt( *ah );
    double ax = ( *ah )*x;
    HOSTRMUL( &ax, &zero, &ax, &zero, &zh, &zl );
    HOSTRDIFF( ah, al, &zh, &zl, &zh, &zl );
    HOSTRMUL( &x, &zero, &half, &zero, &yh, &yl );
    HOSTRMUL( &zh, &zl, &yh, &yl, &zh, &zl );
    HOSTRADD( &ax, &zero, &zh, &zl, ah, al );
}

static __device__ __host__ inline void
gdd_abs( double*ah, double*al )
{
    if( *ah < 0.0 )
    {
        *ah = -( *ah );
        *al = -( *al );
    }
}

static __device__ __host__ inline void
gdd_sign(double *a1h,double*a1l,double*a2h,double*a2l,double*a3h,double*a3l)
{
	*a3h = * a1h;
	*a3l = * a1l;
	gdd_abs(a3h,a3l);
	if(*a2h < 0.0 || (*a2h == 0.0 && *a2l < 0.0))
	{
		*a3h = (-1.0)*(*a3h);
		*a3l = (-1.0)*(*a3l);
	}
}



#endif


