#include "gddcommon.h"
#include "basic_algebra_host.h"


#ifdef __cplusplus
extern "C" {
#endif
void
CalcEachArraySize( int n, int partitionNumber, int*partitions,
                   int *partitionSum )
{
    int quo  = n/partitionNumber;
    int mod = n%partitionNumber;

    int sum = 0;
    int i;

    for( i = 0; i < partitionNumber; i++ )
    {
        partitionSum[i] = sum;

        partitions[i] = quo;
        if( i < mod )
        {
            partitions[i]++;
        }

        sum += partitions[i];
#ifdef _DEBUG_
        printf( "pt[%d]=%d,ps[%d]=%d\n", i, partitions[i], i,
                partitionSum[i] );
#endif
    }
}

void
copyToSplitMatrix( int m, int k, int lda, int numSplit, int *partitions,
                   double* Mhin, double* Mlin, double *Mhout,
                   double*Mlout )
{
/* #undef _OPENMP */
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int colidx, spidx, ptidx;
        for( colidx = 0; colidx < k; colidx++ )
        {
            int sumOfPartition = 0;
            for( spidx = 0; spidx < numSplit; spidx++ )
            {
                int ptrow = partitions[spidx];
#ifdef _OPENMP
 #pragma omp for
#endif
                for( ptidx = 0; ptidx < ptrow; ptidx++ )
                {
                    int rowidx = ptidx+sumOfPartition;


                    int inindex = rowidx+lda*colidx;
                    int outindex = ptidx+ptrow*colidx+sumOfPartition*k;

/*                        printf("in:%d,out:%d\n",inindex,outindex);*/

                    *( Mhout+outindex ) = *( Mhin+inindex );
                    *( Mlout+outindex ) = *( Mlin+inindex );
                }

                sumOfPartition += ptrow;
            }
        }
    }
}

void
copyToComplexSplitMatrix( int m, int k, int lda, int numSplit,
                          int *partitions,
                          cuDoubleComplex* Mhin, cuDoubleComplex* Mlin,
                          cuDoubleComplex *Mhout,
                          cuDoubleComplex*Mlout )
{
/* #undef _OPENMP */
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int colidx, spidx, ptidx;
        for( colidx = 0; colidx < k; colidx++ )
        {
            int sumOfPartition = 0;
            for( spidx = 0; spidx < numSplit; spidx++ )
            {
                int ptrow = partitions[spidx];
#ifdef _OPENMP
 #pragma omp for
#endif
                for( ptidx = 0; ptidx < ptrow; ptidx++ )
                {
                    int rowidx = ptidx+sumOfPartition;


                    int inindex = rowidx+lda*colidx;
                    int outindex = ptidx+ptrow*colidx+sumOfPartition*k;

/*                        printf("in:%d,out:%d\n",inindex,outindex);*/

                    *( Mhout+outindex ) = *( Mhin+inindex );
                    *( Mlout+outindex ) = *( Mlin+inindex );
                }

                sumOfPartition += ptrow;
            }
        }
    }
}

void
copyToComplexConjugateSplitMatrix( int m, int k, int lda, int numSplit,
                                   int *partitions,
                                   cuDoubleComplex* Mhin,
                                   cuDoubleComplex* Mlin,
                                   cuDoubleComplex *Mhout,
                                   cuDoubleComplex*Mlout )
{
/* #undef _OPENMP */
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int colidx, spidx, ptidx;
        for( colidx = 0; colidx < k; colidx++ )
        {
            int sumOfPartition = 0;
            for( spidx = 0; spidx < numSplit; spidx++ )
            {
                int ptrow = partitions[spidx];
#ifdef _OPENMP
 #pragma omp for
#endif
                for( ptidx = 0; ptidx < ptrow; ptidx++ )
                {
                    int rowidx = ptidx+sumOfPartition;


                    int inindex = rowidx+lda*colidx;
                    int outindex = ptidx+ptrow*colidx+sumOfPartition*k;

/*                        printf("in:%d,out:%d\n",inindex,outindex);*/

                    *( Mhout+outindex ) = cuConj( *( Mhin+inindex ) );
                    *( Mlout+outindex ) = cuConj( *( Mlin+inindex ) );
                }

                sumOfPartition += ptrow;
            }
        }
    }
}

void
copyTranspose( int k, int n, int ldb, double* Mhin, double* Mlin,
               double *Mhout,
               double*Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int rowidx, colidx;
        for( rowidx = 0; rowidx < k; rowidx++ )
        {
#ifdef _OPENMP
 #pragma omp for
#endif
            for( colidx = 0; colidx < n; colidx++ )
            {
                int inindex = colidx+ldb*rowidx;
                int outindex = rowidx+colidx*k;
#ifdef _DEBUG_
/*                printf( "in:%d,out:%d\n", inindex, outindex );*/
#endif
                *( Mhout+outindex ) = *( Mhin+inindex );
                *( Mlout+outindex ) = *( Mlin+inindex );
            }
        }
    }
}

void
copyComplexTranspose( int k, int n, int ldb, cuDoubleComplex* Mhin,
                      cuDoubleComplex* Mlin,
                      cuDoubleComplex *Mhout,
                      cuDoubleComplex *Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int rowidx, colidx;
        for( rowidx = 0; rowidx < k; rowidx++ )
        {
#ifdef _OPENMP
 #pragma omp for
#endif
            for( colidx = 0; colidx < n; colidx++ )
            {
                int inindex = colidx+ldb*rowidx;
                int outindex = rowidx+colidx*k;
#ifdef _DEBUG_
/*                printf( "in:%d,out:%d\n", inindex, outindex );*/
#endif
                *( Mhout+outindex ) = *( Mhin+inindex );
                *( Mlout+outindex ) = *( Mlin+inindex );
            }
        }
    }
}

void
copyHermiteTranspose( int k, int n, int ldb, cuDoubleComplex* Mhin,
                      cuDoubleComplex* Mlin,
                      cuDoubleComplex *Mhout,
                      cuDoubleComplex *Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int rowidx, colidx;
        for( rowidx = 0; rowidx < k; rowidx++ )
        {
#ifdef _OPENMP
 #pragma omp for
#endif
            for( colidx = 0; colidx < n; colidx++ )
            {
                int inindex = colidx+ldb*rowidx;
                int outindex = rowidx+colidx*k;
#ifdef _DEBUG_
/*                printf( "in:%d,out:%d\n", inindex, outindex );*/
#endif
                *( Mhout+outindex ) = cuConj( *( Mhin+inindex ) );
                *( Mlout+outindex ) = cuConj( *( Mlin+inindex ) );
            }
        }
    }
}

void
copyFromSymmetricSplitMatrix( int n, int ldc, char*uplo, int numSplit,
                              int * partitions,
                              double *Mhin, double*Mlin, double*Mhout,
                              double*Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int upflag = 1;
        if( *uplo == 'L' || *uplo == 'l' )
        {
            upflag = 0;
        }
        int colidx, spidx, ptidx;
        for( colidx = 0; colidx < n; colidx++ )
        {
            int sumOfPartition = 0;
            for( spidx = 0; spidx < numSplit; spidx++ )
            {
                int ptrow = partitions[spidx];
#ifdef _OPENMP
 #pragma omp for
#endif
                for( ptidx = 0; ptidx < ptrow; ptidx++ )
                {
                    int rowidx = ptidx+sumOfPartition;

                    if( !( ( upflag
                           && rowidx > colidx )
                         || ( !upflag && rowidx < colidx ) ) )
                    {
                        int outindex = rowidx+ldc*colidx;
                        int inindex = ptidx+ptrow*colidx+sumOfPartition*n;

/*                        printf("in:%d,out:%d\n",inindex,outindex);*/

                        *( Mhout+outindex ) = *( Mhin+inindex );
                        *( Mlout+outindex ) = *( Mlin+inindex );
                    }
                }

                sumOfPartition += ptrow;
            }
        }
    }
}

void
copyFromHermitianSplitMatrix( int n, int ldc, char*uplo, int numSplit,
                              int * partitions,
                              cuDoubleComplex *Mhin, cuDoubleComplex*Mlin,
                              cuDoubleComplex*Mhout,
                              cuDoubleComplex*Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int upflag = 1;
        if( *uplo == 'L' || *uplo == 'l' )
        {
            upflag = 0;
        }
        int colidx, spidx, ptidx;
        for( colidx = 0; colidx < n; colidx++ )
        {
            int sumOfPartition = 0;
            for( spidx = 0; spidx < numSplit; spidx++ )
            {
                int ptrow = partitions[spidx];
#ifdef _OPENMP
 #pragma omp for
#endif
                for( ptidx = 0; ptidx < ptrow; ptidx++ )
                {
                    int rowidx = ptidx+sumOfPartition;

                    if( !( ( upflag
                           && rowidx > colidx )
                         || ( !upflag && rowidx < colidx ) ) )
                    {
                        int outindex = rowidx+ldc*colidx;
                        int inindex = ptidx+ptrow*colidx+sumOfPartition*n;

/*                        printf("in:%d,out:%d\n",inindex,outindex);*/

                        *( Mhout+outindex ) = *( Mhin+inindex );
                        *( Mlout+outindex ) = *( Mlin+inindex );
                    }
                }

                sumOfPartition += ptrow;
            }
        }
    }
}

void
copyFromSplitMatrix( int m, int n, int ldc, int numSplit, int * partitions,
                     double *Mhin, double*Mlin, double*Mhout,
                     double*Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int colidx, spidx, ptidx;
        for( colidx = 0; colidx < n; colidx++ )
        {
            int sumOfPartition = 0;
            for( spidx = 0; spidx < numSplit; spidx++ )
            {
                int ptrow = partitions[spidx];
#ifdef _OPENMP
 #pragma omp for
#endif
                for( ptidx = 0; ptidx < ptrow; ptidx++ )
                {
                    int rowidx = ptidx+sumOfPartition;


                    int outindex = rowidx+ldc*colidx;
                    int inindex = ptidx+ptrow*colidx+sumOfPartition*n;

/*                        printf("in:%d,out:%d\n",inindex,outindex);*/

                    *( Mhout+outindex ) = *( Mhin+inindex );
                    *( Mlout+outindex ) = *( Mlin+inindex );
                }

                sumOfPartition += ptrow;
            }
        }
    }
}

void
copyFromComplexSplitMatrix( int m, int n, int ldc, int numSplit,
                            int * partitions,
                            cuDoubleComplex *Mhin, cuDoubleComplex*Mlin,
                            cuDoubleComplex*Mhout,
                            cuDoubleComplex*Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int colidx, spidx, ptidx;
        for( colidx = 0; colidx < n; colidx++ )
        {
            int sumOfPartition = 0;
            for( spidx = 0; spidx < numSplit; spidx++ )
            {
                int ptrow = partitions[spidx];
#ifdef _OPENMP
 #pragma omp for
#endif
                for( ptidx = 0; ptidx < ptrow; ptidx++ )
                {
                    int rowidx = ptidx+sumOfPartition;


                    int outindex = rowidx+ldc*colidx;
                    int inindex = ptidx+ptrow*colidx+sumOfPartition*n;

/*                        printf("in:%d,out:%d\n",inindex,outindex);*/

                    *( Mhout+outindex ) = *( Mhin+inindex );
                    *( Mlout+outindex ) = *( Mlin+inindex );
                }

                sumOfPartition += ptrow;
            }
        }
    }
}

void
copyToTriangleSplitMatrix( char * uplo, char*trans, char*diag, int m,
                           int lda,
                           int numSplit, int *partitions,
                           double* Mhin, double* Mlin, double *Mhout,
                           double*Mlout )
{
/* #undef _OPENMP */
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int btransflag = 0;
        if( *trans == 't' || *trans == 'T' ) btransflag = 1;

        int bupflag  = 0;
        if( *uplo == 'u' || *uplo == 'U' ) bupflag = 1;


        int bdiagflag = 0;
        if( *diag == 'u' || *diag == 'U' ) bdiagflag = 1;


        int colidx, spidx, ptidx;
        for( colidx = 0; colidx < m; colidx++ )
        {
            int sumOfPartition = 0;
            for( spidx = 0; spidx < numSplit; spidx++ )
            {
                int ptrow = partitions[spidx];
#ifdef _OPENMP
 #pragma omp for
#endif
                for( ptidx = 0; ptidx < ptrow; ptidx++ )
                {
                    int    rowidx = ptidx+sumOfPartition;
                    int    inindex;
                    if( !btransflag )
                    {
                        inindex = rowidx+lda*colidx;
                    }
                    else
                    {
                        inindex = colidx+rowidx*lda;
                    }

                    int    outindex = ptidx+ptrow*colidx+sumOfPartition*m;
                    double tmph = *( Mhin+inindex );
                    double tmpl = *( Mlin+inindex );

                    if( bupflag )
                    {
                        if( colidx < rowidx )
                        {
                            tmph = 0.0;
                            tmpl = 0.0;
                        }
                    }
                    else
                    {
                        if( colidx > rowidx )
                        {
                            tmph = 0.0;
                            tmpl = 0.0;
                        }
                    }

                    if( bdiagflag  && ( colidx == rowidx ) )
                    {
                        tmph = 1.0;
                        tmpl = 0.0;
                    }

                    *( Mhout+outindex ) = tmph;
                    *( Mlout+outindex ) = tmpl;
                }

                sumOfPartition += ptrow;
            }
        }
    }
}

void
copyToTriangleMatrix( char * uplo,  char*diag, int m,
                      int lda,
                      double* Mhin, double* Mlin, double *Mhout,
                      double*Mlout )
{
/* #undef _OPENMP */
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int bupflag  = 0;
        if( *uplo == 'u' || *uplo == 'U' ) bupflag = 1;


        int bdiagflag = 0;
        if( *diag == 'u' || *diag == 'U' ) bdiagflag = 1;


        int x, y;

        for( x = 0; x < m; x++ )
        {
#ifdef _OPENMP
 #pragma omp for
#endif
            for( y = 0; y < m; y++ )
            {
                double ph, pl;
                if( bupflag == 1 )
                {
                    if( x > y )
                    {
                        ph = 0.0;
                        pl = 0.0;
                    }
                    else
                    {
                        ph = *( Mhin+x+lda*y );
                        pl = *( Mlin+x+lda*y );
                    }
                }
                else
                {
                    if( x >= y )
                    {
                        ph = *( Mhin+x+lda*y );
                        pl = *( Mlin+x+lda*y );
                    }
                    else
                    {
                        ph = 0.0;
                        pl = 0.0;
                    }
                }

                if( bdiagflag == 1 && x == y )
                {
                    ph = 1.0;
                    pl = 0.0;
                }

                *( Mhout+x+m*y ) = ph;
                *( Mlout+x+m*y ) = pl;
            }
        }
    }
}

void
copyToSymmetricBandSplitMatrix( int n, int k, int lda, char*uplo,
                                int numSplit, int *partitions,
                                double* Mhin,
                                double* Mlin,
                                double *Mhout,
                                double* Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int upflag = 1;

        if( *uplo == 'L' || *uplo == 'l' )
        {
            upflag = 0;
        }

        int colidx, spidx, ptidx;

#ifdef _OPENMP
 #pragma omp for
#endif
        for( colidx = 0; colidx < n; colidx++ )
        {
            int sumOfPartition = 0;
            for( spidx = 0; spidx < numSplit; spidx++ )
            {
                int ptrow = partitions[spidx];
                for( ptidx = 0; ptidx < ptrow; ptidx++ )
                {
                    int rowidx = ptidx+sumOfPartition;
                    int inindex;

                    if( upflag == 1 )
                    {
                        if( rowidx > colidx )
                        {
                            inindex = k-rowidx+rowidx*lda+colidx;
                        }
                        else
                        {
                            inindex = k-colidx+rowidx+colidx*lda;
                        }
                    }
                    else
                    {
                        if( rowidx > colidx )
                        {
                            inindex = -colidx+rowidx+lda*colidx;
                        }
                        else
                        {
                            inindex = -rowidx+rowidx*lda+colidx;
                        }
                    }

                    int    outindex = ptidx+ptrow*colidx+sumOfPartition*n;

                    double ph, pl;
                    if( colidx-rowidx+k >= 0 && colidx-rowidx-k <= 0 )
                    {
                        ph = *( Mhin+inindex );
                        pl = *( Mlin+inindex );
                    }
                    else
                    {
                        ph = 0.0;
                        pl = 0.0;
                    }


                    *( Mhout+outindex ) = ph;
                    *( Mlout+outindex ) = pl;
                }
                sumOfPartition += ptrow;
            }
        }
    }
}

void
copyToHermitianBandSplitMatrix( int n, int k, int lda, char*uplo,
                                int numSplit, int *partitions,
                                cuDoubleComplex* Mhin,
                                cuDoubleComplex* Mlin,
                                cuDoubleComplex *Mhout,
                                cuDoubleComplex* Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int upflag = 1;

        if( *uplo == 'L' || *uplo == 'l' )
        {
            upflag = 0;
        }

        int colidx, spidx, ptidx;

#ifdef _OPENMP
 #pragma omp for
#endif
        for( colidx = 0; colidx < n; colidx++ )
        {
            int sumOfPartition = 0;
            for( spidx = 0; spidx < numSplit; spidx++ )
            {
                int ptrow = partitions[spidx];
                for( ptidx = 0; ptidx < ptrow; ptidx++ )
                {
                    int rowidx = ptidx+sumOfPartition;
                    int inindex;
                    int conjflag = 0;

                    if( upflag == 1 )
                    {
                        if( rowidx > colidx )
                        {
                            inindex = k-rowidx+rowidx*lda+colidx;
                            conjflag = 1;
                        }
                        else
                        {
                            inindex = k-colidx+rowidx+colidx*lda;
                        }
                    }
                    else
                    {
                        if( rowidx > colidx )
                        {
                            inindex = -colidx+rowidx+lda*colidx;
                        }
                        else
                        {
                            inindex = -rowidx+rowidx*lda+colidx;
                            conjflag = 1;
                        }
                    }

                    int             outindex = ptidx+ptrow*colidx+
                                               sumOfPartition*n;

                    cuDoubleComplex ph, pl;
                    if( colidx-rowidx+k >= 0 && colidx-rowidx-k <= 0 )
                    {
                        ph = *( Mhin+inindex );
                        pl = *( Mlin+inindex );
                    }
                    else
                    {
                        ph = make_cuDoubleComplex( 0.0, 0.0 );
                        pl = make_cuDoubleComplex( 0.0, 0.0 );
                    }

                    if( conjflag )
                    {
                        ph = cuConj( ph );
                        pl = cuConj( pl );
                    }
                    if( rowidx == colidx )
                    {
                        ph = make_cuDoubleComplex( cuCreal( ph ), 0.0 );
                        pl = make_cuDoubleComplex( cuCreal( pl ), 0.0 );
                    }


                    *( Mhout+outindex ) = ph;
                    *( Mlout+outindex ) = pl;
                }
                sumOfPartition += ptrow;
            }
        }
    }
}

void
copyToSymmetricSplitMatrix( int n, int lda, char*uplo,
                            int numSplit, int *partitions,
                            double* Mhin,
                            double* Mlin,
                            double *Mhout,
                            double* Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int upflag = 1;

        if( *uplo == 'L' || *uplo == 'l' )
        {
            upflag = 0;
        }

        int colidx, spidx, ptidx;

#ifdef _OPENMP
 #pragma omp for
#endif
        for( colidx = 0; colidx < n; colidx++ )
        {
            int sumOfPartition = 0;
            for( spidx = 0; spidx < numSplit; spidx++ )
            {
                int ptrow = partitions[spidx];
                for( ptidx = 0; ptidx < ptrow; ptidx++ )
                {
                    int rowidx = ptidx+sumOfPartition;
                    int inindex;

                    if( upflag == 1 )
                    {
                        if( rowidx > colidx )
                        {
                            inindex = rowidx*lda+colidx;
                        }
                        else
                        {
                            inindex = rowidx+colidx*lda;
                        }
                    }
                    else
                    {
                        if( rowidx > colidx )
                        {
                            inindex = rowidx+lda*colidx;
                        }
                        else
                        {
                            inindex = rowidx*lda+colidx;
                        }
                    }
                    int outindex = ptidx+ptrow*colidx+sumOfPartition*n;


                    *( Mhout+outindex ) = *( Mhin+inindex );
                    *( Mlout+outindex ) = *( Mlin+inindex );
                }
                sumOfPartition += ptrow;
            }
        }
    }
}

void
copyToHermitianSplitMatrix( int n, int lda, char*uplo,
                            int numSplit, int *partitions,
                            cuDoubleComplex* Mhin,
                            cuDoubleComplex* Mlin,
                            cuDoubleComplex *Mhout,
                            cuDoubleComplex* Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int upflag = 1;

        if( *uplo == 'L' || *uplo == 'l' )
        {
            upflag = 0;
        }

        int colidx, spidx, ptidx;

#ifdef _OPENMP
 #pragma omp for
#endif
        for( colidx = 0; colidx < n; colidx++ )
        {
            int sumOfPartition = 0;
            for( spidx = 0; spidx < numSplit; spidx++ )
            {
                int ptrow = partitions[spidx];
                for( ptidx = 0; ptidx < ptrow; ptidx++ )
                {
                    int rowidx = ptidx+sumOfPartition;
                    int inindex;

                    int conjflag = 0;


                    if( upflag == 1 )
                    {
                        if( rowidx > colidx )
                        {
                            inindex = rowidx*lda+colidx;
                            conjflag = 1;
                        }
                        else
                        {
                            inindex = rowidx+colidx*lda;
                        }
                    }
                    else
                    {
                        if( rowidx > colidx )
                        {
                            inindex = rowidx+lda*colidx;
                        }
                        else
                        {
                            inindex = rowidx*lda+colidx;
                            conjflag = 1;
                        }
                    }
                    int             outindex = ptidx+ptrow*colidx+
                                               sumOfPartition*n;

                    cuDoubleComplex ph, pl;
                    ph = *( Mhin+inindex );
                    pl = *( Mlin+inindex );
                    if( conjflag )
                    {
                        ph = cuConj( ph );
                        pl = cuConj( pl );
                    }
                    if( rowidx == colidx )
                    {
                        ph = make_cuDoubleComplex( cuCreal( ph ), 0.0 );
                        pl = make_cuDoubleComplex( cuCreal( pl ), 0.0 );
                    }


                    *( Mhout+outindex ) = ph;
                    *( Mlout+outindex ) = pl;
                }
                sumOfPartition += ptrow;
            }
        }
    }
}

void
copyToSymmetricMatrix( int n, int lda, char*uplo, double* Mhin,
                       double* Mlin,
                       double *Mhout,
                       double* Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int upflag = 1;

        if( *uplo == 'L' || *uplo == 'l' )
        {
            upflag = 0;
        }
        int x, y;
        for( x = 0; x < n; x++ )
        {
#ifdef _OPENMP
 #pragma omp for
#endif
            for( y = 0; y < n; y++ )
            {
                if( upflag == 1 )
                {
                    if( x > y )
                    {
                        *( Mhout+x+n*y ) = *( Mhin+x*lda+y );
                        *( Mlout+x+n*y ) = *( Mlin+x*lda+y );
                    }
                    else
                    {
                        *( Mhout+x+n*y ) = *( Mhin+x+lda*y );
                        *( Mlout+x+n*y ) = *( Mlin+x+lda*y );
                    }
                }
                else
                {
                    if( x > y )
                    {
                        *( Mhout+x+n*y ) = *( Mhin+x+lda*y );
                        *( Mlout+x+n*y ) = *( Mlin+x+lda*y );
                    }
                    else
                    {
                        *( Mhout+x+n*y ) = *( Mhin+x*lda+y );
                        *( Mlout+x+n*y ) = *( Mlin+x*lda+y );
                    }
                }
            }
        }
    }
}

void
copyToHermitianMatrix( int n, int lda, char*uplo, cuDoubleComplex* Mhin,
                       cuDoubleComplex* Mlin,
                       cuDoubleComplex *Mhout,
                       cuDoubleComplex* Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int upflag = 1;

        if( *uplo == 'L' || *uplo == 'l' )
        {
            upflag = 0;
        }
        int x, y;
        for( x = 0; x < n; x++ )
        {
#ifdef _OPENMP
 #pragma omp for
#endif
            for( y = 0; y < n; y++ )
            {
                if( upflag == 1 )
                {
                    if( x > y )
                    {
                        *( Mhout+x+n*y ) = cuConj( *( Mhin+x*lda+y ) );
                        *( Mlout+x+n*y ) = cuConj( *( Mlin+x*lda+y ) );
                    }
                    else if( x < y )
                    {
                        *( Mhout+x+n*y ) = *( Mhin+x+lda*y );
                        *( Mlout+x+n*y ) = *( Mlin+x+lda*y );
                    }
                    else
                    {
                        *( Mhout+x+n*
                           y ) = make_cuDoubleComplex( cuCreal(
                                                            *( Mhin+x+lda*y ) ),
                                                       0.0 );
                        *( Mlout+x+n*
                           y ) = make_cuDoubleComplex( cuCreal(
                                                            *( Mlin+x+lda*y ) ),
                                                       0.0 );
                    }
                }
                else
                {
                    if( x > y )
                    {
                        *( Mhout+x+n*y ) = *( Mhin+x+lda*y );
                        *( Mlout+x+n*y ) = *( Mlin+x+lda*y );
                    }
                    else if( x < y )
                    {
                        *( Mhout+x+n*y ) = cuConj( *( Mhin+x*lda+y ) );
                        *( Mlout+x+n*y ) = cuConj( *( Mlin+x*lda+y ) );
                    }
                    else
                    {
                        *( Mhout+x+n*
                           y ) = make_cuDoubleComplex( cuCreal(
                                                            *( Mhin+x+lda*y ) ),
                                                       0.0 );
                        *( Mlout+x+n*
                           y ) = make_cuDoubleComplex( cuCreal(
                                                            *( Mlin+x+lda*y ) ),
                                                       0.0 );
                    }
                }
            }
        }
    }
}

void
copyFromSymmetricMatrix( int n, int lda, char*uplo, double* Mhin,
                         double* Mlin,
                         double *Mhout,
                         double* Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int upflag = 1;

        if( *uplo == 'L' || *uplo == 'l' )
        {
            upflag = 0;
        }
        int x, y;
        for( x = 0; x < n; x++ )
        {
#ifdef _OPENMP
 #pragma omp for
#endif
            for( y = 0; y < x+1; y++ )
            {
                if( upflag == 1 )
                {
                    *( Mhout+y+lda*x ) = *( Mhin+y+x*n );
                    *( Mlout+y+lda*x ) = *( Mlin+y+x*n );
                }
                else
                {
                    *( Mhout+x+lda*y ) = *( Mhin+x+y*n );
                    *( Mlout+x+lda*y ) = *( Mlin+x+y*n );
                }
            }
        }
    }
}

void
copyFromHermitianMatrix( int n, int lda, char*uplo, cuDoubleComplex* Mhin,
                         cuDoubleComplex* Mlin,
                         cuDoubleComplex *Mhout,
                         cuDoubleComplex* Mlout )
{
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int upflag = 1;

        if( *uplo == 'L' || *uplo == 'l' )
        {
            upflag = 0;
        }
        int x, y;
        for( x = 0; x < n; x++ )
        {
#ifdef _OPENMP
 #pragma omp for
#endif
            for( y = 0; y < x+1; y++ )
            {
                if( upflag == 1 )
                {
                    *( Mhout+y+lda*x ) = *( Mhin+y+x*n );
                    *( Mlout+y+lda*x ) = *( Mlin+y+x*n );
                }
                else
                {
                    *( Mhout+x+lda*y ) = *( Mhin+x+y*n );
                    *( Mlout+x+lda*y ) = *( Mlin+x+y*n );
                }
            }
        }
    }
}

void
copyToBandMatrix( int m, int n, int lda, int kl, int ku, double * Mhin,
                  double *Mlin, double*Mhout,
                  double*Mlout )
{
    int ldout = kl+ku+1;
    int i, j, k;

    for( j = 0; j < n; j++ )
    {
        k = ku-j;
        for( i = max( 0, j-ku ); i < min( m, j+kl+1 ); i++ )
        {
            *( Mhout+k+i+ldout*j ) = *( Mhin+i+lda*j );
            *( Mlout+k+i+ldout*j ) = *( Mlin+i+lda*j );
        }
    }
}

void
vector_copy( int N, double*xinh, double*xinl, int incxin, double*xouth,
             double*xoutl,
             int incxout )
{
	int i;
    for( i = 0; i < N; i++ )
    {
        *( xouth+i*incxout ) = *( xinh+i*incxin );
        *( xoutl+i*incxout ) = *( xinl+i*incxin );
    }
}

void
copyFromBandMatrix( int m, int n, int kl, int ku, double*Bandh,
                    double*Bandl, int lda, double*Outh,
                    double*Outl )
{
    int x, y;

    for( x = 0; x < m; x++ )
    {
        for( y = 0; y < n; y++ )
        {
            if( x > kl+y  || x < y-ku )
            {
                *( Outh+x+y*m ) = 0.0;
                *( Outl+x+y*m ) = 0.0;
            }
            else
            {
                *( Outh+x+y*m ) = *( Bandh+x+ku-y+y*lda );
                *( Outl+x+y*m ) = *( Bandl+x+ku-y+y*lda );
            }
        }
    }
}

void
zcopyToBandMatrix( int m, int n, int lda, int kl, int ku,
                   cuDoubleComplex * Mhin,
                   cuDoubleComplex *Mlin, cuDoubleComplex*Mhout,
                   cuDoubleComplex*Mlout )
{
    int ldout = kl+ku+1;
    int i, j, k;

    for( j = 0; j < n; j++ )
    {
        k = ku-j;
        for( i = max( 0, j-ku ); i < min( m, j+kl+1 ); i++ )
        {
            *( Mhout+k+i+ldout*j ) = *( Mhin+i+lda*j );
            *( Mlout+k+i+ldout*j ) = *( Mlin+i+lda*j );
        }
    }
}

void
copyToBandSplitMatrix( char *trans, char*diag, int m, int n, int lda,
                       int kl, int ku,
                       int numSplit, int *partitions,
                       double* Mhin, double* Mlin, double *Mhout,
                       double*Mlout )
{
/* #undef _OPENMP */
#ifdef _OPENMP
 #pragma omp parallel
#endif
    {
        int btransflag = 0;
        if( *trans == 't' || *trans == 'T' ) btransflag = 1;

        int bdiagflag = 0;
        if( *diag == 'u' || *diag == 'U' ) bdiagflag = 1;

        int colnum = ( btransflag ? m : n );


        int colidx, spidx, ptidx;
        for( colidx = 0; colidx < colnum; colidx++ )
        {
            int sumOfPartition = 0;
            for( spidx = 0; spidx < numSplit; spidx++ )
            {
                int ptrow = partitions[spidx];
#ifdef _OPENMP
 #pragma omp for
#endif
                for( ptidx = 0; ptidx < ptrow; ptidx++ )
                {
                    int    rowidx = ptidx+sumOfPartition;
                    int    outindex = ptidx+ptrow*colidx+sumOfPartition*
                                      colnum;

                    double ph, pl;
                    if( !btransflag )
                    {
                        int inindex = ku-colidx+rowidx+lda*colidx;
                        if( rowidx-colidx+ku  >= 0 &&  rowidx-colidx-kl <=
                            0 )
                        {
                            ph = *( Mhin+inindex );
                            pl = *( Mlin+inindex );
                        }
                        else
                        {
                            ph = 0.0;
                            pl = 0.0;
                        }
                    }
                    else
                    {
                        int inindex = ku-rowidx+colidx+lda*rowidx;
                        if( colidx-rowidx+ku  >= 0 &&  colidx-rowidx-kl <=
                            0 )
                        {
                            ph = *( Mhin+inindex );
                            pl = *( Mlin+inindex );
                        }
                        else
                        {
                            ph = 0.0;
                            pl = 0.0;
                        }
                    }

                    if( bdiagflag && ( colidx == rowidx ) )
                    {
                        ph = 1.0;
                        pl = 0.0;
                    }

                    *( Mhout+outindex ) = ph;
                    *( Mlout+outindex ) = pl;
                }

                sumOfPartition += ptrow;
            }
        }
    }
}

#ifdef __cplusplus
}


#endif
