#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddasum_host_( int *n, double *xh,double *xl, int*incx,double *reth,double*retl)
{
    double *dxh, *dxl,  *dreth, *dretl;
    int xsize = *n*(*incx);


    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( xsize, sizeof( double ), (void**) &dxh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( xsize, sizeof( double ), (void**) &dxl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( double ), (void**) &dreth )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( double ), (void**) &dretl )
      )
    {
/*	 Memory allocation failed */
        cublasFree( dxh );
        cublasFree( dxl );
        cublasFree( dreth );
        cublasFree( dretl );
        return -1;
    }
    else
    {
        /* Device Memory allocation */
        cublasSetVector( xsize, sizeof( double ), xh, 1, dxh, 1 );
        cublasSetVector( xsize, sizeof( double ), xl, 1, dxl, 1 );
    }

    /* Execute gpu vector add */
    gddasum( n, dxh, dxl, incx,dreth,dretl);

    /* Get From Device Memory */
    cublasGetVector( 1, sizeof( double ), dreth, 1, reth, 1 );
    cublasGetVector( 1, sizeof( double ), dretl, 1, retl, 1 );

    /* Free Device Memories */
    cublasFree( dxh );
    cublasFree( dxl );
    cublasFree( dreth );
    cublasFree( dretl );


    return 0;
}

#ifdef __cplusplus
}
#endif
