#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddaxpy_host_( int *n, double *alphah, double *alphal, double *xh,double *xl, int*incx,double *yh, double *yl,int *incy  )
{
    double *dxh, *dxl, *dyh, *dyl, *dalphah, *dalphal;

    int sizex = (*n)*(*incx);
    int sizey = (*n)*(*incy);


    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizex, sizeof( double ), (void**) &dxh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizex, sizeof( double ), (void**) &dxl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizey, sizeof( double ), (void**) &dyh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizey, sizeof( double ), (void**) &dyl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( double ), (void**) &dalphah )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( double ), (void**) &dalphal )
      )
    {
/*	 Memory allocation failed */
        cublasFree( dxh );
        cublasFree( dxl );
        cublasFree( dyh );
        cublasFree( dyl );
        cublasFree( dalphah );
        cublasFree( dalphal );
        return -1;
    }
    else
    {
        /* Device Memory allocation */
        cublasSetVector( sizex, sizeof( double ), xh, 1, dxh, 1 );
        cublasSetVector( sizex, sizeof( double ), xl, 1, dxl, 1 );
        cublasSetVector( sizey, sizeof( double ), yh, 1, dyh, 1 );
        cublasSetVector( sizey, sizeof( double ), yl, 1, dyl, 1 );
        cublasSetVector( 1, sizeof( double ), alphah, 1, dalphah, 1 );
        cublasSetVector( 1, sizeof( double ), alphal, 1, dalphal, 1 );
    }

    /* Execute gpu vector add */
    gddaxpy( n,  dalphah, dalphal, dxh, dxl, incx, dyh, dyl, incy );

    /* Get From Device Memory */
    cublasGetVector( sizey, sizeof( double ), dyh, 1, yh, 1 );
    cublasGetVector( sizey, sizeof( double ), dyl, 1, yl, 1 );

    /* Free Device Memories */
    cublasFree( dxh );
    cublasFree( dxl );
    cublasFree( dyh );
    cublasFree( dyl );
    cublasFree( dalphah );
    cublasFree( dalphal );


    return 0;
}

#ifdef __cplusplus
}
#endif
