#include "gddcommon.h"

int
gddcopy_host_( int *n, double *dxh, double*dxl, int *incx,
              double *dyh, double * dyl, int *incy )
{
    double *dxh_dev, *dxl_dev, *dyh_dev, *dyl_dev;
    int xsize = *n*( *incx );
    int ysize = *n*( *incy );

    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( xsize , sizeof( double ), (void**) &dxh_dev )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( xsize , sizeof( double ), (void**) &dxl_dev )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( ysize , sizeof( double ), (void**) &dyh_dev )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( ysize , sizeof( double ), (void**) &dyl_dev )
      )
    {
/*	 Memory allocation failed */
        cublasFree( dxh_dev );
        cublasFree( dxl_dev );
        cublasFree( dyh_dev );
        cublasFree( dyl_dev );
        return -1;
    }
    else
    {
        /* Device Memory allocation */
        cublasSetVector( xsize , sizeof( double ), dxh, 1, dxh_dev, 1 );
        cublasSetVector( xsize , sizeof( double ), dxl, 1, dxl_dev, 1 );
        cublasSetVector( ysize , sizeof( double ), dyh, 1, dyh_dev, 1 );
        cublasSetVector( ysize , sizeof( double ), dyl, 1, dyl_dev, 1 );
    }

    /* Execute gpu */
    gddcopy( n, dxh_dev, dxl_dev, incx, dyh_dev, dyl_dev, incy );

    /* Get From Device Memory */
    cublasGetVector( ysize , sizeof( double ), dyh_dev, 1, dyh, 1 );
    cublasGetVector( ysize , sizeof( double ), dyl_dev, 1, dyl, 1 );

    /* Free Device Memories */
    cublasFree( dxh_dev );
    cublasFree( dxl_dev );
    cublasFree( dyh_dev );
    cublasFree( dyl_dev );

    return 0;
}

