#include "gddcommon.h"

int gddgemm_host_
( char* transA, char* transB,
  int* m, int* n, int* k,
  double* alphah, double* alphal,
  double* Ah, double* Al, int* lda,
  double* Bh, double* Bl, int* ldb,
  double* betah, double* betal,
  double* Ch, double* Cl, int* ldc )
{
  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;
  int sizeA, sizeB, sizeC;

  double *dAh, *dAl;
  double *dBh, *dBl;
  double *dCh, *dCl;
  double *dalphah, *dalphal;
  double *dbetah, *dbetal;

  if( *transA == 'N' || *transA == 'n' ){
    /* A(M,K) */
    nrowA = *m, ncolA = *k;
  }
  else{
    /* A(K,M) */
    nrowA = *k, ncolA = *m;
  }

  if( *transB == 'N' || *transB == 'n' ){
    /* B(K,N) */
    nrowB = *k, ncolB = *n;
  }
  else{
    /* B(N,K) */
    nrowB = *n, ncolB = *k;
  }

  {
    /* C(M,N) */
    nrowC = *m, ncolC = *n;
  }

  if( *lda < nrowA || *ldb < nrowB || *ldc < nrowC ){
    printf("Argument Error in gddgemm_host_\n");
    printf("wrong lda or ldb or ldc.\n");
    return -1;
  }

  sizeA = (*lda)*ncolA;
  sizeB = (*ldb)*ncolB;
  sizeC = (*ldc)*ncolC;

  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**) &dAh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**) &dAl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(double), (void**) &dBh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(double), (void**) &dBl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(double), (void**) &dCh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(double), (void**) &dCl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**) &dalphah ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**) &dalphal ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**) &dbetah  ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**) &dbetal  ) ){
      /*	 Memory allocation failed */
    cublasFree( dAh );
    cublasFree( dAl );
    cublasFree( dBh );
    cublasFree( dBl );
    cublasFree( dCh );
    cublasFree( dCl );
    cublasFree( dalphah );
    cublasFree( dalphal );
    cublasFree( dbetah );
    cublasFree( dbetal );
    printf("Device Malloc Error in gddgemm_host_\n");
    return -1;
  }
  else{
    /* Send Matrices(column-majorでFortranと一緒)  */
    cublasSetVector( sizeA, sizeof(double), Ah, 1, dAh, 1 );
    cublasSetVector( sizeA, sizeof(double), Al, 1, dAl, 1 );
    cublasSetVector( sizeB, sizeof(double), Bh, 1, dBh, 1 );
    cublasSetVector( sizeB, sizeof(double), Bl, 1, dBl, 1 );
    cublasSetVector( sizeC, sizeof(double), Ch, 1, dCh, 1 );
    cublasSetVector( sizeC, sizeof(double), Cl, 1, dCl, 1 );
    cublasSetVector( 1, sizeof(double), alphah, 1, dalphah, 1 );
    cublasSetVector( 1, sizeof(double), alphal, 1, dalphal, 1 );
    cublasSetVector( 1, sizeof(double), betah, 1, dbetah, 1 );
    cublasSetVector( 1, sizeof(double), betal, 1, dbetal, 1 );
  }

  /* Execute gpu function */
  gddgemm
    ( transA, transB, m, n, k, dalphah, dalphal, dAh, dAl, lda,
      dBh, dBl, ldb, dbetah, dbetal, dCh, dCl, ldc );

  /* Get From Device Memory */
  cublasGetVector( sizeC, sizeof(double), dCh, 1, Ch, 1 );
  cublasGetVector( sizeC, sizeof(double), dCl, 1, Cl, 1 );

  /* Free Device Memories */
  cublasFree( dAh );
  cublasFree( dAl );
  cublasFree( dBh );
  cublasFree( dBl );
  cublasFree( dCh );
  cublasFree( dCl );
  cublasFree( dalphah );
  cublasFree( dalphal );
  cublasFree( dbetah );
  cublasFree( dbetal );

  return 0;
}

int gddsymm_host_
( char* side, char* uplo,
  int* m, int* n,
  double* alphah, double* alphal,
  double* Ah, double* Al, int* lda,
  double* Bh, double* Bl, int* ldb,
  double* betah, double* betal,
  double* Ch, double* Cl, int* ldc )
{
  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;
  int sizeA, sizeB, sizeC;

  double *dAh, *dAl;
  double *dBh, *dBl;
  double *dCh, *dCl;
  double *dalphah, *dalphal;
  double *dbetah, *dbetal;

  if( *side == 'L' || *side == 'l' ){
    /* C(M,N) += A(M,M)*B(M,N), A is symmetric, B is general */
    nrowA = *m, ncolA = *m;
    nrowB = *m, ncolB = *n;
    nrowC = *m, ncolC = *n;
  }
  else{
    /* C(M,N) += B(M,N)*A(N,N), A is symmetric, B is general */
    nrowB = *m, ncolB = *n;
    nrowA = *n, ncolA = *n;
    nrowC = *m, ncolC = *n;
  }

  if( *lda < nrowA || *ldb < nrowB || *ldc < nrowC ){
    printf("Argument Error in gddsymm_host_\n");
    printf("wrong lda or ldb or ldc.\n");
    return -1;
  }

  sizeA = (*lda)*ncolA;
  sizeB = (*ldb)*ncolB;
  sizeC = (*ldc)*ncolC;

  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**)&dAh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**)&dAl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(double), (void**)&dBh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(double), (void**)&dBl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(double), (void**)&dCh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(double), (void**)&dCl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**)&dalphah ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**)&dalphal ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**)&dbetah  ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**)&dbetal ) ){
    /*	 Memory allocation failed */
    cublasFree( dAh );
    cublasFree( dAl );
    cublasFree( dBh );
    cublasFree( dBl );
    cublasFree( dCh );
    cublasFree( dCl );
    cublasFree( dalphah );
    cublasFree( dalphal );
    cublasFree( dbetah );
    cublasFree( dbetal );
    printf("Device Malloc Error in gddsymm_host_\n");
    return -1;
  }
  else{
    /* Send Matrices(column-majorでFortranと一緒)  */
    cublasSetVector( sizeA, sizeof(double), Ah, 1, dAh, 1 );
    cublasSetVector( sizeA, sizeof(double), Al, 1, dAl, 1 );
    cublasSetVector( sizeB, sizeof(double), Bh, 1, dBh, 1 );
    cublasSetVector( sizeB, sizeof(double), Bl, 1, dBl, 1 );
    cublasSetVector( sizeC, sizeof(double), Ch, 1, dCh, 1 );
    cublasSetVector( sizeC, sizeof(double), Cl, 1, dCl, 1 );
    cublasSetVector( 1, sizeof(double), alphah, 1, dalphah, 1 );
    cublasSetVector( 1, sizeof(double), alphal, 1, dalphal, 1 );
    cublasSetVector( 1, sizeof(double), betah, 1, dbetah, 1 );
    cublasSetVector( 1, sizeof(double), betal, 1, dbetal, 1 );
  }
  
  /* Execute gpu function */
  gddsymm
    ( side, uplo,
      m, n, dalphah, dalphal, dAh, dAl, lda, dBh, dBl, ldb,
      dbetah, dbetal, dCh, dCl, ldc );

  /* Get From Device Memory */
  cublasGetVector( sizeC, sizeof(double), dCh, 1, Ch, 1 );
  cublasGetVector( sizeC, sizeof(double), dCl, 1, Cl, 1 );

  /* Free Device Memories */
  cublasFree( dAh );
  cublasFree( dAl );
  cublasFree( dBh );
  cublasFree( dBl );
  cublasFree( dCh );
  cublasFree( dCl );
  cublasFree( dalphah );
  cublasFree( dalphal );
  cublasFree( dbetah );
  cublasFree( dbetal );
  
  return 0;
}

int gddtrmm_host_
( char* side, char* uplo, char* transA, char* diag,
  int* m, int* n,
  double* alphah, double* alphal,
  double* Ah, double* Al, int* lda,
  double* Bh, double* Bl, int* ldb )
{
  int nrowA, ncolA;
  int nrowB, ncolB;
  int sizeA, sizeB;

  double *dAh, *dAl;
  double *dBh, *dBl;
  double *dalphah, *dalphal;

  if( *side == 'L' || *side == 'l' ){
    /* B(M,N) += A(M,M)*B(M,N), A is symmetric, B is general */
    nrowA = *m, ncolA = *m;
    nrowB = *m, ncolB = *n;
  }
  else{
    /* B(M,N) += B(M,N)*A(N,N), A is symmetric, B is general */
    nrowB = *m, ncolB = *n;
    nrowA = *n, ncolA = *n;
  }

  if( *lda < nrowA || *ldb < nrowB ){
    printf("Argument Error in gddtrmm_host_\n");
    printf("wrong lda or ldb.\n");
    return -1;
  }

  int i,j;

/*    printf("B_host:\n");*/
/*    for (j = 0; j < nrowB; j++) {*/
/*    for (i = 0; i < ncolB; i++) {*/
/*            printf("%f",Bh[j+i*(*lda)]);*/
/*    }*/
/*    printf("\n");*/
/*    }*/





  sizeA = (*lda)*ncolA;
  sizeB = (*ldb)*ncolB;


  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**)&dAh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**)&dAl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(double), (void**)&dBh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(double), (void**)&dBl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**)&dalphah ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**)&dalphal ) ){
    /*	 Memory allocation failed */
    cublasFree( dAh );
    cublasFree( dAl );
    cublasFree( dBh );
    cublasFree( dBl );
    cublasFree( dalphah );
    cublasFree( dalphal );
    printf("Device Malloc Error in gddtrmm_host_\n");
    return -1;
  }
  else{
    /* Send Matrices(column-majorでFortranと一緒)  */
    cublasSetVector( sizeA, sizeof(double), Ah, 1, dAh, 1 );
    cublasSetVector( sizeA, sizeof(double), Al, 1, dAl, 1 );
    cublasSetVector( sizeB, sizeof(double), Bh, 1, dBh, 1 );
    cublasSetVector( sizeB, sizeof(double), Bl, 1, dBl, 1 );
    cublasSetVector( 1, sizeof(double), alphah, 1, dalphah, 1 );
    cublasSetVector( 1, sizeof(double), alphal, 1, dalphal, 1 );
  }


  /* Execute gpu function */
  gddtrmm
    ( side, uplo, transA, diag, 
      m, n, dalphah, dalphal, dAh, dAl, lda, dBh, dBl, ldb );

  /* Get From Device Memory */
  cublasGetVector( sizeB, sizeof(double), dBh, 1, Bh, 1 );
  cublasGetVector( sizeB, sizeof(double), dBl, 1, Bl, 1 );

  /* Free Device Memories */
  cublasFree( dAh );
  cublasFree( dAl );
  cublasFree( dBh );
  cublasFree( dBl );
  cublasFree( dalphah );
  cublasFree( dalphal );
  
  return 0;
}



int gddsyr2k_host_
( char* uplo, char* trans,
  int* n, int* k,
  double* alphah, double* alphal,
  double* Ah, double* Al, int* lda,
  double* Bh, double* Bl, int* ldb,
  double* betah, double* betal,
  double* Ch, double* Cl, int* ldc )
{
  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;
  int sizeA, sizeB, sizeC;

  double *dAh, *dAl;
  double *dBh, *dBl;
  double *dCh, *dCl;
  double *dalphah, *dalphal;
  double *dbetah, *dbetal;

  if( *trans == 'N' || *trans == 'n' ){
    /* C(N,N) += A(N,K) * B(N,K)^T */
    nrowA = *n, ncolA = *k;
    nrowB = *n, ncolB = *k;
    nrowC = *n, ncolC = *n;
  }
  else{
    /* C(N,N) += A(K,N)^T * B(K,N) */
    nrowA = *k, ncolA = *n;
    nrowB = *k, ncolB = *n;
    nrowC = *n, ncolC = *n;
  }

  if( *lda < nrowA || *ldb < nrowB || *ldc < nrowC ){
    printf("Argument Error in gddsyr2k_host_\n");
    printf("wrong lda or ldb or ldc.\n");
    return -1;
  }

  sizeA = (*lda)*ncolA;
  sizeB = (*ldb)*ncolB;
  sizeC = (*ldc)*ncolC;

  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**) &dAh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**) &dAl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(double), (void**) &dBh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(double), (void**) &dBl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(double), (void**) &dCh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(double), (void**) &dCl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**) &dalphah ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**) &dalphal ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**) &dbetah  ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**) &dbetal  ) ){
      /*	 Memory allocation failed */
    cublasFree( dAh );
    cublasFree( dAl );
    cublasFree( dBh );
    cublasFree( dBl );
    cublasFree( dCh );
    cublasFree( dCl );
    cublasFree( dalphah );
    cublasFree( dalphal );
    cublasFree( dbetah );
    cublasFree( dbetal );
    printf("Device Malloc Error in gddsyr2k_host_\n");
    return -1;
  }
  else{
    /* Send Matrices(column-majorでFortranと一緒)  */
    cublasSetVector( sizeA, sizeof(double), Ah, 1, dAh, 1 );
    cublasSetVector( sizeA, sizeof(double), Al, 1, dAl, 1 );
    cublasSetVector( sizeB, sizeof(double), Bh, 1, dBh, 1 );
    cublasSetVector( sizeB, sizeof(double), Bl, 1, dBl, 1 );
    cublasSetVector( sizeC, sizeof(double), Ch, 1, dCh, 1 );
    cublasSetVector( sizeC, sizeof(double), Cl, 1, dCl, 1 );
    cublasSetVector( 1, sizeof(double), alphah, 1, dalphah, 1 );
    cublasSetVector( 1, sizeof(double), alphal, 1, dalphal, 1 );
    cublasSetVector( 1, sizeof(double), betah, 1, dbetah, 1 );
    cublasSetVector( 1, sizeof(double), betal, 1, dbetal, 1 );
  }

  /* Execute gpu function */
  gddsyr2k
    ( uplo, trans, n, k, dalphah, dalphal, dAh, dAl, lda,
      dBh, dBl, ldb, dbetah, dbetal, dCh, dCl, ldc );

  /* Get From Device Memory */
  cublasGetVector( sizeC, sizeof(double), dCh, 1, Ch, 1 );
  cublasGetVector( sizeC, sizeof(double), dCl, 1, Cl, 1 );

  /* Free Device Memories */
  cublasFree( dAh );
  cublasFree( dAl );
  cublasFree( dBh );
  cublasFree( dBl );
  cublasFree( dCh );
  cublasFree( dCl );
  cublasFree( dalphah );
  cublasFree( dalphal );
  cublasFree( dbetah );
  cublasFree( dbetal );

  return 0;
}

int gddsyrk_host_
( char* uplo, char* trans,
  int* n, int* k,
  double* alphah, double* alphal,
  double* Ah, double* Al, int* lda,
  double* betah, double* betal,
  double* Ch, double* Cl, int* ldc )
{
  int nrowA, ncolA;
  int nrowC, ncolC;
  int sizeA, sizeC;

  double *dAh, *dAl;
  double *dCh, *dCl;
  double *dalphah, *dalphal;
  double *dbetah, *dbetal;

  if( *trans == 'N' || *trans == 'n' ){
    /* C(N,N) += A(N,K) * A(N,K)^T */
    nrowA = *n, ncolA = *k;
    nrowC = *n, ncolC = *n;
  }
  else{
    /* C(N,N) += A(K,N)^T * A(K,N) */
    nrowA = *k, ncolA = *n;
    nrowC = *n, ncolC = *n;
  }

  if( *lda < nrowA || *ldc < nrowC ){
    printf("Argument Error in gddsyrk_host_\n");
    printf("wrong lda or ldc.\n");
    return -1;
  }

  sizeA = (*lda)*ncolA;
  sizeC = (*ldc)*ncolC;

  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**) &dAh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**) &dAl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(double), (void**) &dCh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(double), (void**) &dCl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**) &dalphah ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**) &dalphal ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**) &dbetah  ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**) &dbetal  ) ){
      /*	 Memory allocation failed */
    cublasFree( dAh );
    cublasFree( dAl );
    cublasFree( dCh );
    cublasFree( dCl );
    cublasFree( dalphah );
    cublasFree( dalphal );
    cublasFree( dbetah );
    cublasFree( dbetal );
    printf("Device Malloc Error in gddsyrk_host_\n");
    return -1;
  }
  else{
    /* Send Matrices(column-majorでFortranと一緒)  */
    cublasSetVector( sizeA, sizeof(double), Ah, 1, dAh, 1 );
    cublasSetVector( sizeA, sizeof(double), Al, 1, dAl, 1 );
    cublasSetVector( sizeC, sizeof(double), Ch, 1, dCh, 1 );
    cublasSetVector( sizeC, sizeof(double), Cl, 1, dCl, 1 );
    cublasSetVector( 1, sizeof(double), alphah, 1, dalphah, 1 );
    cublasSetVector( 1, sizeof(double), alphal, 1, dalphal, 1 );
    cublasSetVector( 1, sizeof(double), betah, 1, dbetah, 1 );
    cublasSetVector( 1, sizeof(double), betal, 1, dbetal, 1 );
  }

  /* Execute gpu function */
  gddsyrk
    ( uplo, trans, n, k, dalphah, dalphal, dAh, dAl, lda,
      dbetah, dbetal, dCh, dCl, ldc );

  /* Get From Device Memory */
  cublasGetVector( sizeC, sizeof(double), dCh, 1, Ch, 1 );
  cublasGetVector( sizeC, sizeof(double), dCl, 1, Cl, 1 );

  /* Free Device Memories */
  cublasFree( dAh );
  cublasFree( dAl );
  cublasFree( dCh );
  cublasFree( dCl );
  cublasFree( dalphah );
  cublasFree( dalphal );
  cublasFree( dbetah );
  cublasFree( dbetal );

  return 0;
}
