#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddidamax_host_( int *n, double *xh,double *xl, int*incx)
{
    double *dxh, *dxl;
    int xsize = *n*(*incx);


    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( xsize, sizeof( double ), (void**) &dxh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( xsize, sizeof( double ), (void**) &dxl )
      )
    {
/*	 Memory allocation failed */
        cublasFree( dxh );
        cublasFree( dxl );
        return -1;
    }
    else
    {
        /* Device Memory allocation */
        cublasSetVector( xsize, sizeof( double ), xh, 1, dxh, 1 );
        cublasSetVector( xsize, sizeof( double ), xl, 1, dxl, 1 );
    }

    /* Execute gpu vector add */
   int ret =  gddidamax( n, dxh, dxl, incx);


    /* Free Device Memories */
    cublasFree( dxh );
    cublasFree( dxl );


    return ret;
}

#ifdef __cplusplus
}
#endif
