#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddrotm_host_( int *n,  double *xh, double *xl, int*incx, double *yh,
             double *yl, int *incy,
	     double * paramh,double* paraml
              )
{
    double *dxh, *dxl, *dyh, *dyl, *dparamh, *dparaml;

    int     xsize = *n*( *incx );
    int     ysize = *n*( *incy );


    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( xsize, sizeof( double ), (void**) &dxh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( xsize, sizeof( double ), (void**) &dxl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( ysize, sizeof( double ), (void**) &dyh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( ysize, sizeof( double ), (void**) &dyl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 5, sizeof( double ), (void**) &dparamh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 5, sizeof( double ), (void**) &dparaml )
      )
    {
/*	 Memory allocation failed */
        cublasFree( dxh );
        cublasFree( dxl );
        cublasFree( dyh );
        cublasFree( dyl );
        cublasFree( dparamh );
        cublasFree( dparaml );
        return -1;
    }
    else
    {
        /* Device Memory allocation */
        cublasSetVector( xsize, sizeof( double ), xh, 1, dxh, 1 );
        cublasSetVector( xsize, sizeof( double ), xl, 1, dxl, 1 );
        cublasSetVector( ysize, sizeof( double ), yh, 1, dyh, 1 );
        cublasSetVector( ysize, sizeof( double ), yl, 1, dyl, 1 );
        cublasSetVector( 5, sizeof( double ), paramh, 1, dparamh, 1 );
        cublasSetVector( 5, sizeof( double ), paraml, 1, dparaml, 1 );
    }

    /* Execute gpu vector add */
    gddrotm( n, dxh, dxl, incx, dyh, dyl, incy, dparamh, dparaml);

    /* Get From Device Memory */
    cublasGetVector( xsize, sizeof( double ), dxh, 1, xh, 1 );
    cublasGetVector( xsize, sizeof( double ), dxl, 1, xl, 1 );
    cublasGetVector( ysize, sizeof( double ), dyh, 1, yh, 1 );
    cublasGetVector( ysize, sizeof( double ), dyl, 1, yl, 1 );

    /* Free Device Memories */
    cublasFree( dxh );
    cublasFree( dxl );
    cublasFree( dyh );
    cublasFree( dyl );
    cublasFree( dparamh );
    cublasFree( dparaml );


    return 0;
}

#ifdef __cplusplus
}
#endif
