#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddscal_host_( int *n, double *alphah, double *alphal, double *xh,double *xl, int*incx )
{
    double *dxh, *dxl, *dalphah, *dalphal;


    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( *n*(*incx), sizeof( double ), (void**) &dxh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( *n*(*incx), sizeof( double ), (void**) &dxl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( double ), (void**) &dalphah )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( double ), (void**) &dalphal )
      )
    {
/*	 Memory allocation failed */
        cublasFree( dxh );
        cublasFree( dxl );
        cublasFree( dalphah );
        cublasFree( dalphal );
        return -1;
    }
    else
    {
        /* Device Memory allocation */
        cublasSetVector( *n*(*incx), sizeof( double ), xh, 1, dxh, 1 );
        cublasSetVector( *n*(*incx), sizeof( double ), xl, 1, dxl, 1 );
        cublasSetVector( 1, sizeof( double ), alphah, 1, dalphah, 1 );
        cublasSetVector( 1, sizeof( double ), alphal, 1, dalphal, 1 );
    }

    /* Execute gpu vector add */
    gddscal( n,  dalphah, dalphal, dxh, dxl, incx);

    /* Get From Device Memory */
    cublasGetVector( *n*(*incx), sizeof( double ), dxh, 1, xh, 1 );
    cublasGetVector( *n*(*incx), sizeof( double ), dxl, 1, xl, 1 );

    /* Free Device Memories */
    cublasFree( dxh );
    cublasFree( dxl );
    cublasFree( dalphah );
    cublasFree( dalphal );


    return 0;
}

#ifdef __cplusplus
}
#endif
