#include "gddcommon.h"

int
gddswap_host_( int *n, double *dxh, double*dxl, int *incx,
              double *dyh, double * dyl, int *incy )
{
    double *dxh_dev, *dxl_dev, *dyh_dev, *dyl_dev;

    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( *n*(*incx), sizeof( double ), (void**) &dxh_dev )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( *n*(*incx), sizeof( double ), (void**) &dxl_dev )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( *n*(*incy), sizeof( double ), (void**) &dyh_dev )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( *n*(*incy), sizeof( double ), (void**) &dyl_dev )
      )
    {
/*	 Memory allocation failed */
        cublasFree( dxh_dev );
        cublasFree( dxl_dev );
        cublasFree( dyh_dev );
        cublasFree( dyl_dev );
        return -1;
    }
    else
    {
	    // Device Memory allocation
        cublasSetVector( *n*(*incx), sizeof( double ), dxh, 1, dxh_dev, 1 );
        cublasSetVector( *n*(*incx), sizeof( double ), dxl, 1, dxl_dev, 1 );
        cublasSetVector( *n*(*incy), sizeof( double ), dyh, 1, dyh_dev, 1 );
        cublasSetVector( *n*(*incy), sizeof( double ), dyl, 1, dyl_dev, 1 );
    }

    // Execute gpu swap
    gddswap( n, dxh_dev, dxl_dev, incx, dyh_dev, dyl_dev, incy );

    // Get From Device Memory
    cublasGetVector( *n*(*incx), sizeof( double ), dxh_dev, 1, dxh, 1 );
    cublasGetVector( *n*(*incx), sizeof( double ), dxl_dev, 1, dxl, 1 );
    cublasGetVector( *n*(*incy), sizeof( double ), dyh_dev, 1, dyh, 1 );
    cublasGetVector( *n*(*incy), sizeof( double ), dyl_dev, 1, dyl, 1 );

    //Free Device Memories
    cublasFree( dxh_dev );
    cublasFree( dxl_dev );
    cublasFree( dyh_dev );
    cublasFree( dyl_dev );

    return 0;
}

