#include "gddcommon.h"

#ifdef __cplusplus
extern "C" {
#endif
int
gddsyr2_host_( char *uplo, int *n, double *alphah, double*alphal, double * xh, double *xl,
              int *incx,
              double*yh, double*yl,
              int *incy  ,
              double *Ah, double *Al, int *lda)
{

    int     sizeA =  *lda*(*n);
    int     sizex = (*n)*(*incx);
    int     sizey = (*n)*(*incy);

    double *dAh, *dAl, *dalphah, *dalphal, *dxh, *dxl, *dyh, *dyl;

    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizeA, sizeof( double ), (void**) &dAh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizeA, sizeof( double ), (void**) &dAl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizex, sizeof( double ), (void**) &dxh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizex, sizeof( double ), (void**) &dxl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizey, sizeof( double ), (void**) &dyh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizey, sizeof( double ), (void**) &dyl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( double ), (void**) &dalphah )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( double ), (void**) &dalphal )
      )
    {
/*	 Memory allocation failed */
        cublasFree( dAh );
        cublasFree( dAl );
	cublasFree( dxh );
	cublasFree( dxl );
	cublasFree( dyh );
	cublasFree( dyl );
        cublasFree( dalphah );
        cublasFree( dalphal );
        printf( "Device Malloc Error\n" );
        return -1;
    }
    else
    {
        /* Send Matrices(column-majorでFortranと一緒)  */
        cublasSetVector( sizeA, sizeof( double ), Ah, 1, dAh, 1 );
        cublasSetVector( sizeA, sizeof( double ), Al, 1, dAl, 1 );
        cublasSetVector( sizex, sizeof( double ), xh, 1, dxh, 1 );
        cublasSetVector( sizex, sizeof( double ), xl, 1, dxl, 1 );
        cublasSetVector( sizey, sizeof( double ), yh, 1, dyh, 1 );
        cublasSetVector( sizey, sizeof( double ), yl, 1, dyl, 1 );
        cublasSetVector( 1, sizeof( double ), alphah, 1, dalphah, 1 );
        cublasSetVector( 1, sizeof( double ), alphal, 1, dalphal, 1 );
    }

    /* Execute gpu vector add */
    gddsyr2( uplo,  n, dalphah,  dalphal, dxh,
        dxl,  incx,    dyh,  dyl, incy , dAh,  dAl,  lda);

    /* Get From Device Memory */
    cublasGetVector( sizeA, sizeof( double ), dAh, 1, Ah, 1 );
    cublasGetVector( sizeA, sizeof( double ), dAl, 1, Al, 1 );

    /* Free Device Memories */
    cublasFree( dAh );
    cublasFree( dAl );
    cublasFree( dxh );
    cublasFree( dxl );
    cublasFree( dyh );
    cublasFree( dyl );
    cublasFree( dalphah );
    cublasFree( dalphal );

    return 0;
}
#ifdef __cplusplus
}
#endif

