#include "gddcommon.h"

int gddtrsm_host_
( char* side, char* uplo, char* transA, char* diag,
  int* m, int* n,
  double* alphah, double* alphal,
  double* Ah, double* Al, int* lda,
  double* Bh, double* Bl, int* ldb )
{
  int nrowA, ncolA;
  int nrowB, ncolB;
  int sizeA, sizeB;

  double *dAh, *dAl;
  double *dBh, *dBl;
  double *dalphah, *dalphal;

  if( *side == 'L' || *side == 'l' ){
    /* B(M,N) += A(M,M)*B(M,N), A is symmetric, B is general */
    nrowA = *m, ncolA = *m;
    nrowB = *m, ncolB = *n;
  }
  else{
    /* B(M,N) += B(M,N)*A(N,N), A is symmetric, B is general */
    nrowB = *m, ncolB = *n;
    nrowA = *n, ncolA = *n;
  }

  if( *lda < nrowA || *ldb < nrowB ){
    printf("Argument Error in gddtrmm_host_\n");
    printf("wrong lda or ldb.\n");
    return -1;
  }

  sizeA = (*lda)*ncolA;
  sizeB = (*ldb)*ncolB;

  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**)&dAh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(double), (void**)&dAl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(double), (void**)&dBh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(double), (void**)&dBl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**)&dalphah ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(double), (void**)&dalphal ) ){
    /*	 Memory allocation failed */
    cublasFree( dAh );
    cublasFree( dAl );
    cublasFree( dBh );
    cublasFree( dBl );
    cublasFree( dalphah );
    cublasFree( dalphal );
    printf("Device Malloc Error in gddtrmm_host_\n");
    return -1;
  }
  else{
    /* Send Matrices(column-majorでFortranと一緒)  */
    cublasSetVector( sizeA, sizeof(double), Ah, 1, dAh, 1 );
    cublasSetVector( sizeA, sizeof(double), Al, 1, dAl, 1 );
    cublasSetVector( sizeB, sizeof(double), Bh, 1, dBh, 1 );
    cublasSetVector( sizeB, sizeof(double), Bl, 1, dBl, 1 );
    cublasSetVector( 1, sizeof(double), alphah, 1, dalphah, 1 );
    cublasSetVector( 1, sizeof(double), alphal, 1, dalphal, 1 );
  }

  /* Execute gpu function */
  gddtrsm
    ( side, uplo, transA, diag, 
      m, n, dalphah, dalphal, dAh, dAl, lda, dBh, dBl, ldb );

  /* Get From Device Memory */
  cublasGetVector( sizeB, sizeof(double), dBh, 1, Bh, 1 );
  cublasGetVector( sizeB, sizeof(double), dBl, 1, Bl, 1 );

  /* Free Device Memories */
  cublasFree( dAh );
  cublasFree( dAl );
  cublasFree( dBh );
  cublasFree( dBl );
  cublasFree( dalphah );
  cublasFree( dalphal );
  
  return 0;
}



