#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddzdotc_host_( int *n, cuDoubleComplex *xh,cuDoubleComplex *xl, int*incx,cuDoubleComplex *yh, cuDoubleComplex *yl,int *incy  ,cuDoubleComplex *reth,cuDoubleComplex*retl)
{
    cuDoubleComplex *dxh, *dxl, *dyh, *dyl, *dreth, *dretl;
    int xsize = *n*(*incx);
    int ysize = *n*(*incy);


    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( xsize, sizeof( cuDoubleComplex ), (void**) &dxh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( xsize, sizeof( cuDoubleComplex ), (void**) &dxl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( ysize, sizeof( cuDoubleComplex ), (void**) &dyh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( ysize, sizeof( cuDoubleComplex ), (void**) &dyl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dreth )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dretl )
      )
    {
/*	 Memory allocation failed */
        cublasFree( dxh );
        cublasFree( dxl );
        cublasFree( dyh );
        cublasFree( dyl );
        cublasFree( dreth );
        cublasFree( dretl );
        return -1;
    }
    else
    {
        /* Device Memory allocation */
        cublasSetVector( xsize, sizeof( cuDoubleComplex ), xh, 1, dxh, 1 );
        cublasSetVector( xsize, sizeof( cuDoubleComplex ), xl, 1, dxl, 1 );
        cublasSetVector( ysize, sizeof( cuDoubleComplex ), yh, 1, dyh, 1 );
        cublasSetVector( ysize, sizeof( cuDoubleComplex ), yl, 1, dyl, 1 );
    }

    /* Execute gpu vector add */
    gddzdotc( n, dxh, dxl, incx, dyh, dyl, incy ,dreth,dretl);

    /* Get From Device Memory */
    cublasGetVector( 1, sizeof( cuDoubleComplex ), dreth, 1, reth, 1 );
    cublasGetVector( 1, sizeof( cuDoubleComplex ), dretl, 1, retl, 1 );

    /* Free Device Memories */
    cublasFree( dxh );
    cublasFree( dxl );
    cublasFree( dyh );
    cublasFree( dyl );
    cublasFree( dreth );
    cublasFree( dretl );


    return 0;
}

#ifdef __cplusplus
}
#endif
