#include "gddcommon.h"

#ifdef __cplusplus
extern "C" {
#endif
int
gddzhbmv_host_( char*uplo, int *n, int *k, cuDoubleComplex *alphah, cuDoubleComplex*alphal,
              cuDoubleComplex *Ah, cuDoubleComplex *Al, int *lda, cuDoubleComplex * xh, cuDoubleComplex *xl,
              int *incx,
              cuDoubleComplex *betah, cuDoubleComplex*betal, cuDoubleComplex*yh, cuDoubleComplex*yl,
              int *incy  )
{
    int sizeA =  *lda*( *n );

    int sizex, sizey;
    sizex = (*n)*(*incx);
    sizey = (*n)*(*incy);

    cuDoubleComplex *dAh, *dAl, *dalphah, *dalphal, *dbetah, *dbetal, *dxh, *dxl,
    *dyh, *dyl;

    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizeA, sizeof( cuDoubleComplex ), (void**) &dAh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizeA, sizeof( cuDoubleComplex ), (void**) &dAl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizex, sizeof( cuDoubleComplex ), (void**) &dxh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizex, sizeof( cuDoubleComplex ), (void**) &dxl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizey, sizeof( cuDoubleComplex ), (void**) &dyh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizey, sizeof( cuDoubleComplex ), (void**) &dyl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dalphah )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dalphal )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dbetah )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dbetal )
      )
    {
/*	 Memory allocation failed */
        cublasFree( dAh );
        cublasFree( dAl );
        cublasFree( dxh );
        cublasFree( dxl );
        cublasFree( dyh );
        cublasFree( dyl );
        cublasFree( dalphah );
        cublasFree( dalphal );
        cublasFree( dbetah );
        cublasFree( dbetal );
        printf( "Device Malloc Error\n" );
        return -1;
    }
    else
    {
        /* Send Matrices(column-majorでFortranと一緒)  */
        cublasSetVector( sizeA, sizeof( cuDoubleComplex ), Ah, 1, dAh, 1 );
        cublasSetVector( sizeA, sizeof( cuDoubleComplex ), Al, 1, dAl, 1 );
        cublasSetVector( sizex, sizeof( cuDoubleComplex ), xh, 1, dxh, 1 );
        cublasSetVector( sizex, sizeof( cuDoubleComplex ), xl, 1, dxl, 1 );
        cublasSetVector( sizey, sizeof( cuDoubleComplex ), yh, 1, dyh, 1 );
        cublasSetVector( sizey, sizeof( cuDoubleComplex ), yl, 1, dyl, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), alphah, 1, dalphah, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), alphal, 1, dalphal, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), betah, 1, dbetah, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), betal, 1, dbetal, 1 );
    }

    gddzhbmv( uplo,n,k, dalphah,dalphal,dAh,  dAl,  lda, dxh,
        dxl,  incx,dbetah,dbetal,dyh,dyl,incy);

    /* Get From Device Memory */
    cublasGetVector( sizey, sizeof( cuDoubleComplex ), dyh, 1, yh, 1 );
    cublasGetVector( sizey, sizeof( cuDoubleComplex ), dyl, 1, yl, 1 );

    /* Free Device Memories */
    cublasFree( dAh );
    cublasFree( dAl );
    cublasFree( dxh );
    cublasFree( dxl );
    cublasFree( dyh );
    cublasFree( dyl );
    cublasFree( dalphah );
    cublasFree( dalphal );
    cublasFree( dbetah );
    cublasFree( dbetal );

    return 0;
}

#ifdef __cplusplus
}
#endif

