#include "gddcommon.h"

#ifdef __cplusplus
extern "C" {
#endif
int
gddzher2_host_( char *uplo, int *n, cuDoubleComplex *alphah, cuDoubleComplex*alphal, cuDoubleComplex * xh, cuDoubleComplex *xl,
              int *incx,
              cuDoubleComplex*yh, cuDoubleComplex*yl,
              int *incy  ,
              cuDoubleComplex *Ah, cuDoubleComplex *Al, int *lda)
{

    int     sizeA =  *lda*(*n);
    int     sizex = (*n)*(*incx);
    int     sizey = (*n)*(*incy);

    cuDoubleComplex *dAh, *dAl, *dalphah, *dalphal, *dxh, *dxl, *dyh, *dyl;

    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizeA, sizeof( cuDoubleComplex ), (void**) &dAh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizeA, sizeof( cuDoubleComplex ), (void**) &dAl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizex, sizeof( cuDoubleComplex ), (void**) &dxh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizex, sizeof( cuDoubleComplex ), (void**) &dxl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizey, sizeof( cuDoubleComplex ), (void**) &dyh )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( sizey, sizeof( cuDoubleComplex ), (void**) &dyl )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dalphah )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dalphal )
      )
    {
/*	 Memory allocation failed */
        cublasFree( dAh );
        cublasFree( dAl );
	cublasFree( dxh );
	cublasFree( dxl );
	cublasFree( dyh );
	cublasFree( dyl );
        cublasFree( dalphah );
        cublasFree( dalphal );
        printf( "Device Malloc Error\n" );
        return -1;
    }
    else
    {
        /* Send Matrices(column-majorでFortranと一緒)  */
        cublasSetVector( sizeA, sizeof( cuDoubleComplex ), Ah, 1, dAh, 1 );
        cublasSetVector( sizeA, sizeof( cuDoubleComplex ), Al, 1, dAl, 1 );
        cublasSetVector( sizex, sizeof( cuDoubleComplex ), xh, 1, dxh, 1 );
        cublasSetVector( sizex, sizeof( cuDoubleComplex ), xl, 1, dxl, 1 );
        cublasSetVector( sizey, sizeof( cuDoubleComplex ), yh, 1, dyh, 1 );
        cublasSetVector( sizey, sizeof( cuDoubleComplex ), yl, 1, dyl, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), alphah, 1, dalphah, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), alphal, 1, dalphal, 1 );
    }

    /* Execute gpu vector add */
    gddzher2( uplo,  n, dalphah,  dalphal, dxh,
        dxl,  incx,    dyh,  dyl, incy , dAh,  dAl,  lda);

    /* Get From Device Memory */
    cublasGetVector( sizeA, sizeof( cuDoubleComplex ), dAh, 1, Ah, 1 );
    cublasGetVector( sizeA, sizeof( cuDoubleComplex ), dAl, 1, Al, 1 );

    /* Free Device Memories */
    cublasFree( dAh );
    cublasFree( dAl );
    cublasFree( dxh );
    cublasFree( dxl );
    cublasFree( dyh );
    cublasFree( dyl );
    cublasFree( dalphah );
    cublasFree( dalphal );

    return 0;
}
#ifdef __cplusplus
}
#endif

