#include "gddcommon.h"

int gdzgemm_host_
( char* transA, char* transB,
  int*m, int* n, int* k,
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex* Ah, cuDoubleComplex* Al, int* lda,
  cuDoubleComplex* Bh, cuDoubleComplex* Bl, int* ldb,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex* Ch, cuDoubleComplex* Cl, int* ldc )
{
  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;
  int sizeA, sizeB, sizeC;

  cuDoubleComplex *dAh, *dAl;
  cuDoubleComplex *dBh, *dBl;
  cuDoubleComplex *dCh, *dCl;
  cuDoubleComplex *dalphah, *dalphal;
  cuDoubleComplex *dbetah, *dbetal;

  if( *transA == 'N' || *transA == 'n' ){
    /* A(M,K) */
    nrowA = *m, ncolA = *k;
  }
  else{
    /* A(K,M) */
    nrowA = *k, ncolA = *m;
  }

  if( *transB == 'N' || *transB == 'n' ){
    /* B(K,N) */
    nrowB = *k, ncolB = *n;
  }
  else{
    /* B(N,K) */
    nrowB = *n, ncolB = *k;
  }
  {
    /* C(M,N) */
    nrowC = *m, ncolC = *n;
  }

  if( *lda < nrowA || *ldb < nrowB || *ldc < nrowC ){
    printf("Argument Error in gdzgemm_host_.\n");
    printf("wrong lda or ldb or ldc.\n");
    return -1;
  }

  sizeA = (*lda)*ncolA;
  sizeB = (*ldb)*ncolB;
  sizeC = (*ldc)*ncolC;

  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(cuDoubleComplex), (void**)&dAh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(cuDoubleComplex), (void**)&dAl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(cuDoubleComplex), (void**)&dBh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(cuDoubleComplex), (void**)&dBl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**)&dCh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**)&dCl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&dalphah ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&dalphal ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&dbetah  ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&dbetal ) ){
    /*	 Memory allocation failed */
    cublasFree( dAh );
    cublasFree( dAl );
    cublasFree( dBh );
    cublasFree( dBl );
    cublasFree( dCh );
    cublasFree( dCl );
    cublasFree( dalphah );
    cublasFree( dalphal );
    cublasFree( dbetah );
    cublasFree( dbetal );
    printf("Device Malloc Error in gdzgemm_host_.\n");
    return -1;
  }
  else{
    /* Send Matrices(column-majorでFortranと一緒)  */
    cublasSetVector( sizeA, sizeof(cuDoubleComplex), Ah, 1, dAh, 1 );
    cublasSetVector( sizeA, sizeof(cuDoubleComplex), Al, 1, dAl, 1 );
    cublasSetVector( sizeB, sizeof(cuDoubleComplex), Bh, 1, dBh, 1 );
    cublasSetVector( sizeB, sizeof(cuDoubleComplex), Bl, 1, dBl, 1 );
    cublasSetVector( sizeC, sizeof(cuDoubleComplex), Ch, 1, dCh, 1 );
    cublasSetVector( sizeC, sizeof(cuDoubleComplex), Cl, 1, dCl, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), alphah, 1, dalphah, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), alphal, 1, dalphal, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), betah, 1, dbetah, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), betal, 1, dbetal, 1 );
  }
  
  /* Execute gpu function */
  gdzgemm
    ( transA, transB,
      m, n, k, dalphah, dalphal, dAh, dAl, lda, dBh, dBl, ldb,
      dbetah, dbetal, dCh, dCl, ldc );

  /* Get From Device Memory */
  cublasGetVector( sizeC, sizeof(cuDoubleComplex), dCh, 1, Ch, 1 );
  cublasGetVector( sizeC, sizeof(cuDoubleComplex), dCl, 1, Cl, 1 );

  /* Free Device Memories */
  cublasFree( dAh );
  cublasFree( dAl );
  cublasFree( dBh );
  cublasFree( dBl );
  cublasFree( dCh );
  cublasFree( dCl );
  cublasFree( dalphah );
  cublasFree( dalphal );
  cublasFree( dbetah );
  cublasFree( dbetal );
  
  return 0;
}

int gdzhemm_host_
( char* side, char* uplo,
  int* m, int* n,
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex* Ah, cuDoubleComplex* Al, int* lda,
  cuDoubleComplex* Bh, cuDoubleComplex* Bl, int* ldb,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex* Ch, cuDoubleComplex* Cl, int* ldc )
{
  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;
  int sizeA, sizeB, sizeC;

  if( *side == 'L' || *side == 'l' ){
    /* C(M,N) += A(M,M)*B(M,N), A is hermite, B is general */
    nrowA = *m, ncolA = *m;
    nrowB = *m, ncolB = *n;
    nrowC = *m, ncolC = *n;
  }
  else{
    /* C(M,N) += B(M,N)*A(N,N), A is hermite, B is general */
    nrowB = *m, ncolB = *n;
    nrowA = *n, ncolA = *n;
    nrowC = *m, ncolC = *n;
  }

  if( *lda < nrowA || *ldb < nrowB || *ldc < nrowC ){
    printf("Argument Error in gdzhemm_host_\n");
    printf("wrong lda or ldb or ldc.\n");
    return -1;
  }

  sizeA = (*lda)*ncolA;
  sizeB = (*ldb)*ncolB;
  sizeC = (*ldc)*ncolC;

  cuDoubleComplex *dAh, *dAl;
  cuDoubleComplex *dBh, *dBl;
  cuDoubleComplex *dCh, *dCl;
  cuDoubleComplex *dalphah, *dalphal;
  cuDoubleComplex *dbetah, *dbetal;

  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(cuDoubleComplex), (void**)&dAh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(cuDoubleComplex), (void**)&dAl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(cuDoubleComplex), (void**)&dBh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(cuDoubleComplex), (void**)&dBl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**)&dCh ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**)&dCl ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&dalphah ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&dalphal ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&dbetah  ) ||
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**)&dbetal ) ){
    /*	 Memory allocation failed */
    cublasFree( dAh );
    cublasFree( dAl );
    cublasFree( dBh );
    cublasFree( dBl );
    cublasFree( dCh );
    cublasFree( dCl );
    cublasFree( dalphah );
    cublasFree( dalphal );
    cublasFree( dbetah );
    cublasFree( dbetal );
    printf("Device Malloc Error in gdzhemm_host_\n");
    return -1;
  }
  else{
    /* Send Matrices(column-majorでFortranと一緒)  */
    cublasSetVector( sizeA, sizeof(cuDoubleComplex), Ah, 1, dAh, 1 );
    cublasSetVector( sizeA, sizeof(cuDoubleComplex), Al, 1, dAl, 1 );
    cublasSetVector( sizeB, sizeof(cuDoubleComplex), Bh, 1, dBh, 1 );
    cublasSetVector( sizeB, sizeof(cuDoubleComplex), Bl, 1, dBl, 1 );
    cublasSetVector( sizeC, sizeof(cuDoubleComplex), Ch, 1, dCh, 1 );
    cublasSetVector( sizeC, sizeof(cuDoubleComplex), Cl, 1, dCl, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), alphah, 1, dalphah, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), alphal, 1, dalphal, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), betah, 1, dbetah, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), betal, 1, dbetal, 1 );
  }
  
  /* Execute gpu function */
  gdzhemm
    ( side, uplo,
      m, n, dalphah, dalphal, dAh, dAl, lda, dBh, dBl, ldb,
      dbetah, dbetal, dCh, dCl, ldc );

  /* Get From Device Memory */
  cublasGetVector( sizeC, sizeof(cuDoubleComplex), dCh, 1, Ch, 1 );
  cublasGetVector( sizeC, sizeof(cuDoubleComplex), dCl, 1, Cl, 1 );

  /* Free Device Memories */
  cublasFree( dAh );
  cublasFree( dAl );
  cublasFree( dBh );
  cublasFree( dBl );
  cublasFree( dCh );
  cublasFree( dCl );
  cublasFree( dalphah );
  cublasFree( dalphal );
  cublasFree( dbetah );
  cublasFree( dbetal );
  
  return 0;
}



int gdzsyr2k_host_
( char* uplo, char* trans,
  int* n, int* k,
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex* Ah, cuDoubleComplex* Al, int* lda,
  cuDoubleComplex* Bh, cuDoubleComplex* Bl, int* ldb,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex* Ch, cuDoubleComplex* Cl, int* ldc )
{
  int nrowA, ncolA;
  int nrowB, ncolB;
  int nrowC, ncolC;
  int sizeA, sizeB, sizeC;

  if( *trans == 'N' || *trans == 'n' ){
    /* C(N,N) += A(N,K) * B(N,K)^T */
    nrowA = *n, ncolA = *k;
    nrowB = *n, ncolB = *k;
    nrowC = *n, ncolC = *n;
  }
  else{
    /* C(N,N) += A(K,N)^T * B(K,N) */
    nrowA = *k, ncolA = *n;
    nrowB = *k, ncolB = *n;
    nrowC = *n, ncolC = *n;
  }

  if( *lda < nrowA || *ldb < nrowB || *ldc < nrowC ){
    printf("Argument Error in gdzsyr2k_host_\n");
    printf("wrong lda or ldb or ldc.\n");
    return -1;
  }

  sizeA = (*lda)*ncolA;
  sizeB = (*ldb)*ncolB;
  sizeC = (*ldc)*ncolC;

  cuDoubleComplex *dAh, *dAl;
  cuDoubleComplex *dBh, *dBl;
  cuDoubleComplex *dCh, *dCl;
  cuDoubleComplex *dalphah, *dalphal;
  cuDoubleComplex *dbetah, *dbetal;

  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(cuDoubleComplex), (void**) &dAh )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(cuDoubleComplex), (void**) &dAl )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(cuDoubleComplex), (void**) &dBh )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeB, sizeof(cuDoubleComplex), (void**) &dBl )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**) &dCh )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**) &dCl )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**) &dalphah )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**) &dalphal )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**) &dbetah )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**) &dbetal ) ){
      /*	 Memory allocation failed */
    cublasFree( dAh );
    cublasFree( dAl );
    cublasFree( dBh );
    cublasFree( dBl );
    cublasFree( dCh );
    cublasFree( dCl );
    cublasFree( dalphah );
    cublasFree( dalphal );
    cublasFree( dbetah );
    cublasFree( dbetal );
    printf("Device Malloc Error in gdzsyr2k_host_\n");
    return -1;
  }
  else{
    /* Send Matrices(column-majorでFortranと一緒)  */
    cublasSetVector( sizeA, sizeof(cuDoubleComplex), Ah, 1, dAh, 1 );
    cublasSetVector( sizeA, sizeof(cuDoubleComplex), Al, 1, dAl, 1 );
    cublasSetVector( sizeB, sizeof(cuDoubleComplex), Bh, 1, dBh, 1 );
    cublasSetVector( sizeB, sizeof(cuDoubleComplex), Bl, 1, dBl, 1 );
    cublasSetVector( sizeC, sizeof(cuDoubleComplex), Ch, 1, dCh, 1 );
    cublasSetVector( sizeC, sizeof(cuDoubleComplex), Cl, 1, dCl, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), alphah, 1, dalphah, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), alphal, 1, dalphal, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), betah, 1, dbetah, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), betal, 1, dbetal, 1 );
  }

  /* Execute gpu function */
  gdzsyr2k
    ( uplo, trans, n, k, dalphah, dalphal, dAh, dAl, lda,
      dBh, dBl, ldb, dbetah, dbetal, dCh, dCl, ldc );

  /* Get From Device Memory */
  cublasGetVector( sizeC, sizeof(cuDoubleComplex), dCh, 1, Ch, 1 );
  cublasGetVector( sizeC, sizeof(cuDoubleComplex), dCl, 1, Cl, 1 );

  /* Free Device Memories */
  cublasFree( dAh );
  cublasFree( dAl );
  cublasFree( dBh );
  cublasFree( dBl );
  cublasFree( dCh );
  cublasFree( dCl );
  cublasFree( dalphah );
  cublasFree( dalphal );
  cublasFree( dbetah );
  cublasFree( dbetal );

  return 0;
}

int gdzsyrk_host_
( char* uplo, char* trans,
  int* n, int* k,
  cuDoubleComplex* alphah, cuDoubleComplex* alphal,
  cuDoubleComplex* Ah, cuDoubleComplex* Al, int* lda,
  cuDoubleComplex* betah, cuDoubleComplex* betal,
  cuDoubleComplex* Ch, cuDoubleComplex* Cl, int* ldc )
{
  int nrowA, ncolA;
  int nrowC, ncolC;
  int sizeA, sizeC;

  if( *trans == 'N' || *trans == 'n' ){
    /* C(N,N) += A(N,K) * A(N,K)^T */
    nrowA = *n, ncolA = *k;
    nrowC = *n, ncolC = *n;
  }
  else{
    /* C(N,N) += A(K,N)^T * A(K,N) */
    nrowA = *k, ncolA = *n;
    nrowC = *n, ncolC = *n;
  }

  if( *lda < nrowA || *ldc < nrowC ){
    printf("Argument Error in gdzsyrk_host_\n");
    printf("wrong lda or ldc.\n");
    return -1;
  }

  sizeA = (*lda)*ncolA;
  sizeC = (*ldc)*ncolC;

  cuDoubleComplex *dAh, *dAl;
  cuDoubleComplex *dCh, *dCl;
  cuDoubleComplex *dalphah, *dalphal;
  cuDoubleComplex *dbetah, *dbetal;

  if(
     CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(cuDoubleComplex), (void**) &dAh )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeA, sizeof(cuDoubleComplex), (void**) &dAl )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**) &dCh )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( sizeC, sizeof(cuDoubleComplex), (void**) &dCl )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**) &dalphah )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**) &dalphal )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**) &dbetah )
     || CUBLAS_STATUS_SUCCESS !=
     cublasAlloc( 1, sizeof(cuDoubleComplex), (void**) &dbetal ) ){
      /*	 Memory allocation failed */
    cublasFree( dAh );
    cublasFree( dAl );
    cublasFree( dCh );
    cublasFree( dCl );
    cublasFree( dalphah );
    cublasFree( dalphal );
    cublasFree( dbetah );
    cublasFree( dbetal );
    printf("Device Malloc Error in gdzsyrk_host_\n");
    return -1;
  }
  else{
    /* Send Matrices(column-majorでFortranと一緒)  */
    cublasSetVector( sizeA, sizeof(cuDoubleComplex), Ah, 1, dAh, 1 );
    cublasSetVector( sizeA, sizeof(cuDoubleComplex), Al, 1, dAl, 1 );
    cublasSetVector( sizeC, sizeof(cuDoubleComplex), Ch, 1, dCh, 1 );
    cublasSetVector( sizeC, sizeof(cuDoubleComplex), Cl, 1, dCl, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), alphah, 1, dalphah, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), alphal, 1, dalphal, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), betah, 1, dbetah, 1 );
    cublasSetVector( 1, sizeof(cuDoubleComplex), betal, 1, dbetal, 1 );
  }

  /* Execute gpu function */
  gdzsyrk
    ( uplo, trans, n, k, dalphah, dalphal, dAh, dAl, lda,
      dbetah, dbetal, dCh, dCl, ldc );

  /* Get From Device Memory */
  cublasGetVector( sizeC, sizeof(cuDoubleComplex), dCh, 1, Ch, 1 );
  cublasGetVector( sizeC, sizeof(cuDoubleComplex), dCl, 1, Cl, 1 );

  /* Free Device Memories */
  cublasFree( dAh );
  cublasFree( dAl );
  cublasFree( dCh );
  cublasFree( dCl );
  cublasFree( dalphah );
  cublasFree( dalphal );
  cublasFree( dbetah );
  cublasFree( dbetal );

  return 0;
}
