#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddcopy_host_multigpu_( int *n, double *dxh, double*dxl, int *incx,
              double *dyh, double * dyl, int *incy )
{

//NUMBER OF GPUS
    int     ngpu;
    multiple_gpus::bind_cpu_to_gpu_( &ngpu );


    //ARRAYSIZE FOR EACH GPU
    int *   partition;
    int * pointerOfEachPartition;


    partition = (int*) malloc( sizeof( int )*ngpu );
    pointerOfEachPartition = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( *n, ngpu,  partition ,pointerOfEachPartition);




    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
    double *dxh_dev, *dxl_dev, *dyh_dev, *dyl_dev;
    int tid = omp_get_thread_num();
    int xsize = partition[tid]*(*incx);
    int ysize = partition[tid]*(*incy);

    int xOffset = pointerOfEachPartition[tid]*(*incx);
    int yOffset = pointerOfEachPartition[tid]*(*incy);


    if(
        CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( xsize, sizeof( double ), (void**) &dxh_dev )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( xsize, sizeof( double ), (void**) &dxl_dev )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( ysize, sizeof( double ), (void**) &dyh_dev )
      || CUBLAS_STATUS_SUCCESS !=
        cublasAlloc( ysize, sizeof( double ), (void**) &dyl_dev )
      )
    {
/*	 Memory allocation failed */
        cublasFree( dxh_dev );
        cublasFree( dxl_dev );
        cublasFree( dyh_dev );
        cublasFree( dyl_dev );
	printf("Device Malloc Error!\n");
        exit(-1);
    }
    else
    {
	    // Device Memory allocation
        cublasSetVector( xsize, sizeof( double ), dxh+xOffset, 1, dxh_dev, 1 );
        cublasSetVector( xsize, sizeof( double ), dxl+xOffset, 1, dxl_dev, 1 );
        cublasSetVector( ysize, sizeof( double ), dyh+yOffset, 1, dyh_dev, 1 );
        cublasSetVector( ysize, sizeof( double ), dyl+yOffset, 1, dyl_dev, 1 );
    }

#ifdef _DEBUG_
            int devid;
            cudaGetDevice( &devid );
            printf( "TID(%d), Cuda Device = %d\n", tid,devid );
#endif

    // Execute gpu 
    gddcopy( &partition[tid], dxh_dev, dxl_dev, incx, dyh_dev, dyl_dev, incy );

    // Get From Device Memory
    cublasGetVector( ysize, sizeof( double ), dyh_dev, 1, dyh+yOffset, 1 );
    cublasGetVector( ysize, sizeof( double ), dyl_dev, 1, dyl+yOffset, 1 );

    //Free Device Memories
    cublasFree( dxh_dev );
    cublasFree( dxl_dev );
    cublasFree( dyh_dev );
    cublasFree( dyl_dev );
    }

    free(partition);
    free(pointerOfEachPartition);

    return 0;
}

#ifdef __cplusplus
}
#endif

