#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>

#ifdef __cplusplus
extern "C" {
#endif

int
gddgbmv_host_multigpu_( char*trans, int *m, int *n, int*kl, int*ku, double *alphah,
              double*alphal,
              double *Ah, double *Al, int *lda, double * xh, double *xl,
              int *incx,
              double *betah, double*betal, double*yh, double*yl,
              int *incy  )
{
    /* NUMBER OF GPUS */
    int     ngpu;

    /* BIND THREADS TO EACH DEVICE */
    multiple_gpus::bind_cpu_to_gpu_( &ngpu );


    int    sizex, sizey ;
    if(*trans == 'T' || *trans == 't')
    {
	    sizex = *m;
	    sizey = *n;
    }else
    {
	    sizex = *n;
	    sizey = *m;
    }


    /* ARRAYSIZE FOR EACH GPU */
    int *   partition;
    int *   sumPartition;
    partition =    (int*) malloc( sizeof( int )*ngpu );
    sumPartition = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( sizey, ngpu,  partition, sumPartition );


    int     sizeA = ( *m )*( *n );

    double *hAh, *hAl;
    if(
        cudaSuccess !=
        cudaMallocHost( (void **)&hAh, sizeA*sizeof( double ) )
      || cudaSuccess !=
        cudaMallocHost( (void **)&hAl, sizeA*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    char diag = 'N';


    copyToBandSplitMatrix(trans,&diag,*m,*n, *lda,*kl,*ku, ngpu, partition, Ah,
                               Al, hAh,hAl );

 #ifdef _DEBUG_ 
/*    int i, j, k;*/

/*    for( i = 0; i < ngpu; i++ )*/
/*    {*/
/*            for( j = 0; j < partition[i]; j++ )*/
/*            {*/
/*        for( k = 0; k < sizex; k++ )*/
/*        {*/
/*                printf("%e,",*(hAh+j+k*partition[i] + sumPartition[i] * (sizex)));*/
/*            }*/
/*        printf("\n");*/
/*        }*/
/*        printf("\n");*/
/*    }*/
#endif



/*     Working Matrix  */
    omp_set_num_threads( ngpu ); 
 #pragma omp parallel 
    { 
	int     tid = omp_get_thread_num( ); 
 
        double *dAh,*dAl,*dxh, *dxl,*dyh,*dyl; 

	double *dalphah,*dalphal,*dbetah,*dbetal;
 
 
	int offsety = sumPartition[tid]*(*incy);
	int localsizey = partition[tid]*(*incy);

	int localn = sizex; 

	int offsetA = sumPartition[tid]*(localn); 
	int localsizeA = partition[tid]*(localn); 
	int locallda = partition[tid]; 
 
	int localm = partition[tid]; 
 
	/* B,alpha及びbetaを確保 */ 
	if( 
	   CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( localsizey, sizeof( double ), (void**) &dyh ) 
	  || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( localsizey, sizeof( double ), (void**) &dyl ) 
	   || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( sizex*(*incx), sizeof( double ), (void**) &dxh ) 
	   || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( sizex*(*incx), sizeof( double ), (void**) &dxl ) 
	   || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( 1, sizeof( double ), (void**) &dalphah ) 
	   || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( 1, sizeof( double ), (void**) &dalphal ) 
	   || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( 1, sizeof( double ), (void**) &dbetah ) 
	   || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( 1, sizeof( double ), (void**) &dbetal ) 
	  || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( localsizeA, sizeof( double ), (void**) &dAh ) 
	  || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( localsizeA, sizeof( double ), (void**) &dAl ) 
	  ) 
	{ 
	    /*	 Memory allocation failed */ 
	    cublasFree( dyh ); 
	    cublasFree( dyl ); 
	    cublasFree( dxh ); 
	    cublasFree( dxl ); 
	    cublasFree( dalphah ); 
	    cublasFree( dalphal ); 
	    cublasFree( dbetah ); 
	    cublasFree( dbetal ); 
	    cublasFree( dAh ); 
	    cublasFree( dAl ); 
	    exit( -1 ); 
	} 
 
 
	cublasSetVector( sizex*(*incx), sizeof( double ), xh, 1, dxh, 1 ); 
	cublasSetVector( sizex*(*incx), sizeof( double ), xl, 1, dxl, 1 ); 
	cublasSetVector( localsizey, sizeof( double ), yh+offsety, 1, dyh, 1 ); 
	cublasSetVector( localsizey, sizeof( double ), yl+offsety, 1, dyl, 1 ); 
        cublasSetVector( localsizeA, sizeof( double ), hAh+offsetA, 1, dAh, 1); 
        cublasSetVector( localsizeA, sizeof( double ), hAl+offsetA, 1, dAl, 1); 
	cublasSetVector( 1, sizeof( double ), alphah, 1, dalphah, 1 ); 
	cublasSetVector( 1, sizeof( double ), alphal, 1, dalphal, 1 ); 
	cublasSetVector( 1, sizeof( double ), betah, 1, dbetah, 1 ); 
	cublasSetVector( 1, sizeof( double ), betal, 1, dbetal, 1 ); 

 
 
 #ifdef _DEBUG_ 
	    int devid; 
	    cudaGetDevice( &devid ); 
	    printf( "TILE INDEX(%d), Cuda Device = %d\n", tid, 
		    devid ); 
 #endif 
	    char localtrans = 'N';

 
    gddgemv(&localtrans,  &localm,  &localn, dalphah,  dalphal, dAh,  dAl,
  &locallda, dxh,dxl,  incx,  dbetah, dbetal,  dyh,  dyl, incy ); 
#pragma omp barrier
	    cublasGetVector( localsizey, sizeof( double ), dyh, 1, yh+offsety, 
			     1 ); 
	    cublasGetVector( localsizey, sizeof( double ), dyl, 1, yl+offsety, 
			     1 ); 
 
	    cublasFree( dAh ); 
	    cublasFree( dAl ); 
	    cublasFree( dxh ); 
	    cublasFree( dxl ); 
	    cublasFree( dyh ); 
	    cublasFree( dyl ); 
	cublasFree( dalphah ); 
	cublasFree( dalphal ); 
	cublasFree( dbetah ); 
	cublasFree( dbetal ); 
    } 
 
 
    cudaFreeHost( hAh ); 
    cudaFreeHost( hAl ); 
 
    free( partition ); 
    free( sumPartition ); 
 
    return 0; 
}

#ifdef __cplusplus
}
#endif
