#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddidamax_host_multigpu_( int *n,  double *xh, double *xl, int*incx)
{
/* NUMBER OF GPUS */
    int   ngpu;

    multiple_gpus::bind_cpu_to_gpu_( &ngpu );

    /* ARRAYSIZE FOR EACH GPU */
    int * partition;
    int * pointerOfEachPartition;
    int * ret;

    partition = (int*) malloc( sizeof( int )*ngpu );
    pointerOfEachPartition = (int*) malloc( sizeof( int )*ngpu );
    ret = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( *n, ngpu,  partition, pointerOfEachPartition );

    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        double *dxh, *dxl;
        int     tid = omp_get_thread_num( );
        int     xsize = partition[tid]*( *incx );
        int     xOffset = pointerOfEachPartition[tid]*( *incx );


        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( xsize, sizeof( double ), (void**) &dxh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( xsize, sizeof( double ), (void**) &dxl )
          )
        {
/*	 Memory allocation failed */
            cublasFree( dxh );
            cublasFree( dxl );
            printf( "Device Malloc Error!\n" );
            exit( -1 );
        }
        else
        {
            /* Device Memory allocation */
            cublasSetVector( xsize, sizeof( double ), xh+xOffset, 1, dxh,
                             1 );
            cublasSetVector( xsize, sizeof( double ), xl+xOffset, 1, dxl,
                             1 );
        }

        /* Execute gpu  */
#ifdef _DEBUG_
        int devid;
        cudaGetDevice( &devid );
        printf( "TID(%d), Cuda Device = %d\n", tid, devid );
#endif
        ret[tid] = gddidamax( &partition[tid],  dxh, dxl, incx)+pointerOfEachPartition[tid];

        /* Free Device Memories */
        cublasFree( dxh );
        cublasFree( dxl );
    }

    
    int idamax = 0;
    double maxhi =0.0, maxlo = 0.0;
    
    for(int i=0;i<ngpu;i++)
    {
	    int index = ret[i]-1;
	    printf("index:%d\n",index);
	    double tmphi = xh[index*(*incx)];
	    double tmplo = xl[index*(*incx)];
	    gdd_abs(&tmphi,&tmplo);


	    if(maxhi < tmphi || (maxhi == tmphi && maxlo < tmplo))
	    {
	    maxhi = tmphi;
	    maxlo = tmplo;
	    idamax = index;
	    }
    }

    

    free(partition);
    free(pointerOfEachPartition);
    free(ret);
    return idamax+1;
}

#ifdef __cplusplus
}
#endif
