#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddscal_host_multigpu_( int *n, double *alphah, double *alphal, double *xh,
              double *xl,
              int*incx )
{
/* NUMBER OF GPUS */
    int   ngpu;

    multiple_gpus::bind_cpu_to_gpu_( &ngpu );

    /* ARRAYSIZE FOR EACH GPU */
    int * partition;
    int * pointerOfEachPartition;

    partition = (int*) malloc( sizeof( int )*ngpu );
    pointerOfEachPartition = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( *n, ngpu,  partition, pointerOfEachPartition );


    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        double *dxh, *dxl, *dalphah, *dalphal;

        int     tid = omp_get_thread_num( );
        int     xsize = partition[tid]*( *incx );
        int     xOffset = pointerOfEachPartition[tid]*( *incx );


        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( xsize, sizeof( double ), (void**) &dxh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( xsize, sizeof( double ), (void**) &dxl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dalphah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( double ), (void**) &dalphal )
          )
        {
/*	 Memory allocation failed */
            cublasFree( dxh );
            cublasFree( dxl );
            cublasFree( dalphah );
            cublasFree( dalphal );
            printf( "Device Malloc Error!\n" );
            exit( -1 );
        }
        else
        {
            /* Device Memory allocation */
            cublasSetVector( xsize, sizeof( double ), xh+xOffset, 1, dxh, 1 );
            cublasSetVector( xsize, sizeof( double ), xl+xOffset, 1, dxl, 1 );
            cublasSetVector( 1, sizeof( double ), alphah, 1, dalphah, 1 );
            cublasSetVector( 1, sizeof( double ), alphal, 1, dalphal, 1 );
        }

        /* Execute gpu vector add */
#ifdef _DEBUG_
        int devid;
        cudaGetDevice( &devid );
        printf( "TID(%d), Partition:%d Cuda Device = %d\n", tid,partition[tid], devid );
#endif
        gddscal( &partition[tid],  dalphah, dalphal, dxh, dxl, incx );

        /* Get From Device Memory */
        cublasGetVector( xsize, sizeof( double ), dxh, 1, xh+xOffset, 1 );
        cublasGetVector( xsize, sizeof( double ), dxl, 1, xl+xOffset, 1 );

        /* Free Device Memories */
        cublasFree( dxh );
        cublasFree( dxl );
        cublasFree( dalphah );
        cublasFree( dalphal );
    }

    free(partition);
    free(pointerOfEachPartition);


    return 0;
}

#ifdef __cplusplus
}
#endif
