#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>

#ifdef __cplusplus
extern "C" {
#endif

int
gddtbmv_host_multigpu_( char*uplo, char*trans, char*diag, int *n,int *k,
              double *Ah, double *Al, int *lda, double * xh, double *xl,
              int *incx )
{
    /* NUMBER OF GPUS */
    int     ngpu;

    /* BIND THREADS TO EACH DEVICE */
    multiple_gpus::bind_cpu_to_gpu_( &ngpu );


    int    sizex = *n ;

    /* ARRAYSIZE FOR EACH GPU */
    int *   partition;
    int *   sumPartition;
    partition =    (int*) malloc( sizeof( int )*ngpu );
    sumPartition = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( sizex, ngpu,  partition, sumPartition );


    int     sizeA = ( *n )*( *n );

    double *hAh, *hAl;
    if(
        cudaSuccess !=
        cudaMallocHost( (void **)&hAh, sizeA*sizeof( double ) )
      || cudaSuccess !=
        cudaMallocHost( (void **)&hAl, sizeA*sizeof( double ) )
      )
    {
        printf( "Host Malloc Error\n" );
        return -1;
    }

    int kl,ku;

    if(*uplo == 'U' || *uplo == 'u')
    {
	    ku = *k;
	    kl = 0;
    }else {
	    ku = 0;
	    kl = *k;
    }



    copyToBandSplitMatrix(trans,diag,*n,*n, *lda,kl,ku, ngpu, partition, Ah,
                               Al, hAh,
                               hAl );

 #ifdef _DEBUG_ 
/*    int i, j, l;*/

/*    for( i = 0; i < ngpu; i++ )*/
/*    {*/
/*            for( j = 0; j < partition[i]; j++ )*/
/*            {*/
/*        for( l = 0; l < sizex; l++ )*/
/*        {*/
/*                printf("%e,",*(hAh+j+l*partition[i] + sumPartition[i] * (sizex)));*/
/*            }*/
/*        printf("\n");*/
/*        }*/
/*        printf("\n");*/
/*    }*/
#endif


    double alphah = 1.0;
    double alphal = 0.0;
    double betah = 0.0;
    double betal = 0.0;



/*     Working Matrix  */
    omp_set_num_threads( ngpu ); 
 #pragma omp parallel 
    { 
	int     tid = omp_get_thread_num( ); 
 
        double *dAh,*dAl,*dxh, *dxl,*dyh,*dyl; 

	double *dalphah,*dalphal,*dbetah,*dbetal;
 
 

	int localn = sizex; 
	int localm = partition[tid]; 

	int offsety = sumPartition[tid]*(*incx);
	int localsizey = localm;

	int offsetA = sumPartition[tid]*(localn); 
	int localsizeA = localm*(localn); 
	int locallda =   localm; 
 
 
	/* B,alpha及びbetaを確保 */ 
	if( 
	   CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( localsizey, sizeof( double ), (void**) &dyh ) 
	  || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( localsizey, sizeof( double ), (void**) &dyl ) 
	   || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( sizex, sizeof( double ), (void**) &dxh ) 
	   || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( sizex, sizeof( double ), (void**) &dxl ) 
	   || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( 1, sizeof( double ), (void**) &dalphah ) 
	   || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( 1, sizeof( double ), (void**) &dalphal ) 
	   || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( 1, sizeof( double ), (void**) &dbetah ) 
	   || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( 1, sizeof( double ), (void**) &dbetal ) 
	  || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( localsizeA, sizeof( double ), (void**) &dAh ) 
	  || CUBLAS_STATUS_SUCCESS != 
	    cublasAlloc( localsizeA, sizeof( double ), (void**) &dAl ) 
	  ) 
	{ 
	    /*	 Memory allocation failed */ 
	    cublasFree( dyh ); 
	    cublasFree( dyl ); 
	    cublasFree( dxh ); 
	    cublasFree( dxl ); 
	    cublasFree( dalphah ); 
	    cublasFree( dalphal ); 
	    cublasFree( dbetah ); 
	    cublasFree( dbetal ); 
	    cublasFree( dAh ); 
	    cublasFree( dAl ); 
	    exit( -1 ); 
	} 
 
 
	cublasSetVector( sizex, sizeof( double ), xh, *incx, dxh, 1 ); 
	cublasSetVector( sizex, sizeof( double ), xl, *incx, dxl, 1 ); 
        cublasSetVector( localsizeA, sizeof( double ), hAh+offsetA, 1, dAh, 1); 
        cublasSetVector( localsizeA, sizeof( double ), hAl+offsetA, 1, dAl, 1); 
	cublasSetVector( 1, sizeof( double ), &alphah, 1, dalphah, 1 ); 
	cublasSetVector( 1, sizeof( double ), &alphal, 1, dalphal, 1 ); 
	cublasSetVector( 1, sizeof( double ), &betah, 1, dbetah, 1 ); 
	cublasSetVector( 1, sizeof( double ), &betal, 1, dbetal, 1 ); 

 
 
 #ifdef _DEBUG_ 
	    int devid; 
	    cudaGetDevice( &devid ); 
	    printf( "TILE INDEX(%d), Cuda Device = %d\n", tid, 
		    devid ); 
 #endif 
	    char localtrans = 'N';

	    int localinc = 1;

 
    gddgemv(&localtrans,  &localm,  &localn, dalphah,  dalphal, dAh,  dAl,
  &locallda, dxh,dxl,  &localinc,  dbetah, dbetal,  dyh,  dyl, &localinc); 
#pragma omp barrier
	    cublasGetVector( localsizey, sizeof( double ), dyh, 1, xh+offsety, 
			     *incx ); 
	    cublasGetVector( localsizey, sizeof( double ), dyl, 1, xl+offsety, 
			     *incx ); 
 
	    cublasFree( dAh ); 
	    cublasFree( dAl ); 
	    cublasFree( dxh ); 
	    cublasFree( dxl ); 
	    cublasFree( dyh ); 
	    cublasFree( dyl ); 
	cublasFree( dalphah ); 
	cublasFree( dalphal ); 
	cublasFree( dbetah ); 
	cublasFree( dbetal ); 
    } 
 
 
    cudaFreeHost( hAh ); 
    cudaFreeHost( hAl ); 
 
    free( partition ); 
    free( sumPartition ); 
 
    return 0; 
}

#ifdef __cplusplus
}
#endif
