#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddzdotc_host_multigpu_( int *n,  cuDoubleComplex *xh, cuDoubleComplex *xl, int*incx,
                      cuDoubleComplex *yh, cuDoubleComplex *yl, int *incy  ,cuDoubleComplex* reth,cuDoubleComplex*retl)
{
/* NUMBER OF GPUS */
    int   ngpu;

    multiple_gpus::bind_cpu_to_gpu_( &ngpu );

    /* ARRAYSIZE FOR EACH GPU */
    int * partition;
    int * pointerOfEachPartition;
    cuDoubleComplex * hreth,*hretl;

    partition = (int*) malloc( sizeof( int )*ngpu );
    pointerOfEachPartition = (int*) malloc( sizeof( int )*ngpu );
    hreth = (cuDoubleComplex*) malloc( sizeof( cuDoubleComplex )*ngpu );
    hretl = (cuDoubleComplex*) malloc( sizeof( cuDoubleComplex )*ngpu );
    CalcEachArraySize( *n, ngpu,  partition, pointerOfEachPartition );

    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        cuDoubleComplex *dxh, *dxl, *dyh, *dyl,*dreth,*dretl;
        int     tid = omp_get_thread_num( );
        int     xsize = partition[tid]*( *incx );
        int     ysize = partition[tid]*( *incy );
        int     xOffset = pointerOfEachPartition[tid]*( *incx );
        int     yOffset = pointerOfEachPartition[tid]*( *incy );


        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( xsize, sizeof( cuDoubleComplex ), (void**) &dxh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( xsize, sizeof( cuDoubleComplex ), (void**) &dxl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( ysize, sizeof( cuDoubleComplex ), (void**) &dyh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( ysize, sizeof( cuDoubleComplex ), (void**) &dyl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dreth )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dretl )
          )
        {
/*	 Memory allocation failed */
            cublasFree( dxh );
            cublasFree( dxl );
            cublasFree( dyh );
            cublasFree( dyl );
            cublasFree( dreth );
            cublasFree( dretl );
            printf( "Device Malloc Error!\n" );
            exit( -1 );
        }
        else
        {
            /* Device Memory allocation */
            cublasSetVector( xsize, sizeof( cuDoubleComplex ), xh+xOffset, 1, dxh,
                             1 );
            cublasSetVector( xsize, sizeof( cuDoubleComplex ), xl+xOffset, 1, dxl,
                             1 );
            cublasSetVector( ysize, sizeof( cuDoubleComplex ), yh+yOffset, 1, dyh,
                             1 );
            cublasSetVector( ysize, sizeof( cuDoubleComplex ), yl+yOffset, 1, dyl,
                             1 );
        }

        /* Execute gpu vector add */
#ifdef _DEBUG_
        int devid;
        cudaGetDevice( &devid );
        printf( "TID(%d), Cuda Device = %d\n", tid, devid );
#endif
        gddzdotc( &partition[tid],  dxh, dxl, incx,
                     dyh, dyl,incy,
		  dreth,dretl );

        /* Get From Device Memory */
        cublasGetVector( 1, sizeof( cuDoubleComplex ), dreth, 1, hreth+tid, 1 );
        cublasGetVector( 1, sizeof( cuDoubleComplex ), dretl, 1, hretl+tid, 1 );

        /* Free Device Memories */
        cublasFree( dxh );
        cublasFree( dxl );
        cublasFree( dyh );
        cublasFree( dyl );
	cublasFree(dreth);
	cublasFree(dretl);
    }

    *reth = *retl = make_cuDoubleComplex(0.0,0.0);
    for (int i = 0; i < ngpu ; i++) {
	    CADD(&hreth[i],&hretl[i],reth,retl,reth,retl);
    }

    

    free(partition);
    free(pointerOfEachPartition);
    free(hreth);
    free(hretl);
    return 0;
}

#ifdef __cplusplus
}
#endif
