#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>
#include <cuda_runtime.h>


#ifdef __cplusplus
extern "C" {
#endif
int
gddzgerc_host_multigpu_( int *m, int *n, cuDoubleComplex *alphah, cuDoubleComplex*alphal,
                      cuDoubleComplex * xh, cuDoubleComplex *xl,
                      int *incx,
                      cuDoubleComplex*yh, cuDoubleComplex*yl,
                      int *incy,
                      cuDoubleComplex *Ah, cuDoubleComplex *Al,
                      int *lda )
{
/* NUMBER OF GPUS */
    int      ngpu;

    multiple_gpus::bind_cpu_to_gpu_( &ngpu );

    /* ARRAYSIZE FOR EACH GPU */
    int *    partition;
    int *    pointerOfEachPartition;

    partition = (int*) malloc( sizeof( int )*ngpu );
    pointerOfEachPartition = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( *n, ngpu,  partition, pointerOfEachPartition );

    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        cuDoubleComplex *dxh, *dxl, *dyh, *dyl, *dAh, *dAl,*dalphah,*dalphal;
        int     tid = omp_get_thread_num( );
        int     sizex = ( *m )*( *incx );
        int     sizey = partition[tid]*( *incy );
        int     offsety = pointerOfEachPartition[tid]*( *incy );
	int     sizeA = (*lda)*partition[tid];
	int     offsetA = pointerOfEachPartition[tid]*(*lda);


        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizex, sizeof( cuDoubleComplex ), (void**) &dxh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizex, sizeof( cuDoubleComplex ), (void**) &dxl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizey, sizeof( cuDoubleComplex ), (void**) &dyh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizey, sizeof( cuDoubleComplex ), (void**) &dyl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizeA, sizeof( cuDoubleComplex ), (void**) &dAh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( sizeA, sizeof( cuDoubleComplex ), (void**) &dAl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dalphah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dalphal )
          )
        {
/*	 Memory allocation failed */
            cublasFree( dxh );
            cublasFree( dxl );
            cublasFree( dyh );
            cublasFree( dyl );
            cublasFree( dAh );
            cublasFree( dAl );
            cublasFree( dalphah );
            cublasFree( dalphal );
            printf( "Device Malloc Error!\n" );
            exit( -1 );
        }
        else
        {
            /* Device Memory allocation */
            cublasSetVector( sizex, sizeof( cuDoubleComplex ), xh, 1, dxh,1 );
            cublasSetVector( sizex, sizeof( cuDoubleComplex ), xl, 1, dxl,1 );
            cublasSetVector( sizey, sizeof( cuDoubleComplex ), yh+offsety, 1, dyh,1 );
            cublasSetVector( sizey, sizeof( cuDoubleComplex ), yl+offsety, 1, dyl,1 );
            cublasSetVector( sizeA, sizeof( cuDoubleComplex ), Ah+offsetA, 1, dAh,1 );
            cublasSetVector( sizeA, sizeof( cuDoubleComplex ), Al+offsetA, 1, dAl,1 );
            cublasSetVector( 1, sizeof( cuDoubleComplex ), alphah, 1, dalphah,1 );
            cublasSetVector( 1, sizeof( cuDoubleComplex ), alphal, 1, dalphal,1 );
        }


        /* Execute gpu vector add */
#ifdef _DEBUG_
        int devid;
        cudaGetDevice( &devid );
        printf( "TID(%d), Cuda Device = %d\n", tid, devid );
#endif
        gddzgerc( m,&partition[tid], dalphah,dalphal, dxh, dxl, incx,
                     dyh, dyl, incy,
                     dAh,dAl,lda );

        /* Get From Device Memory */
        cublasGetVector( sizeA, sizeof( cuDoubleComplex ), dAh, 1, Ah+offsetA, 1 );
        cublasGetVector( sizeA, sizeof( cuDoubleComplex ), dAl, 1, Al+offsetA, 1 );

        /* Free Device Memories */
        cublasFree( dxh );
        cublasFree( dxl );
        cublasFree( dyh );
        cublasFree( dyl );
        cublasFree( dAh );
        cublasFree( dAl );
        cublasFree( dalphah );
        cublasFree( dalphal );
    }

    free( partition );
    free( pointerOfEachPartition );
    return 0;
}

#ifdef __cplusplus
}
#endif
