#include "gddcommon.h"
#include "multiple_gpu.h"
#include <omp.h>
#include <cutil.h>
#include <cuda.h>

#ifdef __cplusplus
extern "C" {
#endif

int
gddzhbmv_host_multigpu_( char*uplo, int *n,int *kb,cuDoubleComplex *alphah,cuDoubleComplex *alphal,
              cuDoubleComplex *Ah, cuDoubleComplex *Al, int *lda, cuDoubleComplex * xh, cuDoubleComplex *xl,
              int *incx ,cuDoubleComplex *betah,cuDoubleComplex*betal,cuDoubleComplex * yh, cuDoubleComplex *yl,
              int *incy  )
{
    /* NUMBER OF GPUS */
    int      ngpu;
    /* BIND THREADS TO EACH DEVICE */
    multiple_gpus::bind_cpu_to_gpu_( &ngpu );


    int      sizex = *n ;
    int      sizey = *n ;


    /* ARRAYSIZE FOR EACH GPU */
    int *    partition;
    int *    sumPartition;
    partition = (int*) malloc( sizeof( int )*ngpu );
    sumPartition = (int*) malloc( sizeof( int )*ngpu );
    CalcEachArraySize( sizey, ngpu,  partition, sumPartition );


    /* Working Matrix */


    cuDoubleComplex * hAh, *hAl;

    int sizeA = (*n)*(*n);
    if(
            cudaSuccess !=
            cudaMallocHost( (void **)&hAh, sizeA*sizeof( cuDoubleComplex ) )
          || cudaSuccess !=
            cudaMallocHost( (void **)&hAl, sizeA*sizeof( cuDoubleComplex ) )
          )
        {
            printf( "Host Malloc Error\n" );
            return -1;
        }

    copyToHermitianBandSplitMatrix(*n,*kb,*lda,uplo,ngpu,partition,Ah,Al,hAh,hAl);

 #ifdef _DEBUG_ 
/*    int i, j, k;*/

/*    for( i = 0; i < ngpu; i++ )*/
/*    {*/
/*            for( j = 0; j < partition[i]; j++ )*/
/*            {*/
/*        for( k = 0; k < sizex; k++ )*/
/*        {*/
/*                printf("%e,",*(hAh+j+k*partition[i] + sumPartition[i] * (sizex)));*/
/*            }*/
/*        printf("\n");*/
/*        }*/
/*        printf("\n");*/
/*    }*/
#endif




    omp_set_num_threads( ngpu );
#pragma omp parallel
    {
        int     tid = omp_get_thread_num( );

        cuDoubleComplex *dAh,*dAl,*dyh,*dyl,*dxh, *dxl, *dalphah, *dalphal, *dbetah, *dbetal;

	int offsety = sumPartition[tid] *(*incy);
	int localsizey = partition[tid]*(*incy);

	int localsizex = (*n)*(*incx);

	int offsetA = sumPartition[tid]*(*n);
	int localsizeA = partition[tid]*(*n);
	int locallda = partition[tid];

	int localm = partition[tid];
	int localn = *n;

        /* B,alpha及びbetaを確保 */
        if(
            CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dalphah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dalphal )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dbetah )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( 1, sizeof( cuDoubleComplex ), (void**) &dbetal )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( localsizex, sizeof( cuDoubleComplex ), (void**) &dxh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( localsizex, sizeof( cuDoubleComplex ), (void**) &dxl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( localsizey, sizeof( cuDoubleComplex ), (void**) &dyh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( localsizey, sizeof( cuDoubleComplex ), (void**) &dyl )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( localsizeA, sizeof( cuDoubleComplex ), (void**) &dAh )
          || CUBLAS_STATUS_SUCCESS !=
            cublasAlloc( localsizeA, sizeof( cuDoubleComplex ), (void**) &dAl )
          )
        {
            /*	 Memory allocation failed */
            cublasFree( dalphah );
            cublasFree( dalphal );
            cublasFree( dbetah );
            cublasFree( dbetal );
            cublasFree( dxh );
            cublasFree( dxl );
            cublasFree( dyh );
            cublasFree( dyl );
            cublasFree( dAh );
            cublasFree( dAl );
            exit( -1 );
        }


        cublasSetVector( 1, sizeof( cuDoubleComplex ), alphah, 1, dalphah, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), alphal, 1, dalphal, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), betah, 1, dbetah, 1 );
        cublasSetVector( 1, sizeof( cuDoubleComplex ), betal, 1, dbetal, 1 );
        cublasSetVector( localsizex, sizeof( cuDoubleComplex ), xh, 1, dxh, 1 );
        cublasSetVector( localsizex, sizeof( cuDoubleComplex ), xl, 1, dxl, 1 );
        cublasSetVector( localsizey, sizeof( cuDoubleComplex ), yh+offsety, 1, dyh, 1 );
        cublasSetVector( localsizey, sizeof( cuDoubleComplex ), yl+offsety, 1, dyl, 1 );
        cublasSetVector( localsizeA, sizeof( cuDoubleComplex ), hAh+offsetA, 1, dAh, 1 );
        cublasSetVector( localsizeA, sizeof( cuDoubleComplex ), hAl+offsetA, 1, dAl, 1 );


#ifdef _DEBUG_
            int devid;
            cudaGetDevice( &devid );
            printf( "TILE INDEX(%d), Cuda Device = %d\n", tid,
                    devid );
#endif

	    char trans  = 'N';
    gddzgemv(&trans,  &localm,  &localn, dalphah,  dalphal, dAh,  dAl,  &locallda, dxh,
        dxl,  incx,  dbetah, dbetal,  dyh,  dyl, incy );
            cublasGetVector( localsizey, sizeof( cuDoubleComplex ), dyh, 1, yh+offsety,
                             1 );
            cublasGetVector( localsizey, sizeof( cuDoubleComplex ), dyl, 1, yl+offsety,
                             1 );

            cublasFree( dAh );
            cublasFree( dAl );
            cublasFree( dxh );
            cublasFree( dxl );
            cublasFree( dyh );
            cublasFree( dyl );
        cublasFree( dalphah );
        cublasFree( dalphal );
        cublasFree( dbetah );
        cublasFree( dbetal );
    }


    cudaFreeHost( hAh );
    cudaFreeHost( hAl );

    free( partition );
    free( sumPartition );

    return 0;
}

#ifdef __cplusplus
}
#endif
